/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "agent-server.h"
#include "agent-app.h"

struct _AgentAppPrivate
{
  time_t        uptime;
  GMainLoop    *mainloop;
  GKeyFile     *keyfile;
  AgentServer  *server;
};
#define AGENT_APP_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), AGENT_TYPE_APP, AgentAppPrivate))

G_DEFINE_TYPE (AgentApp, agent_app, G_TYPE_OBJECT);

static AgentApp	  *_application = NULL;

static gboolean	agent_conf_has		(AgentApp     *app,
					 const gchar  *group,
					 const gchar  *key);
static gboolean	agent_conf_get_bool	(AgentApp     *app,
					 const gchar  *group,
					 const gchar  *key);
static gint	agent_conf_get_int	(AgentApp     *app,
					 const gchar  *group,
					 const gchar  *key);
static gchar*	agent_conf_get_string	(AgentApp     *app,
					 const gchar  *group,
					 const gchar  *key,
					 gboolean      locale);
#define agent_conf_get_str(app, group, key)	agent_conf_get_string(app, group, key, FALSE)
#define agent_conf_get_lstr(app, group, key)	agent_conf_get_string(app, group, key, TRUE)

/**
 * agent_app_init:
 */
void
agent_app_init (AgentApp *app)
{
  app->priv = AGENT_APP_PRIVATE (app);

  app->priv->uptime   = time (NULL);
  app->priv->mainloop = g_main_loop_new (NULL, FALSE);
  app->priv->keyfile  = g_key_file_new ();
  app->priv->server   = agent_server_new ();
}

/**
 * agent_app_dispose:
 */
void
agent_app_dispose (GObject *object)
{
  AgentApp *app = AGENT_APP (object);

  if (app->priv->server)
    {
      g_object_unref (app->priv->server);
      app->priv->server = NULL;
    }
  G_OBJECT_CLASS (agent_app_parent_class)->dispose (object);
}

/**
 * agent_app_finalize:
 */
void
agent_app_finalize (GObject *object)
{
  AgentApp *app = AGENT_APP (object);

  if (app->priv->keyfile)
    {
      g_key_file_free (app->priv->keyfile);
      app->priv->keyfile = NULL;
    }
  if (app->priv->mainloop)
    {
      g_main_loop_unref (app->priv->mainloop);
      app->priv->mainloop = NULL;
    }
  G_OBJECT_CLASS (agent_app_parent_class)->finalize (object);
}

/**
 * agent_app_class_init:
 */
void
agent_app_class_init (AgentAppClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (AgentAppPrivate));

  gobject_class->dispose  = agent_app_dispose;
  gobject_class->finalize = agent_app_finalize;
}

/**
 * agent_conf_has:
 */
static inline gboolean
agent_conf_has (AgentApp    *app,
		const gchar *group,
		const gchar *key)
{
  g_return_val_if_fail (AGENT_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  return g_key_file_has_key (app->priv->keyfile,
			     group, key, NULL);
}

/**
 * agent_conf_get_bool:
 */
static inline gboolean
agent_conf_get_bool (AgentApp    *app,
		     const gchar *group,
		     const gchar *key)
{
  gboolean ret;
  GError *error = NULL;

  g_return_val_if_fail (AGENT_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  ret = g_key_file_get_boolean (app->priv->keyfile,
				group, key, &error);
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}

/**
 * agent_conf_get_int:
 */
static inline gint
agent_conf_get_int (AgentApp    *app,
		    const gchar *group,
		    const gchar *key)
{
  gint ret;
  GError *error = NULL;

  g_return_val_if_fail (AGENT_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  ret = g_key_file_get_integer (app->priv->keyfile,
				group, key, &error);
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}

/**
 * agent_conf_get_string:
 */
static inline gchar*
agent_conf_get_string (AgentApp    *app,
		       const gchar *group,
		       const gchar *key,
		       gboolean     locale)
{
  gchar *ret;
  GError *error = NULL;

  g_return_val_if_fail (AGENT_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  if (locale)
    {
      ret = g_key_file_get_locale_string (app->priv->keyfile,
					  group, key, NULL, &error);
    }
  else
    {
      ret = g_key_file_get_string (app->priv->keyfile,
				   group, key, &error);
    }
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}

/**
 * agent_app_new:
 */
AgentApp*
agent_app_new (void)
{
  static volatile gsize once = 0;

  if (g_once_init_enter (&once))
    {
      _application = g_object_new (AGENT_TYPE_APP, NULL);
      g_once_init_leave (&once, 1);
    }
  return _application;
}

/**
 * agent_app_get:
 */
AgentApp*
agent_app_get (void)
{
  return _application;
}

/**
 * agent_app_get_uptime:
 */
time_t
agent_app_get_uptime (AgentApp  *app)
{
  g_return_val_if_fail (AGENT_IS_APP (app), 0);
  return app->priv->uptime;
}

/**
 * agent_app_get_server:
 */
AgentServer*
agent_app_get_server (AgentApp *app)
{
  g_return_val_if_fail (AGENT_IS_APP (app), NULL);
  return app->priv->server;
}

/**
 * agent_app_load_config_file:
 */
gboolean
agent_app_load_config_file (AgentApp    *app,
			    const gchar *filename)
{
  GError *error = NULL;

  g_return_val_if_fail (AGENT_IS_APP (app), FALSE);
  g_return_val_if_fail (filename, FALSE);

  g_key_file_load_from_file (app->priv->keyfile, filename,
			     G_KEY_FILE_NONE, &error);
  if (error)
    {
      g_warning (_("Failed to load configure file `%s': %s"),
		 filename, error->message);
      g_error_free (error);
      return FALSE;
    }
  return TRUE;
}

/**
 * agent_app_start:
 *
 * start application main loop. this function will not
 * return until received terminate signal.
 */
gboolean
agent_app_start (AgentApp *app)
{
  gint server_type = SERVER_TYPE_PEER;
  gchar *server_addr = NULL;

  g_return_val_if_fail (AGENT_IS_APP (app), FALSE);

  /* we can startup without config file. */
  if (g_key_file_has_group (app->priv->keyfile, "Agent"))
    {
      /* server type(bus client or peer server) */
      if (agent_conf_has (app, "Agent", "server-type"))
	server_type = agent_conf_get_int (app, "Agent", "server-type");

      if (server_type < SERVER_TYPE_PEER ||
	  server_type > SERVER_TYPE_BUS_SYSTEM)
	{
	  g_critical (_("Value of configure entry `server-type' must "
			"between %d and %d(included)."),
		      SERVER_TYPE_PEER, SERVER_TYPE_BUS_SYSTEM);
	  return FALSE;
	}
      /* read peer server settings. */
      if (server_type == SERVER_TYPE_PEER)
	{
	  gboolean anonymous = TRUE;

	  /* if server type is peer server, need a dbus address. */
	  server_addr = agent_conf_get_str (app, "Agent", "server-addr");
	  if (!server_addr || !server_addr[0])
	    server_addr = g_strdup ("tcp:host=0.0.0.0,port=5433");

	  /* whether or not allow anonymous auth. */
	  if (agent_conf_has (app, "Agent", "auth-anonymous"))
	    anonymous = agent_conf_get_bool (app, "Agent", "auth-anonymous");

	  /* apply settings and start server. */
	  agent_server_set_auth_anonymous (app->priv->server, anonymous);
	}
    }
  if (!agent_server_start (app->priv->server, server_type, server_addr))
    {
      g_free (server_addr);
      return FALSE;
    }
  g_free (server_addr);

  /* enter main loop. */
  g_main_loop_run (app->priv->mainloop);
  return TRUE;
}

/**
 * agent_app_quit:
 *
 * quit application.
 *
 * the only way to call this function is receive a
 * system terminate signal(SIGTERM, SIGHUP)
 */
void
agent_app_quit (AgentApp *app)
{
  g_return_if_fail (AGENT_IS_APP (app));

  agent_server_stop (app->priv->server);

  if (app->priv->mainloop)
    g_main_loop_quit (app->priv->mainloop);
}
