/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>
#include <gio/gio.h>

#include "agent-server.h"
#include "agent-recorder-manager.h"

typedef struct _AgentRecorderServer	AgentRecorderServer;

struct _AgentRecorderServer
{
  guint64   time;
  gchar    *hostname;
  gchar    *username;
  guint32   pid;
  gchar    *ttyname;
  gchar    *ptsname;
  gchar    *command;
  guint16   lookport;
};

struct _AgentRecorderManagerPrivate
{
  GHashTable  *recorder_servers;
  GMutex      *recorder_servers_mutex;
};
#define AGENT_RECORDER_MANAGER_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), AGENT_TYPE_RECORDER_MANAGER, AgentRecorderManagerPrivate))

typedef enum
{
  AGENT_RECORDER_MANAGER_ERROR_DUPLICATION,
  AGENT_RECORDER_MANAGER_ERROR_NOTFOUND,
  AGENT_RECORDER_MANAGER_ERROR_EMPTY,
} AgentRecorderManagerError;

G_DEFINE_TYPE (AgentRecorderManager, agent_recorder_manager, G_TYPE_OBJECT);

static const GDBusErrorEntry	_recorder_manager_error_entries[] =
{
    { AGENT_RECORDER_MANAGER_ERROR_DUPLICATION,	"net.ttyutils.Agent.RecorderManager.Duplication" },
    { AGENT_RECORDER_MANAGER_ERROR_NOTFOUND,	"net.ttyutils.Agent.RecorderManager.Notfound" },
    { AGENT_RECORDER_MANAGER_ERROR_EMPTY,	"net.ttyutils.Agent.RecorderManager.Empty" },
};
#define AGENT_RECORDER_MANAGER_ERROR	(agent_recorder_manager_error_quark ())

/**
 * agent_recorder_manager_error_quark:
 */
static GQuark
agent_recorder_manager_error_quark (void)
{
  static volatile gsize quark_volatile = 0;

  g_dbus_error_register_error_domain ("agent-recorder-manager-error-quark",
				      &quark_volatile,
				      _recorder_manager_error_entries,
				      G_N_ELEMENTS (_recorder_manager_error_entries));
  return (GQuark) quark_volatile;
}

static void		agent_recorder_manager_handle_method_call	(GDBusConnection       *connection,
									 const gchar           *sender,
									 const gchar           *object_path,
									 const gchar           *interface_name,
									 const gchar           *method_name,
									 GVariant              *parameters,
									 GDBusMethodInvocation *invocation,
									 gpointer               user_data);
static GVariant*	agent_recorder_manager_handle_get_property	(GDBusConnection       *connection,
									 const gchar           *sender,
									 const gchar           *object_path,
									 const gchar           *interface_name,
									 const gchar           *property_name,
									 GError               **error,
									 gpointer               user_data);
static gboolean		agent_recorder_manager_handle_set_property	(GDBusConnection       *connection,
									 const gchar           *sender,
									 const gchar           *object_path,
									 const gchar           *interface_name,
									 const gchar           *property_name,
									 GVariant              *value,
									 GError               **error,
									 gpointer               user_data);

GDBusInterfaceVTable	agent_recorder_manager_vtable =
{
  agent_recorder_manager_handle_method_call,
  agent_recorder_manager_handle_get_property,
  agent_recorder_manager_handle_set_property,
};

typedef void	(*RecorderManagerMethodHandler)	(AgentRecorderManager  *recorder_manager,
						 GDBusConnection       *connection,
						 const gchar           *object_path,
						 const gchar           *interface_name,
						 GVariant              *parameters,
						 GDBusMethodInvocation *invocation);

typedef struct _RecorderManagerMethodHandlerEntry	RecorderManagerMethodHandlerEntry;
struct _RecorderManagerMethodHandlerEntry
{
  const gchar *name;
  RecorderManagerMethodHandler handler;
};


/**
 * agent_recorder_server_new:
 */
static AgentRecorderServer*
agent_recorder_server_new (void)
{
  return g_slice_new0 (AgentRecorderServer);
}

/**
 * agent_recorder_server_free:
 */
static void
agent_recorder_server_free (AgentRecorderServer *server)
{
  if (server)
    {
      g_free (server->hostname);
      g_free (server->username);
      g_free (server->ttyname);
      g_free (server->ptsname);
      g_free (server->command);

      g_slice_free (AgentRecorderServer, server);
    }
}

/**
 * agent_recorder_manager_init:
 */
void
agent_recorder_manager_init (AgentRecorderManager *manager)
{
  manager->priv = AGENT_RECORDER_MANAGER_PRIVATE (manager);

  manager->priv->recorder_servers =
   g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
			  (GDestroyNotify) agent_recorder_server_free);
  manager->priv->recorder_servers_mutex = g_mutex_new ();
}

/**
 * agent_recorder_manager_dispose:
 */
void
agent_recorder_manager_dispose (GObject *object)
{
#if 0
  AgentRecorderManager *recorder = AGENT_RECORDER_MANAGER (object);

#endif
  G_OBJECT_CLASS (agent_recorder_manager_parent_class)->dispose (object);
}

/**
 * agent_recorder_manager_finalize:
 */
void
agent_recorder_manager_finalize (GObject *object)
{
  AgentRecorderManager *manager = AGENT_RECORDER_MANAGER (object);

  if (manager->priv->recorder_servers)
    {
      g_hash_table_destroy (manager->priv->recorder_servers);
      manager->priv->recorder_servers = NULL;
    }
  if (manager->priv->recorder_servers_mutex)
    {
      g_mutex_free (manager->priv->recorder_servers_mutex);
      manager->priv->recorder_servers_mutex = NULL;
    }
  G_OBJECT_CLASS (agent_recorder_manager_parent_class)->finalize (object);
}

/**
 * agent_recorder_manager_class_init:
 */
void
agent_recorder_manager_class_init (AgentRecorderManagerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (AgentRecorderManagerPrivate));

  gobject_class->dispose  = agent_recorder_manager_dispose;
  gobject_class->finalize = agent_recorder_manager_finalize;
}

/**
 * agent_recorder_manager_new:
 */
AgentRecorderManager*
agent_recorder_manager_new (void)
{
  return g_object_new (AGENT_TYPE_RECORDER_MANAGER, NULL);
}

/**
 * agent_recorder_manager_login:
 *
 * add new recorder server to list and emit Added signal.
 */
static void
agent_recorder_manager_login (AgentRecorderManager  *manager,
			      GDBusConnection       *connection,
			      const gchar           *object_path,
			      const gchar           *interface_name,
			      GVariant              *parameters,
			      GDBusMethodInvocation *invocation)
{
  AgentRecorderServer *server;
  gchar *key;

  g_return_if_fail (AGENT_IS_RECORDER_MANAGER (manager));
  g_return_if_fail (parameters);
  g_return_if_fail (G_IS_DBUS_METHOD_INVOCATION (invocation));

  server = agent_recorder_server_new ();

  /* get call parameters. */
  g_variant_get (parameters, "(tssusssq)", &server->time,
		 &server->hostname, &server->username,
		 &server->pid, &server->ttyname, &server->ptsname,
		 &server->command, &server->lookport);

  /* use hostname, ttyname, ptsname and pid as hash key. */
  key = g_strdup_printf ("%s:%s:%s:%d", server->hostname,
			 server->ttyname, server->ptsname, server->pid);

  /* remember we are in multithreads environments. */
  g_mutex_lock (manager->priv->recorder_servers_mutex);

  if (g_hash_table_lookup (manager->priv->recorder_servers, key))
    {
      g_mutex_unlock (manager->priv->recorder_servers_mutex);

      g_dbus_method_invocation_return_error (invocation, AGENT_RECORDER_MANAGER_ERROR,
					     AGENT_RECORDER_MANAGER_ERROR_DUPLICATION,
					     _("Duplication recorder server `%s'"), key);
      agent_recorder_server_free (server);
      g_free (key);
      return;
    }
  g_hash_table_insert (manager->priv->recorder_servers, key, server);
  g_mutex_unlock (manager->priv->recorder_servers_mutex);

  g_dbus_method_invocation_return_value (invocation, NULL);
  g_dbus_connection_emit_signal (connection, NULL, object_path, interface_name,
				 "Added", parameters, NULL);
}

/**
 * agent_recorder_manager_logout:
 */
static void
agent_recorder_manager_logout (AgentRecorderManager  *manager,
			       GDBusConnection       *connection,
			       const gchar           *object_path,
			       const gchar           *interface_name,
			       GVariant              *parameters,
			       GDBusMethodInvocation *invocation)
{
  AgentRecorderServer *server;
  gchar *key;

  g_return_if_fail (AGENT_IS_RECORDER_MANAGER (manager));
  g_return_if_fail (parameters);
  g_return_if_fail (G_IS_DBUS_METHOD_INVOCATION (invocation));

  server = agent_recorder_server_new ();

  g_variant_get (parameters, "(suss)", &server->hostname,
		 &server->pid, &server->ttyname, &server->ptsname);

  key = g_strdup_printf ("%s:%s:%s:%d", server->hostname,
			 server->ttyname, server->ptsname, server->pid);
  agent_recorder_server_free (server);

  g_mutex_lock (manager->priv->recorder_servers_mutex);

  if (!g_hash_table_lookup (manager->priv->recorder_servers, key))
    {
      g_mutex_unlock (manager->priv->recorder_servers_mutex);

      g_dbus_method_invocation_return_error (invocation, AGENT_RECORDER_MANAGER_ERROR,
					     AGENT_RECORDER_MANAGER_ERROR_NOTFOUND,
					     _("Recorder server `%s' not found."), key);
      g_free (key);
      return;
    }
  g_hash_table_remove (manager->priv->recorder_servers, key);
  g_free (key);

  g_mutex_unlock (manager->priv->recorder_servers_mutex);

  g_dbus_method_invocation_return_value (invocation, NULL);
  g_dbus_connection_emit_signal (connection, NULL, object_path, interface_name,
				 "Removed", parameters, NULL);
}


/* struct to pass data to hash foreach function. */
typedef struct
{
  AgentRecorderManager *manager;
  GVariantBuilder      *builder;
  gchar                *hostname;
  gint                  count;
} GetListForeachData;

/**
 * agent_recorder_manager_add_server:
 *
 * a hash foreach function to add register server.
 */
static void
agent_recorder_manager_add_server (gpointer key,
				   gpointer value,
				   gpointer user_data)
{
  GetListForeachData *data = user_data;
  AgentRecorderServer *server = value;

  if (data->hostname &&
      g_strcmp0 (server->hostname, data->hostname) != 0)
    return;

  g_variant_builder_add (data->builder, "(tssusssq)",
			 server->time, server->hostname, server->username,
			 server->pid, server->ttyname, server->ptsname,
			 server->command, server->lookport);
  data->count++;
}

/**
 * agent_recorder_manager_get_list:
 */
static void
agent_recorder_manager_get_list (AgentRecorderManager  *manager,
				 GDBusConnection       *connection,
				 const gchar           *object_path,
				 const gchar           *interface_name,
				 GVariant              *parameters,
				 GDBusMethodInvocation *invocation)
{
  GVariantBuilder *builder;
  GetListForeachData data;

  g_return_if_fail (AGENT_IS_RECORDER_MANAGER (manager));
  g_return_if_fail (parameters);
  g_return_if_fail (G_IS_DBUS_METHOD_INVOCATION (invocation));

  builder = g_variant_builder_new (G_VARIANT_TYPE_TUPLE);
  g_variant_builder_open (builder, G_VARIANT_TYPE_ARRAY);

  data.manager  = manager;
  data.builder  = builder;
  data.hostname = NULL;
  data.count = 0;

  g_mutex_lock (manager->priv->recorder_servers_mutex);

  g_hash_table_foreach (manager->priv->recorder_servers,
			agent_recorder_manager_add_server, &data);

  g_mutex_unlock (manager->priv->recorder_servers_mutex);

  if (data.count == 0)
    {
      g_dbus_method_invocation_return_error (invocation,
					     AGENT_RECORDER_MANAGER_ERROR,
					     AGENT_RECORDER_MANAGER_ERROR_EMPTY,
					     _("No registered recorder server."));
    }
  else
    {
      GVariant *result;

      g_variant_builder_close (builder);
      result = g_variant_builder_end (builder);
      g_dbus_method_invocation_return_value (invocation, result);
    }
  g_variant_builder_unref (builder);
}


static RecorderManagerMethodHandlerEntry	_recorder_manager_method_entries[] =
{
    { "Login",		agent_recorder_manager_login },
    { "Logout",		agent_recorder_manager_logout },
    { "GetList",	agent_recorder_manager_get_list },
};

/**
 * agent_recorder_handle_method_call:
 *
 * process method call.
 */
static void
agent_recorder_manager_handle_method_call (GDBusConnection       *connection,
					   const gchar           *sender,
					   const gchar           *object_path,
					   const gchar           *interface_name,
					   const gchar           *method_name,
					   GVariant              *parameters,
					   GDBusMethodInvocation *invocation,
					   gpointer               user_data)
{
  AgentRecorderManager *manager = user_data;
  gint i;

  g_return_if_fail (AGENT_IS_RECORDER_MANAGER (manager));

  if (g_strcmp0 (object_path, "/net/ttyutils/Agent") != 0 ||
      g_strcmp0 (interface_name, "net.ttyutils.Agent.RecorderManager") != 0)
    {
      g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR,
					     G_DBUS_ERROR_SERVICE_UNKNOWN,
					     "Unknown object or interface.");
      return;
    }
  for (i = 0; i < G_N_ELEMENTS (_recorder_manager_method_entries); i++)
    {
      if (g_strcmp0 (method_name, _recorder_manager_method_entries[i].name) == 0)
	{
	  _recorder_manager_method_entries[i].handler (manager, connection,
						       object_path, interface_name,
						       parameters, invocation);
	  return;
	}
    }
  g_dbus_method_invocation_return_error (invocation, G_DBUS_ERROR,
					 G_DBUS_ERROR_UNKNOWN_METHOD,
					 "Unknown method `%s'.", method_name);
}

/**
 * agent_recorder_handle_get_property:
 */
static GVariant*
agent_recorder_manager_handle_get_property (GDBusConnection  *connection,
					    const gchar      *sender,
					    const gchar      *object_path,
					    const gchar      *interface_name,
					    const gchar      *property_name,
					    GError          **error,
					    gpointer          user_data)
{
  AgentRecorderManager *manager = user_data;
  GVariant *ret = NULL;

  g_return_val_if_fail (AGENT_IS_RECORDER_MANAGER (manager), NULL);

  if (g_strcmp0 (property_name, "Count") == 0)
    {
      gint size = g_hash_table_size (manager->priv->recorder_servers);
      ret = g_variant_new_uint32 (size);
    }
  return ret;
}

/**
 * agent_recorder_manager_handle_set_property:
 */
static gboolean
agent_recorder_manager_handle_set_property (GDBusConnection  *connection,
					    const gchar      *sender,
					    const gchar      *object_path,
					    const gchar      *interface_name,
					    const gchar      *property_name,
					    GVariant         *value,
					    GError          **error,
					    gpointer          user_data)
{
  AgentRecorderManager *manager = user_data;

  g_return_val_if_fail (AGENT_IS_RECORDER_MANAGER (manager), FALSE);
#if 0
  if (g_strcmp0 (property_name, "Count") == 0)
    {
      g_object_set (recorder_manager, "count", g_variant_get_int32 (value), NULL);
    }
#endif
  return TRUE;
}
