/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>
#include <gio/gio.h>

#include <ttyutils/ttyutils.h>

#include "agent-app.h"
#include "agent-recorder-manager.h"
#include "agent-server.h"

static const gchar	*_introspection_xml =
"<node>"
"  <interface name='net.ttyutils.Agent.RecorderManager'>"
"    <method name='Login'>"
"      <arg name='time'     type='t' direction='in'/>"
"      <arg name='hostname' type='s' direction='in'/>"
"      <arg name='username' type='s' direction='in'/>"
"      <arg name='pid'      type='u' direction='in'/>"
"      <arg name='ttyname'  type='s' direction='in'/>"
"      <arg name='ptsname'  type='s' direction='in'/>"
"      <arg name='command'  type='s' direction='in'/>"
"      <arg name='lookport' type='q' direction='in'/>"
"    </method>"
"    <method name='Logout'>"
"      <arg name='hostname' type='s' direction='in'/>"
"      <arg name='pid'      type='u' direction='in'/>"
"      <arg name='ttyname'  type='s' direction='in'/>"
"      <arg name='ptsname'  type='s' direction='in'/>"
"    </method>"
"    <method name='GetList'>"
"      <arg name='recorder_servers' type='a(tssusssq)' direction='out'/>"
"    </method>"
"    <signal name='Added'>"
"      <arg name='time'     type='t' direction='out'/>"
"      <arg name='hostname' type='s' direction='out'/>"
"      <arg name='username' type='s' direction='out'/>"
"      <arg name='pid'      type='u' direction='out'/>"
"      <arg name='ttyname'  type='s' direction='out'/>"
"      <arg name='ptsname'  type='s' direction='out'/>"
"      <arg name='command'  type='s' direction='out'/>"
"      <arg name='lookport' type='q' direction='out'/>"
"    </signal>"
"    <signal name='Removed'>"
"      <arg name='hostname' type='s' direction='out'/>"
"      <arg name='pid'      type='u' direction='out'/>"
"      <arg name='ttyname'  type='s' direction='out'/>"
"      <arg name='ptsname'  type='s' direction='out'/>"
"    </signal>"
"    <property name='Count' type='u' access='read'/>"
"  </interface>"
"</node>";

struct _AgentServerPrivate
{
  gboolean              auth_anonymous;
  GDBusNodeInfo        *node_info;
  GDBusServer          *peer_server;
  GDBusConnection      *bus_connection;
  GArray               *interface_ids;
  guint                 own_name_id;
  GSList               *connections;
  AgentRecorderManager *recorder_manager;
};
#define AGENT_SERVER_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), AGENT_TYPE_SERVER, AgentServerPrivate))

G_DEFINE_TYPE (AgentServer, agent_server, G_TYPE_OBJECT);

/* defined in agent-recorder-manager.c */
extern GDBusInterfaceVTable	agent_recorder_manager_vtable;

/**
 * agent_server_init:
 */
void
agent_server_init (AgentServer *server)
{
  GError *error = NULL;

  server->priv = AGENT_SERVER_PRIVATE (server);

  server->priv->node_info =
   g_dbus_node_info_new_for_xml (_introspection_xml, &error);
  if (error)
    g_error ("%s", error->message);

  server->priv->auth_anonymous = TRUE;
  server->priv->peer_server    = NULL;
  server->priv->bus_connection = NULL;
  server->priv->interface_ids  = g_array_new (FALSE, FALSE, sizeof (guint));
  server->priv->own_name_id    = 0;
  server->priv->connections    = NULL;
  server->priv->recorder_manager = agent_recorder_manager_new ();
}

/**
 * agent_server_dispose:
 */
void
agent_server_dispose (GObject *object)
{
  AgentServer *server = AGENT_SERVER (object);

  if (server->priv->connections)
    {
      g_slist_foreach (server->priv->connections, (GFunc) g_object_unref, NULL);
      g_slist_free (server->priv->connections);
      server->priv->connections = NULL;
    }
  G_OBJECT_CLASS (agent_server_parent_class)->dispose (object);
}

/**
 * agent_server_finalize:
 */
void
agent_server_finalize (GObject *object)
{
  AgentServer *server = AGENT_SERVER (object);

  if (server->priv->recorder_manager)
    {
      g_object_unref (server->priv->recorder_manager);
      server->priv->recorder_manager = NULL;
    }
  if (server->priv->peer_server)
    {
      g_object_unref (server->priv->peer_server);
      server->priv->peer_server = NULL;
    }
  if (server->priv->bus_connection)
    {
      g_object_unref (server->priv->bus_connection);
      server->priv->bus_connection = NULL;
    }
  if (server->priv->node_info)
    {
      g_dbus_node_info_unref (server->priv->node_info);
      server->priv->node_info = NULL;
    }
  if (server->priv->interface_ids)
    {
      g_array_free (server->priv->interface_ids, TRUE);
      server->priv->interface_ids = NULL;
    }
  G_OBJECT_CLASS (agent_server_parent_class)->finalize (object);
}

/**
 * agent_server_class_init:
 */
void
agent_server_class_init (AgentServerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (AgentServerPrivate));

  gobject_class->dispose  = agent_server_dispose;
  gobject_class->finalize = agent_server_finalize;
}

/**
 * agent_server_new:
 */
AgentServer*
agent_server_new (void)
{
  return g_object_new (AGENT_TYPE_SERVER, NULL);
}

/**
 * agent_server_set_auth_anonymous:
 */
void
agent_server_set_auth_anonymous (AgentServer *server,
				 gboolean     anonymous)
{
  g_return_if_fail (AGENT_IS_SERVER (server));
  server->priv->auth_anonymous = anonymous;
}

/**
 * agent_server_register_interface:
 */
static void
agent_server_register_interface (AgentServer          *server,
				 GDBusConnection      *connection,
				 GDBusInterfaceInfo   *interface,
				 GDBusInterfaceVTable *vtable,
				 GObject              *object)
{
  guint id;
  GError *error = NULL;

  id = g_dbus_connection_register_object (connection,
					  "/net/ttyutils/Agent",
					  interface, vtable, object,
					  NULL, &error);
  if (error)
    {
      g_critical ("%s", error->message);
      g_error_free (error);
      return;
    }
  g_array_append_val (server->priv->interface_ids, id);
}

/**
 * agent_server_register_objects:
 *
 * register all objects on connection.
 */
static void
agent_server_register_objects (AgentServer     *server,
			       GDBusConnection *connection)
{
  GDBusInterfaceInfo *interface;

  g_return_if_fail (AGENT_IS_SERVER (server));
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));

  /* register AgentRecorderManager object to RecorderManager interface. */
  interface = g_dbus_node_info_lookup_interface (server->priv->node_info,
						 "net.ttyutils.Agent.RecorderManager");
  agent_server_register_interface (server, connection, interface,
				   &agent_recorder_manager_vtable,
				   G_OBJECT (server->priv->recorder_manager));
}

/**
 * agent_server_unregister_objects:
 *
 * unregister all objects.
 */
static void
agent_server_unregister_objects (AgentServer     *server,
				 GDBusConnection *connection)
{
  gint i;

  g_return_if_fail (AGENT_IS_SERVER (server));
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));

  for (i = 0; i < g_array_get_element_size (server->priv->interface_ids); i++)
    {
      guint id;

      id = g_array_index (server->priv->interface_ids, guint, i);
      g_dbus_connection_unregister_object (connection, id);
    }
  g_array_set_size (server->priv->interface_ids, 0);
}

/**
 * on_observer_authorize_authenticated:
 */
static gboolean
on_observer_authorize_authenticated (GDBusAuthObserver *observer,
				     GIOStream         *stream,
				     GCredentials      *credentials,
				     AgentServer       *server)
{
  g_return_val_if_fail (G_IS_DBUS_AUTH_OBSERVER (observer), FALSE);
  g_return_val_if_fail (G_IS_IO_STREAM (stream), FALSE);
  g_return_val_if_fail (AGENT_IS_SERVER (server), FALSE);

  _debug ("auth TODO ...");
  return TRUE;
}

/**
 * on_peer_server_connection_closed:
 */
static void
on_peer_server_connection_closed (GDBusConnection *connection,
				  gboolean         remote_peer_vanished,
				  GError          *error,
				  AgentServer     *server)
{
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));
  g_return_if_fail (AGENT_IS_SERVER (server));

  _debug ("connection closed.");

  agent_server_unregister_objects (server, connection);

  server->priv->connections = g_slist_remove (server->priv->connections, connection);
  g_object_unref (connection);
}

/**
 * on_peer_server_new_connection:
 *
 * response peer client connection.
 */
static void
on_peer_server_new_connection (GDBusServer     *peer_server,
			       GDBusConnection *connection,
			       AgentServer     *server)
{
  g_return_if_fail (G_IS_DBUS_SERVER (peer_server));
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));
  g_return_if_fail (AGENT_IS_SERVER (server));

  server->priv->connections =
   g_slist_prepend (server->priv->connections, g_object_ref (connection));

  g_signal_connect (connection, "closed",
		    G_CALLBACK (on_peer_server_connection_closed), server);

  /* register all service objects. */
  agent_server_register_objects (server, connection);
}

/**
 * agent_server_start_peer:
 *
 * start network server. the @address is a dbus address such as
 * unix:path=/tmp/abc or tcp:host=0.0.0.0,port=3322.
 */
gboolean
agent_server_start_peer (AgentServer  *server,
			 const gchar  *address)
{
  guint flags;
  gchar *guid;
  GDBusAuthObserver *observer;
  GError *error = NULL;

  g_return_val_if_fail (AGENT_IS_SERVER (server), FALSE);
  g_return_val_if_fail (address, FALSE);

  /* authentication observer. */
  observer = g_dbus_auth_observer_new ();
  g_signal_connect (observer, "authorize-authenticated-peer",
		    G_CALLBACK (on_observer_authorize_authenticated), server);

  /* we will accept client connection in independent thread. */
  flags = G_DBUS_SERVER_FLAGS_RUN_IN_THREAD;
  if (server->priv->auth_anonymous)
    flags |= G_DBUS_SERVER_FLAGS_AUTHENTICATION_ALLOW_ANONYMOUS;

  /* server unique identity. */
  guid = g_dbus_generate_guid ();

  /* create peer server and start it. */
  server->priv->peer_server = g_dbus_server_new_sync (address, flags, guid,
						      observer, NULL, &error);
  g_free (guid);
  g_object_unref (observer);

  if (error)
    {
      g_warning (_("Unable to start server at address `%s': %s"),
		 address, error->message);
      g_error_free (error);
      return FALSE;
    }
  g_signal_connect (server->priv->peer_server, "new-connection",
		    G_CALLBACK (on_peer_server_new_connection), server);

  g_dbus_server_start (server->priv->peer_server);

  _message ("Server is listening at: %s",
	    g_dbus_server_get_client_address (server->priv->peer_server));
  return TRUE;
}

/**
 * on_server_bus_acquired:
 *
 * invoke when connected to the specified bus type.
 */
static void
on_server_bus_acquired (GDBusConnection *connection,
			const gchar     *name,
			gpointer         user_data)
{
  AgentServer *server = user_data;

  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));
  g_return_if_fail (AGENT_IS_SERVER (server));

  server->priv->bus_connection = g_object_ref (connection);
  agent_server_register_objects (server, connection);
}

/**
 * on_server_bus_name_acquired:
 *
 * handler to invoke when name is acquired from dbus-daemon.
 */
static void
on_server_bus_name_acquired (GDBusConnection *connection,
			     const gchar     *name,
			     gpointer         user_data)
{
  AgentServer *server = user_data;

  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));
  g_return_if_fail (AGENT_IS_SERVER (server));

  _message ("Server is listening at: %s", name);
}

/**
 * on_server_bus_name_lost:
 *
 * handler to invoke when name is lost.
 */
static void
on_server_bus_name_lost (GDBusConnection *connection,
			 const gchar     *bus_name,
			 gpointer         user_data)
{
  AgentServer *server = user_data;

  g_return_if_fail (AGENT_IS_SERVER (server));

  /* if connection is NULL, means the bus daemon not found.
   * otherwise means we cannot acquire the name beacuse bus
   * daemon security policy. */
  if (!connection)
    g_warning (_("Unable to connect bus daemon."));
  else
    {
      g_warning (_("Unable to acquire name `%s' from bus daemon."), bus_name);
      agent_server_unregister_objects (server, connection);
    }
  g_idle_add ((GSourceFunc) agent_app_quit, agent_app_get ());
}

/**
 * agent_server_start_dbus:
 *
 * start as dbus service.
 */
static gboolean
agent_server_start_dbus (AgentServer *server,
			 GBusType     bus_type)
{
  g_return_val_if_fail (AGENT_IS_SERVER (server), FALSE);

  server->priv->own_name_id = g_bus_own_name (bus_type,
					      "net.ttyutils.Agent",
					      G_BUS_NAME_OWNER_FLAGS_NONE,
					      on_server_bus_acquired,
					      on_server_bus_name_acquired,
					      on_server_bus_name_lost,
					      g_object_ref (server), g_object_unref);
  return TRUE;
}

/**
 * agent_server_start:
 *
 * depend on the address type, the server can be a independent
 * network server, or a service register to dbus-daemon.
 */
gboolean
agent_server_start (AgentServer     *server,
		    AgentServerType  server_type,
		    const gchar     *server_addr)
{
  g_return_val_if_fail (AGENT_IS_SERVER (server), FALSE);

  switch (server_type)
    {
    case SERVER_TYPE_PEER:
      if (!server_addr)
	server_addr = "tcp:host=0.0.0.0,port=5433";
      return agent_server_start_peer (server, server_addr);

    case SERVER_TYPE_BUS_SESSION:
      return agent_server_start_dbus (server, G_BUS_TYPE_SESSION);
    case SERVER_TYPE_BUS_SYSTEM:
      return agent_server_start_dbus (server, G_BUS_TYPE_SYSTEM);

    default:
      g_critical ("Unsupported server address type %d", server_type);
      return FALSE;
    }
}

/**
 * agent_server_stop:
 */
void
agent_server_stop (AgentServer *server)
{
  g_return_if_fail (AGENT_IS_SERVER (server));

  if (server->priv->own_name_id > 0)
    {
      g_bus_unown_name (server->priv->own_name_id);
      server->priv->own_name_id = 0;
    }
  if (server->priv->peer_server &&
      g_dbus_server_is_active (server->priv->peer_server))
    {
      g_dbus_server_stop (server->priv->peer_server);
    }
}
