/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <fcntl.h>
#include <stdio.h>
#include <locale.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <stdlib.h>
#include <execinfo.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <ttyutils/ttyutils.h>

#include "agent-server.h"
#include "agent-app.h"

static gchar		*_opt_config = NULL;
static gchar		*_opt_logfile = NULL;
static gboolean		 _opt_foreground = FALSE;
static gchar		*_opt_pidfile = NULL;
static gboolean		 _opt_version = FALSE;

static GOptionEntry	option_entries[] =
{
    { "config", 'c', 0, G_OPTION_ARG_FILENAME, &_opt_config,
      N_("Use specified configure file"), "FILE" },

    { "logfile", 'l', 0, G_OPTION_ARG_FILENAME, &_opt_logfile,
      N_("Logged message to file"), "FILE" },

    { "foreground", 'f', 0, G_OPTION_ARG_NONE, &_opt_foreground,
      N_("Running as foreground process(default is background)"), NULL },

    { "pidfile", 'p', 0, G_OPTION_ARG_FILENAME, &_opt_pidfile,
      N_("Specify a file to record process ID"), "FILE" },

    { "version", 'v', 0, G_OPTION_ARG_NONE, &_opt_version,
      N_("Output version information"), NULL },

    { NULL, }
};

/**
 * daemon:
 * a substituted daemon function implement
 */
#ifndef HAVE_DAEMON
static int
daemon (int nochdir, int noclose)
{
  switch (fork ())
    {
    case -1:	/* error */
      g_critical ("fork() error: %d(%s)", errno,
		  g_strerror (errno));
      return -1;
    case 0:	/* child */
      break;
    default:	/* parent */
      exit (0);
    }
  setsid ();

  /* change work directory */
  if (!nochdir)
    chdir ("/");

  /* redirect standard input, standard output, and
   * standard error to /dev/null */
  if (!noclose)
    {
      int fd = open ("/dev/null", O_RDWR);
      if (fd < 0)
	{
	  g_warning ("Failed to open '/dev/null': %d(%s)", errno,
		     g_strerror (errno));
	  return -1;
	}
      dup2 (fd, STDIN_FILENO);
      dup2 (fd, STDOUT_FILENO);
      dup2 (fd, STDERR_FILENO);

      if (fd > 2)
	close (fd);
    }
  return 0;
}
#endif

/**
 * write_pid_to_file:
 *
 * write pid to file @filename.
 *
 * this funciton will failed if @filename exists and contains
 * a running pid.
 */
static gboolean
write_pid_to_file (const gchar *filename)
{
  FILE *fp;
  mode_t mask;

  if (g_file_test (filename, G_FILE_TEST_EXISTS))
    {
      gchar *contents;
      gsize length;
      GPid pid;
      GError *error = NULL;

      g_file_get_contents (filename, &contents, &length, &error);
      if (error)
	{
	  g_warning (_("Unable to read `%s': %s"), filename,
		     error->message);
	  g_error_free (error);
	  goto next;
	}
      g_strstrip (contents);
      pid = (GPid) g_ascii_strtod (contents, NULL);
      if (pid == 0)
	goto next;

      /* send a signal to target process, if succeed, means the process
       * is actived, we cannot overwrite it's pid file.
       * if failed beacuse has not permission, we cannot overwirite
       * pidfile too */
      if (kill (pid, SIGUSR1) == 0 || errno == EPERM)
	{
	  g_message (_("Ttyutils Agent(PID %d) is running."), pid);
	  return FALSE;
	}
    }

next:
  /* modify umask to that just accessable for file owner. */
  mask = umask (S_IRWXG | S_IRWXO);
  fp = fopen (filename, "w+");
  umask (mask);

  if (!fp)
    {
      g_warning (_("Failed to open `%s': %s"), filename,
		 g_strerror (errno));
      return FALSE;
    }
  fprintf (fp, "%d", getpid ());
  fclose (fp);
  return TRUE;
}

/**
 * unlink_pidfile:
 */
static void
unlink_pidfile (void)
{
  if (_opt_pidfile)
    g_unlink (_opt_pidfile);
}

/**
 * quit_application:
 */
static gboolean
quit_application (gpointer data)
{
  agent_app_quit (agent_app_get ());
  return FALSE;
}

/**
 * signal_terminate:
 * handler of SIGTERM, SIGINT, SIGHUP
 */
static void
signal_terminate (gint signo)
{
  g_idle_add (quit_application, NULL);
}

/**
 * main:
 */
gint
main (gint argc, gchar *argv[])
{
  GOptionContext *option;
  struct sigaction action;
  AgentApp *app = NULL;
  GError *error = NULL;

  g_set_prgname (argv[0]);
  g_set_application_name ("Ttyutils Agent");

  setlocale (LC_ALL, "");
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  if (!t_check_version ())
    return EXIT_FAILURE;

  /* initialize gthread and gobject system */
  g_thread_init (NULL);
  g_type_init ();

  /* parse command line options */
  option = g_option_context_new (NULL);
  g_option_context_set_summary (option, _("Ttyutils agent server"));
  g_option_context_set_description (option, WEBSITE);
  g_option_context_add_main_entries (option, option_entries, GETTEXT_PACKAGE);
  g_option_context_parse (option, &argc, &argv, &error);
  if (error)
    {
      g_warning (_("Parse command line error: %s"), error->message);
      g_error_free (error);
      g_option_context_free (option);
      return EXIT_FAILURE;
    }
  g_option_context_free (option);

  if (_opt_version)
    {
      t_print_version ();
      return EXIT_SUCCESS;
    }
  /* catch system signals */
  memset (&action, 0, sizeof (action));
  action.sa_handler = signal_terminate;
  action.sa_flags = 0;
  sigemptyset (&action.sa_mask);

  sigaction (SIGTERM, &action, 0);
  sigaction (SIGINT, &action, 0);
  sigaction (SIGHUP, &action, 0);

  action.sa_handler = SIG_IGN;
  sigaction (SIGUSR1, &action, 0);

  /* log message to file if --logfile option is specified */
  if (_opt_logfile)
    {
      if (!t_set_logfile (_opt_logfile))
	return EXIT_FAILURE;

      g_log_set_default_handler (t_glib_log_handler, t_get_logfile ());
      g_set_print_handler (t_glib_print_handler);
      g_set_printerr_handler (t_glib_print_handler);
    }
  /* become a daemon */
  if (!_opt_foreground)
    {
      /* if work directory changed, the argument of
       * --logfile and --config will be search in `/' instead of
       * current direcotry. */
      if (daemon (1, 0) != 0)
	{
	  g_warning ("daemon() error: %s", g_strerror (errno));
	  return EXIT_FAILURE;
	}
      /* if running as daemon, and log file is not specified.
       * we need redirect log messages to syslog */
      if (!_opt_logfile)
	{
	  openlog (g_get_prgname (), LOG_PID | LOG_CONS, LOG_USER);

	  g_log_set_default_handler (t_glib_log_handler, NULL);
	  g_set_print_handler (t_glib_print_handler);
	  g_set_printerr_handler (t_glib_print_handler);
	}
    }
  /* write pid to file if --pidfile option is specified */
  if (_opt_pidfile)
    {
      if (!write_pid_to_file (_opt_pidfile))
	{
	  g_warning (_("Failed to write pid to file `%s'. Quit."),
		     _opt_pidfile);
	  return EXIT_FAILURE;
	}
      g_atexit (unlink_pidfile);
    }
  /* create app object. */
  app = agent_app_new ();

  /* load configure from file
   *
   * if `--logfile' is not appear, will try to open below configure file:
   *
   *  - $SYSCONFDIR/ttyutils/ttyagent.conf
   *  - $HOME/.config/ttyutils/ttyagent.conf
   *
   * Note all of above files is optional(missing is not error,
   * just use built-in default value). and if more than one of
   * above files exists, all exists files will be load(by overwrite
   * previous configure).
   *
   * if `--config' option is appear, it may be a filename(just load
   * it) or `-'(no configure file at all).
   */
  if (!_opt_config)
    {
      gchar *files[3];
      gint i;

      files[0] = g_build_filename (SYSCONFDIR, "ttyutils", "ttyagent.conf", NULL);
      files[1] = g_build_filename (g_get_user_config_dir (),
				   "ttyutils", "ttyagent.conf", NULL);
      files[2] = NULL;

      for (i = 0; files[i]; i++)
	{
	  /* if file exists, but load failed, means configure file has
	   * syntax or similar error occured. */
	  if (g_file_test (files[i], G_FILE_TEST_EXISTS))
	    {
	      if (!agent_app_load_config_file (app, files[i]))
		{
		  g_object_unref (app);
		  return EXIT_FAILURE;
		}
	      break;
	    }
	}
      g_free (files[0]);
      g_free (files[1]);
    }
  else if (g_strcmp0 (_opt_config, "-") != 0)
    {
      /* open user specified configure file */
      if (!agent_app_load_config_file (app, _opt_config))
	{
	  g_object_unref (app);
	  return EXIT_FAILURE;
	}
    }
  if (t_expire_check ())
    g_idle_add ((GSourceFunc) raise, GUINT_TO_POINTER (SIGTERM));

  agent_app_start (app);
  g_object_unref (app);

  return EXIT_SUCCESS;
}
