/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <gdk/gdkkeysyms.h>

#include "glook-app.h"
#include "glook-uri.h"
#include "glook-completion.h"
#include "glook-addrbar.h"

/* bookmark icon name.
 * will be show in the begin of address bar */
#define ADDRBAR_ICON_BOOKMARK		"ttyutils-bookmark"
#define ADDRBAR_ICON_BOOKMARK_SILVER	"ttyutils-bookmark-silver"

enum
{
  PROP_0,
  PROP_PROGRESS,
  PROP_BOOKMARK,
};

struct _GLookAddrbarPriv
{
  GLookBookmark *bookmark;
  gboolean       is_bookmark;
  gboolean       is_progress;
  guint          update_bookmark_timeout_id;
};
#define _ADDRBAR_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_ADDRBAR, GLookAddrbarPriv))

G_DEFINE_TYPE (GLookAddrbar, glook_addrbar, GTK_TYPE_ENTRY);

static gboolean	glook_addrbar_key_press_event		(GtkWidget            *entry,
							 GdkEventKey          *event);
static gboolean	glook_addrbar_button_press_event	(GtkWidget            *entry,
							 GdkEventButton       *event);
static void	glook_addrbar_drag_data_received	(GtkWidget            *widget,
							 GdkDragContext       *drag_context,
							 gint                  x,
							 gint                  y,
							 GtkSelectionData     *data,
							 guint                 info,
							 guint                 time);
static void	on_bookmark_added			(GLookBookmark        *bookmark,
							 const gchar          *address,
							 GLookAddrbar         *addrbar);
static void	on_bookmark_removed			(GLookBookmark        *bookmark,
							 const gchar          *address,
							 GLookAddrbar         *addrbar);
static void	on_bookmark_changed			(GLookBookmark        *bookmark,
							 const gchar          *address,
							 GLookAddrbar         *addrbar);
static void	on_addrbar_text_changed			(GtkEntry             *entry,
							 GLookAddrbar         *addrbar);
static void	on_addrbar_icon_pressed			(GtkEntry             *entry,
							 GtkEntryIconPosition  icon_pos,
							 GdkEventButton       *event,
							 GLookAddrbar         *addrbar);

static void	on_addrbar_activated			(GtkEntry             *entry,
							 GLookAddrbar         *addrbar);
static void	glook_addrbar_load_bookmarks		(GLookAddrbar         *addrbar);
#if 0
static void	glook_addrbar_load_recent		(GLookAddrbar         *addrbar);
#endif

/**
 * glook_addrbar_init:
 */
static void
glook_addrbar_init (GLookAddrbar *addrbar)
{
  GLookCompletion *completion;
  GtkBorder border = {4, 4, 3, 3};

  addrbar->priv = _ADDRBAR_GET_PRIVATE (addrbar);

  g_object_set (addrbar, "inner-border", &border, NULL);

  addrbar->priv->is_bookmark = FALSE;
  addrbar->priv->is_progress = FALSE;
  addrbar->priv->update_bookmark_timeout_id = 0;

  /* initialize internal icons. */
  gtk_entry_set_icon_from_icon_name (GTK_ENTRY (addrbar),
				     GTK_ENTRY_ICON_PRIMARY,
				     ADDRBAR_ICON_BOOKMARK_SILVER);
  gtk_entry_set_icon_sensitive (GTK_ENTRY (addrbar),
				GTK_ENTRY_ICON_PRIMARY, FALSE);
  gtk_entry_set_icon_tooltip_text (GTK_ENTRY (addrbar),
				   GTK_ENTRY_ICON_PRIMARY,
				   _("Bookmark this address"));
  gtk_entry_set_icon_from_stock (GTK_ENTRY (addrbar),
				 GTK_ENTRY_ICON_SECONDARY,
				 GTK_STOCK_GO_FORWARD);
  gtk_entry_set_icon_sensitive (GTK_ENTRY (addrbar),
				GTK_ENTRY_ICON_SECONDARY, FALSE);
  gtk_entry_set_icon_tooltip_text (GTK_ENTRY (addrbar),
				   GTK_ENTRY_ICON_SECONDARY, _("Go to"));
  /* set the completion object */
  completion = glook_completion_new ();
  gtk_entry_set_completion (GTK_ENTRY (addrbar),
			    GTK_ENTRY_COMPLETION (completion));
  g_object_unref (completion);

  /* hold a bookmark object reference. */
  addrbar->priv->bookmark = glook_app_get_bookmark (glook_app_get ());
  g_assert (GLOOK_IS_BOOKMARK (addrbar->priv->bookmark));
  g_object_ref (addrbar->priv->bookmark);

  /* load bookmarks into completion. */
  glook_addrbar_load_bookmarks (addrbar);

  /* connect bookmark object signals. */
  g_signal_connect (addrbar->priv->bookmark, "added",
		    G_CALLBACK (on_bookmark_added), addrbar);
  g_signal_connect (addrbar->priv->bookmark, "removed",
		    G_CALLBACK (on_bookmark_removed), addrbar);
  g_signal_connect (addrbar->priv->bookmark, "changed",
		    G_CALLBACK (on_bookmark_changed), addrbar);

  /* connect to entry's signals */
  g_signal_connect (addrbar, "changed",
		    G_CALLBACK (on_addrbar_text_changed), addrbar);
  g_signal_connect (addrbar, "icon-press",
		    G_CALLBACK (on_addrbar_icon_pressed), addrbar);
  g_signal_connect (addrbar, "activate",
		    G_CALLBACK (on_addrbar_activated), addrbar);
}

/**
 * glook_addrbar_dispose:
 */
static void
glook_addrbar_dispose (GObject *object)
{
  GLookAddrbar *addrbar = GLOOK_ADDRBAR (object);

  if (addrbar->priv->update_bookmark_timeout_id > 0)
    {
      g_source_remove (addrbar->priv->update_bookmark_timeout_id);
      addrbar->priv->update_bookmark_timeout_id = 0;
    }
  if (addrbar->priv->bookmark)
    {
      /* when addrbar dispose, we need disconnect handler from
       * bookmark signals. */
      g_signal_handlers_disconnect_by_func (addrbar->priv->bookmark,
					    on_bookmark_added, addrbar);
      g_signal_handlers_disconnect_by_func (addrbar->priv->bookmark,
					    on_bookmark_removed, addrbar);
      g_signal_handlers_disconnect_by_func (addrbar->priv->bookmark,
					    on_bookmark_changed, addrbar);
      g_object_unref (addrbar->priv->bookmark);
      addrbar->priv->bookmark = NULL;
    }
  G_OBJECT_CLASS (glook_addrbar_parent_class)->dispose (object);
}

/**
 * glook_addrbar_finalize:
 */
static void
glook_addrbar_finalize (GObject *object)
{
  //GLookAddrbar *addrbar = GLOOK_ADDRBAR (object);

  G_OBJECT_CLASS (glook_addrbar_parent_class)->finalize (object);
}

/**
 * glook_addrbar_get_property:
 */
static void
glook_addrbar_get_property (GObject    *object,
			    guint       property_id,
			    GValue     *value,
			    GParamSpec *pspec)
{
  GLookAddrbar *addrbar = GLOOK_ADDRBAR (object);

  switch (property_id)
    {
    case PROP_PROGRESS:
      g_value_set_boolean (value, glook_addrbar_get_progress (addrbar));
      break;
    case PROP_BOOKMARK:
      g_value_set_boolean (value, glook_addrbar_get_bookmark (addrbar));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_addrbar_set_property:
 */
static void
glook_addrbar_set_property (GObject      *object,
			    guint         property_id,
			    const GValue *value,
			    GParamSpec   *pspec)
{
  GLookAddrbar *addrbar = GLOOK_ADDRBAR (object);

  switch (property_id)
    {
    case PROP_PROGRESS:
      glook_addrbar_set_progress (addrbar, g_value_get_boolean (value));
      break;
    case PROP_BOOKMARK:
      glook_addrbar_set_bookmark (addrbar, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_addrbar_class_init:
 */
static void
glook_addrbar_class_init (GLookAddrbarClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GtkIconTheme *theme;

  g_type_class_add_private (klass, sizeof (GLookAddrbarPriv));

  gobject_class->get_property = glook_addrbar_get_property;
  gobject_class->set_property = glook_addrbar_set_property;

  gobject_class->dispose  = glook_addrbar_dispose;
  gobject_class->finalize = glook_addrbar_finalize;

  widget_class->key_press_event = glook_addrbar_key_press_event;
  widget_class->button_press_event = glook_addrbar_button_press_event;
  widget_class->drag_data_received = glook_addrbar_drag_data_received;

  g_object_class_install_property (gobject_class, PROP_PROGRESS,
				   g_param_spec_boolean ("progress", "Progress", "Jobs in operation", FALSE,
							  G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							  G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_BOOKMARK,
				   g_param_spec_boolean ("bookmark", "Bookmark", "Is bookmark", FALSE,
							  G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							  G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_signal_new ("go", GLOOK_TYPE_ADDRBAR, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1, G_TYPE_STRING);
  g_signal_new ("stop", GLOOK_TYPE_ADDRBAR, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
  g_signal_new ("erase", GLOOK_TYPE_ADDRBAR, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
  g_signal_new ("bookmark", GLOOK_TYPE_ADDRBAR, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1, G_TYPE_INT);

  /* check whether system installed required icons */
  theme = gtk_icon_theme_get_default ();

  if (!gtk_icon_theme_has_icon (theme, ADDRBAR_ICON_BOOKMARK))
    {
      g_critical (_("Icon `%s' not found. please check your installation."),
		  ADDRBAR_ICON_BOOKMARK);
    }
  if (!gtk_icon_theme_has_icon (theme, ADDRBAR_ICON_BOOKMARK_SILVER))
    {
      g_critical (_("Icon `%s' not found. please check your installation."),
		  ADDRBAR_ICON_BOOKMARK_SILVER);
    }
}

/**
 * glook_addrbar_insert_completion_item:
 */
static void
glook_addrbar_insert_completion_item (GLookAddrbar *addrbar,
				      const gchar  *address)
{
  GtkEntryCompletion *completion;
  GBookmarkFile *file;
  gchar *title, *icon_name;
  guint visited;

  g_return_if_fail (address);

  file = glook_bookmark_get_file (addrbar->priv->bookmark);

  title = g_bookmark_file_get_title (file, address, NULL);
  g_bookmark_file_get_icon (file, address, &icon_name, NULL, NULL);
  visited = g_bookmark_file_get_visited (file, address, NULL);

  completion = gtk_entry_get_completion (GTK_ENTRY (addrbar));
  glook_completion_insert (GLOOK_COMPLETION (completion),
			   icon_name, address, title, visited, TRUE);
  g_free (title);
  g_free (icon_name);
}

/**
 * glook_addrbar_load_bookmarks:
 */
static void
glook_addrbar_load_bookmarks (GLookAddrbar *addrbar)
{
  GtkEntryCompletion *completion;
  GBookmarkFile *file;
  gchar **uris;
  gsize length, i;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  /* clear the completion first. */
  completion = gtk_entry_get_completion (GTK_ENTRY (addrbar));
  glook_completion_clear (GLOOK_COMPLETION (completion));

  /* load all bookmarks into completion. */
  file = glook_bookmark_get_file (addrbar->priv->bookmark);
  uris = g_bookmark_file_get_uris (file, &length);
  for (i = 0; i < length; i++)
    glook_addrbar_insert_completion_item (addrbar, uris[i]);

  g_strfreev (uris);
}

/**
 * glook_addrbar_load_recent:
 */
#if 0
TODO: add recent support???
static void
glook_addrbar_load_recent (GLookAddrbar *addrbar)
{
  GtkEntryCompletion *completion;
  GtkRecentManager *manager;
  GList *items, *node;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  completion = gtk_entry_get_completion (GTK_ENTRY (addrbar));

  manager = gtk_recent_manager_get_default ();
  gtk_recent_manager_get_items (manager);
  for (node = items; node; node = node->next)
    {
      GtkRecentInfo *info = node->data;
      const gchar *mime = gtk_recent_info_get_mime_type (info);
      if (g_strcmp0 (mime, "application/x-ttylook") == 0)
	{
	  const gchar *uri, *title, *icon_name;
	  guint visited;

	  uri = gtk_recent_info_get_uri (info);
	  title = gtk_recent_info_get_display_name (info);
	  visited = gtk_recent_info_get_visited (info);

	  glook_completion_insert (GLOOK_COMPLETION (completion),
				   "application-x-ttylook", uri, title, visited, TRUE);
	}
    }
}
#endif

/**
 * on_bookmark_added:
 *
 * when bookmark added, we need insert the new bookmark
 * into completion.
 */
static void
on_bookmark_added (GLookBookmark        *bookmark,
		   const gchar          *address,
		   GLookAddrbar         *addrbar)
{
  const gchar *text;

  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  /* check whether current address is bookmark. */
  text = gtk_entry_get_text (GTK_ENTRY (addrbar));
  if (text && text[0])
    glook_addrbar_set_bookmark (addrbar, glook_bookmark_has (bookmark, text));
  else
    glook_addrbar_set_bookmark (addrbar, FALSE);

  /* insert address into completion. */
  glook_addrbar_insert_completion_item (addrbar, address);
}

/**
 * on_bookmark_removed:
 *
 * when bookmark been removed, we need remove the bookmark
 * from completion.
 */
static void
on_bookmark_removed (GLookBookmark *bookmark,
		     const gchar   *address,
		     GLookAddrbar  *addrbar)
{
  const gchar *text;
  GtkEntryCompletion *completion;

  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  /* check whether current address is bookmark. */
  text = gtk_entry_get_text (GTK_ENTRY (addrbar));
  if (text && text[0])
    glook_addrbar_set_bookmark (addrbar, glook_bookmark_has (bookmark, text));
  else
    glook_addrbar_set_bookmark (addrbar, FALSE);

  /* remove the address from completion. */
  completion = gtk_entry_get_completion (GTK_ENTRY (addrbar));
  glook_completion_remove (GLOOK_COMPLETION (completion), address);
}

/**
 * on_bookmark_changed:
 *
 * when bookmark changed, we need update bookmark state.
 */
static void
on_bookmark_changed (GLookBookmark *bookmark,
		     const gchar   *address,
		     GLookAddrbar  *addrbar)
{
  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  /* replace exists item. */
  glook_addrbar_insert_completion_item (addrbar, address);
}

/**
 * on_update_bookmark_timeout:
 */
static gboolean
on_update_bookmark_timeout (gpointer data)
{
  GLookAddrbar *addrbar = data;
  const gchar *text;

  g_return_val_if_fail (GLOOK_IS_ADDRBAR (addrbar), FALSE);

  text = gtk_entry_get_text (GTK_ENTRY (addrbar));
  if (!text || !text[0])
    {
      glook_addrbar_set_bookmark (addrbar, FALSE);
      return FALSE;
    }
  if (glook_bookmark_has (addrbar->priv->bookmark, text))
    glook_addrbar_set_bookmark (addrbar, TRUE);
  else
    glook_addrbar_set_bookmark (addrbar, FALSE);

  return FALSE;
}

/**
 * on_update_bookmark_timeout_destroy:
 */
static void
on_update_bookmark_timeout_destroy (gpointer data)
{
  GLookAddrbar *addrbar = data;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  addrbar->priv->update_bookmark_timeout_id = 0;
}

/**
 * on_addrbar_text_changed:
 *
 * called when text in address bar changed.
 */
void
on_addrbar_text_changed	(GtkEntry     *entry,
			 GLookAddrbar *addrbar)
{
  const gchar *text;
  gchar tooltip[256];

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  text = gtk_entry_get_text (entry);

  /* update go icon tooltip. */
  if (addrbar->priv->is_progress)
    g_snprintf (tooltip, sizeof (tooltip), _("Stop open `%s'"), text);
  else
    g_snprintf (tooltip, sizeof (tooltip), _("Go to `%s'"), text);

  gtk_entry_set_icon_tooltip_text (GTK_ENTRY (addrbar),
				   GTK_ENTRY_ICON_PRIMARY,
				   _("Bookmark this address"));
  gtk_entry_set_icon_tooltip_text (entry, GTK_ENTRY_ICON_SECONDARY, tooltip);

  /* update icon sensitive. */
  gtk_entry_set_icon_sensitive (entry, GTK_ENTRY_ICON_PRIMARY, TRUE);
  gtk_entry_set_icon_sensitive (entry, GTK_ENTRY_ICON_SECONDARY, TRUE);

  if (!text || !text[0])
    {
      gtk_entry_set_icon_sensitive (entry, GTK_ENTRY_ICON_PRIMARY, FALSE);

      /* we cannot deactivate go icon when jobs is in the operation. */
      if (!addrbar->priv->is_progress)
	gtk_entry_set_icon_sensitive (entry, GTK_ENTRY_ICON_SECONDARY, FALSE);

      glook_addrbar_set_bookmark (addrbar, FALSE);
    }
  else
    {
      /* add timeout event to update bookmark status. */
      if (addrbar->priv->update_bookmark_timeout_id > 0)
	g_source_remove (addrbar->priv->update_bookmark_timeout_id);

      addrbar->priv->update_bookmark_timeout_id =
       g_timeout_add_full (G_PRIORITY_DEFAULT, 200,
			   on_update_bookmark_timeout, addrbar,
			   on_update_bookmark_timeout_destroy);
    }
}

/**
 * on_addrbar_bookmark_icon_pressed:
 *
 * handler for user press the bookmark icon.
 */
static void
on_addrbar_bookmark_icon_pressed (GLookAddrbar *addrbar,
				  gint          button)
{
  const gchar *text;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  text = gtk_entry_get_text (GTK_ENTRY (addrbar));
  if (text && text[0])
    {
      /* if current address is bookmark, remove it. */
      if (addrbar->priv->is_bookmark)
	{
	  glook_bookmark_remove (addrbar->priv->bookmark, text);
	  return;
	}
      g_signal_emit_by_name (addrbar, "bookmark", button);
    }
}

/**
 * glook_addrbar_emit_go_with_check:
 *
 * emit `go' signal only if address is valid uri.
 */
static void
glook_addrbar_emit_go_with_check (GLookAddrbar *addrbar)
{
  gchar *text;
  GLookUri *uri;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  if (gtk_entry_get_text_length (GTK_ENTRY (addrbar)) == 0)
    return;

  text = g_strdup (gtk_entry_get_text (GTK_ENTRY (addrbar)));
  g_strstrip (text);
  if (text[0])
    {
      uri = glook_uri_new_from_string (text);
      if (uri)
	{
	  glook_uri_free (uri);
	  g_signal_emit_by_name (addrbar, "go", text);
	}
      else
	{
	  gtk_entry_set_icon_from_stock (GTK_ENTRY (addrbar),
					 GTK_ENTRY_ICON_PRIMARY,
					 GTK_STOCK_DIALOG_WARNING);
	  gtk_entry_set_icon_tooltip_text (GTK_ENTRY (addrbar),
					   GTK_ENTRY_ICON_PRIMARY,
					   _("Address incorrect"));
	  gtk_editable_select_region (GTK_EDITABLE (addrbar), 0, -1);
	}
    }
  g_free (text);
}

/**
 * on_address_go_icon_pressed:
 */
static void
on_address_go_icon_pressed (GLookAddrbar *addrbar)
{
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  if (addrbar->priv->is_progress)
    {
      g_signal_emit_by_name (addrbar, "stop");
      return;
    }
  glook_addrbar_emit_go_with_check (addrbar);
}

/**
 * on_addrbar_icon_pressed:
 *
 * called when user click the primary or secondary icon.
 */
void
on_addrbar_icon_pressed (GtkEntry             *entry,
			 GtkEntryIconPosition  icon_pos,
			 GdkEventButton       *event,
			 GLookAddrbar         *addrbar)
{
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  if (event->type != GDK_BUTTON_PRESS)
    return;

  switch (icon_pos)
    {
    case GTK_ENTRY_ICON_PRIMARY:	/* bookmark */
      on_addrbar_bookmark_icon_pressed (addrbar, event->button);
      break;
    case GTK_ENTRY_ICON_SECONDARY:	/* go/stop */
      if (event->button == 1)
	on_address_go_icon_pressed (addrbar);
      break;
    default:
      g_assert_not_reached ();
    }
}

/**
 * on_addrbar_activated:
 *
 * called when addrbar been activated(press ENTER).
 */
static void
on_addrbar_activated (GtkEntry     *entry,
		      GLookAddrbar *addrbar)
{
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  /* open the address. */
  glook_addrbar_emit_go_with_check (addrbar);
}

/**
 * glook_addrbar_key_press_event:
 */
static gboolean
glook_addrbar_key_press_event (GtkWidget   *widget,
			       GdkEventKey *event)
{
  GtkWidgetClass *widget_class;

  g_return_val_if_fail (GLOOK_IS_ADDRBAR (widget), FALSE);

  switch (event->keyval)
    {
    case GDK_Escape:
      gtk_editable_select_region (GTK_EDITABLE (widget), 0, -1);
      g_signal_emit_by_name (widget, "erase");
      return TRUE;
    case GDK_Down:
      /* emit a changed signal to popup the completion
       * window. see gtkentry.c for detail. */
      g_signal_emit_by_name (widget, "changed", 0);
      return TRUE;
    }
  widget_class = GTK_WIDGET_CLASS (glook_addrbar_parent_class);
  if (!widget_class->key_press_event)
    return FALSE;

  return widget_class->key_press_event (widget, event);
}

/**
 * glook_addrbar_button_press_event:
 */
static gboolean
glook_addrbar_button_press_event (GtkWidget      *widget,
				  GdkEventButton *event)
{
  GtkWidgetClass *widget_class;

  g_return_val_if_fail (GLOOK_IS_ADDRBAR (widget), FALSE);

  if (event->button == 1 && event->type == GDK_2BUTTON_PRESS &&
      gtk_entry_get_text_length (GTK_ENTRY (widget)) == 0)
    {
      g_signal_emit_by_name (widget, "changed", 0);
      return TRUE;
    }
  widget_class = GTK_WIDGET_CLASS (glook_addrbar_parent_class);
  if (!widget_class->button_press_event)
    return FALSE;

  return widget_class->button_press_event (widget, event);
}

/**
 * glook_addrbar_drag_data_received:
 */
static void
glook_addrbar_drag_data_received (GtkWidget            *widget,
				  GdkDragContext       *drag_context,
				  gint                  x,
				  gint                  y,
				  GtkSelectionData     *data,
				  guint                 info,
				  guint                 time)
{
  GLookAddrbar *addrbar = GLOOK_ADDRBAR (widget);
  GtkWidgetClass *widget_class;
  gchar *text;
  GLookUri *uri;

  text = (gchar*) gtk_selection_data_get_text (data);
  if (!text)
    {
      gtk_drag_finish (drag_context, FALSE, FALSE, time);
      return;
    }
  /* check the text is acceptable. */
  uri = glook_uri_new_from_string (text);
  if (!uri)
    {
      gtk_drag_finish (drag_context, FALSE, FALSE, time);
      g_free (text);
      return;
    }
  glook_uri_free (uri);
  g_free (text);

  /* clear the entry first. */
  gtk_entry_set_text (GTK_ENTRY (widget), "");

  /* call parent method. */
  widget_class = GTK_WIDGET_CLASS (glook_addrbar_parent_class);
  if (widget_class->drag_data_received)
    widget_class->drag_data_received (widget, drag_context, x, y, data, info, time);

  /* open the address immediate. */
  glook_addrbar_emit_go_with_check (addrbar);
}

/**
 * glook_addrbar_new:
 */
GtkWidget*
glook_addrbar_new (void)
{
  return g_object_new (GLOOK_TYPE_ADDRBAR, NULL);
}

/**
 * glook_addrbar_set_progress:
 *
 * set address bar has a job in the operation.
 * address bar will change to go icon to stop icon.
 * and click the stop icon will cancel current operation.
 */
void
glook_addrbar_set_progress (GLookAddrbar *addrbar,
			    gboolean      progress)
{
  gchar tooltip[256];
  const gchar *icon_name;

  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  if (progress)
    {
      g_snprintf (tooltip, sizeof (tooltip), _("Stop open `%s'"),
		  gtk_entry_get_text (GTK_ENTRY (addrbar)));
      icon_name = GTK_STOCK_STOP;
    }
  else
    {
      g_snprintf (tooltip, sizeof (tooltip), _("Go to `%s'"),
		  gtk_entry_get_text (GTK_ENTRY (addrbar)));
      icon_name = GTK_STOCK_GO_FORWARD;
    }
  /* change the go icon to/from stop icon */
  gtk_entry_set_icon_from_stock (GTK_ENTRY (addrbar),
				 GTK_ENTRY_ICON_SECONDARY, icon_name);
  /* change the secondary icon's tooltip. */
  gtk_entry_set_icon_tooltip_text (GTK_ENTRY (addrbar),
				   GTK_ENTRY_ICON_SECONDARY, tooltip);

  addrbar->priv->is_progress = progress;
  g_object_notify (G_OBJECT (addrbar), "progress");
}

/**
 * glook_addrbar_get_progress:
 *
 * get there whether has a job in the operation.
 */
gboolean
glook_addrbar_get_progress (GLookAddrbar *addrbar)
{
  g_return_val_if_fail (GLOOK_IS_ADDRBAR (addrbar), FALSE);
  return addrbar->priv->is_progress;
}

/**
 * glook_addrbar_set_bookmark:
 */
void
glook_addrbar_set_bookmark (GLookAddrbar *addrbar,
			    gboolean      bookmark)
{
  g_return_if_fail (GLOOK_IS_ADDRBAR (addrbar));

  gtk_entry_set_icon_from_icon_name (GTK_ENTRY (addrbar),
				     GTK_ENTRY_ICON_PRIMARY,
				     bookmark ?
				     ADDRBAR_ICON_BOOKMARK :
				     ADDRBAR_ICON_BOOKMARK_SILVER);
  addrbar->priv->is_bookmark = bookmark;
  g_object_notify (G_OBJECT (addrbar), "bookmark");
}

/**
 * glook_addrbar_get_bookmark:
 */
gboolean
glook_addrbar_get_bookmark (GLookAddrbar *addrbar)
{
  g_return_val_if_fail (GLOOK_IS_ADDRBAR (addrbar), FALSE);
  return addrbar->priv->is_bookmark;
}
