/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-misc.h"
#include "glook-tab.h"
#include "glook-agent-tab.h"

/* agent ui glade file name. */
#ifdef DEBUG
# define AGENT_GLADE_FILE	SRCDIR G_DIR_SEPARATOR_S "glade" G_DIR_SEPARATOR_S "agent.xml"
#else
# define AGENT_GLADE_FILE	GLADEDIR G_DIR_SEPARATOR_S "agent.xml"
#endif

enum
{
  PROP_0,
  PROP_VISIBLE_COLUMNS,
};

enum
{
  HOST_COL_HOSTNAME,
  HOST_COL_COUNT,
  HOST_COL_NODE,
  HOST_N_COLUMNS,
};

enum
{
  SERVER_COL_SERVER,
  SERVER_COL_PID,
  SERVER_COL_USERNAME,
  SERVER_COL_TTYNAME,
  SERVER_COL_PTSNAME,
  SERVER_COL_COMMAND,
  SERVER_COL_TIME,
  SERVER_COL_LOOKPORT,
  SERVER_COL_TOOLTIP,
  SERVER_N_COLUMNS,
};

typedef struct _GLookServer	GLookServer;
struct _GLookServer
{
  gint     refcount;
  guint64  time;
  gchar   *hostname;
  gchar   *username;
  guint32  pid;
  gchar   *ttyname;
  gchar   *ptsname;
  gchar   *command;
  guint16  lookport;
};

struct _GLookAgentTabPriv
{
  GtkBuilder     *builder;
  GtkActionGroup *server_actions;
  GtkUIManager   *server_ui;
  GNode          *servers;
  GDBusProxy     *proxy;
  GCancellable   *proxy_cancellable;
  gulong          proxy_cancel_id;
  guint           proxy_watch_id;
};
#define _AGENT_TAB_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_AGENT_TAB, GLookAgentTabPriv))

G_DEFINE_TYPE (GLookAgentTab, glook_agent_tab, GLOOK_TYPE_TAB);

static void	glook_agent_tab_progress		(GLookTab         *tab);
static void	glook_agent_tab_open			(GLookTab         *tab);
static void	glook_agent_tab_stop			(GLookTab         *tab);
static gboolean	glook_agent_tab_is_active		(GLookTab         *tab);
static gboolean	glook_agent_tab_zoom			(GLookTab         *tab,
							 gint              factor);
static void	glook_agent_tab_reload			(GLookTab         *tab);

static void	load_servers_into_host_view		(GLookAgentTab    *tab);
static void	unload_servers_from_host_view		(GLookAgentTab    *tab);
static void	load_servers_into_server_view		(GLookAgentTab    *tab,
							 GNode            *node);
static void	unload_servers_from_server_view		(GLookAgentTab    *tab);

static void	on_host_tree_selection_changed		(GtkTreeSelection *selection,
							 GLookAgentTab    *tab);
static void	on_server_tree_selection_changed	(GtkTreeSelection *selection,
							 GLookAgentTab    *tab);

static void	glook_agent_tab_open_server		(GLookAgentTab    *tab,
							 gboolean          newtab);
static void	on_server_open_new_tab_activated	(GtkAction        *action,
							 GLookAgentTab    *tab);
static void	on_server_open_new_window_activated	(GtkAction        *action,
							 GLookAgentTab    *tab);
static void	on_server_visible_column_toggled	(GtkToggleAction  *action,
							 GLookAgentTab    *tab);
static void	on_server_copy_link_address_activated	(GtkAction        *action,
							 GLookAgentTab    *tab);

static const gchar	*_introspection_xml =
"<node>"
"  <interface name='net.ttyutils.Agent.RecorderManager'>"
"    <method name='GetList'>"
"      <arg name='recorder-servers' type='a(tssusssq)' direction='out'/>"
"    </method>"
"    <signal name='Added'>"
"      <arg name='time'     type='t' direction='out'/>"
"      <arg name='hostname' type='s' direction='out'/>"
"      <arg name='username' type='s' direction='out'/>"
"      <arg name='pid'      type='u' direction='out'/>"
"      <arg name='ttyname'  type='s' direction='out'/>"
"      <arg name='ptsname'  type='s' direction='out'/>"
"      <arg name='command'  type='s' direction='out'/>"
"      <arg name='lookport' type='q' direction='out'/>"
"    </signal>"
"    <signal name='Removed'>"
"      <arg name='hostname' type='s' direction='out'/>"
"      <arg name='pid'      type='u' direction='out'/>"
"      <arg name='ttyname'  type='s' direction='out'/>"
"      <arg name='ptsname'  type='s' direction='out'/>"
"    </signal>"
"    <property name='Count' type='u' access='read'/>"
"  </interface>"
"</node>";

/**
 * glook_proxy_get_interface_info:
 */
static GDBusInterfaceInfo*
glook_proxy_get_interface_info (void)
{
  static GDBusInterfaceInfo *info = NULL;
  static gsize has_info = 0;

  if (g_once_init_enter (&has_info))
    {
      GDBusNodeInfo *data;
      data = g_dbus_node_info_new_for_xml (_introspection_xml, NULL);
      info = data->interfaces[0];
      g_once_init_leave (&has_info, 1);
    }
  return info;
}

static GtkActionEntry _server_actions[] =
{
    { "OpenNewTab", NULL, N_("Open in new _tab"), NULL,
      N_("Open selected server in new tab"),
      G_CALLBACK (on_server_open_new_tab_activated),
    },
    { "OpenNewWindow", NULL, N_("Open in new _window"), NULL,
      N_("Open selected server in new window"),
      G_CALLBACK (on_server_open_new_window_activated),
    },
    { "CopyLinkAddress", NULL, N_("Copy _link address"), NULL,
      N_("Copy link address to clipboard"),
      G_CALLBACK (on_server_copy_link_address_activated),
    },

    { "ColumnMenu", NULL, N_("_Visible columns"), },
};

static GtkToggleActionEntry _server_column_actions[] =
{
    { "PidColumn", NULL, "_PID", NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
    { "UserColumn", NULL, N_("_User"), NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
    { "TtyColumn", NULL, "TT_Y", NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
    { "PtsColumn", NULL, "PT_S", NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
    { "CommandColumn", NULL, N_("_Command"), NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
    { "TimeColumn", NULL, N_("_Time"), NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
    { "LookPortColumn", NULL, N_("L_ook Port"), NULL, NULL,
      G_CALLBACK (on_server_visible_column_toggled), TRUE,
    },
};

static const gchar *_server_menu_ui =
"<ui>"
"<popup name='ServerMenu'>"
"  <menuitem action='OpenNewTab'/>"
"  <menuitem action='OpenNewWindow'/>"
"  <separator/>"
"  <menu action='ColumnMenu'>"
"    <menuitem action='PidColumn'/>"
"    <menuitem action='UserColumn'/>"
"    <menuitem action='TtyColumn'/>"
"    <menuitem action='PtsColumn'/>"
"    <menuitem action='CommandColumn'/>"
"    <menuitem action='TimeColumn'/>"
"    <menuitem action='LookPortColumn'/>"
"  </menu>"
"  <separator/>"
"  <menuitem action='CopyLinkAddress'/>"
"</popup>"
"</ui>";


/**
 * glook_server_new:
 */
static GLookServer*
glook_server_new (void)
{
  GLookServer *server = g_slice_new0 (GLookServer);
  server->refcount = 1;
  return server;
}

/**
 * glook_server_ref:
 */
GLookServer*
glook_server_ref (GLookServer *server)
{
  g_atomic_int_inc (&server->refcount);
  return server;
}

/**
 * glook_server_unref:
 */
static void
glook_server_unref (GLookServer *server)
{
  if (server && g_atomic_int_dec_and_test (&server->refcount))
    {
      g_free (server->hostname);
      g_free (server->username);
      g_free (server->ttyname);
      g_free (server->ptsname);
      g_free (server->command);

      g_slice_free (GLookServer, server);
    }
}

/**
 * glook_server_is_equal:
 *
 * 2 server is equal if them has same hostname, ttyname,
 * ptsname and pid.
 */
static gboolean
glook_server_is_equal (GLookServer *server1,
		       GLookServer *server2)
{
  g_return_val_if_fail (server1, FALSE);
  g_return_val_if_fail (server2, FALSE);

  return (g_strcmp0 (server1->hostname, server2->hostname) == 0 &&
	  g_strcmp0 (server1->ttyname, server2->ttyname) == 0 &&
	  g_strcmp0 (server1->ptsname, server2->ptsname) == 0 &&
	  server1->pid == server2->pid);
}

/**
 * bind_server_visible_column_settings:
 */
static void
bind_server_visible_column_settings (GLookAgentTab *tab,
				     GtkBuilder    *builder)
{
  GObject *object;
  GtkAction *action;
  GSettings *settings;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (GTK_IS_BUILDER (builder));

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_AGENTTAB);

  /* pid column and action. */
  object = gtk_builder_get_object (builder, "PidColumn");
  g_settings_bind (settings, "pid-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "PidColumn");
  g_settings_bind (settings, "pid-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
  /* user column and action. */
  object = gtk_builder_get_object (builder, "UserColumn");
  g_settings_bind (settings, "user-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "UserColumn");
  g_settings_bind (settings, "user-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
  /* tty column and action. */
  object = gtk_builder_get_object (builder, "TtyColumn");
  g_settings_bind (settings, "tty-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "TtyColumn");
  g_settings_bind (settings, "tty-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
  /* pts column and action. */
  object = gtk_builder_get_object (builder, "PtsColumn");
  g_settings_bind (settings, "pts-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "PtsColumn");
  g_settings_bind (settings, "pts-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
  /* command column and action. */
  object = gtk_builder_get_object (builder, "CommandColumn");
  g_settings_bind (settings, "command-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "CommandColumn");
  g_settings_bind (settings, "command-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
  /* time column and action. */
  object = gtk_builder_get_object (builder, "TimeColumn");
  g_settings_bind (settings, "time-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "TimeColumn");
  g_settings_bind (settings, "time-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
  /* look port column and action. */
  object = gtk_builder_get_object (builder, "LookPortColumn");
  g_settings_bind (settings, "lookport-column-visible",
		   object, "visible", G_SETTINGS_BIND_DEFAULT);
  action = gtk_action_group_get_action (tab->priv->server_actions, "LookPortColumn");
  g_settings_bind (settings, "lookport-column-visible",
		   G_OBJECT (action), "active", G_SETTINGS_BIND_DEFAULT);
}

/**
 * glook_agent_tab_init:
 */
static void
glook_agent_tab_init (GLookAgentTab *tab)
{
  GtkTreeSelection *selection;
  GObject *toplevel;
  GObject *treeview;
  GError *error = NULL;

  g_object_set (tab, "label", _("Sessions"),
		"image", "application-x-ttylook", NULL);

  tab->priv = _AGENT_TAB_GET_PRIVATE (tab);

  tab->priv->servers = NULL;
  tab->priv->proxy = NULL;
  tab->priv->proxy_cancellable = g_cancellable_new ();
  tab->priv->proxy_cancel_id = 0;
  tab->priv->proxy_watch_id = 0;

  tab->priv->server_actions = gtk_action_group_new ("ServerActionGroup");;
  gtk_action_group_set_translation_domain (tab->priv->server_actions,
					   GETTEXT_PACKAGE);
  gtk_action_group_add_actions (tab->priv->server_actions, _server_actions,
				G_N_ELEMENTS (_server_actions), tab);
  gtk_action_group_add_toggle_actions (tab->priv->server_actions,
				       _server_column_actions,
				       G_N_ELEMENTS (_server_column_actions), tab);
  tab->priv->server_ui = gtk_ui_manager_new ();
  gtk_ui_manager_insert_action_group (tab->priv->server_ui, tab->priv->server_actions, 0);
  gtk_ui_manager_add_ui_from_string (tab->priv->server_ui, _server_menu_ui, -1, &error);
  if (error)
    {
      g_error ("%s", error->message);
      return;
    }
  /* the glade file is the ui definitions file.
   * application cannot continue if this file is missing. */
  if (!g_file_test (AGENT_GLADE_FILE, G_FILE_TEST_EXISTS))
    {
      g_critical (_("File `%s' not found. please check your "
		    "installation and try again."), AGENT_GLADE_FILE);
      return;
    }
  tab->priv->builder = gtk_builder_new ();
  gtk_builder_add_from_file (tab->priv->builder, AGENT_GLADE_FILE, &error);
  if (error)
    {
      g_critical (_("Failed to parse file '%s': %s"),
		  AGENT_GLADE_FILE, error->message);
      g_critical (_("Please check your installation and try again."));
      g_error_free (error);
      g_object_unref (tab->priv->builder);
      tab->priv->builder = NULL;
      return;
    }
  /* connect all widgets signals */
  gtk_builder_connect_signals (tab->priv->builder, tab);

  /* add toplevel box */
  toplevel = gtk_builder_get_object (tab->priv->builder, "ToplevelBox");
  gtk_container_add (GTK_CONTAINER (tab), GTK_WIDGET (toplevel));

  /* bind server treeview visible column settings. */
  bind_server_visible_column_settings (tab, tab->priv->builder);

  /* connect host treeview select changed signal */
  treeview = gtk_builder_get_object (tab->priv->builder, "HostTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  g_signal_connect (selection, "changed",
		    G_CALLBACK (on_host_tree_selection_changed), tab);

  /* connect server treeview select changed signal */
  treeview = gtk_builder_get_object (tab->priv->builder, "ServerTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  g_signal_connect (selection, "changed",
		    G_CALLBACK (on_server_tree_selection_changed), tab);
  /* DND support. */
  gtk_tree_view_enable_model_drag_source (GTK_TREE_VIEW (treeview),
					  GDK_BUTTON1_MASK, NULL, 0,
					  GDK_ACTION_COPY);
  gtk_drag_source_add_text_targets (GTK_WIDGET (treeview));
}

/**
 * glook_agent_tab_get_property:
 */
static void
glook_agent_tab_get_property (GObject    *object,
			      guint       property_id,
			      GValue     *value,
			      GParamSpec *pspec)
{
  //GLookAgentTab *tab = GLOOK_AGENT_TAB (object);

  switch (property_id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_agent_tab_set_property:
 */
static void
glook_agent_tab_set_property (GObject      *object,
			      guint         property_id,
			      const GValue *value,
			      GParamSpec   *pspec)
{
  //GLookAgentTab *tab = GLOOK_AGENT_TAB (object);

  switch (property_id)
    {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_agent_tab_dispose:
 */
static void
glook_agent_tab_dispose (GObject *object)
{
  GLookAgentTab *tab = GLOOK_AGENT_TAB (object);

  if (tab->priv->proxy_watch_id > 0)
    {
      g_bus_unwatch_name (tab->priv->proxy_watch_id);
      tab->priv->proxy_watch_id = 0;
    }
  if (tab->priv->proxy)
    {
      g_object_unref (tab->priv->proxy);
      tab->priv->proxy = NULL;
    }
  if (tab->priv->builder)
    {
      g_object_unref (tab->priv->builder);
      tab->priv->builder = NULL;
    }
  G_OBJECT_CLASS (glook_agent_tab_parent_class)->dispose (object);
}

/**
 * glook_agent_tab_free_node:
 */
static gboolean
glook_agent_tab_free_node (GNode    *node,
			   gpointer  data)
{
  GLookServer *server = node->data;
  glook_server_unref (server);
  return FALSE;
}

/**
 * glook_agent_tab_finalize:
 */
static void
glook_agent_tab_finalize (GObject *object)
{
  GLookAgentTab *tab = GLOOK_AGENT_TAB (object);

  if (tab->priv->proxy_cancellable)
    {
      g_object_unref (tab->priv->proxy_cancellable);
      tab->priv->proxy_cancellable = NULL;
    }
  if (tab->priv->server_actions)
    {
      g_object_unref (tab->priv->server_actions);
      tab->priv->server_actions = NULL;
    }
  if (tab->priv->server_ui)
    {
      g_object_unref (tab->priv->server_ui);
      tab->priv->server_ui = NULL;
    }
  if (tab->priv->servers)
    {
      g_node_traverse (tab->priv->servers, G_IN_ORDER, G_TRAVERSE_LEAVES, -1,
		       glook_agent_tab_free_node, NULL);
      g_node_destroy (tab->priv->servers);
      tab->priv->servers = NULL;
    }
  G_OBJECT_CLASS (glook_agent_tab_parent_class)->finalize (object);
}

/**
 * glook_agent_tab_class_init:
 */
static void
glook_agent_tab_class_init (GLookAgentTabClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GLookTabClass *parent_class = GLOOK_TAB_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookAgentTabPriv));

  gobject_class->get_property = glook_agent_tab_get_property;
  gobject_class->set_property = glook_agent_tab_set_property;
  gobject_class->dispose  = glook_agent_tab_dispose;
  gobject_class->finalize = glook_agent_tab_finalize;

  parent_class->progress  = glook_agent_tab_progress;
  parent_class->open      = glook_agent_tab_open;
  parent_class->stop      = glook_agent_tab_stop;
  parent_class->is_active = glook_agent_tab_is_active;
  parent_class->zoom      = glook_agent_tab_zoom;
  parent_class->reload    = glook_agent_tab_reload;
}

/**
 * glook_agent_tab_add_server:
 *
 * add a GLookServer struct to tree.
 */
static void
glook_agent_tab_add_server (GLookAgentTab *tab,
			    GLookServer   *server)
{
  GNode *node, *child;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (server);

  node = g_node_first_child (tab->priv->servers);
  while (node)
    {
      GLookServer *server2 = node->data;

      /* insert new server node(level 3) to exists node
       * if there has same host exists. */
      if (g_strcmp0 (server->hostname, server2->hostname) == 0)
	{
	  gboolean exists = FALSE;

	  /* check whether or not the node exists. */
	  child = g_node_first_child (node);
	  while (child)
	    {
	      GLookServer *server3 = child->data;

	      if (glook_server_is_equal (server3, server))
		{
		  /* change time. */
		  server3->time = server->time;
		  exists = TRUE;
		  break;
		}
	      child = g_node_next_sibling (child);
	    }
	  if (!exists)
	    g_node_append_data (node, server);

	  return;
	}
      node = g_node_next_sibling (node);
    }
  /* create a new host node(level 2). and insert the new
   * server node(level 3) below it.
   * here we use server as parent node data too. */
  node = g_node_append_data (tab->priv->servers, server);
  g_node_append_data (node, server);
}

/**
 * glook_agent_tab_remove_server:
 *
 * remove a server from internal tree.
 */
static void
glook_agent_tab_remove_server (GLookAgentTab *tab,
			       GLookServer   *server)
{
  GNode *node, *child;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (server);

  node = g_node_first_child (tab->priv->servers);
  while (node)
    {
      GLookServer *server2 = node->data;

      if (g_strcmp0 (server->hostname, server2->hostname) == 0)
	{
	  child = g_node_first_child (node);
	  while (child)
	    {
	      GLookServer *server3 = child->data;

	      if (glook_server_is_equal (server3, server))
		{
		  g_node_destroy (child);
		  break;
		}
	      child = g_node_next_sibling (child);
	    }
	  /* if node has not children, remove the node itself too. */
	  if (g_node_n_children (node) == 0)
	    {
	      g_node_destroy (node);
	    }
	  return;
	}
      node = g_node_next_sibling (node);
    }
}

/**
 * glook_agent_tab_show_info:
 */
static void
glook_agent_tab_show_info (GLookAgentTab  *tab,
			   GtkMessageType  type,
			   const gchar    *format,
			   ...)
{
  va_list args;
  gchar buff[1024];
  GObject *info_bar;
  GObject *info_label;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  va_start (args, format);
  g_vsnprintf (buff, sizeof (buff), format, args);
  va_end (args);

  info_label = gtk_builder_get_object (tab->priv->builder, "InfoLabel");
  gtk_label_set_label (GTK_LABEL (info_label), buff);

  info_bar = gtk_builder_get_object (tab->priv->builder, "Infobar");
  gtk_info_bar_set_message_type (GTK_INFO_BAR (info_bar), type);
  gtk_widget_show (GTK_WIDGET (info_bar));
}

/**
 * glook_agent_tab_finish:
 *
 * emit `finish' signal.
 */
static void
glook_agent_tab_finish (GLookAgentTab *tab)
{
  GObject *treeview;
  GdkWindow *window;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  /* reset treeview cursor. */
  treeview = gtk_builder_get_object (tab->priv->builder, "HostTreeview");
  window = gtk_widget_get_window (GTK_WIDGET (treeview));
  if (GDK_IS_WINDOW (window))
    gdk_window_set_cursor (window, NULL);

  treeview = gtk_builder_get_object (tab->priv->builder, "ServerTreeview");
  window = gtk_widget_get_window (GTK_WIDGET (treeview));
  if (GDK_IS_WINDOW (window))
    gdk_window_set_cursor (window, NULL);

  g_signal_emit_by_name (tab, "finished");

  gtk_widget_grab_focus (GTK_WIDGET (treeview));
}

/**
 * on_agent_proxy_g_signal:
 *
 * response agent server signal.
 */
static void
on_agent_proxy_g_signal (GDBusProxy  *proxy,
			 const gchar *sender_name,
			 const gchar *signal_name,
			 GVariant    *parameters,
			 gpointer     user_data)
{
  GLookAgentTab *tab = user_data;
  GLookServer *server;
  GObject *paned;
  GObject *info_bar;

  g_return_if_fail (G_IS_DBUS_PROXY (proxy));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (g_strcmp0 (signal_name, "Added") == 0)
    {
      server = glook_server_new ();

      g_variant_get (parameters, "(tssusssq)",
		     &server->time, &server->hostname,
		     &server->username, &server->pid,
		     &server->ttyname, &server->ptsname,
		     &server->command, &server->lookport);

      glook_agent_tab_add_server (tab, server);
    }
  else if (g_strcmp0 (signal_name, "Removed") == 0)
    {
      server = glook_server_new ();

      g_variant_get (parameters, "(suss)",
		     &server->hostname, &server->pid,
		     &server->ttyname, &server->ptsname);

      glook_agent_tab_remove_server (tab, server);
      glook_server_unref (server);
    }
  else
    {
      _warning ("unhandled signal `%s'", signal_name);
      return;
    }
  /* update host treeview. we just clear the treeview first and
   * reinsert all servers, this maybe slow if has many servers. */
  unload_servers_from_server_view (tab);
  unload_servers_from_host_view (tab);

  paned = gtk_builder_get_object (tab->priv->builder, "Paned");
  info_bar = gtk_builder_get_object (tab->priv->builder, "Infobar");

  if (g_node_n_children (tab->priv->servers) > 0)
    {
      gtk_widget_set_sensitive (GTK_WIDGET (paned), TRUE);
      load_servers_into_host_view (tab);
    }
  else
    {
      gtk_widget_set_sensitive (GTK_WIDGET (paned), FALSE);
    }
  gtk_widget_hide (GTK_WIDGET (info_bar));
}

/**
 * on_agent_proxy_g_properties_changed:
 */
static void
on_agent_proxy_g_properties_changed (GDBusProxy         *proxy,
				     GVariant           *changed_properties,
				     const gchar* const *invalidated_properties,
				     gpointer            user_data)
{
  g_return_if_fail (G_IS_DBUS_PROXY (proxy));
  _debug ("proxy property changed");
}

/**
 * on_proxy_list_all_ready:
 */
static void
on_proxy_list_all_ready (GObject      *object,
			 GAsyncResult *res,
			 gpointer      user_data)
{
  GDBusProxy *proxy = G_DBUS_PROXY (object);
  GLookAgentTab *tab = user_data;
  GVariant *result, *child;
  GVariantIter iter;
  GObject *paned;
  GError *error = NULL;

  g_return_if_fail (G_IS_DBUS_PROXY (proxy));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  result = g_dbus_proxy_call_finish (proxy, res, &error);
  if (error)
    {
      g_warning (_("Agent response: %s"), error->message);

      /* show server response message. */
      g_dbus_error_strip_remote_error (error);
      glook_agent_tab_show_info (tab, GTK_MESSAGE_WARNING,
				 "%s", error->message);
      g_error_free (error);
      glook_agent_tab_finish (tab);
      return;
    }
  /* the result was a tuple contains a array, so we go
   * into array first. */
  g_variant_iter_init (&iter, result);
  child = g_variant_iter_next_value (&iter);
  g_variant_iter_init (&iter, child);

  while (1)
    {
      GLookServer *server = glook_server_new ();

      /* parse result and insert into internal tree. */
      if (g_variant_iter_next (&iter, "(tssusssq)", &server->time,
			       &server->hostname, &server->username,
			       &server->pid, &server->ttyname,
			       &server->ptsname, &server->command,
			       &server->lookport))
	{
	  glook_agent_tab_add_server (tab, server);
	}
      else
	{
	  glook_server_unref (server);
	  break;
	}
    }
  /* load all received servers into treeview. */
  if (g_node_n_children (tab->priv->servers) > 0)
    load_servers_into_host_view (tab);

  paned = gtk_builder_get_object (tab->priv->builder, "Paned");
  gtk_widget_set_sensitive (GTK_WIDGET (paned), TRUE);

  glook_agent_tab_finish (tab);
}

/**
 * on__agent_proxy_ready:
 */
static void
on_create_proxy_ready (GObject      *source,
		       GAsyncResult *res,
		       gpointer      user_data)
{
  GLookAgentTab *tab = user_data;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  tab->priv->proxy = g_dbus_proxy_new_finish (res, &error);
  if (error)
    {
      g_warning (_("Failed to connect agent server: %s"), error->message);
      g_error_free (error);
    }
  g_signal_connect (tab->priv->proxy, "g-signal",
		    G_CALLBACK (on_agent_proxy_g_signal), tab);
  g_signal_connect (tab->priv->proxy, "g-properties-changed",
		    G_CALLBACK (on_agent_proxy_g_properties_changed), tab);

  g_dbus_proxy_call (tab->priv->proxy, "GetList",
		     NULL, G_DBUS_CALL_FLAGS_NO_AUTO_START,
		     -1, tab->priv->proxy_cancellable,
		     on_proxy_list_all_ready, tab);
}

/**
 * on_agent_proxy_connection_closed:
 */
static void
on_agent_proxy_connection_closed (GDBusConnection *connection,
				  gboolean         remote_peer_vanished,
				  GError          *error,
				  GLookAgentTab   *tab)
{
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (error)
    g_message (_("Disconnect from agent: %s"), error->message);

  if (remote_peer_vanished)
    {
      glook_agent_tab_show_info (tab, GTK_MESSAGE_INFO,
				 _("Agent server has showdown."));
    }
  if (tab->priv->proxy)
    {
      g_object_unref (tab->priv->proxy);
      tab->priv->proxy = NULL;
    }
}

/**
 * glook_agent_tab_create_proxy:
 */
static void
glook_agent_tab_create_proxy (GLookAgentTab   *tab,
			      GDBusConnection *connection,
			      const gchar     *bus_name)
{
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  /* listen to connection close signal. */
  g_signal_connect (connection, "closed",
		    G_CALLBACK (on_agent_proxy_connection_closed), tab);

  g_dbus_proxy_new (connection,
		    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START |
		    G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
		    glook_proxy_get_interface_info (),
		    bus_name, "/net/ttyutils/Agent",
		    "net.ttyutils.Agent.RecorderManager",
		    tab->priv->proxy_cancellable,
		    on_create_proxy_ready, tab);
}

/**
 * on_connection_new_for_address_ready:
 */
static void
on_connection_new_for_address_ready (GObject      *source,
				     GAsyncResult *res,
				     gpointer      user_data)
{
  GLookAgentTab *tab = user_data;
  GDBusConnection *connection;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  connection = g_dbus_connection_new_for_address_finish (res, &error);
  if (error)
    {
      g_warning (_("Failed to create connection: %s"), error->message);

      if (!g_cancellable_is_cancelled (tab->priv->proxy_cancellable))
	glook_agent_tab_show_info (tab, GTK_MESSAGE_WARNING, error->message);

      g_error_free (error);

      glook_agent_tab_finish (tab);
      g_object_unref (tab);
      return;
    }
  glook_agent_tab_create_proxy (tab, connection, NULL);
  g_object_unref (connection);

  g_object_unref (tab);
}

/**
 * on_watch_bus_name_appeared:
 *
 * called when bus name appeared on the bus.
 */
static void
on_watch_bus_name_appeared (GDBusConnection *connection,
			    const gchar     *bus_name,
			    const gchar     *name_owner,
			    gpointer         user_data)
{
  GLookAgentTab *tab = user_data;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  glook_agent_tab_create_proxy (tab, connection, bus_name);
}

/**
 * on_watch_bus_name_vanished:
 */
static void
on_watch_bus_name_vanished (GDBusConnection *connection,
			    const gchar     *bus_name,
			    gpointer         user_data)
{
  GLookAgentTab *tab = user_data;
  const gchar *msg;
  GtkMessageType msgtype;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (tab->priv->proxy)
    {
      msg = _("Agent server has shut down.");
      msgtype = GTK_MESSAGE_INFO;

      g_object_unref (tab->priv->proxy);
      tab->priv->proxy = NULL;
    }
  else
    {
      msg = _("Can not find agent server on the bus.");
      msgtype = GTK_MESSAGE_WARNING;
    }
  glook_agent_tab_show_info (tab, msgtype, msg);
  glook_agent_tab_finish (tab);
}

/**
 * on_observer_authorize_authenticated_peer:
 */
static gboolean
on_observer_authorize_authenticated_peer (GDBusAuthObserver *observer,
					  GIOStream         *iostream,
					  GCredentials      *credentials,
					  gpointer           user_data)       
{
  g_debug ("auth: TODO");
  return TRUE;
}

/**
 * glook_agent_tab_proxy:
 *
 * start proxy remote agent server Recorder interface.
 * agent server maybe is a dbus service or a peer server.
 */
void
glook_agent_tab_proxy (GLookAgentTab *tab)
{
  GLookUri *uri;
  gchar *bus_address;
  GBusType bus_type;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  uri = glook_tab_get_uri (GLOOK_TAB (tab));
  g_assert (glook_uri_get_scheme_type (uri) == GLOOK_URI_SCHEME_AGENT);

  /* get dbus address from URI. */
  bus_address = glook_uri_get_dbus_address (uri, &bus_type);
  if (bus_address)
    {
      GDBusAuthObserver *observer;

      /* authentication observer. */
      observer = g_dbus_auth_observer_new ();
      g_signal_connect (observer, "authorize-authenticated-peer",
			G_CALLBACK (on_observer_authorize_authenticated_peer), tab);

      g_dbus_connection_new_for_address (bus_address,
					 G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT,
					 observer, tab->priv->proxy_cancellable,
					 on_connection_new_for_address_ready,
					 g_object_ref (tab));
      g_free (bus_address);
      g_object_unref (observer);
    }
  else if (bus_type == G_BUS_TYPE_SESSION ||
	   bus_type == G_BUS_TYPE_SYSTEM)
    {
      /* watch agent bus name. */
      tab->priv->proxy_watch_id = g_bus_watch_name (bus_type, "net.ttyutils.Agent",
						    G_BUS_NAME_WATCHER_FLAGS_NONE,
						    on_watch_bus_name_appeared,
						    on_watch_bus_name_vanished,
						    g_object_ref (tab), g_object_unref);
    }
  else
    {
      const gchar *msg = _("Agent address unknown or unrecognizable.");
      g_warning ("%s", msg);

      glook_agent_tab_show_info (tab, GTK_MESSAGE_ERROR, msg);
      glook_agent_tab_finish (tab);
    }
}

/**
 * glook_agent_proxy_idle:
 */
static gboolean
glook_agent_proxy_idle (gpointer data)
{
  GLookAgentTab *tab = data;

  /* remove the cancel event handler. */
  g_cancellable_disconnect (tab->priv->proxy_cancellable,
			    tab->priv->proxy_cancel_id);
  g_cancellable_reset (tab->priv->proxy_cancellable);
  tab->priv->proxy_cancel_id = 0;

  /* change tab appear to busy state. and notify external can
   * cancel our job. */
  g_signal_emit_by_name (tab, "progress");

  /* remove bus name watch first. */
  if (tab->priv->proxy_watch_id)
    {
      g_bus_unwatch_name (tab->priv->proxy_watch_id);
      tab->priv->proxy_watch_id = 0;
    }
  /* start proxy ttyagent interface. */
  glook_agent_tab_proxy (tab);
  return FALSE;
}

/**
 * glook_agent_tab_proxy_cancelled:
 */
static void
glook_agent_tab_proxy_cancelled (GCancellable  *cancellable,
				 GLookAgentTab *tab)
{
  GObject *info_bar;
  GObject *paned;

  g_return_if_fail (G_IS_CANCELLABLE (cancellable));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  /* clear all treeview. */
  unload_servers_from_server_view (tab);
  unload_servers_from_host_view (tab);

  /* clear exists servers before request new servers. */
  if (tab->priv->servers)
    {
      g_node_traverse (tab->priv->servers,
		       G_IN_ORDER, G_TRAVERSE_LEAVES, -1,
		       glook_agent_tab_free_node, NULL);
      g_node_destroy (tab->priv->servers);
    }
  tab->priv->servers = g_node_new (NULL);

  info_bar = gtk_builder_get_object (tab->priv->builder, "Infobar");
  paned = gtk_builder_get_object (tab->priv->builder, "Paned");

  gtk_widget_hide (GTK_WIDGET (info_bar));
  gtk_widget_set_sensitive (GTK_WIDGET (paned), FALSE);

  g_idle_add_full (G_PRIORITY_HIGH_IDLE, glook_agent_proxy_idle,
		   g_object_ref (tab), g_object_unref);
}

/**
 * glook_agent_tab_progress:
 *
 * set busy cursor effect.
 */
static void
glook_agent_tab_progress (GLookTab *tab)
{
  GLookAgentTab *agent_tab = GLOOK_AGENT_TAB (tab);
  GObject *treeview;
  GdkWindow *window;
  GdkCursor *cursor;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  cursor = gdk_cursor_new (GDK_WATCH);

  /* set treeview cursor. */
  treeview = gtk_builder_get_object (agent_tab->priv->builder,
				     "HostTreeview");
  window = gtk_widget_get_window (GTK_WIDGET (treeview));
  if (GDK_IS_WINDOW (window))
    gdk_window_set_cursor (window, cursor);

  treeview = gtk_builder_get_object (agent_tab->priv->builder,
				     "ServerTreeview");
  window = gtk_widget_get_window (GTK_WIDGET (treeview));
  if (GDK_IS_WINDOW (window))
    gdk_window_set_cursor (window, cursor);

  gdk_cursor_unref (cursor);

  /* call parent class method. */
  GLOOK_TAB_CLASS (glook_agent_tab_parent_class)->progress (tab);
}

/**
 * glook_agent_tab_open:
 *
 * connect to agent server.
 */
static void
glook_agent_tab_open (GLookTab *tab)
{
  GLookAgentTab *agent_tab = GLOOK_AGENT_TAB (tab);

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (GTK_IS_BUILDER (agent_tab->priv->builder));

  /* prevent start another connect before previous connect has
   * not started. */
  if (agent_tab->priv->proxy_cancel_id > 0)
    return;

  /* destroy proxy if exists. */
  if (agent_tab->priv->proxy)
    {
      g_object_unref (agent_tab->priv->proxy);
      agent_tab->priv->proxy = NULL;
    }
  g_cancellable_reset (agent_tab->priv->proxy_cancellable);

  /* before open connection, we have to disconnect exists
   * connection first. and new connect action will be start
   * just after disconnect. */
  agent_tab->priv->proxy_cancel_id =
   g_cancellable_connect (agent_tab->priv->proxy_cancellable,
			  G_CALLBACK (glook_agent_tab_proxy_cancelled),
			  g_object_ref (agent_tab), g_object_unref);
  g_cancellable_cancel (agent_tab->priv->proxy_cancellable);
}

/**
 * glook_agent_tab_stop:
 */
static void
glook_agent_tab_stop (GLookTab *tab)
{
  GLookAgentTab *agent_tab = GLOOK_AGENT_TAB (tab);

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (GTK_IS_BUILDER (agent_tab->priv->builder));

  /* remove bus name watch id. */
  if (agent_tab->priv->proxy_watch_id > 0)
    {
      g_bus_unwatch_name (agent_tab->priv->proxy_watch_id);
      agent_tab->priv->proxy_watch_id = 0;
    }
  /* cancel current operation */
  if (!g_cancellable_is_cancelled (agent_tab->priv->proxy_cancellable))
    {
      g_cancellable_cancel (agent_tab->priv->proxy_cancellable);
      glook_agent_tab_finish (agent_tab);
    }
}

/**
 * glook_agent_tab_is_active:
 */
static gboolean
glook_agent_tab_is_active (GLookTab *tab)
{
  GLookAgentTab *agent_tab = GLOOK_AGENT_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_AGENT_TAB (tab), FALSE);
  g_return_val_if_fail (GTK_IS_BUILDER (agent_tab->priv->builder), FALSE);

  /* agent has not persisted connection. when tab still
   * hold some servers, we think it is actived. */
  if (g_node_n_children (agent_tab->priv->servers) > 0)
    return TRUE;

  return G_IS_DBUS_PROXY (agent_tab->priv->proxy);
}

/**
 * glook_agent_tab_reload:
 *
 * reload tab. this is is a virtual method of parent.
 */
static void
glook_agent_tab_reload (GLookTab *tab)
{
  GLookAgentTab *agent_tab = GLOOK_AGENT_TAB (tab);

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (GTK_IS_BUILDER (agent_tab->priv->builder));

  glook_agent_tab_stop (tab);
  glook_agent_tab_open (tab);
}

/**
 * treeview_remove_column:
 *
 * remove column from treeview temporary.
 */
static void
treeview_remove_column (GtkTreeViewColumn *column,
			GtkTreeView       *treeview)
{
  g_object_ref (column);
  gtk_tree_view_remove_column (treeview, column);
}

/**
 * treeview_insert_column:
 *
 * reinsert just removed column.
 */
static void
treeview_insert_column (GtkTreeViewColumn *column,
			GtkTreeView       *treeview)
{
  gtk_tree_view_insert_column (treeview, column, -1);
  g_object_unref (column);
}

/**
 * glook_agent_tab_zoom_treeview:
 */
static void
glook_agent_tab_zoom_treeview (GLookTab    *tab,
			       GtkTreeView *treeview,
			       gint         factor)
{
  GList *columns, *col;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));
  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));

  columns = gtk_tree_view_get_columns (treeview);

  /* remove all columns before zoom. */
  g_list_foreach (columns, (GFunc) treeview_remove_column, treeview);

  for (col = columns; col; col = col->next)
    {
      GList *renderers, *cell;

      /* get all packing cells and set they scale property. */
      renderers = gtk_cell_layout_get_cells (GTK_CELL_LAYOUT (col->data));
      for (cell = renderers; cell; cell = cell->next)
	{
	  if (GTK_IS_CELL_RENDERER_TEXT (cell->data))
	    g_object_set (cell->data, "scale", (gdouble) factor / 100, NULL);
	}
      g_list_free (renderers);
    }
  /* insert all columns. */
  g_list_foreach (columns, (GFunc) treeview_insert_column, treeview);
  g_list_free (columns);

  /* resizes all columns to their optimal width. */
  gtk_tree_view_columns_autosize (treeview);
}

/**
 * glook_agent_tab_zoom:
 *
 * zoom. this is a virtual method of parent.
 */
static gboolean
glook_agent_tab_zoom (GLookTab *tab,
		      gint      factor)
{
  GLookAgentTab *agent_tab = GLOOK_AGENT_TAB (tab);
  GObject *treeview;

  g_return_val_if_fail (GLOOK_IS_AGENT_TAB (tab), FALSE);
  g_return_val_if_fail (GTK_IS_BUILDER (agent_tab->priv->builder), FALSE);

  if (factor > 0)
    {
      /* zoom host treeview */
      treeview = gtk_builder_get_object (agent_tab->priv->builder,
					 "HostTreeview");
      glook_agent_tab_zoom_treeview (tab, GTK_TREE_VIEW (treeview), factor);

      /* zoom server treeview */
      treeview = gtk_builder_get_object (agent_tab->priv->builder,
					 "ServerTreeview");
      glook_agent_tab_zoom_treeview (tab, GTK_TREE_VIEW (treeview), factor);
    }
  return TRUE;
}

/**
 * insert_server_to_host_view:
 */
static void
insert_server_to_host_view (GNode    *node,
			    gpointer  data)
{
  GLookServer *server = node->data;
  GtkListStore  *store = data;

  g_return_if_fail (server);
  g_return_if_fail (GTK_IS_LIST_STORE (store));

  gtk_list_store_insert_with_values (store, NULL, -1,
				     HOST_COL_HOSTNAME, server->hostname,
				     HOST_COL_COUNT, g_node_n_children (node),
				     HOST_COL_NODE, node,
				     -1);
}

/**
 * load_servers_into_host_view:
 *
 * load servers to host treeview.
 */
static void
load_servers_into_host_view (GLookAgentTab *tab)
{
  GObject *hostview;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (g_node_n_children (tab->priv->servers) == 0)
    return;

  hostview = gtk_builder_get_object (tab->priv->builder, "HostTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (hostview));

  /* detach model from treeview before insert data to it. */
  g_object_ref (model);
  gtk_tree_view_set_model (GTK_TREE_VIEW (hostview), NULL);

  /* insert reginfo to model, one by one */
  g_node_children_foreach (tab->priv->servers, G_TRAVERSE_ALL,
			   insert_server_to_host_view, model);

  /* reattach model to treeview and set search column. */
  gtk_tree_view_set_model (GTK_TREE_VIEW (hostview), model);
  g_object_unref (model);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (hostview), HOST_COL_HOSTNAME);

  /* select the first row */
  if (gtk_tree_model_get_iter_first (model, &iter))
    {
      GtkTreeSelection *selection;

      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (hostview));
      gtk_tree_selection_select_iter (selection, &iter);
    }
}

/**
 * unload_servers_from_host_view:
 *
 * clear host treeview.
 */
static void
unload_servers_from_host_view (GLookAgentTab *tab)
{
  GObject *hostview;
  GtkTreeModel *model;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  hostview = gtk_builder_get_object (tab->priv->builder, "HostTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (hostview));

  /* detach model from treeview before insert data to it. */
  g_object_ref (model);
  gtk_tree_view_set_model (GTK_TREE_VIEW (hostview), NULL);

  /* very easy */
  gtk_list_store_clear (GTK_LIST_STORE (model));

  /* reattach model to treeview and set search column. */
  gtk_tree_view_set_model (GTK_TREE_VIEW (hostview), model);
  g_object_unref (model);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (hostview), HOST_COL_HOSTNAME);
}

/**
 * insert_server_to_server_view:
 *
 * a helper function of load_servers_into_server_view().
 */
static void
insert_server_into_server_view (GNode    *node,
				gpointer  data)
{
  GLookServer *server = node->data;
  GtkListStore *model = data;
  time_t time_;
  struct tm *tmp;
  gint today;
  gchar strtime[256];
  const gchar *ttyname, *ptsname;
  gchar *tooltip;

  g_return_if_fail (node);
  g_return_if_fail (GTK_IS_LIST_STORE (model));

  time_ = time (NULL);
  tmp = localtime (&time_);
  today = tmp->tm_yday;

  tmp = localtime ((time_t*) &server->time);

  /* format the time to string. when the time is today sometime,
   * we just show time(no date), otherwise show date+time. */
  if (tmp->tm_yday == today)
    strftime (strtime, sizeof (strtime), "%T", tmp);
  else
    strftime (strtime, sizeof (strtime), "%m-%d %R", tmp);

  /* strip the `/dev/' prefix if possible. */
  ttyname = server->ttyname;
  if (ttyname && (strlen (ttyname) > strlen ("/dev/")) &&
      g_str_has_prefix (ttyname, "/dev/"))
    {
      ttyname += strlen ("/dev/");
    }
  ptsname = server->ptsname;
  if (ptsname && (strlen (ptsname) > strlen ("/dev/")) &&
      g_str_has_prefix (ptsname, "/dev/"))
    {
      ptsname += strlen ("/dev/");
    }
  tooltip = g_strdup_printf ("look:host=%s,port=%d",
			      server->hostname, server->lookport);
  gtk_list_store_insert_with_values (model, NULL, -1,
				     SERVER_COL_SERVER, server,
				     SERVER_COL_PID, server->pid,
				     SERVER_COL_USERNAME, server->username,
				     SERVER_COL_TTYNAME, ttyname,
				     SERVER_COL_PTSNAME, ptsname,
				     SERVER_COL_COMMAND, server->command,
				     SERVER_COL_TIME, strtime,
				     SERVER_COL_LOOKPORT, server->lookport,
				     SERVER_COL_TOOLTIP, tooltip,
				     -1);
  g_free (tooltip);
}

/**
 * load_servers_into_server_view:
 *
 * load server into server treeview.
 */
static void
load_servers_into_server_view (GLookAgentTab *tab,
			       GNode         *node)
{
  GObject *serverview;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (g_node_n_children (tab->priv->servers) == 0)
    return;

  serverview = gtk_builder_get_object (tab->priv->builder, "ServerTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (serverview));

  /* detach store from treeview before insert data to it. */
  g_object_ref (model);
  gtk_tree_view_set_model (GTK_TREE_VIEW (serverview), NULL);

  /* insert reginfo to treeview, one by one */
  g_node_children_foreach (node, G_TRAVERSE_ALL,
			   insert_server_into_server_view, model);

  /* reattach store to treeview and set search column. */
  gtk_tree_view_set_model (GTK_TREE_VIEW (serverview), model);
  g_object_unref (model);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (serverview),
				   SERVER_COL_USERNAME);
  /* select the first row */
  if (gtk_tree_model_get_iter_first (model, &iter))
    {
      GtkTreeSelection *selection;
      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (serverview));
      gtk_tree_selection_select_iter (selection, &iter);
    }
}

/**
 * unload_servers_from_server_view:
 *
 * clear server treeview contents.
 */
static void
unload_servers_from_server_view (GLookAgentTab *tab)
{
  GObject *serverview;
  GtkTreeModel *model;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  serverview = gtk_builder_get_object (tab->priv->builder, "ServerTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (serverview));

  /* detach store from treeview before insert data to it. */
  g_object_ref (model);
  gtk_tree_view_set_model (GTK_TREE_VIEW (serverview), NULL);

  gtk_list_store_clear (GTK_LIST_STORE (model));

  /* reattach store to treeview and set search column. */
  gtk_tree_view_set_model (GTK_TREE_VIEW (serverview), model);
  g_object_unref (model);
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (serverview),
				   SERVER_COL_USERNAME);
}

/**
 * on_host_tree_selection_changed:
 *
 * when host treeview selection changed, we need
 * update the server treeview.
 */
static void
on_host_tree_selection_changed (GtkTreeSelection *selection,
				GLookAgentTab    *tab)
{
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GTK_IS_TREE_SELECTION (selection));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      GNode *node;

      gtk_tree_model_get (model, &iter, HOST_COL_NODE, &node, -1);
      if (node)
	{
	  /* reload server treeview contents. */
	  unload_servers_from_server_view (tab);
	  load_servers_into_server_view (tab, node);
	}
    }
}

/**
 * on_server_tree_selection_changed:
 */
static void
on_server_tree_selection_changed (GtkTreeSelection *selection,
				  GLookAgentTab    *tab)
{
  GtkAction *action;
  gboolean sensitive = TRUE;

  g_return_if_fail (GTK_IS_TREE_SELECTION (selection));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (gtk_tree_selection_count_selected_rows (selection) == 0)
    sensitive = FALSE;

  /* update actions sensitive state. */
  action = gtk_action_group_get_action (tab->priv->server_actions,
					"OpenNewTab");
  gtk_action_set_sensitive (action, sensitive);
  action = gtk_action_group_get_action (tab->priv->server_actions,
					"OpenNewWindow");
  gtk_action_set_sensitive (action, sensitive);
  action = gtk_action_group_get_action (tab->priv->server_actions,
					"CopyLinkAddress");
  gtk_action_set_sensitive (action, sensitive);
}

/**
 * glook_agent_tab_new:
 */
GtkWidget*
glook_agent_tab_new (void)
{
  return g_object_new (GLOOK_TYPE_AGENT_TAB, NULL);
}

/**
 * show_window:
 */
static gboolean
show_window (gpointer data)
{
  GtkWidget *window = data;

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  if (!gtk_widget_get_visible (window))
    gtk_widget_show (window);

  return FALSE;
}

/**
 * get_selected_server_address:
 *
 * the result must be free with g_free().
 */
static gchar*
get_selected_server_address (GLookAgentTab *tab)
{
  GObject *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_val_if_fail (GLOOK_IS_AGENT_TAB (tab), NULL);

  treeview = gtk_builder_get_object (tab->priv->builder, "ServerTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      GLookServer *server = NULL;

      gtk_tree_model_get (model, &iter, SERVER_COL_SERVER, &server, -1);

      /* construct a `look:host=xx,port=xx' uri address. */
      return g_strdup_printf ("look:host=%s,port=%d",
			      server->hostname, server->lookport);
    }
  return NULL;
}

/**
 * glook_agent_tab_open_server:
 *
 * open selected server. if newtab is TRUE, will open
 * in new tab, otherwise will create a new window to open it.
 */
static void
glook_agent_tab_open_server (GLookAgentTab *tab,
			     gboolean       newtab)
{
  GtkWidget *window;
  gchar *address;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  address = get_selected_server_address (tab);
  if (!address)
    return;

  if (newtab)
    {
      window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
      glook_window_open_uri_string (GLOOK_WINDOW (window), address);
    }
  else
    {
      window = glook_app_create_window (glook_app_get (), FALSE);
      glook_window_open_uri_string (GLOOK_WINDOW (window), address);
      g_idle_add (show_window, window);
    }
  g_free (address);
}

/**
 * on_server_open_new_tab_activated:
 *
 * create a new tab and open selected server.
 */
static void
on_server_open_new_tab_activated (GtkAction     *action,
				  GLookAgentTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  glook_agent_tab_open_server (tab, TRUE);
}

/**
 * on_server_open_new_window_activated:
 *
 * create a new window and open selected server.
 */
static void
on_server_open_new_window_activated (GtkAction     *action,
				     GLookAgentTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  glook_agent_tab_open_server (tab, FALSE);
}

/**
 * on_server_treeview_button_pressed:
 *
 * popup context menu when right-click the mouse button.
 */
gboolean
on_server_treeview_button_pressed (GtkTreeView    *treeview,
				   GdkEventButton *event,
				   GLookAgentTab  *tab)
{
  GtkTreeSelection *selection;
  GtkTreePath *path;
  GtkWidget *menu;

  g_return_val_if_fail (GTK_IS_TREE_VIEW (treeview), FALSE);
  g_return_val_if_fail (GLOOK_IS_AGENT_TAB (tab), FALSE);

  if (event->button != 3 || event->type != GDK_BUTTON_PRESS)
    return FALSE;

  selection = gtk_tree_view_get_selection (treeview);

  /* get tree path for row which behind mouse point. */
  if (gtk_tree_view_get_path_at_pos (treeview, (gint) event->x,
				     (gint) event->y, &path,
				     NULL, NULL, NULL))
    {
      if (!gtk_tree_selection_path_is_selected (selection, path))
	{
	  gtk_tree_selection_unselect_all (selection);
	  gtk_tree_selection_select_path (selection, path);
	}
      gtk_tree_path_free (path);

      menu = gtk_ui_manager_get_widget (tab->priv->server_ui,
					"/ServerMenu");
      gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		      event->button, event->time);
      return TRUE;
    }
  return FALSE;
}

/**
 * on_server_visible_column_toggled:
 *
 * show/hide a column.
 */
static void
on_server_visible_column_toggled (GtkToggleAction *action,
				  GLookAgentTab   *tab)
{
  GObject *treeview;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  /* resizes all columns to their optimal width. */
  treeview = gtk_builder_get_object (tab->priv->builder, "ServerTreeview");
  gtk_tree_view_columns_autosize (GTK_TREE_VIEW (treeview));
}

/**
 * on_server_copy_link_address_activated:
 *
 * copy selected server address to clipboard.
 */
static void
on_server_copy_link_address_activated (GtkAction     *action,
				       GLookAgentTab *tab)
{
  GtkClipboard *clipboard;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
  if (clipboard)
    {
      gchar *address = get_selected_server_address (tab);
      if (address)
	{
	  gtk_clipboard_set_text (clipboard, address, -1);
	  g_free (address);
	}
    }
}

/**
 * on_server_treeview_row_activated:
 *
 * row activated, will open the selected server in new tab.
 */
void
on_server_treeview_row_activated (GtkTreeView       *treeview,
				  GtkTreePath       *path,
				  GtkTreeViewColumn *column,
				  GLookAgentTab     *tab)
{
  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (path);
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  /* open selected server in new tab. */
  glook_agent_tab_open_server (tab, TRUE);
}

/**
 * on_server_treeview_drag_data_get:
 */
void
on_server_treeview_drag_data_get (GtkWidget        *widget,
				  GdkDragContext   *drag_context,
				  GtkSelectionData *data,
				  guint             info,
				  guint             time,
				  GLookAgentTab    *tab)
{
  gchar *address;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  address = get_selected_server_address (tab);
  gtk_selection_data_set_text (data, address, -1);
  g_free (address);
}

/**
 * on_server_treeview_drag_begin:
 */
void
on_server_treeview_drag_begin (GtkWidget      *widget,
			       GdkDragContext *drag_context,
			       GLookAgentTab  *tab)
{
  GdkPixmap *pixmap;
  gchar *address;

  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  /* make a uri drop icon to replace default treeview drag icon. */
  address = get_selected_server_address (tab);
  if (address)
    {
      pixmap = glook_create_text_drag_icon (widget, address,
					    strlen (address));
      if (pixmap)
	{
	  gtk_drag_set_icon_pixmap (drag_context,
				    gdk_drawable_get_colormap (pixmap),
				    pixmap, NULL, -2, -2);
	  g_object_unref (pixmap);
	}
      g_free (address);
    }
}

/**
 * on_info_bar_responsed:
 *
 * response info bar.
 */
void
on_info_bar_responsed (GtkWidget *infobar,
		       gint       resp_id,
		       gpointer   user_data)
{
  GLookAgentTab *tab = user_data;

  g_return_if_fail (GTK_IS_INFO_BAR (infobar));
  g_return_if_fail (GLOOK_IS_AGENT_TAB (tab));

  if (resp_id == GTK_RESPONSE_CLOSE)
    {
      GtkWidget *tab_head;
      GtkWidget *close_button;

      /* activate tab head's close button. */
      tab_head = glook_tab_get_tab_head (GLOOK_TAB (tab));
      close_button = g_object_get_data (G_OBJECT (tab_head), "close-button");
      gtk_button_clicked (GTK_BUTTON (close_button));
    }
}
