/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>
#include <gio/gio.h>
#define G_SETTINGS_ENABLE_BACKEND
#include <gio/gsettingsbackend.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-notebook.h"
#include "glook-tab.h"
#include "glook-blank-tab.h"
#include "glook-bookmark.h"
#include "glook-accel-map.h"
#include "glook-window.h"

struct _GLookAppPriv
{
  GtkWindowGroup   *window_group;
  GLookBookmark    *bookmark;
  GLookAccelMap    *accel_map;
  GSettingsBackend *settings_backend;
  GSettings        *settings[N_SETTINGS];
  GdkRectangle      workarea;
};
#define _APP_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_APP, GLookAppPriv))

G_DEFINE_TYPE (GLookApp, glook_app, G_TYPE_OBJECT);

/* global glook app instance. */
static GLookApp		*_app = NULL;

static void	on_disable_mnemonics_changed	(GSettings   *settings,
						 const gchar *key,
						 GLookApp    *app);
static void	on_disable_accelerators_changed	(GSettings   *settings,
						 const gchar *key,
						 GLookApp    *app);
static void	glook_app_query_workarea	(GLookApp    *app);

/**
 * glook_host_init:
 */
static void
glook_app_init (GLookApp *app)
{
  const gchar *p;
  gchar *path;

  app->priv = _APP_GET_PRIVATE (app);

  app->priv->window_group = gtk_window_group_new ();
  app->priv->settings_backend = NULL;
  app->priv->bookmark = glook_bookmark_new ();

  /* load bookmark from file. */
  path = g_build_filename (g_get_user_config_dir (),
			   "ttyutils", "glook-bookmarks.xbel", NULL);
  glook_bookmark_set_path (app->priv->bookmark, path);
  g_free (path);

  /* query desktop workarea size. */
  glook_app_query_workarea (app);

  /* use keyfile settings backend. */
  p = g_getenv ("GSETTINGS_BACKEND");
  if (!p)
    p = g_getenv ("TTYUTILS_SETTINGS_BACKEND");

  if (p && (g_strcmp0 (p, "keyfile") == 0))
    {
      gchar *path = g_build_filename (g_get_user_config_dir (),
				      "ttyutils", "ttyglook.conf", NULL);

      /* we need not create the parent since g_keyfile_settings_new ()
       * will create it. */
      app->priv->settings_backend =
       g_keyfile_settings_backend_new (path, "/apps/ttyutils/glook/", NULL);

      g_free (path);

      /* we have create settings object for every schema beacuse when
       * we cannot get child schema from other settings backend. */
      app->priv->settings[SETTINGS_GENERAL] =
       g_settings_new_with_backend ("net.ttyutils.glook.general",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_AGENTTAB] =
       g_settings_new_with_backend ("net.ttyutils.glook.general.agenttab",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_ACCELERATOR] =
       g_settings_new_with_backend ("net.ttyutils.glook.accelerator",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_TERMINAL] =
       g_settings_new_with_backend ("net.ttyutils.glook.terminal",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_PROFILE1] =
       g_settings_new_with_backend ("net.ttyutils.glook.terminal.profile1",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_PROFILE2] =
       g_settings_new_with_backend ("net.ttyutils.glook.terminal.profile2",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_PROFILE3] =
       g_settings_new_with_backend ("net.ttyutils.glook.terminal.profile3",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_PROFILE4] =
       g_settings_new_with_backend ("net.ttyutils.glook.terminal.profile4",
				    app->priv->settings_backend);
      app->priv->settings[SETTINGS_PROFILE5] =
       g_settings_new_with_backend ("net.ttyutils.glook.terminal.profile5",
				    app->priv->settings_backend);
    }
  else
    {
      app->priv->settings[SETTINGS_GENERAL] =
       g_settings_new ("net.ttyutils.glook.general");
      app->priv->settings[SETTINGS_AGENTTAB] =
       g_settings_new ("net.ttyutils.glook.general.agenttab");
      app->priv->settings[SETTINGS_ACCELERATOR] =
       g_settings_new ("net.ttyutils.glook.accelerator");
      app->priv->settings[SETTINGS_TERMINAL] =
       g_settings_new ("net.ttyutils.glook.terminal");
      app->priv->settings[SETTINGS_PROFILE1] =
       g_settings_new ("net.ttyutils.glook.terminal.profile1");
      app->priv->settings[SETTINGS_PROFILE2] =
       g_settings_new ("net.ttyutils.glook.terminal.profile2");
      app->priv->settings[SETTINGS_PROFILE3] =
       g_settings_new ("net.ttyutils.glook.terminal.profile3");
      app->priv->settings[SETTINGS_PROFILE4] =
       g_settings_new ("net.ttyutils.glook.terminal.profile4");
      app->priv->settings[SETTINGS_PROFILE5] =
       g_settings_new ("net.ttyutils.glook.terminal.profile5");
    }
  /* bind accelerator settings. */
  app->priv->accel_map = glook_accel_map_new (app->priv->settings[SETTINGS_ACCELERATOR]);

  /* disable accelerators and mnemonic is global settings. */
  g_signal_connect (app->priv->settings[SETTINGS_ACCELERATOR],
		    "changed::disable-accelerators",
		    G_CALLBACK (on_disable_accelerators_changed), app);
  g_signal_connect (app->priv->settings[SETTINGS_ACCELERATOR],
		    "changed::disable-mnemonics",
		    G_CALLBACK (on_disable_mnemonics_changed), app);
}

/**
 * glook_app_dispose:
 */
static void
glook_app_dispose (GObject *object)
{
  GLookApp *app = GLOOK_APP (object);

  if (app->priv->window_group)
    {
      g_object_unref (app->priv->window_group);
      app->priv->window_group = NULL;
    }
  if (app->priv->bookmark)
    {
      glook_bookmark_save_to_file (app->priv->bookmark, NULL);
      g_object_unref (app->priv->bookmark);
      app->priv->bookmark = NULL;
    }
  if (app->priv->accel_map)
    {
      g_object_unref (app->priv->accel_map);
      app->priv->accel_map = NULL;
    }
  G_OBJECT_CLASS (glook_app_parent_class)->dispose (object);
}

/**
 * glook_app_finalize:
 */
static void
glook_app_finalize (GObject *object)
{
  GLookApp *app = GLOOK_APP (object);
  gint i;

  for (i = 0; i < N_SETTINGS; i++)
    {
      if (app->priv->settings[i])
	{
	  g_object_unref (app->priv->settings[i]);
	  app->priv->settings[i] = NULL;
	}
    }
  if (app->priv->settings_backend)
    {
      g_object_unref (app->priv->settings_backend);
      app->priv->settings_backend = NULL;
    }
  G_OBJECT_CLASS (glook_app_parent_class)->finalize (object);
}

/**
 * glook_app_class_init:
 */
static void
glook_app_class_init (GLookAppClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookAppPriv));

  gobject_class->dispose  = glook_app_dispose;
  gobject_class->finalize = glook_app_finalize;
}

/**
 * glook_app_query_workarea:
 */
static void
glook_app_query_workarea (GLookApp *app)
{
  GdkWindow *root;

  g_return_if_fail (GLOOK_IS_APP (app));

  root = gdk_get_default_root_window ();
  if (root)
    {
      GdkAtom prop, type;
      gint format, length;
      guchar *data = NULL;
      guint32 *p;

      prop = gdk_atom_intern_static_string ("_NET_WORKAREA");
      if (!gdk_property_get (root, prop, GDK_NONE, 0, 16, FALSE,
			     &type, &format, &length, &data))
	{
	  g_warning ("Unable to get workarea size.");
	  return;
	}
      p = (guint32*) data;

      app->priv->workarea.x = p[0];
      app->priv->workarea.y = p[1];
      app->priv->workarea.width = p[2];
      app->priv->workarea.height = p[3];

      g_free (data);
#if 0
      g_debug ("workarea = %d,%d,%d,%d",
	       app->priv->workarea.x,
	       app->priv->workarea.y,
	       app->priv->workarea.width,
	       app->priv->workarea.height);
#endif
    }
}

/**
 * glook_app_new:
 */
GLookApp*
glook_app_new (void)
{
  g_assert (!_app);
  _app = g_object_new (GLOOK_TYPE_APP, NULL);
  return _app;
}

/**
 * glook_app_get:
 */
GLookApp*
glook_app_get (void)
{
  g_assert (_app);
  return _app;
}

/**
 * glook_app_get_window_group:
 */
GtkWindowGroup*
glook_app_get_window_group (GLookApp *app)
{
  g_return_val_if_fail (GLOOK_IS_APP (app), NULL);
  return app->priv->window_group;
}

/**
 * glook_app_get_bookmark:
 */
GLookBookmark*
glook_app_get_bookmark (GLookApp *app)
{
  g_return_val_if_fail (GLOOK_IS_APP (app), NULL);
  return app->priv->bookmark;
}

/**
 * glook_app_get_settings:
 */
GSettings*
glook_app_get_settings (GLookApp         *app,
			GLookAppSettings  settings)
{
  g_return_val_if_fail (GLOOK_IS_APP (app), NULL);
  return app->priv->settings[settings];
}

/**
 * glook_app_get_profile_settings:
 */
GSettings*
glook_app_get_profile_settings (GLookApp    *app,
				const gchar *profile)
{
  GLookAppSettings settings;

  g_return_val_if_fail (GLOOK_IS_APP (app), NULL);
  g_return_val_if_fail (profile, NULL);

  if (g_strcmp0 (profile, "profile1") == 0)
    settings = SETTINGS_PROFILE1;
  else if (g_strcmp0 (profile, "profile2") == 0)
    settings = SETTINGS_PROFILE2;
  else if (g_strcmp0 (profile, "profile3") == 0)
    settings = SETTINGS_PROFILE3;
  else if (g_strcmp0 (profile, "profile4") == 0)
    settings = SETTINGS_PROFILE4;
  else if (g_strcmp0 (profile, "profile5") == 0)
    settings = SETTINGS_PROFILE5;
  else
    {
      g_critical ("settings for profile %s not found.", profile);
      return NULL;
    }
  return glook_app_get_settings (glook_app_get (), settings);
}

/**
 * on_disable_accelerators_chagned:
 *
 * disable/enable accelerator keys.
 */
static void
on_disable_accelerators_changed (GSettings   *settings,
				 const gchar *key,
				 GLookApp    *app)
{
  GtkSettings *gtksettings;
  gboolean disable;
  gchar *menu_accel;

  g_return_if_fail (G_IS_SETTINGS (settings));
  g_return_if_fail (GLOOK_IS_APP (app));

  disable = g_settings_get_boolean (settings, "disable-accelerators");

  gtksettings = gtk_settings_get_default ();

  g_object_set (gtksettings, "gtk-enable-accels", !disable, NULL);

  /* disable/enable menubar activate key(default F10) too.
   * before disable, we should save current value for later restore. */
  menu_accel = g_object_get_data (G_OBJECT (app), "gtk-menu-bar-accel");
  if (!menu_accel)
    {
      g_object_get (gtksettings, "gtk-menu-bar-accel", &menu_accel, NULL);
      g_object_set_data_full (G_OBJECT (app),
			      "gtk-menu-bar-accel", menu_accel,
			      (GDestroyNotify) g_free);
    }
  if (disable)
    g_object_set (gtksettings, "gtk-menu-bar-accel", NULL, NULL);
  else
    g_object_set (gtksettings, "gtk-menu-bar-accel", menu_accel, NULL);
}

/**
 * on_disable_mnemonics_changed:
 */
static void
on_disable_mnemonics_changed (GSettings   *settings,
			      const gchar *key,
			      GLookApp    *app)
{
  GtkSettings *gtksettings;
  gboolean disable;

  g_return_if_fail (G_IS_SETTINGS (settings));
  g_return_if_fail (GLOOK_IS_APP (app));

  disable = g_settings_get_boolean (settings, "disable-mnemonics");

  gtksettings = gtk_settings_get_default ();
  g_object_set (gtksettings, "gtk-enable-mnemonics", !disable, NULL);
}

/**
 * glook_app_set_accelerators:
 */
void
glook_app_set_accelerators (GLookApp *app)
{
  GSettings *settings;
  GtkSettings *gtksettings;
  gchar *menu_accel;

  g_return_if_fail (GLOOK_IS_APP (app));

  settings = glook_app_get_settings (app, SETTINGS_ACCELERATOR);
  gtksettings = gtk_settings_get_default ();

  if (g_settings_get_boolean (settings, "disable-mnemonics"))
    g_object_set (gtksettings, "gtk-enable-mnemonics", FALSE, NULL);

  if (g_settings_get_boolean (settings, "disable-accelerators"))
    {
      g_object_get (gtksettings, "gtk-menu-bar-accel", &menu_accel, NULL);
      if (menu_accel)
	{
	  g_object_set_data_full (G_OBJECT (app),
				  "gtk-menu-bar-accel", menu_accel,
				  (GDestroyNotify) g_free);
	}
      g_object_set (gtksettings, "gtk-menu-bar-accel", NULL, NULL);
      g_object_set (gtksettings, "gtk-enable-accels", FALSE, NULL);
    }
}

/**
 * glook_app_run:
 */
void
glook_app_run (GLookApp *app)
{
  g_return_if_fail (GLOOK_IS_APP (app));

  gtk_main ();
}

/**
 * on_app_window_destroy:
 *
 * called when top window destroy.
 * only window created by glook_app_create_window() will
 * go into here.
 */
static void
on_app_window_destroy (GLookWindow *window,
		       GLookApp    *app)
{
  GList *windows, *node;

  /* from the window from window group */
  gtk_window_group_remove_window (app->priv->window_group,
				  GTK_WINDOW (window));

  /* quit the application if it is last window */
  windows = gtk_window_group_list_windows (app->priv->window_group);
  for (node = windows; node; node = node->next)
    {
      if (GLOOK_IS_WINDOW (node->data))
	{
	  g_list_free (windows);
	  return;
	}
    }
  glook_app_quit (app);
}

/**
 * glook_app_quit:
 */
void
glook_app_quit (GLookApp *app)
{
  GList *windows, *node;
  GtkWidget *active_window = NULL;
  gint i = 0;

  g_return_if_fail (GLOOK_IS_APP (app));

  windows = gtk_window_group_list_windows (app->priv->window_group);

  /* if has more than one windows open, we give user prompt. */
  for (node = windows; node; node = node->next)
    {
      if (GLOOK_IS_WINDOW (node->data))
	{
	  if (!active_window && gtk_window_is_active (GTK_WINDOW (node->data)))
	    active_window = node->data;

	  i++;
	}
    }
  if (i > 1)
    {
      GtkWidget *dialog;

      dialog = gtk_message_dialog_new (GTK_WINDOW (active_window),
				       GTK_DIALOG_MODAL |
				       GTK_DIALOG_DESTROY_WITH_PARENT,
				       GTK_MESSAGE_QUESTION,
				       GTK_BUTTONS_YES_NO,
				       _("Close all windows?"));
      gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						_("Exit the application will "
						  "close all open %d windows."), i);
      gtk_window_set_keep_above (GTK_WINDOW (dialog), TRUE);
      if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_YES)
	{
	  gtk_widget_destroy (dialog);
	  return;
	}
      gtk_widget_destroy (dialog);
    }
  /* close all windows. */
  for (node = windows; node; node = node->next)
    {
      if (GLOOK_IS_WINDOW (node->data))
	{
	  g_signal_handlers_disconnect_by_func (node->data,
						on_app_window_destroy, app);
	  gtk_widget_set_sensitive (GTK_WIDGET (node->data), FALSE);
	  while (gtk_events_pending ())
	    gtk_main_iteration ();

	  gtk_widget_destroy (GTK_WIDGET (node->data));
	}
    }
  gtk_main_quit ();
}

/**
 * on_app_window_delete_event:
 */
static gboolean
on_app_window_delete_event (GLookWindow *window,
			    GdkEvent    *event,
			    GLookApp    *app)
{
  GSettings *settings;
  GtkWidget *notebook;
  GtkWidget *tab;
  gint i;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (GLOOK_IS_APP (app), FALSE);

  notebook = glook_window_get_notebook (window);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);
  if (g_settings_get_boolean (settings, "close-window-warning"))
    {
      if ((i = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook))) > 1)
	{
	  GtkWidget *dialog;

	  dialog = gtk_message_dialog_new (GTK_WINDOW (window),
					   GTK_DIALOG_MODAL |
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_MESSAGE_QUESTION,
					   GTK_BUTTONS_YES_NO,
					   _("Close all tabs?"));
	  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						    _("Close the window will close "
						      "all open %d tabs."), i);
	  if (gtk_dialog_run (GTK_DIALOG (dialog)) != GTK_RESPONSE_YES)
	    {
	      gtk_widget_destroy (dialog);
	      return TRUE;
	    }
	  gtk_widget_destroy (dialog);
	}
    }
  /* close all tabs. */
  for (i = 0; i < gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook)); i++)
    {
      tab = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i);
      glook_tab_stop (GLOOK_TAB (tab));
    }
  /* disable user interactive */
  gtk_widget_set_sensitive (GTK_WIDGET (window), FALSE);

  while (gtk_events_pending ())
    gtk_main_iteration ();

  glook_window_delete (window);
  return FALSE;
}

/**
 * glook_app_create_window:
 *
 * create a top level window.
 */
GtkWidget*
glook_app_create_window (GLookApp *app,
			 gboolean  with_blank_tab)
{
  GtkWidget *window;
  GList *toplevels, *node;

  g_return_val_if_fail (GLOOK_IS_APP (app), NULL);

  window = glook_window_new ();

  /* make new window has position offset from exists window. */
  toplevels = gtk_window_list_toplevels ();
  if (toplevels)
    {
      gint x, y;

      for (node = toplevels; node; node = node->next)
	{
	  if (gtk_window_is_active (GTK_WINDOW (node->data)))
	    {
	      gtk_window_get_position (GTK_WINDOW (node->data), &x, &y);
	      gtk_window_move (GTK_WINDOW (window), x + 10, y + 10);
	      break;
	    }
	}
      if (node == NULL)
	{
	  gtk_window_get_position (GTK_WINDOW (toplevels->data), &x, &y);
	  gtk_window_move (GTK_WINDOW (window), x + 10, y + 10);
	}
      g_list_free (toplevels);
    }
  /* add window to window group */
  gtk_window_group_add_window (app->priv->window_group,
			       GTK_WINDOW (window));
  /* add a blank tab to window? */
  if (with_blank_tab)
    glook_window_add_blank_tab (GLOOK_WINDOW (window), TRUE);

  /* when window been destroy, we remove it from window group */
  g_signal_connect (G_OBJECT (window), "delete-event",
		    G_CALLBACK (on_app_window_delete_event), app);
  g_signal_connect_after (G_OBJECT (window), "destroy",
			  G_CALLBACK (on_app_window_destroy), app);
  return window;
}

/**
 * glook_app_get_workarea:
 */
GdkRectangle*
glook_app_get_workarea (GLookApp *app)
{
  g_return_val_if_fail (GLOOK_IS_APP (app), NULL);
  return &app->priv->workarea;
}
