/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-window.h"
#include "glook-ui.h"
#include "glook-addrbar.h"
#include "glook-completion.h"

enum
{
  COL_ICON,
  COL_URI,
  COL_URI_CASEFOLD,
  COL_TITLE,
  COL_TIME,
  N_COLUMNS
};

struct _GLookCompletionPriv
{
  GtkTreeModel *model;
  gchar        *match_key;
};
#define _COMPLETION_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_COMPLETION, GLookCompletionPriv))

G_DEFINE_TYPE (GLookCompletion, glook_completion, GTK_TYPE_ENTRY_COMPLETION);

static void	glook_completion_pack_columns	(GLookCompletion    *completion);
static gboolean	glook_completion_find_uri	(GLookCompletion    *completion,
						 const gchar        *uri,
						 GtkTreeIter        *iter);
static gboolean	glook_completion_match		(GtkEntryCompletion *completion,
						 const gchar        *key,
						 GtkTreeIter        *iter,
						 gpointer            user_data);
/**
 * glook_completion_init:
 */
static void
glook_completion_init (GLookCompletion *completion)
{
  completion->priv = _COMPLETION_GET_PRIVATE (completion);

  completion->priv->match_key = NULL;

  /* create the list store to hold potential match strings */
  completion->priv->model =
   GTK_TREE_MODEL (gtk_list_store_new (N_COLUMNS,
				       GDK_TYPE_PIXBUF,
				       G_TYPE_STRING,
				       G_TYPE_STRING,
				       G_TYPE_STRING,
				       G_TYPE_ULONG));
  g_object_set (completion,
		"inline-selection", TRUE,
		"model", completion->priv->model,
		"text-column", COL_URI,
		"minimum-key-length", 0, NULL);

  /* add all visible cells */
  glook_completion_pack_columns (completion);

  /* replace default match function */
  gtk_entry_completion_set_match_func (GTK_ENTRY_COMPLETION (completion),
				       glook_completion_match, completion, NULL);
}

/**
 * glook_completion_dispose:
 */
static void
glook_completion_dispose (GObject *object)
{
  GLookCompletion *completion = GLOOK_COMPLETION (object);

  if (completion->priv->model)
    {
      g_object_unref (completion->priv->model);
      completion->priv->model = NULL;
    }
  G_OBJECT_CLASS (glook_completion_parent_class)->dispose (object);
}

/**
 * glook_completion_finalize:
 */
static void
glook_completion_finalize (GObject *object)
{
  GLookCompletion *completion = GLOOK_COMPLETION (object);

  if (completion->priv->match_key)
    {
      g_free (completion->priv->match_key);
      completion->priv->match_key = NULL;
    }
  G_OBJECT_CLASS (glook_completion_parent_class)->finalize (object);
}

/**
 * glook_completion_class_init:
 */
static void
glook_completion_class_init (GLookCompletionClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookCompletionPriv));

  gobject_class->dispose  = glook_completion_dispose;
  gobject_class->finalize = glook_completion_finalize;
}

/**
 * glook_completion_match:
 */
static gboolean
glook_completion_match (GtkEntryCompletion *entry_completion,
			const gchar        *key,
			GtkTreeIter        *iter,
			gpointer            user_data)
{
  GLookCompletion *completion = user_data;
  gchar *casefold;
  gboolean ret = FALSE;

  g_return_val_if_fail (GLOOK_IS_COMPLETION (completion), FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (iter, FALSE);

  gtk_tree_model_get (GTK_TREE_MODEL (completion->priv->model),
		      iter, COL_URI_CASEFOLD, &casefold, -1);
  if (casefold)
    {
      if (g_strstr_len (casefold, -1, key))
	ret = TRUE;

      g_free (casefold);
    }
  g_free (completion->priv->match_key);
  completion->priv->match_key = g_strdup (key);

  return ret;
}

/**
 * glook_completion_render_uri:
 */
static void
glook_completion_render_uri (GtkCellLayout   *layout,
			     GtkCellRenderer *cell,
			     GtkTreeModel    *model,
			     GtkTreeIter     *iter,
			     gpointer         data)
{
  GLookCompletion *completion = data;
  gchar *uri, *title, *casefold, *p, *p2;
  GString *string;

  g_return_if_fail (GLOOK_IS_COMPLETION (completion));
  g_return_if_fail (GTK_IS_CELL_LAYOUT (layout));
  g_return_if_fail (GTK_IS_TREE_MODEL (model));
  g_return_if_fail (iter);

  gtk_tree_model_get (model, iter, COL_URI, &uri,
		      COL_URI_CASEFOLD, &casefold, COL_TITLE, &title, -1);

  string = g_string_new ("<span foreground='forestgreen'>");

  if (completion->priv->match_key &&
      strlen (completion->priv->match_key) > 0)
    {
      gint keylen = strlen (completion->priv->match_key);

      p = g_strstr_len (casefold, -1, completion->priv->match_key);
      if (p)
	{
	  gint offset = p - casefold;

	  p2 = g_markup_escape_text (uri, offset);
	  g_string_append (string, p2);
	  g_free (p2);

	  g_string_append (string, "<span weight='bold' style='italic'>");
	  p2 = g_markup_escape_text (uri + offset, keylen);
	  g_string_append (string, p2);
	  g_free (p2);
	  g_string_append (string, "</span>");

	  p2 = g_markup_escape_text (uri + offset + keylen, -1);
	  g_string_append (string, p2);
	  g_free (p2);
	}
      else
	{
	  p2 = g_markup_escape_text (uri, -1);
	  g_string_append (string, p2);
	  g_free (p2);
	}
    }
  else
    {
      p2 = g_markup_escape_text (uri, -1);
      g_string_append (string, p2);
      g_free (p2);
    }
  g_string_append (string, "</span> - ");

  /* title is optional. */
  if (title)
    {
      g_string_append (string, "<span foreground='dim gray' weight='600'>");
      p2 = g_markup_escape_text (title, -1);
      g_string_append (string, p2);
      g_free (p2);
      g_string_append (string, "</span>");
    }
  g_free (uri);
  g_free (casefold);
  g_free (title);

  g_object_set (cell, "markup", string->str, NULL);
  g_string_free (string, TRUE);
}

/**
 * glook_completion_render_time:
 */
static void
glook_completion_render_time (GtkCellLayout   *layout,
			      GtkCellRenderer *cell,
			      GtkTreeModel    *model,
			      GtkTreeIter     *iter,
			      gpointer         data)
{
  time_t time_;
  struct tm *tmp;
  gint today;
  gchar strtime[128], markup[256];

  g_return_if_fail (GTK_IS_CELL_LAYOUT (layout));
  g_return_if_fail (GTK_IS_TREE_MODEL (model));
  g_return_if_fail (iter);

  time_ = time (NULL);
  tmp = localtime (&time_);
  today = tmp->tm_yday;

  gtk_tree_model_get (model, iter, COL_TIME, &time_, -1);

  tmp = localtime (&time_);

  /* if time is today, just show time, otherwise show data + time. */
  if (today == tmp->tm_yday)
    strftime (strtime, sizeof (strtime), "%X", tmp);
  else
    strftime (strtime, sizeof (strtime), "%x", tmp);

  g_snprintf (markup, sizeof (markup),
	      "<span foreground=\"dim gray\">%s</span>", strtime);

  g_object_set (cell, "markup", markup, NULL);
}

/**
 * glook_completion_pack_columns:
 */
static void
glook_completion_pack_columns (GLookCompletion *completion)
{
  GtkCellRenderer *renderer;

  g_return_if_fail (GLOOK_IS_COMPLETION (completion));

  /* icon cell */
  renderer = gtk_cell_renderer_pixbuf_new ();
  g_object_set (renderer, "cell-background", "grey93", "xpad", 4, NULL);
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (completion), renderer, FALSE);
  gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (completion),
				 renderer, "pixbuf", COL_ICON);
  /* uri and title cell */
  renderer = gtk_cell_renderer_text_new ();
  g_object_set (renderer, "ellipsize", PANGO_ELLIPSIZE_END, "scale", 0.9, NULL);
  gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (completion), renderer, TRUE);
  gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (completion), renderer,
				      glook_completion_render_uri, completion, NULL);
  /* time cell */
  renderer = gtk_cell_renderer_text_new ();
  g_object_set (renderer, "alignment", PANGO_ALIGN_RIGHT, "scale", 0.9, NULL);
  gtk_cell_layout_pack_end (GTK_CELL_LAYOUT (completion), renderer, FALSE);
  gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (completion), renderer,
				      glook_completion_render_time, completion, NULL);
}

/**
 * glook_completion_find_uri:
 *
 * this function is slowed if store has many lines.
 * any improve or suggest is welcome.
 */
static gboolean
glook_completion_find_uri (GLookCompletion *completion,
			   const gchar     *uri,
			   GtkTreeIter     *iter)
{
  GtkTreeModel *model;
  GtkTreeIter iter2;
  gboolean valid;

  g_return_val_if_fail (GLOOK_IS_COMPLETION (completion), FALSE);
  g_return_val_if_fail (uri, FALSE);

  model = completion->priv->model;

  valid = gtk_tree_model_get_iter_first (model, &iter2);
  while (valid)
    {
      gchar *uri2 = NULL;

      gtk_tree_model_get (model, &iter2, COL_URI, &uri2, -1);
      if (uri2)
	{
	  if (g_ascii_strcasecmp (uri, uri2) == 0)
	    {
	      if (iter)  *iter = iter2;

	      g_free (uri2);
	      return TRUE;
	    }
	  g_free (uri2);
	}
      valid = gtk_tree_model_iter_next (model, &iter2);
    }
  return FALSE;
}

/**
 * glook_completion_new:
 */
GLookCompletion*
glook_completion_new (void)
{
  return g_object_new (GLOOK_TYPE_COMPLETION, NULL);
}

/**
 * glook_completion_insert:
 *
 * insert a completion entry to match.
 */
void
glook_completion_insert (GLookCompletion *completion,
			 const gchar     *icon_name,
			 const gchar     *uri,
			 const gchar     *title,
			 time_t           time_,
			 gboolean         replace)
{
  GtkListStore *store;
  GtkTreeIter iter;
  gchar *casefold;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_COMPLETION (completion));
  g_return_if_fail (uri);

  store = GTK_LIST_STORE (completion->priv->model);

  /* check whether or not the uri exists. */
  if (glook_completion_find_uri (completion, uri, &iter))
    {
      if (!replace)
	{
	  g_debug ("completion already has uri `%s'.", uri);
	  return;
	}
      gtk_list_store_remove (store, &iter);
    }
  casefold = g_utf8_casefold (uri, -1);

  if (time_ == 0)
    time_ = time (NULL);

  gtk_list_store_insert_with_values (store, &iter, 0,
				     COL_URI, uri,
				     COL_URI_CASEFOLD, casefold,
				     COL_TITLE, title,
				     COL_TIME, time_, -1);
  g_free (casefold);

  /* icon is optional. */
  if (icon_name)
    {
      GdkPixbuf *pixbuf;

      /* render the icon name in icon theme */
      pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
					 icon_name, 16,
					 GTK_ICON_LOOKUP_USE_BUILTIN |
					 GTK_ICON_LOOKUP_FORCE_SIZE, &error);
      if (error)
	{
	  g_warning (_("Failed to load icon `%s': %s"), icon_name,
		     error->message);
	  g_error_free (error);
	  return;
	}
      gtk_list_store_set (store, &iter, COL_ICON, pixbuf, -1);
      g_object_unref (pixbuf);
    }
}

/**
 * glook_completion_remove:
 *
 * remove a match entry from completion.
 */
void
glook_completion_remove (GLookCompletion *completion,
			 const gchar     *uri)
{
  GtkListStore *store;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_COMPLETION (completion));
  g_return_if_fail (uri);

  store = GTK_LIST_STORE (completion->priv->model);

  if (!glook_completion_find_uri (completion, uri, &iter))
    {
      g_message ("address `%s' not found.", uri);
      return;
    }
  gtk_list_store_remove (store, &iter);
}

/**
 * glook_completion_clear:
 */
void
glook_completion_clear (GLookCompletion *completion)
{
  g_return_if_fail (GLOOK_IS_COMPLETION (completion));
  gtk_list_store_clear (GTK_LIST_STORE (completion->priv->model));
}
