/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-bookmark.h"
#include "glook-tab.h"
#include "glook-local-tab.h"

enum
{
  BOOKMARK_COL_GROUP,
  BOOKMARK_COL_ICONNAME,
  BOOKMARK_COL_TITLE,
  BOOKMARK_COL_ADDRESS,
  BOOKMARK_COL_DESCRIPTION,
  BOOKMARK_COL_VISITED,
  BOOKMARK_N_COLUMNS,
};

static void	glook_local_bookmark_reload		(GLookLocalTab    *tab);
static void	on_bookmark_tree_selection_changed	(GtkTreeSelection *selection,
							 GLookLocalTab    *tab);
static void	on_bookmark_added			(GLookBookmark    *bookmark,
							 const gchar      *address,
							 GLookLocalTab    *tab);
static void	on_bookmark_removed			(GLookBookmark    *bookmark,
							 const gchar      *address,
							 GLookLocalTab    *tab);
static void	on_bookmark_changed			(GLookBookmark    *bookmark,
							 const gchar      *address,
							 GLookLocalTab    *tab);
static void	on_bookmark_treeview_destroy		(GtkWidget        *treeview,
							 GLookLocalTab    *tab);

/**
 * on_bookmark_toplevel_map:
 */
void
on_bookmark_toplevel_map (GtkWidget     *widget,
			  GLookLocalTab *tab)
{
  GtkWidget *treeview;
  GtkTreeSelection *selection;
  GLookBookmark *bookmark;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect the handler. */
  g_signal_handlers_disconnect_by_func (widget, on_bookmark_toplevel_map, tab);

  /* load bookmark into treeview. */
  glook_local_bookmark_reload (tab);

  treeview = glook_local_tab_get_widget (tab, "BookmarkTreeview");

  /* connect selection change signal. */
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);
  g_signal_connect (selection, "changed",
		    G_CALLBACK (on_bookmark_tree_selection_changed), tab);
  gtk_widget_grab_focus (treeview);

  /* when bookmark changed, we need update the treeview.
   * when treeview destroy, we need disconnect from bookmark signal. */
  bookmark = glook_app_get_bookmark (glook_app_get ());
  g_signal_connect (bookmark, "added",
		    G_CALLBACK (on_bookmark_added), tab);
  g_signal_connect (bookmark, "removed",
		    G_CALLBACK (on_bookmark_removed), tab);
  g_signal_connect (bookmark, "changed",
		    G_CALLBACK (on_bookmark_changed), tab);
  g_signal_connect (treeview, "destroy",
		    G_CALLBACK (on_bookmark_treeview_destroy), tab);
}

/**
 * glook_local_bookmark_create_group:
 */
static void
glook_local_bookmark_create_group (GLookLocalTab *tab,
				   GtkTreeModel  *model,
				   GtkTreeIter   *iter,
				   const gchar   *group,
				   const gchar   *icon_name)
{
  gboolean valid;

  g_return_if_fail (GLOOK_LOCAL_TAB (tab));
  g_return_if_fail (GTK_IS_TREE_MODEL (model));
  g_return_if_fail (iter);

  valid = gtk_tree_model_get_iter_first (model, iter);
  while (valid)
    {
      gchar *group2 = NULL;

      gtk_tree_model_get (model, iter, BOOKMARK_COL_GROUP, &group2, -1);

      /* find exists group. */
      if (g_strcmp0 (group, group2) == 0)
	{
	  g_free (group2);
	  return;
	}
      g_free (group2);

      valid = gtk_tree_model_iter_next (model, iter);
    }
  /* create new group. */
  gtk_tree_store_insert_with_values (GTK_TREE_STORE (model), iter, NULL, -1,
				     BOOKMARK_COL_GROUP, group,
				     BOOKMARK_COL_TITLE, group,
				     BOOKMARK_COL_ICONNAME, icon_name, -1);
}

/**
 * glook_local_bookmark_add_address:
 */
static void
glook_local_bookmark_add_address (GLookLocalTab *tab,
				  GBookmarkFile *file,
				  GtkTreeModel  *model,
				  const gchar   *address)
{
  gchar *icon_name = NULL;
  gchar *title, *desc;
  time_t visited;
  gchar **groups;
  gint j;
  GtkTreeIter group_iter;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (address);

  title = g_bookmark_file_get_title (file, address, NULL);
  desc = g_bookmark_file_get_description (file, address, NULL);
  visited = g_bookmark_file_get_visited (file, address, NULL);
  g_bookmark_file_get_icon (file, address, &icon_name, NULL, NULL);

  groups = g_bookmark_file_get_groups (file, address, NULL, NULL);
  if (groups)
    {
      for (j = 0; groups[j]; j++)
	{
	  glook_local_bookmark_create_group (tab, model, &group_iter,
					     groups[j], icon_name);

	  gtk_tree_store_insert_with_values (GTK_TREE_STORE (model),
					     NULL, &group_iter, -1,
					     BOOKMARK_COL_ICONNAME, icon_name, 
					     BOOKMARK_COL_GROUP, groups[j], 
					     BOOKMARK_COL_TITLE, title, 
					     BOOKMARK_COL_ADDRESS, address, 
					     BOOKMARK_COL_DESCRIPTION, desc, 
					     BOOKMARK_COL_VISITED, (guint) visited, 
					     -1);
	}
      g_strfreev (groups);
    }
  else
    {
      g_critical (_("URI `%s' has not group."), address);
    }
  g_free (icon_name);
  g_free (title);
  g_free (desc);
}

/**
 * glook_local_bookmark_reload:
 *
 * load bookmarks into treeview.
 */
static void
glook_local_bookmark_reload (GLookLocalTab *tab)
{
  GLookBookmark *bookmark;
  GBookmarkFile *file;
  gchar **uris;
  gsize length, i;
  GObject *model;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  model = glook_local_tab_get_object (tab, "BookmarkStore");

  /* clear the store first. */
  gtk_tree_store_clear (GTK_TREE_STORE (model));

  bookmark = glook_app_get_bookmark (glook_app_get ());
  file = glook_bookmark_get_file (bookmark);

  uris = g_bookmark_file_get_uris (file, &length);
  for (i = 0; i < length; i++)
    glook_local_bookmark_add_address (tab, file, GTK_TREE_MODEL (model), uris[i]);

  g_strfreev (uris);
}

/**
 * on_bookmark_tree_selection_changed:
 */
static void
on_bookmark_tree_selection_changed (GtkTreeSelection *selection,
				    GLookLocalTab    *tab)
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkWidget *table;
  GtkWidget *title_entry;
  GtkWidget *desc_entry;
  GtkWidget *visited_entry;
  gchar *title = NULL, *desc = NULL;
  guint visited = 0;
  struct tm *tmp;
  gchar strtime[256] = { '\0', };

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  table = glook_local_tab_get_widget (tab, "BookmarkTable");
  title_entry = glook_local_tab_get_widget (tab, "BookmarkTitleEntry");
  desc_entry = glook_local_tab_get_widget (tab, "BookmarkDescEntry");
  visited_entry = glook_local_tab_get_widget (tab, "BookmarkVisitedEntry");

  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  if (gtk_tree_model_iter_has_child (model, &iter))
    {
      gtk_tree_model_get (model, &iter, BOOKMARK_COL_TITLE, &title, -1);
      gtk_entry_set_text (GTK_ENTRY (title_entry), title ? title : "");
      gtk_entry_set_text (GTK_ENTRY (desc_entry), "");
      gtk_entry_set_text (GTK_ENTRY (visited_entry), "");
      g_free (title);

      gtk_widget_set_sensitive (table, FALSE);
      return;
    }
  else
    {
      gtk_widget_set_sensitive (table, TRUE);

      gtk_tree_model_get (model, &iter, BOOKMARK_COL_TITLE, &title,
			  BOOKMARK_COL_DESCRIPTION, &desc,
			  BOOKMARK_COL_VISITED, &visited, -1);
      if (visited > 0)
	{
	  tmp = localtime ((time_t*) &visited);
	  strftime (strtime, sizeof (strtime), "%x %X", tmp);
	}
      gtk_entry_set_text (GTK_ENTRY (title_entry), title ? title : "");
      gtk_entry_set_text (GTK_ENTRY (desc_entry), desc ? desc : "");
      gtk_entry_set_text (GTK_ENTRY (visited_entry), strtime);

      g_free (title);
      g_free (desc);
    }
}


typedef struct
{
  GtkTreeIter *iter;
  const gchar *address;
  gboolean     found;
} ForeachData;

/**
 * get_iter_by_address_foreach:
 */
static gboolean
get_iter_by_address_foreach (GtkTreeModel *model,
			     GtkTreePath  *path,
			     GtkTreeIter  *iter,
			     gpointer      user_data)
{
  ForeachData *data = user_data;
  gchar *address;

  g_return_val_if_fail (GTK_IS_TREE_MODEL (model), TRUE);

  /* skip group node. */
  if (gtk_tree_path_get_depth (path) == 1)
    return FALSE;

  gtk_tree_model_get (model, iter, BOOKMARK_COL_ADDRESS, &address, -1);
  if (g_strcmp0 (address, data->address) == 0)
    {
      *data->iter = *iter;
      data->found = TRUE;
      g_free (address);
      return TRUE;
    }
  g_free (address);

  return FALSE;
}

/**
 * glook_bookmark_get_iter:
 */
static gboolean
glook_bookmark_get_iter (GLookLocalTab *tab,
			 GtkTreeModel  *model,
			 GtkTreeIter   *iter,
			 const gchar   *address)
{
  ForeachData data;

  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), FALSE);
  g_return_val_if_fail (GTK_IS_TREE_MODEL (model), FALSE);

  data.iter = iter;
  data.address = address;
  data.found = FALSE;

  gtk_tree_model_foreach (model, get_iter_by_address_foreach, &data);
  return data.found;
}

/**
 * on_bookmark_added:
 */
static void
on_bookmark_added (GLookBookmark *bookmark,
		   const gchar   *address,
		   GLookLocalTab *tab)
{
  GObject *model;
  GBookmarkFile *file;

  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  model = glook_local_tab_get_object (tab, "BookmarkStore");
  file = glook_bookmark_get_file (bookmark);

  glook_local_bookmark_add_address (tab, file, GTK_TREE_MODEL (model), address);
}

/**
 * on_bookmark_removed:
 */
static void
on_bookmark_removed (GLookBookmark *bookmark,
		     const gchar   *address,
		     GLookLocalTab *tab)
{
  GObject *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  model = glook_local_tab_get_object (tab, "BookmarkStore");

  if (glook_bookmark_get_iter (tab, GTK_TREE_MODEL (model),
			       &iter, address))
    {
      gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);
    }
}

/**
 * on_bookmark_changed:
 */
static void
on_bookmark_changed (GLookBookmark *bookmark,
		     const gchar   *address,
		     GLookLocalTab *tab)
{
  GObject *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_BOOKMARK (bookmark));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  model = glook_local_tab_get_object (tab, "BookmarkStore");

  if (glook_bookmark_get_iter (tab, GTK_TREE_MODEL (model),
			       &iter, address))
    {
      GBookmarkFile *file;
      gchar *title, *desc;
      time_t visited;

      file = glook_bookmark_get_file (bookmark);
      title = g_bookmark_file_get_title (file, address, NULL);
      desc = g_bookmark_file_get_description (file, address, NULL);
      visited = g_bookmark_file_get_visited (file, address, NULL);

      gtk_tree_store_set (GTK_TREE_STORE (model), &iter,
			  BOOKMARK_COL_TITLE, title,
			  BOOKMARK_COL_DESCRIPTION, desc,
			  BOOKMARK_COL_VISITED, visited, -1);
       g_free (title);
       g_free (desc);
    }
}

/**
 * on_bookmark_treeview_destroy:
 */
static void
on_bookmark_treeview_destroy (GtkWidget     *treeview,
			      GLookLocalTab *tab)
{
  GLookBookmark *bookmark;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect from bookmark signal. */
  bookmark = glook_app_get_bookmark (glook_app_get ());
  g_signal_handlers_disconnect_by_func (bookmark, on_bookmark_added, tab);
  g_signal_handlers_disconnect_by_func (bookmark, on_bookmark_removed, tab);
  g_signal_handlers_disconnect_by_func (bookmark, on_bookmark_changed, tab);
}

/**
 * on_bookmark_title_entry_changed:
 */
void
on_bookmark_title_entry_changed (GtkEntry      *entry,
				 GLookLocalTab *tab)
{
  GtkWidget *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GLookBookmark *bookmark;
  GBookmarkFile *file;
  gchar *address;
  const gchar *title;

  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "BookmarkTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  gtk_tree_model_get (model, &iter, BOOKMARK_COL_ADDRESS, &address, -1);

  bookmark = glook_app_get_bookmark (glook_app_get ());
  file = glook_bookmark_get_file (bookmark);

  /* change bookmark title, this will not trigger bookmark changed signal.
   * so we need update the treeview by ourself. */
  title = gtk_entry_get_text (entry);
  g_bookmark_file_set_title (file, address, title);
  g_free (address);

  /* update the treeview. */
  gtk_tree_store_set (GTK_TREE_STORE (model), &iter,
		      BOOKMARK_COL_TITLE, title, -1);
}

/**
 * on_bookmark_desc_entry_changed:
 */
void
on_bookmark_desc_entry_changed (GtkEntry      *entry,
				GLookLocalTab *tab)
{
  GtkWidget *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GLookBookmark *bookmark;
  GBookmarkFile *file;
  gchar *address;
  const gchar *desc;

  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "BookmarkTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  gtk_tree_model_get (model, &iter, BOOKMARK_COL_ADDRESS, &address, -1);

  bookmark = glook_app_get_bookmark (glook_app_get ());
  file = glook_bookmark_get_file (bookmark);

  /* change bookmark title, this will not trigger bookmark changed signal.
   * so we need update the treeview by ourself. */
  desc = gtk_entry_get_text (entry);
  g_bookmark_file_set_description (file, address, desc);
  g_free (address);

  /* update the treeview. */
  gtk_tree_store_set (GTK_TREE_STORE (model), &iter,
		      BOOKMARK_COL_DESCRIPTION, desc, -1);
}


/* popup menu ui */
static const gchar *_bookmark_popup_menu_ui =
"<ui>"
"<popup name='BookmarkPopupMenu'>"
"  <menuitem action='OpenNewTabAction'/>"
"  <menuitem action='OpenNewWindowAction'/>"
"  <separator/>"
"  <menuitem action='CopyAddressAction'/>"
"  <separator/>"
"  <menuitem action='DeleteAction'/>"
"</popup>"
"</ui>";

/**
 * on_bookmark_treeview_button_pressed:
 */
gboolean
on_bookmark_treeview_button_pressed (GtkTreeView    *treeview,
				     GdkEventButton *event,
				     GLookLocalTab  *tab)
{
  GtkTreeSelection *selection;
  GtkTreePath *path;

  g_return_val_if_fail (GTK_IS_TREE_VIEW (treeview), TRUE);
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), TRUE);

  if (event->button != 3 || event->type != GDK_BUTTON_PRESS)
    return FALSE;

  selection = gtk_tree_view_get_selection (treeview);

  /* get tree path for row which behind mouse point. */
  if (gtk_tree_view_get_path_at_pos (treeview, (gint) event->x,
				     (gint) event->y, &path,
				     NULL, NULL, NULL))
    {
      GtkUIManager *ui;
      GObject *action_group;
      GtkWidget *menu;
      GError *error = NULL;

      if (!gtk_tree_selection_path_is_selected (selection, path))
	{
	  gtk_tree_selection_unselect_all (selection);
	  gtk_tree_selection_select_path (selection, path);
	}
      if (gtk_tree_path_get_depth (path) == 1)
	{
	  gtk_tree_path_free (path);
	  return FALSE;
	}
      gtk_tree_path_free (path);

      action_group = glook_local_tab_get_object (tab, "PopupMenuActionGroup");

      ui = gtk_ui_manager_new ();
      gtk_ui_manager_insert_action_group (ui, GTK_ACTION_GROUP (action_group), 0);
      gtk_ui_manager_add_ui_from_string (ui, _bookmark_popup_menu_ui, -1, &error);
      if (error)
	{
	  g_error ("%s", error->message);
	  return TRUE;
	}
      menu = gtk_ui_manager_get_widget (ui, "/BookmarkPopupMenu");
      gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		      event->button, event->time);
      g_object_unref (ui);
    }
  return TRUE;
}

/**
 * on_bookmark_treeview_row_activated:
 */
void
on_bookmark_treeview_row_activated (GtkTreeView       *treeview,
				    GtkTreePath       *path,
				    GtkTreeViewColumn *column,
				    GLookLocalTab     *tab)
{
  GObject *action;

  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  action = glook_local_tab_get_object (tab, "OpenNewTabAction");
  gtk_action_activate (GTK_ACTION (action));
}

/**
 * show_window:
 */
static gboolean
show_window (gpointer data)
{
  GtkWidget *window = data;

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  if (!gtk_widget_get_visible (window))
    gtk_widget_show (window);

  return FALSE;
}

/**
 * glook_local_bookmark_open:
 */
static void
glook_local_bookmark_open (GLookLocalTab *tab,
			   gboolean       newtab)
{
  GtkWidget *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkWidget *window;
  gchar *address;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "BookmarkTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  gtk_tree_model_get (model, &iter, BOOKMARK_COL_ADDRESS, &address, -1);
  if (!address)
    {
      GtkTreePath *treepath;
      treepath = gtk_tree_model_get_path (model, &iter);

      if (gtk_tree_view_row_expanded (GTK_TREE_VIEW (treeview), treepath))
	gtk_tree_view_collapse_row (GTK_TREE_VIEW (treeview), treepath);
      else
	gtk_tree_view_expand_row (GTK_TREE_VIEW (treeview), treepath, FALSE);

      gtk_tree_path_free (treepath);
      return;
    }
  if (newtab)
    {
      window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
      glook_window_open_uri_string (GLOOK_WINDOW (window), address);
    }
  else
    {
      window = glook_app_create_window (glook_app_get (), FALSE);
      glook_window_open_uri_string (GLOOK_WINDOW (window), address);
      g_idle_add (show_window, window);
    }
  g_free (address);
}

/**
 * on_bookmark_open_new_tab_activate:
 */
void
on_bookmark_open_new_tab_activate (GtkAction     *action,
				   GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_local_bookmark_open (tab, TRUE);
}

/**
 * on_bookmark_open_new_window_activate:
 */
void
on_bookmark_open_new_window_activate (GtkAction     *action,
				      GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_local_bookmark_open (tab, FALSE);
}

/**
 * on_bookmark_copy_address_activate:
 */
void
on_bookmark_copy_address_activate (GtkAction     *action,
				   GLookLocalTab *tab)
{
  GtkWidget *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gchar *address;
  GtkClipboard *clipboard;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "BookmarkTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
  if (clipboard)
    {
      gtk_tree_model_get (model, &iter, BOOKMARK_COL_ADDRESS, &address, -1);
      if (address)
	{
	  gtk_clipboard_set_text (clipboard, address, -1);
	  g_free (address);
	}
    }
}

/**
 * on_bookmark_backup_activate:
 */
void
on_bookmark_backup_activate (GtkAction     *action,
			     GLookLocalTab *tab)
{
  GtkWidget *dialog, *window;
  GtkFileChooser *chooser;
  GtkFileFilter *filter;
  const gchar *dir;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  dialog = gtk_file_chooser_dialog_new (_("Backup bookmarks into file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("Bookmark files"));
  gtk_file_filter_add_mime_type (filter, "application/x-xbel");
  gtk_file_chooser_add_filter (chooser, filter);
  gtk_file_chooser_set_filter (chooser, filter);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (chooser, filter);

  gtk_file_chooser_set_do_overwrite_confirmation (chooser, TRUE);

  dir = g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS);
  gtk_file_chooser_set_current_folder (chooser, dir);
  gtk_file_chooser_set_current_name (chooser, "ttyutils-glook-bookmark.xbel");

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      GLookBookmark *bookmark;
      gchar *filename = gtk_file_chooser_get_filename (chooser);

      bookmark = glook_app_get_bookmark (glook_app_get ());
      glook_bookmark_save_to_file (bookmark, filename);
      g_free (filename);
    }
  gtk_widget_destroy (dialog);
}

/**
 * on_bookmark_delete_activate:
 */
void
on_bookmark_delete_activate (GtkAction     *action,
			     GLookLocalTab *tab)
{
  GtkWidget *treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gchar *address;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "BookmarkTreeview");

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  gtk_tree_model_get (model, &iter, BOOKMARK_COL_ADDRESS, &address, -1);
  if (!address)
    {
      /* we can safely remove the group if it has not children. */
      if (!gtk_tree_model_iter_has_child (model, &iter))
	gtk_tree_store_remove (GTK_TREE_STORE (model), &iter);

      return;
    }
  glook_bookmark_remove (glook_app_get_bookmark (glook_app_get ()), address);
  g_free (address);
}

/**
 * on_bookmark_edit_button_clicked:
 */
void
on_bookmark_edit_button_clicked (GtkToggleButton *button,
				 GLookLocalTab   *tab)
{
  gboolean toggled;
  GtkWidget *table;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  toggled = gtk_toggle_button_get_active (button);
  table = glook_local_tab_get_widget (tab, "BookmarkTable");
  gtk_widget_set_visible (table, toggled);
}

/**
 * on_bookmark_delete_button_clicked:
 */
void
on_bookmark_delete_button_clicked (GtkWidget     *button,
				   GLookLocalTab *tab)
{
  GObject *action;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  action = glook_local_tab_get_object (tab, "DeleteAction");
  gtk_action_activate (GTK_ACTION (action));
}

/**
 * on_bookmark_backup_button_clicked:
 */
void
on_bookmark_backup_button_clicked (GtkWidget     *button,
				   GLookLocalTab *tab)
{
  GObject *action;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  action = glook_local_tab_get_object (tab, "BackupAction");
  gtk_action_activate (GTK_ACTION (action));
}

/**
 * on_bookmark_load_button_clicked:
 */
void
on_bookmark_load_button_clicked (GtkWidget     *button,
				 GLookLocalTab *tab)
{
  GtkWidget *dialog, *window;
  GtkFileChooser *chooser;
  GtkFileFilter *filter;
  const gchar *dir;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  dialog = gtk_file_chooser_dialog_new (_("Add bookmarks from file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_OPEN,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("Bookmark files"));
  gtk_file_filter_add_mime_type (filter, "application/x-xbel");
  gtk_file_chooser_add_filter (chooser, filter);
  gtk_file_chooser_set_filter (chooser, filter);

  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (chooser, filter);

  dir = g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS);
  gtk_file_chooser_set_current_folder (chooser, dir);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      GLookBookmark *bookmark;
      gchar *filename = gtk_file_chooser_get_filename (chooser);

      bookmark = glook_app_get_bookmark (glook_app_get ());
      glook_bookmark_add_from_file (bookmark, filename);
      g_free (filename);
    }
  gtk_widget_destroy (dialog);
}
