/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>
#include <gdk/gdkkeysyms.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-statusbar.h"
#include "glook-ui.h"
#include "glook-tab.h"
#include "glook-local-tab.h"

typedef struct _GLookLogMessage	GLookLogMessage;

struct _GLookLogMessage
{
  gint    refcount;
  gchar  *domain;
  gint    level;
  gchar  *message;
  time_t  time;
};

/* use recursion mutex to prevent log message from library function, like
 * g_return_if... () family function to block application.
 */
static GList		*_log_message_list = NULL;
static GStaticRecMutex	 _log_message_mutex = G_STATIC_REC_MUTEX_INIT;

/* remember all messages in window statusbar stack. */
static guint		 _window_message_context_id = 0;
static guint		 _window_message_timeout = 0;
static guint             _window_message_count = 0;

/**
 * glook_log_message_new:
 */
static GLookLogMessage*
glook_log_message_new (void)
{
  GLookLogMessage *logmsg = g_slice_new0 (GLookLogMessage);

  logmsg->refcount = 1;
  return logmsg;
}

/**
 * glook_log_message_ref:
 */
static GLookLogMessage*
glook_log_message_ref (GLookLogMessage *logmsg)
{
  if (logmsg)
    g_atomic_int_inc (&logmsg->refcount);

  return logmsg;
}

/**
 * glook_log_message_unref:
 */
static void
glook_log_message_unref (GLookLogMessage *logmsg)
{
  if (logmsg && g_atomic_int_dec_and_test (&logmsg->refcount))
    {
      g_free (logmsg->domain);
      g_free (logmsg->message);

      g_slice_free (GLookLogMessage, logmsg);
    }
}

/**
 * glook_local_log_cleanup:
 */
void
glook_local_log_cleanup (void)
{
  g_static_rec_mutex_lock (&_log_message_mutex);

  if (_log_message_list)
    {
      g_list_foreach (_log_message_list,
		      (GFunc) glook_log_message_unref, NULL);
      g_list_free (_log_message_list);
      _log_message_list = NULL;
    }
  g_static_rec_mutex_unlock (&_log_message_mutex);
}

/**
 * window_pop_message:
 */
static void
window_pop_message (GLookWindow *window,
		    gpointer     data)
{
  if (GLOOK_IS_WINDOW (window))
    {
      GtkWidget *statusbar;

      statusbar = glook_window_get_statusbar (window);
      gtk_statusbar_pop (GTK_STATUSBAR (statusbar),
			 _window_message_context_id);
    }
}

/**
 * remove_window_message_timeout:
 */
static gboolean
remove_window_message_timeout (gpointer data)
{
  GtkWindowGroup *window_group;
  GList *windows;

  window_group = glook_app_get_window_group (glook_app_get ());
  windows = gtk_window_group_list_windows (window_group);
  if (windows)
    {
      g_list_foreach (windows, (GFunc) window_pop_message, NULL);
      g_list_free (windows);

      _window_message_count--;
    }
  return _window_message_count > 0;
}

/**
 * remove_window_message_timeout_destroy:
 */
static void
remove_window_message_timeout_destroy (gpointer data)
{
  _window_message_timeout = 0;
}

/**
 * window_push_message:
 *
 * push message into window statusbar.
 */
static void
window_push_message (GLookWindow     *window,
		     GLookLogMessage *logmsg)
{
  g_return_if_fail (logmsg);

  if (GLOOK_IS_WINDOW (window))
    {
      GtkWidget *statusbar;
      GtkMessageType message_type;

      if (logmsg->level & G_LOG_LEVEL_WARNING)
	message_type = GTK_MESSAGE_WARNING;
      else if (logmsg->level & G_LOG_LEVEL_CRITICAL)
	message_type = GTK_MESSAGE_ERROR;
      else
	message_type = GTK_MESSAGE_INFO;

      statusbar = glook_window_get_statusbar (window);

      if (_window_message_context_id == 0)
	{
	  _window_message_context_id =
	   gtk_statusbar_get_context_id (GTK_STATUSBAR (statusbar),
					 "window-statusbar-message-context");
	}
      glook_statusbar_push (GLOOK_STATUSBAR (statusbar),
			    _window_message_context_id,
			    message_type, logmsg->message);
    }
}

/**
 * window_show_message_idle:
 *
 * show message in all window statusbar.
 */
static gboolean
window_show_message_idle (gpointer data)
{
  GLookLogMessage *logmsg = data;
  GtkWindowGroup *window_group;
  GList *windows;

  g_return_val_if_fail (logmsg, FALSE);

  window_group = glook_app_get_window_group (glook_app_get ());
  windows = gtk_window_group_list_windows (window_group);
  if (windows)
    {
      g_list_foreach (windows, (GFunc) window_push_message, logmsg);
      g_list_free (windows);

      /* auto remove the message after 3 seconds. */
      if (_window_message_timeout == 0)
	{
	  _window_message_timeout =
	   g_timeout_add_seconds_full (G_PRIORITY_DEFAULT, 3,
				       remove_window_message_timeout, NULL,
				       remove_window_message_timeout_destroy);
	}
      _window_message_count++;
    }
  return FALSE;
}

/**
 * _glook_local_log_handler:
 *
 * redirect glib message to internal list.
 */
static void
_glook_local_log_handler (const gchar    *domain,
			  GLogLevelFlags  level,
			  const gchar    *message,
			  gpointer        user_data)
{
  GLookLogMessage *logmsg;

  if (level & G_LOG_FLAG_FATAL)
    {
      if (isatty (STDOUT_FILENO))
	{
	  printf ("%s-Fatal: %s\n", domain, message);
	  g_on_error_stack_trace (NULL);
	}
      return;
    }
  logmsg = glook_log_message_new ();

  logmsg->domain  = g_strdup (domain);
  logmsg->level   = level;
  logmsg->message = g_strdup (message);
  logmsg->time    = time (NULL);

  g_static_rec_mutex_lock (&_log_message_mutex);

  /* append to message list. */
  _log_message_list = g_list_append (_log_message_list, logmsg);
#ifdef DEBUG
  g_print ("%s\n", logmsg->message);
#endif
  /* show message to window statusbar. only message in Ttyutils
   * domain get show. */
  if (((logmsg->level & G_LOG_LEVEL_MESSAGE) ||
       (logmsg->level & G_LOG_LEVEL_WARNING) ||
       (logmsg->level & G_LOG_LEVEL_CRITICAL)) &&
#ifdef DEBUG
      g_pattern_match_simple ("*", logmsg->domain))
#else
      g_pattern_match_simple ("Ttyutils*", logmsg->domain))
#endif
    {
      g_idle_add_full (G_PRIORITY_HIGH_IDLE, window_show_message_idle,
		       glook_log_message_ref (logmsg),
		       (GDestroyNotify) glook_log_message_unref);
    }
  /* prevent unlimit erlarge. */
  if (g_list_length (_log_message_list) >= 1000)
    {
      GList *node = _log_message_list;
      _log_message_list = node->next;

      glook_log_message_unref (node->data);
      g_list_free_1 (node);
    }
  g_static_rec_mutex_unlock (&_log_message_mutex);
}

/**
 * glook_local_log_redirect:
 *
 * redirect glib log messages.
 */
void
glook_local_log_redirect (void)
{
  g_log_set_default_handler (_glook_local_log_handler, NULL);
}

/**
 * get_current_log_level:
 */
static GLogLevelFlags
get_current_log_level_flags (GLookLocalTab *tab)
{
  GtkWidget *combo;
  GLogLevelFlags flags;

  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), G_LOG_LEVEL_MASK);

  flags = G_LOG_LEVEL_MASK;

  combo = glook_local_tab_get_widget (tab, "LogLevelCombo");

  switch (gtk_combo_box_get_active (GTK_COMBO_BOX (combo)))
    {
    case 1:	/* message */
      flags &= ~(G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG);
      break;
    case 2:	/* warning */
      flags &= ~(G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG | G_LOG_LEVEL_MESSAGE);
      break;
    case 3:	/* critical */
      flags &= ~(G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG |
		 G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_WARNING);
      break;
    }
  return flags;
}

/**
 * glook_local_log_format_time:
 */
static const gchar*
glook_local_log_format_time (time_t time_)
{
  static gchar strtime[128];
  time_t curr_time;
  struct tm *tmp;
  gint today;

  curr_time = time (NULL);
  tmp = localtime (&curr_time);
  today = tmp->tm_yday;

  tmp = localtime (&time_);

  if (today == tmp->tm_yday)
    strftime (strtime, sizeof (strtime), "%X", tmp);
  else
    strftime (strtime, sizeof (strtime), "%x", tmp);

  return strtime;
}

/**
 * glook_log_output_message:
 *
 * output message to text buffer.
 */
static void
glook_local_log_output_message (GLookLogMessage *logmsg,
				GLookLocalTab   *tab)
{
  GtkWidget *entry;
  const gchar *filter_text;
  const gchar *tagname;
  GObject *buffer;
  GtkTextIter iter;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* filter messages by level. */
  if (!(logmsg->level & get_current_log_level_flags (tab)))
    return;

  /* filter messages by text. */
  entry = glook_local_tab_get_widget (tab, "LogFilterEntry");
  filter_text = gtk_entry_get_text (GTK_ENTRY (entry));
  if (filter_text && filter_text[0] &&
      !g_strstr_len (logmsg->message, -1, filter_text))
    return;

  switch (logmsg->level)
    {
    case G_LOG_LEVEL_DEBUG:
    case G_LOG_LEVEL_INFO:
      tagname = "tag-debug";
      break;
    case G_LOG_LEVEL_MESSAGE:
      tagname = "tag-message";
      break;
    case G_LOG_LEVEL_WARNING:
      tagname = "tag-warning";
      break;
    default:
      tagname = "tag-critical";
      break;
    }
  buffer = glook_local_tab_get_object (tab, "LogTextBuffer");
  gtk_text_buffer_get_end_iter (GTK_TEXT_BUFFER (buffer), &iter);

  gtk_text_buffer_insert_with_tags_by_name (GTK_TEXT_BUFFER (buffer), &iter,
					    glook_local_log_format_time (logmsg->time), -1,
					    "tag-scale", "tag-time", NULL);
  gtk_text_buffer_insert (GTK_TEXT_BUFFER (buffer), &iter, ": ", -1);
  gtk_text_buffer_insert_with_tags_by_name (GTK_TEXT_BUFFER (buffer),
					    &iter, logmsg->message, -1,
					    "tag-scale", tagname, NULL);
  gtk_text_buffer_insert_with_tags_by_name (GTK_TEXT_BUFFER (buffer), &iter,
					    "\n", 1, "tag-scale", NULL);
}

/**
 * glook_local_log_autoscroll_idle:
 *
 * scroll to bottom of scrolled window.
 */
static gboolean
glook_local_log_autoscroll_idle (gpointer data)
{
  GLookLocalTab *tab = data;

  if (GLOOK_IS_LOCAL_TAB (tab))
    {
      GtkWidget *scrolled;
      GtkAdjustment *vadj;
      gdouble value;

      scrolled = glook_local_tab_get_widget (tab, "LogScrolled");
      g_object_get (scrolled, "vadjustment", &vadj, NULL);

      value = gtk_adjustment_get_upper (vadj) - gtk_adjustment_get_page_size (vadj);
      gtk_adjustment_set_value (vadj, value);

      g_object_unref (vadj);
    }
  return FALSE;
}

/**
 * glook_local_log_refresh:
 */
void
glook_local_log_refresh (GLookLocalTab *tab)
{
  GLookLogMessage *lastmsg;
  GList *node = _log_message_list;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (!_log_message_list)
    return;

  g_static_rec_mutex_lock (&_log_message_mutex);

  /* try to continue from preivous output message. */
  lastmsg = g_object_get_data (G_OBJECT (tab), "last-log-message");
  if (lastmsg)
    {
      node = g_list_find (_log_message_list, lastmsg);
      if (node)
	node = node->next;
    }
  if (node)
    {
      GtkWidget *textview;
      GObject *buffer;
      GtkWidget *button;

      /* detach buffer from textview. */
      textview = glook_local_tab_get_widget (tab, "LogTextview");
      buffer = glook_local_tab_get_object (tab, "LogTextBuffer");

      g_object_ref (buffer);
      gtk_text_view_set_buffer (GTK_TEXT_VIEW (textview), NULL);

      g_list_foreach (node, (GFunc) glook_local_log_output_message, tab);

      /* reattach buffer to textview */
      gtk_text_view_set_buffer (GTK_TEXT_VIEW (textview), GTK_TEXT_BUFFER (buffer));
      g_object_unref (buffer);

      /* autoscroll to the bottom of view */
      button = glook_local_tab_get_widget (tab, "AutoScrollButton");
      if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button)))
	g_idle_add (glook_local_log_autoscroll_idle, tab);
    }
  /* save last message, next refresh will continue from here. */
  node = g_list_last (_log_message_list);
  g_object_set_data_full (G_OBJECT (tab), "last-log-message",
			  glook_log_message_ref (node->data),
			  (GDestroyNotify) glook_log_message_unref);

  g_static_rec_mutex_unlock (&_log_message_mutex);
}

/**
 * log_textview_refresh_timeout:
 */
static gboolean
log_textview_refresh_timeout (gpointer data)
{
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (data), FALSE);

  glook_local_log_refresh (data);
  return TRUE;
}

/**
 * remove_refresh_timeout:
 */
static void
remove_refresh_timeout (GLookLocalTab *tab)
{
  guint timeout_id;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  timeout_id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (tab),
						    "refresh-timeout"));
  if (timeout_id > 0)
    g_source_remove (timeout_id);
}

/**
 * on_log_toplevel_map:
 */
void
on_log_toplevel_map (GtkWidget     *box,
			 GLookLocalTab *tab)
{
  guint timeout_id;
  GObject *buffer;
  GtkWidget *textview;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect the handler. */
  g_signal_handlers_disconnect_by_func (box, on_log_toplevel_map, tab);

  buffer = glook_local_tab_get_object (tab, "LogTextBuffer");

  /* create buffer tags. */
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (buffer), "tag-debug",
			      "foreground", "grey25", NULL);
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (buffer), "tag-message",
			      "foreground", "black", NULL);
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (buffer), "tag-warning",
			      "foreground", "blue", NULL);
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (buffer), "tag-critical",
			      "foreground", "red", NULL);
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (buffer), "tag-time",
			      "foreground", "grey30", NULL);
  gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (buffer), "tag-scale",
			      "scale", 1.0, NULL);

  /* add timeout event to refresh log textview. */
  timeout_id = g_timeout_add (500, log_textview_refresh_timeout, tab);

  g_object_set_data (G_OBJECT (tab), "refresh-timeout",
		     GUINT_TO_POINTER (timeout_id));

  /* remove the timeout event when widget destroy. */
  g_signal_connect (tab, "destroy",
		    G_CALLBACK (remove_refresh_timeout), NULL);

  /* take textview back from drag dest. */
  textview = glook_local_tab_get_widget (tab, "LogTextview");
  gtk_widget_grab_focus (textview);
}

/**
 * on_log_save_button_clicked:
 *
 * save log messages to file.
 */
void
on_log_save_button_clicked (GtkWidget     *button,
			    GLookLocalTab *tab)
{
  GtkWidget *dialog, *window;
  GtkFileChooser *chooser;
  const gchar *dir;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (!_log_message_list)
    {
      g_message (_("No message to save."));
      return;
    }
  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  dialog = gtk_file_chooser_dialog_new (_("Save log messages to file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  gtk_file_chooser_set_do_overwrite_confirmation (chooser, TRUE);
  dir = g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS);
  gtk_file_chooser_set_current_folder (chooser, dir);
  gtk_file_chooser_set_current_name (chooser, "ttyutils-glook.log");

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename;
      GString *string;
      GList *node;
      GError *error = NULL;

      filename = gtk_file_chooser_get_filename (chooser);
      if (!filename)
	{
	  gtk_widget_destroy (dialog);
	  return;
	}
      string = g_string_new (NULL);

      g_static_rec_mutex_lock (&_log_message_mutex);

      for (node = _log_message_list; node; node = node->next)
	{
	  GLookLogMessage *logmsg = node->data;
	  const gchar *level;

	  switch (logmsg->level)
	    {
	    case G_LOG_LEVEL_DEBUG:
	      level = _("Debug");
	      break;
	    case G_LOG_LEVEL_INFO:
	      level = _("Info");
	      break;
	    case G_LOG_LEVEL_MESSAGE:
	      level = _("Message");
	      break;
	    case G_LOG_LEVEL_WARNING:
	      level = _("Warning");
	      break;
	    default:
	      level = _("Critical");
	      break;
	    }
	  g_string_append_printf (string, "%s: <%s>-%s: %s\n",
				  glook_local_log_format_time (logmsg->time),
				  logmsg->domain, level, logmsg->message);
	}
      g_static_rec_mutex_unlock (&_log_message_mutex);

      g_file_set_contents (filename, string->str, -1, &error);
      if (error)
	{
	  g_message (_("Failed to save log message to file `%s': %s"),
		     filename, error->message);
	  g_error_free (error);
	}
      g_free (filename);
      g_string_free (string, TRUE);
    }
  gtk_widget_destroy (dialog);
}

/**
 * on_log_clear_button_clicked:
 *
 * clear log messages.
 */
void
on_log_clear_button_clicked (GtkWidget     *button,
			     GLookLocalTab *tab)
{
  GObject *buffer;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  g_object_set_data (G_OBJECT (tab), "last-log-message", NULL);
  glook_local_log_cleanup ();

  buffer = glook_local_tab_get_object (tab, "LogTextBuffer");
  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (buffer), "", 0);
}

/**
 * on_log_auto_scroll_button_toggled:
 */
void
on_log_auto_scroll_button_toggled (GtkToggleButton *button,
				   GLookLocalTab   *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
}

/**
 * on_log_filter_entry_changed:
 *
 * refilter the log message.
 */
void
on_log_filter_entry_changed (GtkWidget     *entry,
			     GLookLocalTab *tab)
{
  GObject *buffer;

  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  buffer = glook_local_tab_get_object (tab, "LogTextBuffer");
  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (buffer), "", 0);

  g_object_set_data (G_OBJECT (tab), "last-log-message", NULL);
}

/**
 * on_log_filter_entry_icon_pressed:
 *
 * clear the filter text.
 */
void
on_log_filter_entry_icon_pressed (GtkEntry             *entry,
				  GtkEntryIconPosition  position,
				  GdkEvent             *event,
				  GLookLocalTab        *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (position == GTK_ENTRY_ICON_SECONDARY)
    gtk_entry_set_text (entry, "");
}

/**
 * on_log_filter_entry_key_pressed:
 */
gboolean
on_log_filter_entry_key_pressed (GtkEntry      *entry,
				 GdkEventKey   *event,
				 GLookLocalTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), FALSE);

  /* clear the filter text when press ESC key. */
  if (event->keyval == GDK_Escape)
    {
      gtk_entry_set_text (entry, "");
      return TRUE;
    }
  return FALSE;
}

/**
 * on_log_level_combo_changed:
 */
void
on_log_level_combo_changed (GtkWidget     *combo,
			    GLookLocalTab *tab)
{
  GObject *buffer;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  buffer = glook_local_tab_get_object (tab, "LogTextBuffer");
  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (buffer), "", 0);

  g_object_set_data (G_OBJECT (tab), "last-log-message", NULL);
  glook_local_log_refresh (tab);
}

/**
 * glook_local_log_zoom:
 */
gboolean
glook_local_log_zoom (GLookLocalTab *tab,
		      gint           factor)
{
  GObject *buffer;
  GtkTextTagTable *tag_table;
  GtkTextTag *tag;
  gdouble scale;

  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), FALSE);

  buffer = glook_local_tab_get_object (tab, "LogTextBuffer");
  g_return_val_if_fail (GTK_IS_TEXT_BUFFER (buffer), FALSE);

  tag_table = gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (buffer));
  tag = gtk_text_tag_table_lookup (tag_table, "tag-scale");

  scale = (gdouble) factor / 100;
  g_object_set (tag, "scale", scale, NULL);

  /* clear the buffer and refresh. */
  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (buffer), "", 0);
  g_object_set_data (G_OBJECT (tab), "last-log-message", NULL);
  glook_local_log_refresh (tab);

  return TRUE;
}

/**
 * on_log_textview_scroll_event:
 */
gboolean
on_log_textview_scroll_event (GtkWidget      *textview,
			      GdkEventScroll *event,
			      GLookLocalTab  *tab)
{
  GtkWidget *window;

  g_return_val_if_fail (GLOOK_IS_LOCAL_TAB (tab), FALSE);

  window = gtk_widget_get_toplevel (textview);

  /* Ctrl + wheel mice zoom in/out the textview. */
  if (GLOOK_IS_WINDOW (window) && event->state & GDK_CONTROL_MASK)
    {
      GtkUIManager *uiman;
      GtkAction *action;

      uiman = glook_window_get_uiman (GLOOK_WINDOW (window));

      if (event->direction == GDK_SCROLL_DOWN ||
	  event->direction == GDK_SCROLL_RIGHT)
	action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomOut");
      else
	action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomIn");

      gtk_action_activate (action);
      return TRUE;
    }
  return FALSE;
}

