/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>
#include <gio/gio.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-accel-map.h"
#include "glook-tab.h"
#include "glook-local-tab.h"
#include "glook-profile-dialog.h"

/* notebook page number, must keep sync with ui file. */
enum
{
  PAGE_NO_GENERAL     = 0,
  PAGE_NO_ACCELERATOR = 1,
  PAGE_NO_ENCODING    = 2,
  PAGE_NO_TERMINAL    = 3,
  PAGE_NO_MAX         = 4,
};

/* categories treeview columns, must keep sync with ui file. */
enum
{
  CATEGORIES_COL_TITLE,
  CATEGORIES_COL_PAGEN,
  CATEGORIES_COL_TOOLTIP,
  CATEGORIES_N_COLS
};

/* accelerator treeview columns, must keep sync with ui file. */
enum
{
  ACCEL_COL_ACTION,		/* action name */
  ACCEL_COL_KEY,		/* accel key */
  ACCEL_COL_MODS,		/* accel modifier */
  ACCEL_COL_VISIBLE,		/* accelerator column visible */
  ACCEL_COL_SETTING_KEY,	/* settings key */
  ACCEL_N_COLS
};

/* terminal profile treeview columns, must keep sync with ui file. */
enum
{
  PROFILE_COL_DEFAULT,
  PROFILE_COL_PROFILE,
  PROFILE_COL_DISPLAY_NAME,
  PROFILE_COL_DESCRIPTION,
  PROFILE_N_COLS
};

typedef struct _GLookAccelItem		GLookAccelItem;

struct _GLookAccelItem
{
  const gchar *action_name;
  const gchar *setting_key;
};

static GLookAccelItem	application_actions[] =
{
    { N_("New tab"),              "new-tab-accel" },
    { N_("New window"),           "new-window-accel" },
    { N_("Playback file"),        "play-file-accel" },
    { N_("Show Control"),         "show-control-accel" },
    { N_("Close window"),         "close-window-accel" },
    { N_("Activate address bar"), "address-bar-accel" },
    { N_("Quit"),                 "quit-accel" },
};
static GLookAccelItem	tab_actions[] =
{
    { N_("Copy"),                "copy-accel" },
    { N_("Paste"),               "paste-accel" },
    { N_("Select all"),          "select-all-accel" },
    { N_("Reload"),              "reload-accel" },
    { N_("Zoom in"),             "zoom-in-accel" },
    { N_("Zoom out"),            "zoom-out-accel" },
    { N_("Zoom 100"),            "zoom-100-accel" },
    { N_("Writable"),            "writable-accel" },
    { N_("Save Terminal Screen"),"save-screen-accel" },
    { N_("Suspension"),          "suspension-accel" },
    { N_("Disconnect"),          "disconnect-accel" },
    { N_("Close tab"),           "close-tab-accel" },
    { N_("Switch to tab 1"),     "tab-1-accel" },
    { N_("Switch to tab 2"),     "tab-2-accel" },
    { N_("Switch to tab 3"),     "tab-3-accel" },
    { N_("Switch to tab 4"),     "tab-4-accel" },
    { N_("Switch to tab 5"),     "tab-5-accel" },
    { N_("Switch to tab 6"),     "tab-6-accel" },
    { N_("Switch to tab 7"),     "tab-7-accel" },
    { N_("Switch to tab 8"),     "tab-8-accel" },
    { N_("Switch to last tab"),  "tab-9-accel" },
    { N_("Switch to left tab"),  "left-tab-accel" },
    { N_("Switch to right tab"), "right-tab-accel" },
    { N_("Playback/Pause"),      "playback-accel" },
    { N_("Forward"),             "forward-accel" },
    { N_("Backward"),            "backward-accel" },
};
static GLookAccelItem	help_actions[] =
{
    { N_("Get start"),        "get-start-accel" },
    { N_("Online resources"), "resources-accel" },
};

static void	on_selected_encoding_selection_changed	(GtkTreeSelection *selection,
							 GLookLocalTab    *tab);

/**
 * on_accel_settings_changed:
 *
 * when accelerator settings changed, we need update
 * the treeview.
 */
static void
on_accel_settings_changed (GSettings     *settings,
			   const gchar   *key,
			   GtkWidget     *treeview)
{
  GtkTreeModel *model;
  GtkTreeIter iter, child_iter;
  gboolean valid, child_valid;

  g_return_if_fail (G_IS_SETTINGS (settings));
  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while (valid)
    {
      child_valid = gtk_tree_model_iter_children (model, &child_iter, &iter);
      while (child_valid)
	{
	  gchar *setting_key;

	  gtk_tree_model_get (model, &child_iter,
			      ACCEL_COL_SETTING_KEY, &setting_key, -1);
	  if (g_strcmp0 (setting_key, key) == 0)
	    {
	      gchar *accelerator;
	      guint accel_key;
	      GdkModifierType accel_mods;

	      accelerator = g_settings_get_string (settings, key);
	      gtk_accelerator_parse (accelerator, &accel_key, &accel_mods);

	      gtk_tree_store_set (GTK_TREE_STORE (model), &child_iter,
				  ACCEL_COL_KEY, accel_key,
				  ACCEL_COL_MODS, accel_mods, -1);
	      g_free (accelerator);
	      g_free (setting_key);
	      return;
	    }
	  g_free (setting_key);
	  child_valid = gtk_tree_model_iter_next (model, &child_iter);
	}
      valid = gtk_tree_model_iter_next (model, &iter);
    }
}

/**
 * disconnect_settings_changed_handler:
 */
static void
disconnect_settings_changed_handler (GtkWidget *treeview,
				     gpointer   user_data)
{
  GSettings *accel_settings;

  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));

  accel_settings = glook_app_get_settings (glook_app_get (),
					   SETTINGS_ACCELERATOR); 
  g_signal_handlers_disconnect_by_func (accel_settings,
					on_accel_settings_changed, treeview);
}

/**
 * accelerator_treeview_load:
 *
 * load all accelerators into treeview.
 */
static void
accelerator_treeview_load (GLookLocalTab *tab)
{
  struct {
    gchar          *action_name;
    GLookAccelItem *actions;
    gsize           n_actions;
  } items[] = {
      { N_("Application"), application_actions, G_N_ELEMENTS (application_actions) },
      { N_("Tab"),         tab_actions,         G_N_ELEMENTS (tab_actions) },
      { N_("Help"),        help_actions,        G_N_ELEMENTS (help_actions) },
  };
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GSettings *settings;
  gint i, j;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "AcceleratorTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  /* detach model from treeview. */
  g_object_ref (model);
  gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), NULL);

  /* clear tree store first. */
  gtk_tree_store_clear (GTK_TREE_STORE (model));

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_ACCELERATOR);

  for (i = 0; i < G_N_ELEMENTS (items); i++)
    {
      /* insert parent node. */
      gtk_tree_store_insert_with_values (GTK_TREE_STORE (model), &iter, NULL, -1,
					 ACCEL_COL_ACTION, gettext (items[i].action_name),
					 ACCEL_COL_VISIBLE, FALSE, -1);
      /* insert all child node. */
      for (j = 0; j < items[i].n_actions; j++)
	{
	  gchar *accelerator;
	  guint key = 0;
	  GdkModifierType mods = 0;

	  accelerator = g_settings_get_string (settings, items[i].actions[j].setting_key);
	  if (accelerator)
	    {
	      gtk_accelerator_parse (accelerator, &key, &mods);
	      g_free (accelerator);
	    }
	  gtk_tree_store_insert_with_values (GTK_TREE_STORE (model), NULL, &iter, -1,
					     ACCEL_COL_ACTION, gettext (items[i].actions[j].action_name),
					     ACCEL_COL_KEY, key,
					     ACCEL_COL_MODS, mods,
					     ACCEL_COL_VISIBLE, TRUE,
					     ACCEL_COL_SETTING_KEY, items[i].actions[j].setting_key,
					     -1);
	}
    }
  /* reattach model to treeview. */
  gtk_tree_view_set_model (GTK_TREE_VIEW (treeview), model);
  g_object_unref (model);

  /* expand treeview and set search column. */
  gtk_tree_view_expand_all (GTK_TREE_VIEW (treeview));
  gtk_tree_view_set_search_column (GTK_TREE_VIEW (treeview), ACCEL_COL_ACTION);

  /* connect to settings changed signal to update accelerator treeview. */
  g_signal_connect (settings, "changed",
		    G_CALLBACK (on_accel_settings_changed), treeview);
  g_signal_connect (treeview, "destroy",
		    G_CALLBACK (disconnect_settings_changed_handler), NULL);
}

/**
 * bind_widgets_settings:
 *
 * bind widgets property with gsettings.
 */
static void
bind_widgets_settings (GLookLocalTab *tab)
{
  GObject *object;
  GSettings *general_settings;
  GSettings *accel_settings;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  general_settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);
  accel_settings = glook_app_get_settings (glook_app_get (), SETTINGS_ACCELERATOR);

  /* agent address settings. */
  object = glook_local_tab_get_object (tab, "AgentAddressEntry");
  g_settings_bind (general_settings, "agent-address",
		   object, "text", G_SETTINGS_BIND_DEFAULT);
  /* connect agent on startup settings. */
  object = glook_local_tab_get_object (tab, "ConnectAgentCheck");
  g_settings_bind (general_settings, "startup-connect-agent",
		   object, "active", G_SETTINGS_BIND_DEFAULT);
  /* open bookmark on startup settings. */
  object = glook_local_tab_get_object (tab, "OpenBookmarkCheck");
  g_settings_bind (general_settings, "startup-open-bookmark",
		   object, "active", G_SETTINGS_BIND_DEFAULT);
  /* open terminal on startup settings. */
  object = glook_local_tab_get_object (tab, "OpenTerminalCheck");
  g_settings_bind (general_settings, "startup-open-terminal",
		   object, "active", G_SETTINGS_BIND_DEFAULT);
  /* window urgency hint settings. */
  object = glook_local_tab_get_object (tab, "WindowUrgencyHintCheck");
  g_settings_bind (general_settings, "window-urgency-hint",
		   object, "active", G_SETTINGS_BIND_DEFAULT);
  /* close window warning settings. */
  object = glook_local_tab_get_object (tab, "CloseWindowWarningCheck");
  g_settings_bind (general_settings, "close-window-warning",
		   object, "active", G_SETTINGS_BIND_DEFAULT);

  /* disable accelerator and mnemonic settings. */
  object = glook_local_tab_get_object (tab, "AcceleratorDisableButton");
  g_settings_bind (accel_settings, "disable-accelerators",
		   object, "active", G_SETTINGS_BIND_DEFAULT);
  object = glook_local_tab_get_object (tab, "MnemonicDisableButton");
  g_settings_bind (accel_settings, "disable-mnemonics",
		   object, "active", G_SETTINGS_BIND_DEFAULT);
}

/**
 * on_categories_tree_selection_changed:
 *
 * switch notebook page.
 */
static void
on_categories_tree_selection_changed (GtkTreeSelection *selection,
				      GLookLocalTab    *tab)
{
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GTK_IS_TREE_SELECTION (selection));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      GtkWidget *notebook;
      gint page;

      gtk_tree_model_get (model, &iter, CATEGORIES_COL_PAGEN, &page, -1);

      notebook = glook_local_tab_get_widget (tab, "PreferenceNotebook");
      gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), page);
    }
}

/**
 * connect_widget_signals:
 *
 * connect object's signals that is not connectable in glade file.
 */
static void
connect_widget_signals (GLookLocalTab *tab)
{
  GObject *object;
  GtkTreeSelection *selection;
  GtkTreePath *treepath;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* when categories treeview selection changed. we need switch
   * the notebook page. */
  object = glook_local_tab_get_object (tab, "CategoriesTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (object));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);

  g_signal_connect (selection, "changed",
		    G_CALLBACK (on_categories_tree_selection_changed), tab);

  treepath = g_object_get_data (object, "selected-page");
  if (treepath)
    {
      gtk_tree_selection_select_path (selection, treepath);
      g_object_set_data (object, "selected-page", NULL);
    }
  else
    {
      treepath = gtk_tree_path_new_first ();
      gtk_tree_selection_select_path (selection, treepath);
      gtk_tree_path_free (treepath);
    }
  /* select the first row in usable encoding treeview. */
  object = glook_local_tab_get_object (tab, "UsableEncodingTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (object));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);

  treepath = gtk_tree_path_new_first ();
  gtk_tree_selection_select_path (selection, treepath);
  gtk_tree_path_free (treepath);

  /* when selected encoding changed. we need update the remove button
   * sensitive. */
  object = glook_local_tab_get_object (tab, "SelectedEncodingTreeview");
  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (object));
  gtk_tree_selection_set_mode (selection, GTK_SELECTION_BROWSE);

  g_signal_connect (selection, "changed",
		    G_CALLBACK (on_selected_encoding_selection_changed), tab);
}

/**
 * on_preference_toplevel_map:
 * 
 * use this event handler as initial function.
 */
void
on_preference_toplevel_map (GtkWidget     *paned,
			    GLookLocalTab *tab)
{
  GtkWidget *treeview;

  g_return_if_fail (GTK_IS_PANED (paned));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect signal handler. */
  g_signal_handlers_disconnect_by_func (paned, on_preference_toplevel_map, tab);

  /* load data into accelerator treeview */
  accelerator_treeview_load (tab);

  /* bind widget property with gsettings key and connect widget signals */
  bind_widgets_settings (tab);
  connect_widget_signals (tab);

  treeview = glook_local_tab_get_widget (tab, "CategoriesTreeview");
  gtk_widget_grab_focus (treeview);
}


typedef struct _AccelCheckData	AccelCheckData;
struct _AccelCheckData
{
  guint           key;
  GdkModifierType mods;
  gboolean        exists;
};

/**
 * check_accelerator_exists:
 *
 * when modify accelerator of action, we need check whether
 * new accelerator already be occupied by other action.
 */
static void
check_accelerator_exists (gpointer        user_data,
			  const gchar    *accel_path,
			  guint           key,
			  GdkModifierType mods,
			  gboolean        changed)
{
  AccelCheckData *data = user_data;

  if (data->exists)
    return;

  if (data->key == key && data->mods == mods)
    {
      gchar *name = gtk_accelerator_name (key, mods);

      g_message (_("accelerator `%s' has been occupied by `%s'"),
		 name, accel_path);
      g_free (name);

      data->exists = TRUE;
    }
}

/**
 * change_accelerator:
 *
 * modify action's accelerator.
 */
static void
change_accelerator (GLookLocalTab   *tab,
		    const gchar     *tree_path,
		    guint            accel_key,
		    GdkModifierType  accel_mods)
{
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GSettings *settings;
  gchar *setting_key, *accelerator;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* check whether accelerator already exists. */
  if (accel_key != 0)
    {
      AccelCheckData data;

      data.exists = FALSE;
      data.key = accel_key;
      data.mods = accel_mods;

      gtk_accel_map_foreach (&data, check_accelerator_exists);
      if (data.exists)
	return;
    }
  treeview = glook_local_tab_get_widget (tab, "AcceleratorTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  if (!gtk_tree_model_get_iter_from_string (model, &iter, tree_path))
    return;

  gtk_tree_model_get (model, &iter, ACCEL_COL_SETTING_KEY, &setting_key, -1);

  accelerator = gtk_accelerator_name (accel_key, accel_mods);

  /* change settings */
  settings = glook_app_get_settings (glook_app_get (), SETTINGS_ACCELERATOR);
  g_settings_set_string (settings, setting_key, accelerator);
  g_free (setting_key);
  g_free (accelerator);

  /* change tree store. */
  gtk_tree_store_set (GTK_TREE_STORE (model), &iter,
		      ACCEL_COL_KEY, accel_key,
		      ACCEL_COL_MODS, accel_mods, -1);
}

/**
 * on_preference_accelerator_renderer_cleared:
 */
void
on_preference_accelerator_renderer_cleared (GtkCellRendererAccel *renderer,
					    const gchar          *path_string,
					    GLookLocalTab        *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  change_accelerator (tab, path_string, 0, 0);
}

/**
 * on_preference_accelerator_renderer_edited:
 */
void
on_preference_accelerator_renderer_edited (GtkCellRendererAccel *renderer,
					   const gchar          *path_string,
					   guint                 accel_key,
					   GdkModifierType       accel_mods,
					   guint                 hardware_keycode,
					   GLookLocalTab        *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  change_accelerator (tab, path_string, accel_key, accel_mods);
}

/**
 * on_preference_accelerator_save_button_clicked:
 *
 * save accelerators into file.
 */
void
on_preference_accelerator_save_button_clicked (GtkWidget     *button,
					       GLookLocalTab *tab)
{
  GtkWidget *dialog, *window;
  GtkFileChooser *chooser;
  const gchar *dir;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  dialog = gtk_file_chooser_dialog_new (_("Save accelerator to file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  gtk_file_chooser_set_do_overwrite_confirmation (chooser, TRUE);
  dir = g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS);
  gtk_file_chooser_set_current_folder (chooser, dir);
  gtk_file_chooser_set_current_name (chooser, "ttyutils-glook-accelerator.scm");

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename = gtk_file_chooser_get_filename (chooser);
      gtk_accel_map_save (filename);
      g_free (filename);
    }
  gtk_widget_destroy (dialog);
}

/**
 * on_preference_accelerator_load_button_clicked:
 *
 * load accelerators list from file.
 */
void
on_preference_accelerator_load_button_clicked (GtkWidget     *button,
					       GLookLocalTab *tab)
{
  GtkWidget *dialog, *window;
  GtkFileChooser *chooser;
  const gchar *dir;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  dialog = gtk_file_chooser_dialog_new (_("Load accelerator from file"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_OPEN,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  dir = g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS);
  gtk_file_chooser_set_current_folder (chooser, dir);

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename = gtk_file_chooser_get_filename (chooser);
      gtk_accel_map_load (filename);
      g_free (filename);
    }
  gtk_widget_destroy (dialog);
}

/**
 * on_preference_accelerator_reset_button_clicked:
 *
 * reset all accelerators to default.
 */
void
on_preference_accelerator_reset_button_clicked (GtkWidget     *button,
						GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_accel_map_reset ();
}

/**
 * on_encoding_selected_treeview_selection_changed:
 *
 * callback for selected encodings treeview selection changed
 */
static void
on_selected_encoding_selection_changed (GtkTreeSelection *selection,
					GLookLocalTab    *tab)
{
  GtkWidget *button;

  button = glook_local_tab_get_widget (tab, "EncodingRemoveButton");

  if (gtk_tree_selection_get_selected (selection, NULL, NULL))
    gtk_widget_set_sensitive (button, TRUE);
  else
    gtk_widget_set_sensitive (button, FALSE);
}

/**
 * on_preference_encoding_add_button_clicked:
 *
 * add avaliable encoding to selected encoding.
 */
void
on_preference_encoding_add_button_clicked (GtkWidget     *button,
					   GLookLocalTab *tab)
{
  GtkWidget *usable_treeview;
  GtkWidget *selected_treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model, *model2;
  GtkTreeIter iter;
  GtkTreePath *path;
  gchar *enc, *lang;
  gboolean vaild;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  usable_treeview = glook_local_tab_get_widget (tab, "UsableEncodingTreeview");
  selected_treeview = glook_local_tab_get_widget (tab, "SelectedEncodingTreeview");

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (usable_treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  gtk_tree_model_get (model, &iter, 0, &lang, 1, &enc, -1);

  model2 = gtk_tree_view_get_model (GTK_TREE_VIEW (selected_treeview));

  /* check whether encoding already exists */
  vaild = gtk_tree_model_get_iter_first (model2, &iter);
  while (vaild)
    {
      gchar *enc2;

      gtk_tree_model_get (model2, &iter, 1, &enc2, -1);
      if (g_ascii_strcasecmp (enc, enc2) == 0)
	{
	  g_free (enc2);
	  goto end;
	}
      g_free (enc2);
      vaild = gtk_tree_model_iter_next (model2, &iter);
    }
  gtk_list_store_insert_with_values (GTK_LIST_STORE (model2), &iter, -1,
				     0, lang, 1, enc, -1);
end:

  g_free (enc);
  g_free (lang);

  /* select inserted or exists row, and ensure it is visible */
  path = gtk_tree_model_get_path (model2, &iter);
  gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (selected_treeview),
				path, NULL, TRUE, 0.5, 0);
  gtk_tree_path_free (path);

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (selected_treeview));
  gtk_tree_selection_select_iter (selection, &iter);
}

/**
 * on_preference_encoding_remove_button_clicked:
 *
 * remove selected encoding.
 */
void
on_preference_encoding_remove_button_clicked (GtkWidget     *button,
					      GLookLocalTab *tab)
{
  GtkWidget *selected_treeview;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkTreePath *path;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  selected_treeview = glook_local_tab_get_widget (tab, "SelectedEncodingTreeview");

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (selected_treeview));
  if (!gtk_tree_selection_get_selected (selection, &model, &iter))
    return;

  path = gtk_tree_model_get_path (model, &iter);
  gtk_list_store_remove (GTK_LIST_STORE (model), &iter);
  gtk_tree_selection_select_path (selection, path);

  /* ensure selection is not empty if there has remain rows */
  if (!gtk_tree_selection_get_selected (selection, NULL, NULL))
    {
      if (gtk_tree_path_prev (path))
	gtk_tree_selection_select_path (selection, path);
    }
  if (gtk_tree_selection_get_selected (selection, NULL, NULL))
    {
      gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (selected_treeview),
				    path, NULL, TRUE, 0.5, 0);
    }
  gtk_tree_path_free (path);
}

/**
 * update_selected_encoding_settings:
 */
static void
update_selected_encoding_settings (GLookLocalTab *tab,
				   GtkTreeModel  *model)
{
  GtkTreeIter iter;
  gboolean valid;
  GSettings *settings;
  GVariantBuilder builder;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (GTK_IS_TREE_MODEL (model));

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);

  valid = gtk_tree_model_get_iter_first (model, &iter);

  /* if has not selected encoding, clear the settings and quit. */
  if (!valid)
    {
      g_settings_set_value (settings, "selected-encodings",
			    g_variant_new_array (G_VARIANT_TYPE ("{ss}"), NULL, 0));
      return;
    }
  g_variant_builder_init (&builder, G_VARIANT_TYPE_ARRAY);

  while (valid)
    {
      gchar *language, *encoding;

      gtk_tree_model_get (model, &iter, 0, &language, 1, &encoding, -1);

      g_variant_builder_add (&builder, "{ss}", language, encoding);
      g_free (language);
      g_free (encoding);

      valid = gtk_tree_model_iter_next (model, &iter);
    }
  g_settings_set_value (settings, "selected-encodings",
			g_variant_builder_end (&builder));
}

/**
 * on_preference_selected_encoding_row_inserted:
 *
 * when row insert into selected encoding model, we need
 * update the settings.
 */
void
on_preference_selected_encoding_row_inserted (GtkTreeModel  *model,
					      GtkTreePath   *treepath,
					      GtkTreeIter   *iter,
					      GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_TREE_MODEL (model));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  update_selected_encoding_settings (tab, model);
}

/**
 * on_preference_selected_encoding_row_deleted:
 *
 * when row deleted from selected encoding model, we need
 * update the settings.
 */
void
on_preference_selected_encoding_row_deleted (GtkTreeModel  *model,
					     GtkTreePath   *treepath,
					     GLookLocalTab *tab)
{
  g_return_if_fail (GTK_IS_TREE_MODEL (model));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  update_selected_encoding_settings (tab, model);
}

/**
 * on_preference_selected_encoding_treeview_map:
 *
 * when selected encoding treeview map, we insert selected encoding
 * from settings to treeview model.
 */
void
on_preference_selected_encoding_treeview_map (GtkTreeView   *treeview,
					      GLookLocalTab *tab)
{
  GSettings *settings;
  GVariant *value;
  GVariantIter iter;
  GtkTreeModel *model;
  gchar *language, *encoding;

  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect the handler. */
  g_signal_handlers_disconnect_by_func (treeview, on_preference_selected_encoding_treeview_map, tab);

  /* get selected encoding settings. */
  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);
  value = g_settings_get_value (settings, "selected-encodings");
  if (!value)
    return;

  model = gtk_tree_view_get_model (treeview);

  /* insert into treeview model, one by one. */
  g_variant_iter_init (&iter, value);
  while (g_variant_iter_next (&iter, "{ss}", &language, &encoding))
    {
      gtk_list_store_insert_with_values (GTK_LIST_STORE (model), NULL, -1,
					 0, language, 1, encoding, -1);
      g_free (language);
      g_free (encoding);
    }
  g_variant_unref (value);
}

/**
 * on_preference_selected_encoding_treeview_row_activated:
 */
void
on_preference_selected_encoding_treeview_row_activated (GtkTreeView       *treeview,
							GtkTreePath       *path,
							GtkTreeViewColumn *column,
							GLookLocalTab     *tab)
{
  GtkWidget *button;

  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* remove selected encoding. */
  button = glook_local_tab_get_widget (tab, "EncodingRemoveButton");
  gtk_button_clicked (GTK_BUTTON (button));
}

/**
 * on_preference_usable_encoding_treeview_row_activated:
 */
void
on_preference_usable_encoding_treeview_row_activated (GtkTreeView       *treeview,
						      GtkTreePath       *path,
						      GtkTreeViewColumn *column,
						      GLookLocalTab     *tab)
{
  GtkWidget *button;

  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* add avaliable encoding into selected encoding. */
  button = glook_local_tab_get_widget (tab, "EncodingAddButton");
  gtk_button_clicked (GTK_BUTTON (button));
}

/**
 * glook_local_preference_select_page:
 *
 * select specified notebook page.
 */
void
glook_local_preference_select_page (GLookLocalTab *tab,
				    gint           page_no,
				    const gchar   *focus_name)
{
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (page_no >= 0 && page_no < PAGE_NO_MAX);

  treeview = glook_local_tab_get_widget (tab, "CategoriesTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while (valid)
    {
      gint page;

      gtk_tree_model_get (model, &iter, CATEGORIES_COL_PAGEN, &page, -1);

      if (page == page_no)
	{
	  if (gtk_widget_get_mapped (treeview))
	    {
	      GtkTreeSelection *selection;
	      selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
	      gtk_tree_selection_select_iter (selection, &iter);
	    }
	  else
	    {
	      GtkTreePath *treepath;
	      treepath = gtk_tree_model_get_path (model, &iter);
	      g_object_set_data_full (G_OBJECT (treeview),
				      "selected-page", treepath,
				      (GDestroyNotify) gtk_tree_path_free);
	    }
	  break;
	}
      valid = gtk_tree_model_iter_next (model, &iter);
    }
  if (focus_name)
    {
      GtkWidget *widget = glook_local_tab_get_widget (tab, focus_name);
      if (GTK_IS_WIDGET (widget) && gtk_widget_get_mapped (widget))
	gtk_widget_grab_focus (widget);
    }
}

/**
 * glook_local_preference_select_general_page:
 */
void
glook_local_preference_select_general_page (GLookLocalTab *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_local_preference_select_page (tab, PAGE_NO_GENERAL,
				      "AgentAddressEntry");
}

/**
 * glook_local_preference_select_accelerator_page:
 */
void
glook_local_preference_select_accelerator_page (GLookLocalTab *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_local_preference_select_page (tab, PAGE_NO_ACCELERATOR,
				      "AcceleratorTreeview");
}

/**
 * glook_local_preference_select_encoding_page:
 */
void
glook_local_preference_select_encoding_page (GLookLocalTab *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_local_preference_select_page (tab, PAGE_NO_ENCODING,
				      "UsableEncodingTreeview");
}

/**
 * glook_local_preference_select_terminal_page:
 */
void
glook_local_preference_select_terminal_page (GLookLocalTab *tab)
{
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  glook_local_preference_select_page (tab, PAGE_NO_TERMINAL,
				      "TerminalProfileTreeview");
}

/**
 * on_preference_terminal_profile_treeview_map:
 */
void
on_preference_terminal_profile_treeview_map (GtkWidget     *treeview,
					     GLookLocalTab *tab)
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;
  GSettings *settings;
  gchar *default_profile;

  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  /* disconnect the handler */
  g_signal_handlers_disconnect_by_func (treeview,
					on_preference_terminal_profile_treeview_map, tab);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_TERMINAL);
  default_profile = g_settings_get_string (settings, "default-profile");

  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));
  valid = gtk_tree_model_get_iter_first (model, &iter);
  while (valid)
    {
      gchar *profile, *name, *desc;
      gboolean active;

      gtk_tree_model_get (model, &iter, PROFILE_COL_PROFILE, &profile, -1);

      settings = glook_app_get_profile_settings (glook_app_get (), profile);
      name = g_settings_get_string (settings, "name");
      desc = g_settings_get_string (settings, "description");

      active = (g_strcmp0 (profile, default_profile) == 0);
      g_free (profile);

      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
			  PROFILE_COL_DEFAULT, active,
			  PROFILE_COL_DISPLAY_NAME, name,
			  PROFILE_COL_DESCRIPTION, desc, -1);
      g_free (name);
      g_free (desc);

      valid = gtk_tree_model_iter_next (model, &iter);
    }
  g_free (default_profile);
}

/**
 * on_preference_terminal_profile_default_cell_toggled:
 */
void
on_preference_terminal_profile_default_cell_toggled (GtkCellRendererToggle *cell,
						     const gchar           *path,
						     GLookLocalTab         *tab)
{
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter, iter2;
  gboolean active, valid;
  gchar *profile;
  GSettings *settings;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (path);

  treeview = glook_local_tab_get_widget (tab, "TerminalProfileTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  if (!gtk_tree_model_get_iter_from_string (model, &iter, path))
    return;

  gtk_tree_model_get (model, &iter, PROFILE_COL_DEFAULT, &active, -1);
  if (active)
    return;

  valid = gtk_tree_model_get_iter_first (model, &iter2);
  while (valid)
    {
      gtk_list_store_set (GTK_LIST_STORE (model), &iter2,
			  PROFILE_COL_DEFAULT, FALSE, -1);
      valid = gtk_tree_model_iter_next (model, &iter2);
    }
  gtk_list_store_set (GTK_LIST_STORE (model), &iter, PROFILE_COL_DEFAULT, TRUE, -1);

  /* update settings. */
  settings = glook_app_get_settings (glook_app_get (), SETTINGS_TERMINAL);
  gtk_tree_model_get (model, &iter, PROFILE_COL_PROFILE, &profile, -1);
  g_settings_set_string (settings, "default-profile", profile);
  g_free (profile);
}

/**
 * on_preference_terminal_profile_name_edited:
 */
void
on_preference_terminal_profile_name_edited (GtkCellRenderer *renderer,
					    gchar           *path,
					    gchar           *new_text,
					    GLookLocalTab   *tab)
{
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (path);

  if (!new_text || strlen (new_text) == 0)
    return;

  treeview = glook_local_tab_get_widget (tab, "TerminalProfileTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  if (gtk_tree_model_get_iter_from_string (model, &iter, path))
    {
      gchar *profile;
      GSettings *settings;

      gtk_tree_model_get (model, &iter, PROFILE_COL_PROFILE, &profile, -1);

      settings = glook_app_get_profile_settings (glook_app_get (), profile);
      g_free (profile);

      g_settings_set_string (settings, "name", new_text);

      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
			  PROFILE_COL_DISPLAY_NAME, new_text, -1);
    }
}

/**
 * on_preference_terminal_profile_description_edited:
 */
void
on_preference_terminal_profile_description_edited (GtkCellRenderer *renderer,
						   gchar           *path,
						   gchar           *new_text,
						   GLookLocalTab   *tab)
{
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (path);

  if (!new_text || strlen (new_text) == 0)
    return;

  treeview = glook_local_tab_get_widget (tab, "TerminalProfileTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  if (gtk_tree_model_get_iter_from_string (model, &iter, path))
    {
      gchar *profile;
      GSettings *settings;

      gtk_tree_model_get (model, &iter, PROFILE_COL_PROFILE, &profile, -1);

      settings = glook_app_get_profile_settings (glook_app_get (), profile);
      g_free (profile);

      g_settings_set_string (settings, "description", new_text);

      gtk_list_store_set (GTK_LIST_STORE (model), &iter,
			  PROFILE_COL_DESCRIPTION, new_text, -1);
    }
}

/**
 * on_terminal_profile_dialog_responsed:
 */
static void
on_terminal_profile_dialog_responsed (GtkWidget     *dialog,
				      gint           resp_id,
				      GLookLocalTab *tab)
{
  const gchar *profile;
  GtkWidget *treeview;
  GtkTreeModel *model;
  GtkTreeIter iter;
  gboolean valid;

  g_return_if_fail (GTK_IS_DIALOG (dialog));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  if (resp_id == GTK_RESPONSE_HELP)
    return;

  /* update profile name and desciption */
  profile = glook_profile_dialog_get_profile (GLOOK_PROFILE_DIALOG (dialog));

  treeview = glook_local_tab_get_widget (tab, "TerminalProfileTreeview");
  model = gtk_tree_view_get_model (GTK_TREE_VIEW (treeview));

  valid = gtk_tree_model_get_iter_first (model, &iter);
  while (valid)
    {
      GSettings *settings;
      gchar *profile2;

      gtk_tree_model_get (model, &iter, PROFILE_COL_PROFILE, &profile2, -1);
      if (g_strcmp0 (profile, profile2) == 0)
	{
	  gchar *name, *desc;

	  settings = glook_app_get_profile_settings (glook_app_get (), profile);
	  name = g_settings_get_string (settings, "name");
	  desc = g_settings_get_string (settings, "description");

	  gtk_list_store_set (GTK_LIST_STORE (model), &iter,
			      PROFILE_COL_DISPLAY_NAME, name,
			      PROFILE_COL_DESCRIPTION, desc, -1);
	  g_free (name);
	  g_free (desc);
	  break;
	}
      valid = gtk_tree_model_iter_next (model, &iter);
    }
  gtk_widget_destroy (dialog);
}

/**
 * open_terminal_profile_dialog:
 */
static void
open_terminal_profile_dialog (GLookLocalTab *tab,
			      GtkTreeView   *treeview)
{
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;

  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));
  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));

  selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
  if (gtk_tree_selection_get_selected (selection, &model, &iter))
    {
      GtkWidget *dialog, *window;
      gchar *profile;

      gtk_tree_model_get (model, &iter, PROFILE_COL_PROFILE, &profile, -1);

      dialog = glook_profile_dialog_new ();

      window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
      gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
      gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);

      glook_profile_dialog_set_profile (GLOOK_PROFILE_DIALOG (dialog), profile,
					glook_app_get_profile_settings (glook_app_get (), profile));
      g_free (profile);

      g_signal_connect (dialog, "response",
			G_CALLBACK (on_terminal_profile_dialog_responsed), tab);
      gtk_widget_show (dialog);
    }
}

/**
 * on_preference_terminal_profile_treeview_row_activated:
 */
void
on_preference_terminal_profile_treeview_row_activated (GtkTreeView       *treeview,
						       GtkTreePath       *path,
						       GtkTreeViewColumn *column,
						       GLookLocalTab     *tab)
{
  g_return_if_fail (GTK_IS_TREE_VIEW (treeview));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  open_terminal_profile_dialog (tab, treeview);
}

/**
 * on_preference_terminal_profile_edit_button_clicked:
 */
void
on_preference_terminal_profile_edit_button_clicked (GtkWidget     *button,
						    GLookLocalTab *tab)
{
  GtkWidget *treeview;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GLOOK_IS_LOCAL_TAB (tab));

  treeview = glook_local_tab_get_widget (tab, "TerminalProfileTreeview");
  open_terminal_profile_dialog (tab, GTK_TREE_VIEW (treeview));
}
