/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include "glook-misc.h"


typedef struct _TerminalCommand	TerminalCommand;
struct _TerminalCommand
{
  const gchar *exec;
  const gchar *options;
};
static TerminalCommand	_terminals[] =
{
    { "gnome-terminal", "--geometry=${COLS}x${ROWS}+0+0 --title=\"${TITLE}\" --command=\"${COMMAND}\"" },
    { "xterm",		"-title \"${TITLE}\" -e \"${COMMAND}\"" },
    { "vte", 		"--geometry=${COLS}x${ROWS}+0+0 --command=\"${COMMAND}\"" },
    { "konsole",	"-e ${COMMAND}" },
};

/**
 * glook_macro_substitute_eval:
 *
 * return TRUE to stop subsititute.
 */
static gboolean
glook_macro_substitute_eval (const GMatchInfo *info,
			     GString          *result,
			     gpointer          data)
{
  GHashTable *table = data;
  gchar *match;
  const gchar *val;

  g_return_val_if_fail (info, TRUE);
  g_return_val_if_fail (result, TRUE);

  match = g_match_info_fetch (info, 0);
  val = g_hash_table_lookup (table, match);
  g_free (match);

  if (val)
    {
      g_string_append (result, val);
      return FALSE;
    }
  g_error (_("Macro `%s' unrecognizable."), match);
  return TRUE;
}

/**
 * glook_macro_substitute:
 */
static gchar*
glook_macro_substitute (const gchar *options,
			guint16      nrows,
			guint16      ncols,
			const gchar *title,
			const gchar *command)
{
  GRegex *regex;
  GHashTable *table;
  gchar *res;
  GError *error = NULL;

  g_return_val_if_fail (command, NULL);

  table = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_free);

  g_hash_table_insert (table, "${ROWS}", g_strdup_printf ("%d", nrows));
  g_hash_table_insert (table, "${COLS}", g_strdup_printf ("%d", ncols));
  g_hash_table_insert (table, "${TITLE}", g_strdup (title));
  g_hash_table_insert (table, "${COMMAND}", g_strdup (command));

  regex = g_regex_new ("\\$\\{[a-zA-Z0-9]+\\}",
		       G_REGEX_CASELESS, 0, &error);
  if (error)
    {
      g_error ("%s", error->message);
      return NULL;
    }
  res = g_regex_replace_eval (regex, options, -1, 0, 0,
			      glook_macro_substitute_eval,
			      table, &error);
  g_regex_unref (regex);
  g_hash_table_destroy (table);

  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return NULL;
    }
  return res;
}

/**
 * glook_launch_terminal:
 */
gboolean
glook_launch_terminal (GtkWidget   *window,
		       const gchar *command,
		       const gchar *title)
{
  gint i;
  gchar *prog, *options, *cmdline;
  GdkScreen *screen;
  GError *error = NULL;

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (command, FALSE);

  /* TODO:
   * get user prefer terminal from system configure system.
   * Currently this settings is store in GConf configure
   * system with key '/desktop/gnome/applications/terminal/exec'.
   * and GSettings unable to read GConf configure now. */

  for (i = 0; i < G_N_ELEMENTS (_terminals); i++)
    {
      prog = g_find_program_in_path (_terminals[i].exec);
      if (prog)
	break;
    }
  if (!prog)
    {
      g_warning (_("No terminal application can been found."));
      return FALSE;
    }
  /* subsititute macros to construct a command line to execute. */
  options = glook_macro_substitute (_terminals[i].options,
				    40, 80, title, command);
  cmdline = g_strdup_printf ("%s %s", prog, options);
  g_free (prog);
  g_free (options);

  /* spawn the command line. */
  screen = gtk_window_get_screen (GTK_WINDOW (window));
  gdk_spawn_command_line_on_screen (screen, cmdline, &error);
  g_free (cmdline);
  if (error)
    {
      g_warning (_("Unable to execute `%s': %s"), command,
		 error->message);
      g_error_free (error);
      return FALSE;
    }
  return TRUE;
}

/**
 * glook_launch_url:
 */
static gboolean
glook_launch_url (GtkWidget   *window,
		  const gchar *url,
		  gboolean     email_link)
{
  gchar *prog, *command;
  GdkScreen *screen;
  GError *error = NULL;

  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (url, FALSE);

  prog = g_find_program_in_path ("gnome-open");
  if (!prog)
    {
      if (email_link)
	prog = g_find_program_in_path ("xdg-email");
      else
	prog = g_find_program_in_path ("xdg-open");
    }
  if (!prog)
    {
      g_warning (_("No appropriate application can been found."));
      return FALSE;
    }
  command = g_strdup_printf ("%s %s", prog, url);
  g_free (prog);

  screen = gtk_window_get_screen (GTK_WINDOW (window));
  gdk_spawn_command_line_on_screen (screen, command, &error);
  g_free (command);
  if (error)
    {
      g_warning (_("Unable to open `%s': %s"), url, error->message);
      g_error_free (error);
      return FALSE;
    }
  return TRUE;
}

/**
 * glook_open_url:
 */
void
glook_open_url (GtkWidget   *window,
		const gchar *url)
{
  g_return_if_fail (GTK_IS_WINDOW (window));
  g_return_if_fail (url);

  /* we first try gtk_show_uri(). */
  if (!gtk_show_uri (gtk_window_get_screen (GTK_WINDOW (window)),
		     url, GDK_CURRENT_TIME, NULL))
    {
      glook_launch_url (window, url, FALSE);
    }
}

/**
 * url_link_activate_func:
 */
static void
url_link_activate_func (GtkAboutDialog *dialog,
			const gchar    *link_,
			gpointer        data)
{
  glook_open_url (GTK_WIDGET (dialog), link_);
}

/**
 * email_link_activate_func:
 */
static void
email_link_activate_func (GtkAboutDialog *dialog,
			  const gchar    *link_,
			  gpointer        data)
{
  g_return_if_fail (link_);

  if (!gtk_show_uri (gtk_window_get_screen (GTK_WINDOW (dialog)),
		     link_, GDK_CURRENT_TIME, NULL))
    {
      glook_launch_url (GTK_WIDGET (dialog), link_, TRUE);
    }
}

/**
 * glook_about:
 */
void
glook_about (GtkWidget *window)
{
  const gchar *license =
   "Copyright (c) 2008-2010 Xiaohu Wu<xiaohu417@gmail.com>\n"
   "All rights reserved.\n"
   "\n"
   "Redistribution and use in source and binary forms, with or without "
   "modification, are permitted provided that the following conditions "
   "are met:\n"
   "  1. Redistributions of source code must retain the above copyright "
   "notice, this list of conditions and the following disclaimer.\n"
   "  2. Redistributions in binary form must reproduce the above copyright "
   "notice, this list of conditions and the following disclaimer in the "
   "documentation and/or other materials provided with the distribution.\n"
   "  3. The name of the author may not be used to endorse or promote "
   "products derived from this software without specific prior written "
   "permission.\n"
   "\n"
   "THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY "
   "EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, "
   "THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A "
   "PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE "
   "AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, "
   "EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED "
   "TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, "
   "DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND "
   "ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, "
   "OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT "
   "OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF "
   "SUCH DAMAGE.";

  gchar *authors[] =
    {
      "Xiaohu Wu<xiaohu417@gmail.com>",
      NULL
    };
  gchar *documenters[] =
    {
      "Xiaohu Wu<xiaohu417@gmail.com>",
      NULL
    };
  gtk_about_dialog_set_email_hook (email_link_activate_func, NULL, NULL);
  gtk_about_dialog_set_url_hook (url_link_activate_func, NULL, NULL);

  gtk_show_about_dialog (GTK_WINDOW (window),
			 "program-name", g_get_application_name (),
			 "logo-icon-name", "application-x-ttylook",
			 "version", VERSION,
			 "authors", authors,
			 "documenters", documenters,
			 "comments", _("Watch and control user terminal session"),
			 "copyright",
			 _("Copyright (C) 2008-2010 Xiaohu Wu, all right reserved."),
			 "license", license,
			 "wrap-license", TRUE,
			 "translator-credits",
			 _("Xiaohu Wu<xiaohu417@gmail.com> Simplified Chinese 2008-2010"),
			 "website", WEBSITE,
			 NULL);
}


/* below code is take from Gtk+ source.
 *
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (C) 2004-2006 Christian Hammond
 * Copyright (C) 2008 Cody Russell
 * Copyright (C) 2008 Red Hat, Inc.
 */
#define DRAG_ICON_MAX_WIDTH 250
#define DRAG_ICON_MAX_HEIGHT 250
#define DRAG_ICON_LAYOUT_BORDER 5
#define DRAG_ICON_MAX_LINES 7
#define ELLIPSIS_CHARACTER "\xe2\x80\xa6"

static void
append_n_lines (GString *str, const gchar *text, GSList *lines, gint n_lines)
{
  PangoLayoutLine *line;
  gint i;

  for (i = 0; i < n_lines; i++)
    {
      line = lines->data;
      g_string_append_len (str, &text[line->start_index], line->length);
      lines = lines->next;
    }
}

static void
limit_layout_lines (PangoLayout *layout)
{
  const gchar *text;
  GString     *str;
  GSList      *lines, *elem;
  gint         n_lines;

  n_lines = pango_layout_get_line_count (layout);
  
  if (n_lines >= DRAG_ICON_MAX_LINES)
    {
      text  = pango_layout_get_text (layout);
      str   = g_string_new (NULL);
      lines = pango_layout_get_lines_readonly (layout);

      /* get first lines */
      elem = lines;
      append_n_lines (str, text, elem,
                      DRAG_ICON_MAX_LINES / 2);

      g_string_append (str, "\n" ELLIPSIS_CHARACTER "\n");

      /* get last lines */
      elem = g_slist_nth (lines, n_lines - DRAG_ICON_MAX_LINES / 2);
      append_n_lines (str, text, elem,
                      DRAG_ICON_MAX_LINES / 2);

      pango_layout_set_text (layout, str->str, -1);
      g_string_free (str, TRUE);
    }
}

/*
 * glook_create_text_drag_icon:
 * Creates a drag and drop icon from @text.
 *
 * take from GTK source.
 *
 * Returns: a #GdkPixmap to use as DND icon
 */
GdkPixmap *
glook_create_text_drag_icon (GtkWidget *widget, 
			     gchar     *text,
			     gsize      len)
{
  GdkDrawable  *drawable = NULL;
  PangoContext *context;
  PangoLayout  *layout;
  gint          pixmap_height, pixmap_width;
  gint          layout_width, layout_height;
  GtkStyle     *style;

  g_return_val_if_fail (widget != NULL, NULL);
  g_return_val_if_fail (text != NULL, NULL);

  context = gtk_widget_get_pango_context (widget);
  layout  = pango_layout_new (context);

  pango_layout_set_text (layout, text, len);
  pango_layout_set_wrap (layout, PANGO_WRAP_WORD_CHAR);
  pango_layout_get_size (layout, &layout_width, &layout_height);

  layout_width = MIN (layout_width, DRAG_ICON_MAX_WIDTH * PANGO_SCALE);
  pango_layout_set_width (layout, layout_width);

  limit_layout_lines (layout);

  /* get again layout extents, they may have changed */
  pango_layout_get_size (layout, &layout_width, &layout_height);

  pixmap_width  = layout_width  / PANGO_SCALE + DRAG_ICON_LAYOUT_BORDER * 2;
  pixmap_height = layout_height / PANGO_SCALE + DRAG_ICON_LAYOUT_BORDER * 2;

  drawable = gdk_pixmap_new (gtk_widget_get_window (widget),
                             pixmap_width  + 2,
                             pixmap_height + 2,
                             -1);

  style = gtk_widget_get_style (widget);
  gdk_draw_rectangle (drawable,
                      style->base_gc [gtk_widget_get_state (widget)],
                      TRUE,
                      0, 0,
                      pixmap_width + 1,
                      pixmap_height + 1);

  gdk_draw_layout (drawable,
                   style->text_gc [gtk_widget_get_state (widget)],
                   1 + DRAG_ICON_LAYOUT_BORDER,
                   1 + DRAG_ICON_LAYOUT_BORDER,
                   layout);

  gdk_draw_rectangle (drawable,
                      style->black_gc,
                      FALSE,
                      0, 0,
                      pixmap_width + 1,
                      pixmap_height + 1);

  g_object_unref (layout);

  return drawable;
}
