/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <pwd.h>
#include <errno.h>

#include <glib/gi18n.h>

#include "glook-app.h"
#include "glook-window.h"
#include "glook-tab.h"
#include "glook-terminal.h"
#include "glook-terminal-actions.h"
#include "glook-terminal-tab.h"

#ifdef G_OS_UNIX

enum
{
  PROP_0,
  PROP_ENCODING_UI,
};

struct _GLookTerminalTabPriv
{
  GtkWidget      *scrolled;
  GtkWidget      *terminal;
  GPid            child_pid;
  GtkUIManager   *popup_menu_ui;
  GtkActionGroup *ui_actions;
  guint           ui_merge_id;
  GtkAccelGroup  *ui_accel_group;
  GtkActionGroup *encoding_actions;
  gchar          *encoding_ui;
  guint           encoding_merge_id;
};
#define _TERMINAL_TAB_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_TERMINAL_TAB, GLookTerminalTabPriv))

G_DEFINE_TYPE (GLookTerminalTab, glook_terminal_tab, GLOOK_TYPE_TAB);

static void	glook_terminal_tab_open		(GLookTab     *tab);
static void	glook_terminal_tab_stop		(GLookTab     *tab);
static gboolean	glook_terminal_tab_is_active	(GLookTab     *tab);
static void	glook_terminal_tab_reload	(GLookTab     *tab);
static gboolean	glook_terminal_tab_zoom		(GLookTab     *tab,
						 gint          factor);
static void	glook_terminal_tab_merge_ui	(GLookTab      *tab,
						 GtkWidget     *window);
static void	glook_terminal_tab_unmerge_ui	(GLookTab      *tab,
						 GtkWidget     *window);
static void	glook_terminal_tab_geometry_hint(GLookTab     *tab);

static void	glook_terminal_tab_set_encoding_ui	(GLookTerminalTab *tab,
							 const gchar      *encoding_ui);
const gchar*	glook_terminal_tab_get_encoding_ui	(GLookTerminalTab *tab);

static void	on_terminal_tab_mapped		(GLookTerminalTab *tab,
						 gpointer          user_data);

static void	on_terminal_char_size_changed	(GLookTerminal    *terminal,
						 guint             width,
						 guint             height,
						 GLookTerminalTab *tab);
static void	on_terminal_paste_activated	(GtkAction        *action,
						 GLookTerminalTab *tab);
static void	on_terminal_show_popup_menu	(GLookTerminal          *terminal,
						 GLookTerminalPopupInfo *info,
						 GLookTerminalTab       *tab);
static void	on_terminal_char_size_realized	(GLookTerminal    *terminal,
						 GLookTerminalTab *tab);
static void	on_terminal_child_exited	(GLookTerminal    *terminal,
						 GLookTerminalTab *tab);
static void	on_terminal_icon_title_changed	(GLookTerminal    *terminal,
						 GLookTerminalTab *tab);
static void	on_terminal_window_title_changed(GLookTerminal    *terminal,
						 GLookTerminalTab *tab);

/* actions will be merge to tab menu. */
static GtkActionEntry _tabmenu_actions[] =
{
    { "TerminalPaste", GTK_STOCK_PASTE, NULL, NULL,
      N_("Paste text from clipboard to look server"),
      G_CALLBACK (on_terminal_paste_activated),
    },
};

static const gchar *_tabmenu_ui_string =
"<ui>"
"  <menubar name='Menubar'>"
"    <menu action='TabMenu'>"
"      <placeholder name='TabMenuPlaceholder'>"
"        <menu action='TerminalEncodingMenu'>"
"          <menuitem action='TerminalEncodingCurrent'/>"
"          <placeholder name='EncodingMenuPlaceholder'/>"
"          <separator />"
"          <menuitem action='TerminalEncodingSetting'/>"
"        </menu>"
"        <separator />"
"        <menuitem action='TerminalEditProfile'/>"
"        <menu action='TerminalProfileMenu'>"
"          <menuitem action='TerminalProfile1'/>"
"          <menuitem action='TerminalProfile2'/>"
"          <menuitem action='TerminalProfile3'/>"
"          <menuitem action='TerminalProfile4'/>"
"          <menuitem action='TerminalProfile5'/>"
"          <separator />"
"          <menuitem action='TerminalManageProfile'/>"
"        </menu>"
"        <separator />"
"        <menuitem action='TerminalSaveScreen'/>"
"        <menuitem action='TerminalReset'/>"
"        <menuitem action='TerminalResetClear'/>"
"      </placeholder>"
"    </menu>"
"  </menubar>"
"</ui>";

static const gchar *_popup_menu_ui_string =
"<ui>"
"  <popup name='TerminalPopupMenu'>"
"    <menuitem action='TerminalOpenLink'/>"
"    <menuitem action='TerminalMailto'/>"
"    <menuitem action='TerminalCallto'/>"
"    <menuitem action='TerminalCopyLink'/>"
"    <separator />"
"    <menuitem action='TerminalCopy'/>"
"    <menuitem action='TerminalPaste'/>"
"    <menuitem action='TerminalSelectAll'/>"
"    <separator />"
"    <menuitem action='TerminalEditProfile'/>"
"    <menu action='TerminalProfileMenu'>"
"      <menuitem action='TerminalProfile1'/>"
"      <menuitem action='TerminalProfile2'/>"
"      <menuitem action='TerminalProfile3'/>"
"      <menuitem action='TerminalProfile4'/>"
"      <menuitem action='TerminalProfile5'/>"
"      <separator />"
"      <menuitem action='TerminalManageProfile'/>"
"    </menu>"
"    <separator />"
"    <menuitem action='TerminalSaveScreen'/>"
"  </popup>"
"</ui>";

/**
 * glook_terminal_tab_init:
 */
static void
glook_terminal_tab_init (GLookTerminalTab *tab)
{
  GSettings *settings;
  gchar *profile;
  GError *error = NULL;

  g_object_set (tab, "image", "ttyutils-terminal",
		"label", _("Terminal"), NULL);

  tab->priv = _TERMINAL_TAB_GET_PRIVATE (tab);
  tab->priv->child_pid = -1;
  tab->priv->ui_merge_id = 0;
  tab->priv->ui_accel_group = gtk_accel_group_new ();
  tab->priv->encoding_actions = gtk_action_group_new ("TerminalTabEncodingActionGroup");;
  tab->priv->encoding_ui = NULL;
  tab->priv->encoding_merge_id = 0;

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_TERMINAL);
  profile = g_settings_get_string (settings, "default-profile");

  tab->priv->terminal = glook_terminal_new ();
  glook_terminal_set_profile (GLOOK_TERMINAL (tab->priv->terminal), profile);

  /* create a scroll window to packing terminal. */
  tab->priv->scrolled = gtk_scrolled_window_new (NULL, NULL);
  g_object_set (tab->priv->scrolled,
		"border-width", 0,
		"hscrollbar-policy", GTK_POLICY_NEVER,
		"vscrollbar-policy", GTK_POLICY_NEVER,
		"shadow-type", GTK_SHADOW_IN, NULL);
  gtk_widget_show (tab->priv->scrolled);

  gtk_container_add (GTK_CONTAINER (tab->priv->scrolled), tab->priv->terminal);
  gtk_container_add (GTK_CONTAINER (tab), tab->priv->scrolled);

  tab->priv->ui_actions = gtk_action_group_new ("TerminalTabActionGroup");
  gtk_action_group_set_translation_domain (tab->priv->ui_actions,
					   GETTEXT_PACKAGE);
  /* insert terminal actions. */
  glook_terminal_add_actions (tab->priv->ui_actions, tab->priv->terminal,
			      tab->priv->scrolled, profile);
  g_free (profile);

  gtk_action_group_add_actions (tab->priv->ui_actions, _tabmenu_actions,
				G_N_ELEMENTS (_tabmenu_actions), tab);
  /* create context menu ui. */
  tab->priv->popup_menu_ui = gtk_ui_manager_new ();
  gtk_ui_manager_insert_action_group (tab->priv->popup_menu_ui,
				      tab->priv->ui_actions, 0);
  gtk_ui_manager_add_ui_from_string (tab->priv->popup_menu_ui,
				     _popup_menu_ui_string, -1, &error);
  if (error)
    {
      g_error ("terminal context menu ui error: %s", error->message);
      return;
    }
  g_signal_connect (tab, "map", G_CALLBACK (on_terminal_tab_mapped), NULL);

  g_signal_connect (tab->priv->terminal, "show-popup-menu",
		    G_CALLBACK (on_terminal_show_popup_menu), tab);
  g_signal_connect(tab->priv->terminal, "char-size-changed",
		   G_CALLBACK(on_terminal_char_size_changed), tab);
  g_signal_connect(tab->priv->terminal, "realize",
		   G_CALLBACK(on_terminal_char_size_realized), tab);
  g_signal_connect (tab->priv->terminal, "child-exited",
		    G_CALLBACK (on_terminal_child_exited), tab);
  g_signal_connect (tab->priv->terminal, "icon-title-changed",
		    G_CALLBACK (on_terminal_icon_title_changed), tab);
  g_signal_connect (tab->priv->terminal, "window-title-changed",
		    G_CALLBACK (on_terminal_window_title_changed), tab);
  gtk_widget_show_all (GTK_WIDGET (tab));
}

/**
 * glook_terminal_tab_dispose:
 */
static void
glook_terminal_tab_dispose (GObject *object)
{
  GLookTerminalTab *tab = GLOOK_TERMINAL_TAB (object);

  if (tab->priv->ui_accel_group)
    {
      g_object_unref (tab->priv->ui_accel_group);
      tab->priv->ui_accel_group = NULL;
    }
  if (tab->priv->ui_actions)
    {
      g_object_unref (tab->priv->ui_actions);
      tab->priv->ui_actions = NULL;
    }
  if (tab->priv->encoding_actions)
    {
      g_object_unref (tab->priv->encoding_actions);
      tab->priv->encoding_actions = NULL;
    }
  G_OBJECT_CLASS (glook_terminal_tab_parent_class)->dispose (object);
}

/**
 * glook_terminal_tab_finalize:
 */
static void
glook_terminal_tab_finalize (GObject *object)
{
  GLookTerminalTab *tab = GLOOK_TERMINAL_TAB (object);

  if (tab->priv->popup_menu_ui)
    {
      g_object_unref (tab->priv->popup_menu_ui);
      tab->priv->popup_menu_ui = NULL;
    }
  if (tab->priv->encoding_ui)
    {
      g_free (tab->priv->encoding_ui);
      tab->priv->encoding_ui = NULL;
    }
  G_OBJECT_CLASS (glook_terminal_tab_parent_class)->finalize (object);
}

/**
 * glook_terminal_tab_get_property:
 */
static void
glook_terminal_tab_get_property (GObject    *object,
				 guint       property_id,
				 GValue     *value,
				 GParamSpec *pspec)
{
  GLookTerminalTab *tab = GLOOK_TERMINAL_TAB (object);

  switch (property_id)
    {
    case PROP_ENCODING_UI:
      g_value_set_string (value, glook_terminal_tab_get_encoding_ui (tab));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_terminal_tab_set_property:
 */
static void
glook_terminal_tab_set_property (GObject      *object,
				 guint         property_id,
				 const GValue *value,
				 GParamSpec   *pspec)
{
  GLookTerminalTab *tab = GLOOK_TERMINAL_TAB (object);

  switch (property_id)
    {
    case PROP_ENCODING_UI:
      glook_terminal_tab_set_encoding_ui (tab, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_terminal_tab_class_init:
 */
static void
glook_terminal_tab_class_init (GLookTerminalTabClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GLookTabClass *tab_class = GLOOK_TAB_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookTerminalTabPriv));

  gobject_class->get_property = glook_terminal_tab_get_property;
  gobject_class->set_property = glook_terminal_tab_set_property;
  gobject_class->dispose  = glook_terminal_tab_dispose;
  gobject_class->finalize = glook_terminal_tab_finalize;

  tab_class->open = glook_terminal_tab_open;
  tab_class->stop = glook_terminal_tab_stop;
  tab_class->is_active = glook_terminal_tab_is_active;
  tab_class->reload = glook_terminal_tab_reload;
  tab_class->zoom = glook_terminal_tab_zoom;
  tab_class->merge_ui = glook_terminal_tab_merge_ui;
  tab_class->unmerge_ui = glook_terminal_tab_unmerge_ui;
  tab_class->geometry_hint = glook_terminal_tab_geometry_hint;

  g_object_class_install_property (gobject_class, PROP_ENCODING_UI,
				   g_param_spec_string ("encoding-ui", "encoding-ui", "Encoding UI", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
}

/**
 * glook_terminal_tab_set_encoding_ui:
 */
static void
glook_terminal_tab_set_encoding_ui (GLookTerminalTab *tab,
				    const gchar      *encoding_ui)
{
  gchar *new_ui;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  new_ui = g_strdup (encoding_ui);
  g_free (tab->priv->encoding_ui);
  tab->priv->encoding_ui = new_ui;

  /* remerge encoding ui. */
  if (tab->priv->encoding_merge_id > 0)
    {
      GtkWidget *window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

      glook_window_unmerge_ui (GLOOK_WINDOW (window), NULL,
			       tab->priv->encoding_merge_id);
      if (tab->priv->encoding_ui)
	{
	  tab->priv->encoding_merge_id =
	   glook_window_merge_ui (GLOOK_WINDOW (window),
				  tab->priv->encoding_actions, NULL,
				  tab->priv->encoding_ui, -1);
	  if (tab->priv->encoding_merge_id < 0)
	    g_critical ("Failed to merge look tab encoding ui.");
	}
    }
}

/**
 * glook_terminal_tab_get_encoding_ui:
 */
const gchar*
glook_terminal_tab_get_encoding_ui (GLookTerminalTab *tab)
{
  g_return_val_if_fail (GLOOK_IS_TERMINAL_TAB (tab), NULL);
  return tab->priv->encoding_ui;
}

/**
 * remove_encoding_action:
 */
static void
remove_encoding_action (GtkAction      *action,
			GtkActionGroup *group)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GTK_IS_ACTION_GROUP (group));

  gtk_action_group_remove_action (group, action);
}

/**
 * on_encoding_action_activated:
 */
static void
on_encoding_action_activated (GtkAction        *action,
			      GLookTerminalTab *tab)
{
  const gchar *encoding;

  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  encoding = g_object_get_data (G_OBJECT (action), "encoding");
  vte_terminal_set_encoding (VTE_TERMINAL (tab->priv->terminal), encoding);
}

/**
 * glook_terminal_tab_build_encoding_ui:
 *
 * convert selected-encoding settings to a mergable ui string.
 */
gboolean
glook_terminal_tab_build_encoding_ui (GValue   *value,
				      GVariant *variant,
				      gpointer  user_data)
{
  GLookTerminalTab *tab = user_data;
  GList *actions;
  GString *ui_string;
  GVariantIter iter;
  gchar *language, *encoding;

  g_return_val_if_fail (GLOOK_IS_TERMINAL_TAB (tab), FALSE);

  /* clear encoding action group first. */
  actions = gtk_action_group_list_actions (tab->priv->encoding_actions);
  if (actions)
    {
      g_list_foreach (actions, (GFunc) remove_encoding_action,
		      tab->priv->encoding_actions);
      g_list_free (actions);
    }
  /* build encoding ui string. */
  ui_string = g_string_new ("<menubar name='Menubar'>"
			    "<menu action='TabMenu'>"
			    "<placeholder name='TabMenuPlaceholder'>"
			    "<menu action='TerminalEncodingMenu'>"
			    "<placeholder name='EncodingMenuPlaceholder'>");

  /* create encoding action and insert into encoding action group. */
  g_variant_iter_init (&iter, variant);
  while (g_variant_iter_next (&iter, "{ss}", &language, &encoding))
    {
      GtkAction *action;
      gchar name[256], label[512];

      g_snprintf (name, sizeof (name), "%sAction", encoding);
      g_snprintf (label, sizeof (label), "%s (_%s)", language, encoding);

      action = gtk_action_new (name, label, NULL, NULL);

      /* add ui string. */
      g_string_append_printf (ui_string, "<menuitem action='%s'/>", name);

      /* associated encoding with action. */
      g_object_set_data_full (G_OBJECT (action), "encoding", encoding, g_free);
      g_free (language);

      /* when action activated, changed terminal encoding. */
      g_signal_connect (action, "activate",
			G_CALLBACK (on_encoding_action_activated), tab);

      gtk_action_group_add_action (tab->priv->encoding_actions, action);
      g_object_unref (action);
    }
  g_string_append (ui_string, "</placeholder></menu></placeholder></menu></menubar>");

  g_value_set_string (value, ui_string->str);
  g_string_free (ui_string, TRUE);

  return TRUE;
}

/**
 * on_terminal_tab_mapped:
 */
static void
on_terminal_tab_mapped (GLookTerminalTab *tab,
			gpointer          user_data)
{
  GSettings *settings;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  /* disconnect to prevent bind more than once. */
  g_signal_handlers_disconnect_by_func (tab, on_terminal_tab_mapped, NULL);

  settings = glook_app_get_settings (glook_app_get (), SETTINGS_GENERAL);

  /* when selected-encoding settings changed, we need rebuild
   * the encoding ui. */
  g_settings_bind_with_mapping (settings, "selected-encodings",
				tab, "encoding-ui", G_SETTINGS_BIND_GET,
				glook_terminal_tab_build_encoding_ui,
				NULL, tab, NULL);
}

/**
 * get_term_scroll_get_border_size:
 */
static void
terminal_scrolled_get_border_size (GLookTerminalTab *tab,
				   gint             *width,
				   gint             *height)
{
  GtkAllocation alloc1, alloc2;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  /* get terminal scrolled window and terminal allocated size. */
  gtk_widget_get_allocation (tab->priv->scrolled, &alloc1);
  gtk_widget_get_allocation (tab->priv->terminal, &alloc2);

  /* terminal scrolled's border =
   * terminal scrolled allocated size - terminal allocaed size. */
  if (width)  *width  = alloc1.width  - alloc2.width;
  if (height) *height = alloc1.height - alloc2.height;
}

/**
 * get_terminal_maximum_size:
 */
static void
get_terminal_maximum_size (GLookTerminalTab *tab,
			   gint             *width,
			   gint             *height)
{
  GtkWidget *window;
  GtkAllocation tab_alloc;
  GdkRectangle window_frame, *workarea;
  gint scroll_width, scroll_height;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  workarea = glook_app_get_workarea (glook_app_get ());

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!GLOOK_IS_WINDOW (window))
    {
      *width  = workarea->width;
      *height = workarea->height;
      return;
    }
  /* take window frame into account. */
  glook_window_get_frame_extents (GLOOK_WINDOW (window), &window_frame);
  gtk_widget_get_allocation (GTK_WIDGET (tab), &tab_alloc);

  if (tab_alloc.width > 1 || tab_alloc.height > 1)
    {
      *width  = workarea->width - (window_frame.width - tab_alloc.width);
      *height = workarea->height - (window_frame.height - tab_alloc.height);
    }
  else
    {
      *width  = workarea->width;
      *height = workarea->height;
      return;
    }
  terminal_scrolled_get_border_size (tab, &scroll_width, &scroll_height);
  *width  -= scroll_width;
  *height -= scroll_height;
}

/**
 * glook_terminal_tab_fork_command:
 */
static gboolean
glook_terminal_tab_fork_command (GLookTerminalTab *tab,
				 gchar           **argv,
				 gchar           **envv,
				 const gchar      *workdir,
				 gboolean          lastlog,
				 gboolean          utmp,
				 gboolean          wtmp)
{
  VteTerminal *terminal;

  g_return_val_if_fail (GLOOK_IS_TERMINAL_TAB (tab), FALSE);

  terminal = VTE_TERMINAL (tab->priv->terminal);

#if VTE_MAJOR_VERSION == 0 && VTE_MINOR_VERSION == 24
  tab->priv->child_pid = vte_terminal_fork_command (terminal, argv[0], argv,
						    envv, workdir,
						    lastlog, utmp, wtmp);
  return (tab->priv->child_pid > 0);
#else
    {
      VtePtyFlags flags = VTE_PTY_DEFAULT;
      GError *error = NULL;

      if (!lastlog)
	flags |= VTE_PTY_NO_LASTLOG;
      if (!utmp)
	flags |= VTE_PTY_NO_UTMP;
      if (!wtmp)
	flags |= VTE_PTY_NO_WTMP;

      vte_terminal_fork_command_full (terminal, flags, workdir,
				      argv, envv, 0, NULL, NULL,
				      &tab->priv->child_pid, &error);
      if (error)
	{
	  g_warning ("%s", error->message);
	  g_error_free (error);
	  return FALSE;
	}
      return TRUE;
    }
#endif
}

/**
 * string_to_boolean:
 */
static inline gboolean
string_to_boolean (const gchar *string)
{
  return (g_ascii_strcasecmp (string, "true") == 0 ||
	  g_ascii_strcasecmp (string, "yes") == 0 ||
	  g_ascii_strcasecmp (string, "1") == 0);
}

/**
 * glook_terminal_tab_open:
 */
static void
glook_terminal_tab_open (GLookTab *tab)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);
  GLookUri *uri;
  const gchar *value;
  gchar **argv;
  gchar *workdir = NULL;
  gboolean lastlog = TRUE, utmp = TRUE, wtmp = TRUE;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  uri = glook_tab_get_uri (tab);
  g_return_if_fail (uri);

  /* read command line from uri. */
  value = glook_uri_lookup_attr (uri, "command");
  if (value)
    {
      gint argc;
      GError *error = NULL;

      g_shell_parse_argv (value, &argc, &argv, &error);
      if (error)
	{
	  g_warning ("%s", error->message);
	  g_error_free (error);
	  return;
	}
    }
  else
    {
      struct passwd *pswd = getpwuid (getuid ());
      if (!pswd)
	{
	  g_warning ("%s", g_strerror (errno));
	  return;
	}
      argv = g_new0 (gchar*, 2);

      argv[0] = g_strdup (pswd->pw_shell);
      argv[1] = NULL;
    }
  /* read work directory from uri. */
  if ((value = glook_uri_lookup_attr (uri, "workdir")))
    workdir = g_strdup (value);

  /* read lastlog from uri. */
  if ((value = glook_uri_lookup_attr (uri, "lastlog")))
    lastlog = string_to_boolean (value);

  /* read utmp from uri. */
  if ((value = glook_uri_lookup_attr (uri, "utmp")))
    utmp = string_to_boolean (value);

  /* read wtmp from uri. */
  if ((value = glook_uri_lookup_attr (uri, "wtmp")))
    wtmp = string_to_boolean (value);

  if (!glook_terminal_tab_fork_command (term_tab, argv, NULL, workdir,
					lastlog, utmp, wtmp))
    {
      g_warning ("fork terminal child process failed.");
      g_strfreev (argv);
      g_free (workdir);
      return;
    }
  g_strfreev (argv);
  g_free (workdir);

  gtk_widget_grab_focus (term_tab->priv->terminal);
}

/**
 * glook_terminal_tab_stop:
 */
static void
glook_terminal_tab_stop (GLookTab *tab)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);
#if VTE_MAJOR_VERSION == 0 && VTE_MINOR_VERSION == 24
  gint pty;
#else
  VtePty *pty;
#endif

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

#if VTE_MAJOR_VERSION == 0 && VTE_MINOR_VERSION == 24
  pty = vte_terminal_get_pty (VTE_TERMINAL (term_tab->priv->terminal));
  if (pty > 0)
    close (pty);
#else
  pty = vte_terminal_get_pty_object (VTE_TERMINAL (term_tab->priv->terminal));
  vte_pty_close (pty);
#endif
}

/**
 * glook_terminal_tab_is_active:
 */
static gboolean
glook_terminal_tab_is_active (GLookTab *tab)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_TERMINAL_TAB (tab), FALSE);
  return (term_tab->priv->child_pid > 0);
}

/**
 * glook_terminal_tab_reload:
 */
static void
glook_terminal_tab_reload (GLookTab *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  if (glook_terminal_tab_is_active (tab))
    return;

  glook_terminal_tab_open (tab);
}

/**
 * glook_terminal_tab_zoom:
 */
static gboolean
glook_terminal_tab_zoom (GLookTab *tab,
			 gint      factor)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);

  g_return_val_if_fail (GLOOK_IS_TERMINAL_TAB (tab), FALSE);
  g_return_val_if_fail (factor > 0, FALSE);

  glook_terminal_zoom (GLOOK_TERMINAL (term_tab->priv->terminal), factor);
  return TRUE;
}

/**
 * glook_terminal_tab_merge_ui:
 */
static void
glook_terminal_tab_merge_ui (GLookTab  *tab,
			     GtkWidget *window)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (GTK_WINDOW (window));

  /* merge tab menu. */
  term_tab->priv->ui_merge_id =
   glook_window_merge_ui (GLOOK_WINDOW (window),
			  term_tab->priv->ui_actions,
			  term_tab->priv->ui_accel_group,
			  _tabmenu_ui_string, -1);
  if (term_tab->priv->ui_merge_id < 0)
    g_critical ("Failed to merge look tab ui.");

  /* merge encoding menu. */
  if (term_tab->priv->encoding_ui)
    {
      term_tab->priv->encoding_merge_id =
       glook_window_merge_ui (GLOOK_WINDOW (window),
			      term_tab->priv->encoding_actions, NULL,
			      term_tab->priv->encoding_ui, -1);
      if (term_tab->priv->encoding_merge_id < 0)
	g_critical ("Failed to merge look tab encoding ui.");
    }
}

/**
 * glook_terminal_tab_unmerge_ui:
 */
static void
glook_terminal_tab_unmerge_ui (GLookTab  *tab,
			       GtkWidget *window)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (GTK_WINDOW (window));

  if (term_tab->priv->encoding_merge_id > 0)
    {
      glook_window_unmerge_ui (GLOOK_WINDOW (window), NULL,
			       term_tab->priv->encoding_merge_id);
      term_tab->priv->encoding_merge_id = 0;
    }
  if (term_tab->priv->ui_merge_id > 0)
    {
      glook_window_unmerge_ui (GLOOK_WINDOW (window),
			       term_tab->priv->ui_accel_group,
			       term_tab->priv->ui_merge_id);
      term_tab->priv->ui_merge_id = 0;
    }
}

/**
 * glook_terminal_tab_geometry_hint:
 */
static void
glook_terminal_tab_geometry_hint (GLookTab *tab)
{
  GLookTerminalTab *term_tab = GLOOK_TERMINAL_TAB (tab);
  VteTerminal *terminal;
  GtkWidget *window;
  GdkGeometry geometry;
  guint width, height;
  GtkBorder *inner_border;
  gint max_width, max_height;
  gint xpad = 0, ypad = 0;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  terminal = VTE_TERMINAL (term_tab->priv->terminal);

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!gtk_widget_get_realized (window))
    return;

  gtk_widget_style_get (GTK_WIDGET (terminal),
			"inner-border", &inner_border, NULL);
  if (inner_border)
    {
      xpad = inner_border->left + inner_border->right;
      ypad = inner_border->top + inner_border->bottom;
      gtk_border_free (inner_border);
    }
  width  = vte_terminal_get_char_width (terminal);
  height = vte_terminal_get_char_height (terminal);

  get_terminal_maximum_size (term_tab, &max_width, &max_height);

  geometry.width_inc   = width;
  geometry.height_inc  = height;
  geometry.base_width  = xpad;
  geometry.base_height = ypad;
  geometry.min_width   = xpad + width * 4;
  geometry.min_height  = ypad + height * 2;
  geometry.max_width   = max_width;
  geometry.max_height  = max_height;

  gtk_window_set_geometry_hints (GTK_WINDOW (window), GTK_WIDGET (terminal),
				 &geometry,
				 GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE |
				 GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE);
}

/**
 * on_terminal_paste_activated:
 */
static void
on_terminal_paste_activated (GtkAction        *action,
			     GLookTerminalTab *tab)
{
  g_return_if_fail (GTK_IS_ACTION (action));
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  vte_terminal_paste_clipboard (VTE_TERMINAL (tab->priv->terminal));
}

/**
 * popup_menu_clipboard_targets_received:
 */
static void
popup_menu_clipboard_targets_received (GtkClipboard *clipboard,
				       GdkAtom      *targets,
				       gint          ntargets,
				       gpointer      data)
{
  GLookTerminalTab *tab = GLOOK_TERMINAL_TAB (data);
  GLookTerminalPopupInfo *info;
  GtkAction *open_link_action;
  GtkAction *copy_link_action;
  GtkAction *mailto_action;
  GtkAction *callto_action;
  GtkAction *action;
  GtkWidget *menu;

  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  info = glook_terminal_get_popup_info (GLOOK_TERMINAL (tab->priv->terminal));

  /* change link actions visible. */
  open_link_action = gtk_action_group_get_action (tab->priv->ui_actions,
						  "TerminalOpenLink");
  copy_link_action = gtk_action_group_get_action (tab->priv->ui_actions,
						  "TerminalCopyLink");
  mailto_action = gtk_action_group_get_action (tab->priv->ui_actions,
					       "TerminalMailto");
  callto_action = gtk_action_group_get_action (tab->priv->ui_actions,
					       "TerminalCallto");
  gtk_action_set_visible (open_link_action, FALSE);
  gtk_action_set_visible (copy_link_action, FALSE);
  gtk_action_set_visible (mailto_action, FALSE);
  gtk_action_set_visible (callto_action, FALSE);

  if (info->address)
    {
      switch (info->uri_type)
	{
	case TERMINAL_URI_TYPE_URI:
	case TERMINAL_URI_TYPE_WEB:
	  gtk_action_set_visible (open_link_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	case TERMINAL_URI_TYPE_MAIL:
	  gtk_action_set_visible (mailto_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	case TERMINAL_URI_TYPE_CALL:
	  gtk_action_set_visible (callto_action, TRUE);
	  gtk_action_set_visible (copy_link_action, TRUE);
	  break;
	default:
	  break;
	}
    }
  /* change copy/paste action sensitive. */
  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalCopy");
  if (vte_terminal_get_has_selection (VTE_TERMINAL (tab->priv->terminal)))
    gtk_action_set_sensitive (action, TRUE);
  else
    gtk_action_set_sensitive (action, FALSE);

  action = gtk_action_group_get_action (tab->priv->ui_actions,
					"TerminalPaste");
  if (targets && gtk_targets_include_text (targets, ntargets))
    gtk_action_set_sensitive (action, TRUE);
  else
    gtk_action_set_sensitive (action, FALSE);

  menu = gtk_ui_manager_get_widget (tab->priv->popup_menu_ui,
				    "/TerminalPopupMenu");
  /* select the first item when popup activate by accelerator. */
  if (info->button == 0)
    gtk_menu_shell_select_first (GTK_MENU_SHELL (menu), FALSE);

  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 
		  info->button, info->timestamp);
}

/**
 * on_terminal_show_popup_menu:
 */
static void
on_terminal_show_popup_menu (GLookTerminal          *terminal,
			     GLookTerminalPopupInfo *info,
			     GLookTerminalTab       *tab)
{
  GtkClipboard *clipboard;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (info);
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  clipboard = gtk_widget_get_clipboard (GTK_WIDGET (terminal),
					GDK_SELECTION_CLIPBOARD);
  gtk_clipboard_request_targets (clipboard,
				 popup_menu_clipboard_targets_received, tab);
}

/**
 * on_terminal_char_size_realized:
 */
static void
on_terminal_char_size_realized (GLookTerminal    *terminal,
				GLookTerminalTab *tab)
{
  glook_terminal_tab_geometry_hint (GLOOK_TAB (tab));
}

/**
 * on_terminal_char_size_changed:
 */
static void
on_terminal_char_size_changed (GLookTerminal    *terminal,
			       guint             width,
			       guint             height,
			       GLookTerminalTab *tab)
{
  GtkWidget *window;
  GdkGeometry geometry;
  GtkBorder *inner_border;
  gint max_width, max_height;
  gint xpad = 0, ypad = 0;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));
  if (!gtk_widget_get_realized (window))
    return;

  gtk_widget_style_get (GTK_WIDGET (terminal),
			"inner-border", &inner_border, NULL);
  if (inner_border)
    {
      xpad = inner_border->left + inner_border->right;
      ypad = inner_border->top + inner_border->bottom;
      gtk_border_free (inner_border);
    }
  get_terminal_maximum_size (tab, &max_width, &max_height);

  geometry.width_inc   = width;
  geometry.height_inc  = height;
  geometry.base_width  = xpad;
  geometry.base_height = ypad;
  geometry.min_width   = xpad + width * 4;
  geometry.min_height  = ypad + height * 2;
  geometry.max_width   = max_width;
  geometry.max_height  = max_height;

  gtk_window_set_geometry_hints (GTK_WINDOW (window), GTK_WIDGET (terminal),
				 &geometry,
				 GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE |
				 GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE);
}

/**
 * on_terminal_child_exited:
 */
static void
on_terminal_child_exited (GLookTerminal    *terminal,
			  GLookTerminalTab *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  //tab->priv->child_pid = -1;
  glook_tab_close (GLOOK_TAB (tab));
}

/**
 * on_terminal_icon_title_changed:
 */
static void
on_terminal_icon_title_changed (GLookTerminal    *terminal,
				GLookTerminalTab *tab)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));
}

/**
 * on_terminal_window_title_changed:
 */
static void
on_terminal_window_title_changed (GLookTerminal    *terminal,
				  GLookTerminalTab *tab)
{
  const gchar *title, *p;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GLOOK_IS_TERMINAL_TAB (tab));

  window = gtk_widget_get_toplevel (GTK_WIDGET (tab));

  title = vte_terminal_get_window_title (VTE_TERMINAL (terminal));
  p = strchr (title, ':');

  if (p && strlen (p) > 1)
    {
      glook_tab_set_label (GLOOK_TAB (tab), p + 1);
      gtk_window_set_title (GTK_WINDOW (window), p + 1);
    }
  else
    {
      glook_tab_set_label (GLOOK_TAB (tab), title);
      gtk_window_set_title (GTK_WINDOW (window), title);
    }
}

/**
 * glook_terminal_tab_new:
 */
GtkWidget*
glook_terminal_tab_new (void)
{
  return g_object_new (GLOOK_TYPE_TERMINAL_TAB, NULL);
}

#endif /* G_OS_UNIX */
