/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-misc.h"
#include "glook-profile-dialog.h"
#include "glook-window.h"
#include "glook-ui.h"
#include "glook-terminal.h"

typedef struct _GLookTerminalRegex	GLookTerminalRegex;
typedef struct _GLookTerminalMatchTag	GLookTerminalMatchTag;
typedef struct _GLookTerminalSettings	GLookTerminalSettings;

enum
{
  PROP_0,
  PROP_PROFILE,
};

/* DND target */
enum
{
  TARGET_COLOR,
};

struct _GLookTerminalPriv
{
  gint       default_font_size;
  GSList    *match_tags;
  gchar     *profile;
  GSettings *settings;
  guint      connect_ids[128];
  GLookTerminalPopupInfo *popup_info;
};
#define _TERMINAL_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_TERMINAL, GLookTerminalPriv))

struct _GLookTerminalRegex
{
  const gchar          *pattern;
  GLookTerminalUriType  uri_type;
  GdkCursorType         cursor;
  GRegex               *regex;
};

struct _GLookTerminalMatchTag
{
  gint                 tag_id;
  GLookTerminalUriType uri_type;
};

struct _GLookTerminalSettings
{
  const gchar *settings_key;
  const gchar *prop_name;
};

/* take from gnome-terminal with few modify. */
#define USERCHARS	"-[:alnum:]"
#define USERCHARS_CLASS "[" USERCHARS "]"
#define PASSCHARS_CLASS "[-[:alnum:]\\Q,?;.:/!%$^*&~\"#'\\E]"
#define HOSTCHARS_CLASS "[-[:alnum:]]"
#define HOST		HOSTCHARS_CLASS "+(\\." HOSTCHARS_CLASS "+)*"
#define PORT		"(?:\\:[[:digit:]]{1,5})?"
#define PATHCHARS_CLASS "[-[:alnum:]\\Q_$.+!*(),;:@&=?/~#%\\E]"
#define SCHEME		"(?:news:|telnet:|nntp:|file:\\/|https?:|ftps?:|webcal:|look:|agent:|local:)"
#define USERPASS	USERCHARS_CLASS "+(?:" PASSCHARS_CLASS "+)?"
#define URLPATH		"/" PATHCHARS_CLASS "*[^\\Q]'.}>) \t\r\n,\"\\E]"

/* built-ins regex patterns and associated uri type. */
static const GLookTerminalRegex	_regex_patterns[] =
{
  { SCHEME "//(?:" USERPASS "\\@)?" HOST PORT "(?:" URLPATH ")?",
    TERMINAL_URI_TYPE_URI, GDK_HAND2 },
  { "(?:www|ftp)" HOSTCHARS_CLASS "*\\." HOST PORT "(?:" URLPATH ")?",
    TERMINAL_URI_TYPE_WEB, GDK_HAND2 },
  { "(?:callto:|h323:|sip:)" USERCHARS_CLASS "[" USERCHARS ".]*(?:" PORT "/[a-z0-9]+)?\\@" HOST, 
    TERMINAL_URI_TYPE_CALL, GDK_HAND1 },
  { "(?:mailto:)?" USERCHARS_CLASS "[" USERCHARS ".]*\\@" HOSTCHARS_CLASS "+\\." HOST,
    TERMINAL_URI_TYPE_MAIL, GDK_HAND2 },
  { "news:[[:alnum:]\\Q^_{|}~!\"#$%&'()*+,./;:=?`\\E]+",
    TERMINAL_URI_TYPE_URI, GDK_HAND2 },
};
/* precompiled regexes array. */
static GArray	*_regex_builtins = NULL;

/* settings and property binding. */
static GLookTerminalSettings	_terminal_settings[] =
{
    /* settings			 property */
    { "audible-bell",		"audible-bell" },
    { "visible-bell",		"visible-bell" },
    { "allow-bold",		"allow-bold" },
    { "word-chars",		"word-chars" },
    { "scroll-on-output",	"scroll-on-output" },
    { "scroll-on-keystroke",	"scroll-on-keystroke" },
    { "scrollback",		"scrollback-lines" },
};


G_DEFINE_TYPE (GLookTerminal, glook_terminal, VTE_TYPE_TERMINAL);

static void	glook_terminal_install_gregex		(GLookTerminal    *terminal);

static void	glook_terminal_realize			(GtkWidget        *terminal);
static void	glook_terminal_style_set		(GtkWidget        *terminal,
							 GtkStyle         *prev_style);
static gboolean	glook_terminal_button_pressed		(GtkWidget        *terminal,
							 GdkEventButton   *event);
static gboolean	glook_terminal_popup_menu		(GtkWidget        *terminal);
static gboolean	glook_terminal_scroll_event		(GtkWidget        *terminal,
							 GdkEventScroll   *event);
static void	glook_terminal_drag_data_received	(GtkWidget        *widget,
							 GdkDragContext   *drag_context,
							 gint              x,
							 gint              y,
							 GtkSelectionData *data,
							 guint             info,
							 guint             time);

static void	glook_terminal_set_font			(GLookTerminal    *terminal);
static void	glook_terminal_set_colors		(GLookTerminal    *terminal);
static void	glook_terminal_set_background		(GLookTerminal    *terminal);

static void	glook_terminal_resize_window		(VteTerminal      *terminal,
							 guint             width,
							 guint             height);
static void	glook_terminal_move_window		(VteTerminal      *terminal,
							 guint             x,
							 guint             y);
static void	glook_terminal_refresh_window		(VteTerminal      *terminal);
static void	glook_terminal_iconify_window		(VteTerminal      *terminal);
static void	glook_terminal_deiconify_window		(VteTerminal      *terminal);
static void	glook_terminal_raise_window		(VteTerminal      *terminal);
static void	glook_terminal_lower_window		(VteTerminal      *terminal);
static void	glook_terminal_maximize_window		(VteTerminal      *terminal);
static void	glook_terminal_restore_window		(VteTerminal      *terminal);


/**
 * glook_terminal_popup_info_new:
 */
GLookTerminalPopupInfo*
glook_terminal_popup_info_new (GLookTerminal *terminal)
{
  GLookTerminalPopupInfo *info;

  info = g_slice_new0 (GLookTerminalPopupInfo);

  info->refcount = 1;
  info->terminal = g_object_ref (terminal);

  return info;
}

/**
 * glook_terminal_popup_info_ref:
 */
GLookTerminalPopupInfo*
glook_terminal_popup_info_ref (GLookTerminalPopupInfo *info)
{
  g_return_val_if_fail (info, NULL);

  g_atomic_int_inc (&info->refcount);
  return info;
}

/**
 * glook_terminal_popup_info_unref:
 */
void
glook_terminal_popup_info_unref (GLookTerminalPopupInfo *info)
{
  g_return_if_fail (info);

  if (g_atomic_int_dec_and_test (&info->refcount))
    {
      g_object_unref (info->terminal);
      g_free (info->address);

      g_slice_free (GLookTerminalPopupInfo, info);
    }
}


/**
 * glook_terminal_init:
 */
static void
glook_terminal_init (GLookTerminal *terminal)
{
  const GtkTargetEntry target_table[] =
    {
	{ "application/x-color", 0, TARGET_COLOR, },
    };
  GtkTargetList *target_list;
  GtkTargetEntry *targets;
  gint n_targets;
  PangoFontDescription *fontdesc;

  terminal->priv = _TERMINAL_GET_PRIVATE (terminal);

  terminal->priv->default_font_size = 0;
  terminal->priv->match_tags = NULL;
  terminal->priv->profile = NULL;
  terminal->priv->popup_info = NULL;

  memset (terminal->priv->connect_ids, 0, sizeof (terminal->priv->connect_ids)); 

  /* default font. */
  fontdesc = pango_font_description_new ();
  pango_font_description_set_family_static (fontdesc, "monospace");

  /* set terminal default properties. */
  g_object_set (terminal,
		"pointer-autohide", TRUE,
		"cursor-blink-mode", VTE_CURSOR_BLINK_OFF,
		"font-desc", fontdesc,
		"word-chars", "-A-Za-z0-9,./?%&#:_",
		"scrollback-lines", 100, NULL);

  pango_font_description_free (fontdesc);

  /* setup DND. */
  target_list = gtk_target_list_new (NULL, 0);
  gtk_target_list_add_text_targets (target_list, 0);
  gtk_target_list_add_table (target_list, target_table, G_N_ELEMENTS (target_table));

  targets = gtk_target_table_new_from_list (target_list, &n_targets);
  gtk_drag_dest_set (GTK_WIDGET (terminal), GTK_DEST_DEFAULT_ALL,
		     targets, n_targets, GDK_ACTION_COPY | GDK_ACTION_MOVE);
  gtk_target_table_free (targets, n_targets);
  gtk_target_list_unref (target_list);

  /* install regex to match uri. */
  glook_terminal_install_gregex (terminal);
}

/**
 * glook_terminal_dispose:
 */
static void
glook_terminal_dispose (GObject *object)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (object);
  gint i;

  for (i = 0; i < G_N_ELEMENTS (terminal->priv->connect_ids); i++)
    {
      if (terminal->priv->connect_ids[i] > 0)
	{
	  g_signal_handler_disconnect (terminal->priv->settings,
				       terminal->priv->connect_ids[i]);
	  terminal->priv->connect_ids[i] = 0;
	}
    }
  if (terminal->priv->settings)
    {
      g_object_unref (terminal->priv->settings);
      terminal->priv->settings = NULL;
    }
  G_OBJECT_CLASS (glook_terminal_parent_class)->dispose (object);
}

/**
 * glook_terminal_finalize:
 */
static void
glook_terminal_finalize (GObject *object)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (object);

  if (terminal->priv->match_tags)
    {
      g_slist_foreach (terminal->priv->match_tags, (GFunc) g_free, NULL);
      g_slist_free (terminal->priv->match_tags);
    }
  if (terminal->priv->profile)
    {
      g_free (terminal->priv->profile);
      terminal->priv->profile = NULL;
    }
  if (terminal->priv->popup_info)
    {
      glook_terminal_popup_info_unref (terminal->priv->popup_info);
      terminal->priv->popup_info = NULL;
    }
  G_OBJECT_CLASS (glook_terminal_parent_class)->finalize (object);
}

/**
 * glook_terminal_get_property:
 */
static void
glook_terminal_get_property (GObject    *object,
			     guint       property_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (object);

  switch (property_id)
    {
    case PROP_PROFILE:
      g_value_set_string (value, glook_terminal_get_profile (terminal));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_terminal_set_property:
 */
static void
glook_terminal_set_property (GObject      *object,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (object);

  switch (property_id)
    {
    case PROP_PROFILE:
      glook_terminal_set_profile (terminal, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * glook_terminal_uri_clicked:
 */
static gboolean
glook_terminal_uri_clicked (GLookTerminal *terminal,
			    const gchar   *address,
			    gint           uri_type,
			    guint          modified)
{
  GtkWidget *window;

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), FALSE);

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
  glook_open_url (window, address);

  return TRUE;
}

/**
 * glook_terminal_class_init:
 */
static void
glook_terminal_class_init (GLookTerminalClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  VteTerminalClass *vte_class = VTE_TERMINAL_CLASS (klass);
  gint i;

  g_type_class_add_private (klass, sizeof (GLookTerminalPriv));

  gobject_class->get_property = glook_terminal_get_property;
  gobject_class->set_property = glook_terminal_set_property;
  gobject_class->dispose  = glook_terminal_dispose;
  gobject_class->finalize = glook_terminal_finalize;

  widget_class->realize = glook_terminal_realize;
  widget_class->style_set = glook_terminal_style_set;
  widget_class->button_press_event = glook_terminal_button_pressed;
  widget_class->popup_menu = glook_terminal_popup_menu;
  widget_class->scroll_event = glook_terminal_scroll_event;
  widget_class->drag_data_received = glook_terminal_drag_data_received;

  vte_class->resize_window    = glook_terminal_resize_window;
  vte_class->move_window      = glook_terminal_move_window;
  vte_class->refresh_window   = glook_terminal_refresh_window;
  vte_class->iconify_window   = glook_terminal_iconify_window;
  vte_class->deiconify_window = glook_terminal_deiconify_window;
  vte_class->raise_window     = glook_terminal_raise_window;
  vte_class->lower_window     = glook_terminal_lower_window;
  vte_class->maximize_window  = glook_terminal_maximize_window;
  vte_class->restore_window   = glook_terminal_restore_window;

  klass->uri_clicked = glook_terminal_uri_clicked;

  g_object_class_install_property (gobject_class, PROP_PROFILE,
				   g_param_spec_string ("profile", "Profile", "Terminal profile", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_signal_new ("uri-clicked", GLOOK_TYPE_TERMINAL, G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (GLookTerminalClass, uri_clicked),
		g_signal_accumulator_true_handled, NULL,
		t_marshal_BOOLEAN__STRING_INT_UINT,
		G_TYPE_BOOLEAN, 3,
		G_TYPE_STRING, G_TYPE_INT, G_TYPE_UINT);
  g_signal_new ("show-popup-menu", GLOOK_TYPE_TERMINAL, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__POINTER,
		G_TYPE_NONE, 1, G_TYPE_POINTER);
  g_signal_new ("drag-text-received", GLOOK_TYPE_TERMINAL, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__STRING,
		G_TYPE_NONE, 1, G_TYPE_STRING);
  g_signal_new ("size-changed", GLOOK_TYPE_TERMINAL, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		t_marshal_VOID__INT_INT,
		G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);

  /* precompile all built-ins regexes.
   * we think these regexes is _static_, and don't free them until quit. */
  g_assert (_regex_builtins == NULL);
  _regex_builtins = g_array_new (FALSE, TRUE, sizeof (GLookTerminalRegex));

  for (i = 0; i < G_N_ELEMENTS (_regex_patterns); i++)
    {
      GLookTerminalRegex regex;
      GError *error = NULL;

      regex.regex = g_regex_new (_regex_patterns[i].pattern,
				  G_REGEX_CASELESS | G_REGEX_OPTIMIZE,
				  G_REGEX_MATCH_NOTEMPTY, &error);
      if (error)
	{
	  g_error ("compile regex `%s' error: %s",
		   _regex_patterns[i].pattern, error->message);
	  g_error_free (error);
	  break;
	}
      regex.uri_type = _regex_patterns[i].uri_type;
      regex.cursor = _regex_patterns[i].cursor;

      g_array_append_val (_regex_builtins, regex);
    }
}

/**
 * glook_terminal_resize_window:
 */
static void
glook_terminal_resize_window (VteTerminal *terminal,
			      guint        width,
			      guint        height)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GLOOK_IS_WINDOW (window) && width > 2 && height > 2)
    {
      GtkBorder *inner_border;
      glong char_width, char_height;
      glong cols, rows;

      gtk_widget_style_get (GTK_WIDGET (terminal), "inner-border",
			    &inner_border, NULL);
      if (inner_border != NULL)
	{
	  width  -= inner_border->left + inner_border->right;
	  height -= inner_border->top + inner_border->bottom;
	  gtk_border_free (inner_border);
	}
      char_width  = vte_terminal_get_char_width  (VTE_TERMINAL (terminal));
      char_height = vte_terminal_get_char_height (VTE_TERMINAL (terminal));

      cols = width  / char_width;
      rows = height / char_height;

      g_signal_emit_by_name (terminal, "size-changed", rows, cols);
      return;
    }
  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->resize_window)
    klass->resize_window (terminal, width, height);
}

/**
 * glook_terminal_move_window:
 */
static void
glook_terminal_move_window (VteTerminal *terminal,
			    guint        x,
			    guint        y)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GTK_IS_WINDOW (window))
    {
      GdkWindow *gdkwindow = gtk_widget_get_window (window);
      if (gdkwindow)
	gdk_window_move (gdkwindow, x, y);

      return;
    }
  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->move_window)
    klass->move_window (terminal, x, y);
}

/**
 * glook_terminal_refresh_window:
 */
static void
glook_terminal_refresh_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GTK_IS_WINDOW (window))
    {
      GdkWindow *gdkwindow;
      GtkAllocation allocation;
      GdkRectangle rect;

      gdkwindow = gtk_widget_get_window (window);
      if (gdkwindow)
	{
	  gtk_widget_get_allocation (window, &allocation);

	  rect.x = rect.y = 0;
	  rect.width = allocation.width;
	  rect.height = allocation.height;

	  gdk_window_invalidate_rect (gdkwindow, &rect, TRUE);
	}
      return;
    }
  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->refresh_window)
    klass->refresh_window (terminal);
}

/**
 * glook_terminal_iconify_window:
 */
static void
glook_terminal_iconify_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GLOOK_IS_WINDOW (window))
    {
      gtk_window_iconify (GTK_WINDOW (window));
      return;
    }
  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->iconify_window)
    klass->iconify_window (terminal);
}

/**
 * glook_terminal_deiconify_window:
 */
static void
glook_terminal_deiconify_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GLOOK_IS_WINDOW (window))
    {
      gtk_window_deiconify (GTK_WINDOW (window));
      return;
    }
  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->deiconify_window)
    klass->deiconify_window (terminal);
}

/**
 * glook_terminal_raise_window:
 */
static void
glook_terminal_raise_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GLOOK_IS_WINDOW (window))
    {
      gtk_window_present (GTK_WINDOW (window));
      return;
    }
  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->raise_window)
    klass->raise_window (terminal);
}

/**
 * glook_terminal_lower_window:
 */
static void
glook_terminal_lower_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->lower_window)
    klass->lower_window (terminal);

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GLOOK_IS_WINDOW (window))
    {
      GdkWindow *gdkwindow = gtk_widget_get_window (window);
      if (gdkwindow)
	gdk_window_lower (gdkwindow);
    }
}

/**
 * glook_terminal_maximize_window:
 */
static void
glook_terminal_maximize_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->maximize_window)
    klass->maximize_window (terminal);

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));
  if (GLOOK_IS_WINDOW (window))
    gtk_window_maximize (GTK_WINDOW (window));
}

/**
 * glook_terminal_restore_window:
 */
static void
glook_terminal_restore_window (VteTerminal *terminal)
{
  VteTerminalClass *klass;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  klass = VTE_TERMINAL_CLASS (glook_terminal_parent_class);
  if (klass->restore_window)
    klass->restore_window (terminal);

  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (GLOOK_IS_WINDOW (window))
    gtk_window_unmaximize (GTK_WINDOW (window));
}

/**
 * glook_terminal_install_gregex:
 *
 * insert all built-in uri match regex. each pattern has
 * a associated uri type.
 */
static void
glook_terminal_install_gregex (GLookTerminal *terminal)
{
  gint i, tag_id;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  /* add all builtins regex to terminal for match. */
  for (i = 0; i < _regex_builtins->len; i++)
    {
      GLookTerminalRegex *regex;
      GLookTerminalMatchTag *tag;

      regex = &g_array_index (_regex_builtins, GLookTerminalRegex, i);

      tag_id = vte_terminal_match_add_gregex (VTE_TERMINAL (terminal),
					      regex->regex, 0);
      vte_terminal_match_set_cursor_type (VTE_TERMINAL (terminal), tag_id,
					  regex->cursor);

      tag = g_new0 (GLookTerminalMatchTag, 1);
      tag->tag_id = tag_id;
      tag->uri_type = regex->uri_type;

      terminal->priv->match_tags =
       g_slist_prepend (terminal->priv->match_tags, tag);
    }
}

/**
 * glook_terminal_check_match:
 *
 * check whether col and row has a matched string(a uri).
 * if has return the matched string and associated uri type.
 * otherwise return NULL.
 */
static gchar*
glook_terminal_check_match (GLookTerminal        *terminal,
			    gint                  col,
			    gint                  row,
			    GLookTerminalUriType *uri_type)
{
  GSList *node;
  gchar *matched;
  gint tag_id;

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), NULL);

  matched = vte_terminal_match_check (VTE_TERMINAL (terminal),
				      col, row, &tag_id);
  if (!matched)
    return NULL;

  /* find uri type associated with tag id. */
  for (node = terminal->priv->match_tags; node; node = node->next)
    {
      GLookTerminalMatchTag *tag = node->data;
      if (tag->tag_id == tag_id)
	{
	  if (uri_type)
	    *uri_type = tag->uri_type;

	  return matched;
	}
    }
  g_free (matched);
  return NULL;
}

/**
 * glook_terminal_realize:
 */
static void
glook_terminal_realize (GtkWidget *widget)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (widget);
  gint effect;
  GtkWidget *window;

  GTK_WIDGET_CLASS (glook_terminal_parent_class)->realize (widget);

  g_assert (terminal->priv->settings);

  effect = g_settings_get_int (terminal->priv->settings, "background-effect");
  window = gtk_widget_get_toplevel (widget);

  if (effect == BACKGROUND_EFFECT_TRANSPARENT &&
      GLOOK_IS_WINDOW (window) &&
      !glook_window_get_rgba_colormap (GLOOK_WINDOW (window)))
    vte_terminal_set_background_transparent (VTE_TERMINAL (widget), TRUE);
  else
    vte_terminal_set_background_transparent (VTE_TERMINAL (widget), FALSE);
}

/**
 * glook_terminal_style_set:
 *
 * when style changed, we need update the theme color and font.
 */
static void
glook_terminal_style_set (GtkWidget *widget,
			  GtkStyle  *prev_style)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (widget);
  void (* style_set) (GtkWidget*, GtkStyle*) =
   GTK_WIDGET_CLASS (glook_terminal_parent_class)->style_set;

  if (style_set)
    style_set (widget, prev_style);

  if (gtk_widget_get_realized (widget))
    {
      glook_terminal_set_font (terminal);
      glook_terminal_set_colors (terminal);
    }
}

/**
 * on_terminal_button_pressed:
 *
 * process button press event.
 */
static gboolean
glook_terminal_button_pressed (GtkWidget      *terminal,
			       GdkEventButton *event)
{
  guint modified;
  glong char_width, char_height;
  gint col, row;
  GtkBorder *inner_border;
  gchar *matched;
  GLookTerminalUriType uri_type = 0;
  GtkWidgetClass *widget_class;

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), FALSE);

  modified = event->state & gtk_accelerator_get_default_mod_mask ();

  /* get column and row corresponds to mouse pointer position. */
  glook_terminal_get_char_size (GLOOK_TERMINAL (terminal),
				&char_width, &char_height);
  gtk_widget_style_get (GTK_WIDGET (terminal), "inner-border",
			&inner_border, NULL);
  if (inner_border)
    {
      col = (event->x - inner_border->left) / char_width;
      row = (event->y - inner_border->top)  / char_height;
      gtk_border_free (inner_border);
    }
  else
    {
      col = event->x / char_width;
      row = event->y / char_height;
    }
  /* check whether click on a link address. */
  matched = glook_terminal_check_match (GLOOK_TERMINAL (terminal),
					col, row, &uri_type);
  /* open the link with flavor program if click on a uri link
   * with CTRL + left|middle button. */
  if (matched && (event->button == 1 || event->button == 2) &&
      (modified & GDK_CONTROL_MASK))
    {
      gboolean result = FALSE;

      g_signal_emit_by_name (terminal, "uri-clicked", matched,
			     uri_type, modified, &result);
      g_free (matched);

      /* don't do anything else such as select with the click. */
      if (result)
	return TRUE;
    }
  /* popup context menu if press right button without Ctrl, Shift
   * and so on. */
  if (event->button == 3 && (modified & (GDK_SHIFT_MASK |
					 GDK_CONTROL_MASK |
					 GDK_MOD1_MASK)) == 0)
    {
      GLookTerminalPopupInfo *info;
      GLookTerminal *gterminal = GLOOK_TERMINAL (terminal);

      info = glook_terminal_popup_info_new (GLOOK_TERMINAL (terminal));
      info->button = event->button;
      info->modified = modified;
      info->timestamp = event->time;
      info->address = matched;
      info->uri_type = uri_type;

      if (gterminal->priv->popup_info)
	glook_terminal_popup_info_unref (gterminal->priv->popup_info);

      gterminal->priv->popup_info = info;

      g_signal_emit_by_name (terminal, "show-popup-menu", info);
      return TRUE;
    }
  /* default behavior is to let the terminal widget deal with it. */
  widget_class = GTK_WIDGET_CLASS (glook_terminal_parent_class);
  if (widget_class->button_press_event)
    return widget_class->button_press_event (terminal, event);

  return FALSE;
}

/**
 * glook_terminal_popup_menu:
 */
static gboolean
glook_terminal_popup_menu (GtkWidget *terminal)
{
  GLookTerminalPopupInfo *info;
  GLookTerminal *gterminal = GLOOK_TERMINAL (terminal);

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), FALSE);

  info = glook_terminal_popup_info_new (GLOOK_TERMINAL (terminal));
  info->button = 0;
  info->timestamp = gtk_get_current_event_time ();

  if (gterminal->priv->popup_info)
    glook_terminal_popup_info_unref (gterminal->priv->popup_info);

  gterminal->priv->popup_info = info;

  g_signal_emit_by_name (terminal, "show-popup-menu", info);
  return TRUE;
}

/**
 * glook_terminal_scroll_event:
 */
static gboolean
glook_terminal_scroll_event (GtkWidget      *terminal,
			     GdkEventScroll *event)
{
  GtkWidget *window;
  GtkWidgetClass *widget_class;

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), FALSE);

  window = gtk_widget_get_toplevel (terminal);

  /* Ctrl + wheel mice zoom in/out the terminal. */
  if (GLOOK_IS_WINDOW (window) && event->state & GDK_CONTROL_MASK)
    {
      GtkUIManager *uiman;
      GtkAction *action;

      uiman = glook_window_get_uiman (GLOOK_WINDOW (window));

      if (event->direction == GDK_SCROLL_DOWN ||
	  event->direction == GDK_SCROLL_RIGHT)
	action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomOut");
      else
	action = glook_ui_get_window_action (GLOOK_UI (uiman), "ZoomIn");

      gtk_action_activate (action);
      return TRUE;
    }
  /* default behavior is to let the terminal widget deal with it. */
  widget_class = GTK_WIDGET_CLASS (glook_terminal_parent_class);
  if (widget_class->scroll_event)
    return widget_class->scroll_event (terminal, event);

  return FALSE;
}

/**
 * drag_text_received_idle:
 */
static gboolean
drag_text_received_idle (gpointer data)
{
  GLookTerminal *terminal = data;
  gchar *text;

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), FALSE);

  text = g_object_get_data (G_OBJECT (terminal), "drag-received-text");
  g_signal_emit_by_name (terminal, "drag-text-received", text);
  g_object_set_data (G_OBJECT (terminal), "drag-received-text", NULL);

  return FALSE;
}

/**
 * glook_terminal_drag_received_color:
 */
static void
glook_terminal_drag_received_color (GLookTerminal    *terminal,
				    GtkSelectionData *data)
{
  guint16 *rawdata;
  gint length;
  GdkColor color;
  gchar *color_string;

  rawdata = (guint16*) gtk_selection_data_get_data (data);
  length = gtk_selection_data_get_length (data);

  /* We accept drops with the wrong format, since the KDE color
   * chooser incorrectly drops application/x-color with format 8. */
  if (length != 8)
    return;

  color.red   = rawdata[0];
  color.green = rawdata[1];
  color.blue  = rawdata[2];
  /* FIXME: use opacity from data[3] */

  color_string = gdk_color_to_string (&color);
  g_settings_set_string (terminal->priv->settings,
			 "background-color", color_string);
  g_free (color_string);

  g_settings_set_boolean (terminal->priv->settings,
			  "use-system-color", FALSE);
}

/**
 * glook_terminal_drag_data_received:
 */
static void
glook_terminal_drag_data_received (GtkWidget        *widget,
				   GdkDragContext   *drag_context,
				   gint              x,
				   gint              y,
				   GtkSelectionData *data,
				   guint             info,
				   guint             time)
{
  GLookTerminal *terminal = GLOOK_TERMINAL (widget);
  GdkAtom target;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  target = gtk_selection_data_get_target (data);

#if 0
    {
      GList *tmp = drag_context->targets;

      g_print ("info: %d\n", info);
      while (tmp != NULL)
	{
	  GdkAtom atom = GDK_POINTER_TO_ATOM (tmp->data);
	  g_print ("Target: %s\n", gdk_atom_name (atom));        
	  tmp = tmp->next;
	}
      g_print ("Chosen target: %s\n", gdk_atom_name (target));
    }
#endif

  if (gtk_targets_include_text (&target, 1))
    {
      gchar *text = (gchar*) gtk_selection_data_get_text (data);
      if (text)
	{
	  g_object_set_data_full (G_OBJECT (terminal),
				  "drag-received-text", text, g_free);
	  g_idle_add (drag_text_received_idle, terminal);
	}
      gtk_drag_finish (drag_context, TRUE, FALSE, time);
      return;
    }
  switch (info)
    {
    case TARGET_COLOR:
      glook_terminal_drag_received_color (terminal, data);
      break;
    default:
      g_assert_not_reached ();
    }
  gtk_drag_finish (drag_context, TRUE, FALSE, time);
}

/**
 * glook_terminal_new:
 */
GtkWidget*
glook_terminal_new (void)
{
  return g_object_new (GLOOK_TYPE_TERMINAL, NULL);
}

/**
 * int_to_enum_settings_mapping:
 *
 * convert gsettings value to terminal enum property.
 *
 * this is a gsettings mapping function. see
 * g_settings_bind_with_mapping() for detail.
 */
static gboolean
int_to_enum_settings_mapping (GValue   *value,
			      GVariant *variant,
			      gpointer  user_data)
{
  gint venum;

  g_return_val_if_fail (value && variant, FALSE);

  g_variant_get (variant, "i", &venum);
  g_value_set_enum (value, venum);
  return TRUE;
}

/**
 * font_name_to_desc_settings_mapping:
 */
static gboolean
font_name_to_desc_settings_mapping (GValue   *value,
				    GVariant *variant,
				    gpointer  user_data)
{
  GSettings *settings = user_data;
  gboolean use_system_font;
  gchar *font_name;
  PangoFontDescription *font_desc;

  use_system_font = g_settings_get_boolean (settings, "use-system-font");
  if (use_system_font)
    font_name = g_settings_get_string (settings, "system-font");
  else
    g_variant_get (variant, "s", &font_name);

  font_desc = pango_font_description_from_string (font_name);
  g_free (font_name);

  g_value_set_boxed (value, font_desc);
  pango_font_description_free (font_desc);

  return TRUE;
}

/**
 * glook_terminal_set_font:
 */
static void
glook_terminal_set_font (GLookTerminal *terminal)
{
  gboolean use_system_font;
  gchar *font_name;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_assert (terminal->priv->settings);

  use_system_font = g_settings_get_boolean (terminal->priv->settings,
					    "use-system-font");
  if (use_system_font)
    font_name = g_settings_get_string (terminal->priv->settings, "system-font");
  else
    font_name = g_settings_get_string (terminal->priv->settings, "font");

  vte_terminal_set_font_from_string (VTE_TERMINAL (terminal), font_name);
  g_free (font_name);
}

/**
 * on_use_system_font_settings_changed:
 */
static void
on_use_system_font_settings_changed (GSettings     *settings,
				     const gchar   *key,
				     GLookTerminal *terminal)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  glook_terminal_set_font (terminal);
}

/**
 * glook_terminal_set_colors:
 */
static void
glook_terminal_set_colors (GLookTerminal *terminal)
{
  GdkColor fg, bg, palette[16];
  gchar *value, palette_key[32];
  gint i;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_assert (terminal->priv->settings);

  /* we can get the foreground and background from system default
   * settings or from user customize. */
  if (g_settings_get_boolean (terminal->priv->settings,
			      "use-system-color"))
    {
      GtkStyle *style = gtk_widget_get_style (GTK_WIDGET (terminal));

      fg = style->text[GTK_STATE_NORMAL];
      bg = style->base[GTK_STATE_NORMAL];
    }
  else
    {
      value = g_settings_get_string (terminal->priv->settings,
				     "foreground-color");
      gdk_color_parse (value, &fg);
      g_free (value);

      value = g_settings_get_string (terminal->priv->settings,
				     "background-color");
      gdk_color_parse (value, &bg);
      g_free (value);
    }
  /* get palette colors. */
  for (i = 1; i <= 16; i++)
    {
      g_snprintf (palette_key, sizeof (palette_key), "palette%d", i);
      value = g_settings_get_string (terminal->priv->settings, palette_key);
      gdk_color_parse (value, &palette[i - 1]);
      g_free (value);
    }
  /* change terminal colors. */
  vte_terminal_set_colors (VTE_TERMINAL (terminal), &fg, &bg, palette, 16);
  vte_terminal_set_background_tint_color (VTE_TERMINAL (terminal), &bg);
}

/**
 * on_color_settings_changed:
 */
static void
on_color_settings_changed (GSettings     *settings,
			   const gchar   *key,
			   GLookTerminal *terminal)
{
  g_return_if_fail (G_IS_SETTINGS (settings));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  glook_terminal_set_colors (terminal);
}

/**
 * glook_terminal_set_background:
 */
static void
glook_terminal_set_background (GLookTerminal *terminal)
{
  VteTerminal *vterminal = VTE_TERMINAL (terminal);
  gint effect;
  gboolean image_set = FALSE;
  GtkWidget *window;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_assert (terminal->priv->settings);

  effect = g_settings_get_int (terminal->priv->settings,
			       "background-effect");
  /* change terminal background image. */
  if (effect == BACKGROUND_EFFECT_IMAGE)
    {
      gchar *file;
      gboolean scroll;

      file = g_settings_get_string (terminal->priv->settings,
				    "background-image");
      if (file && file[0])
	{
	  scroll = g_settings_get_boolean (terminal->priv->settings,
					   "background-scroll");
	  vte_terminal_set_background_image_file (vterminal, file);
	  vte_terminal_set_scroll_background (vterminal, scroll);
	  image_set = TRUE;
	}
      g_free (file);
    }
  if (!image_set)
    {
      vte_terminal_set_background_image (vterminal, NULL);
      vte_terminal_set_scroll_background (vterminal, FALSE);
    }
  /* set terminal background saturation and opacity. */
  if (effect == BACKGROUND_EFFECT_IMAGE ||
      effect == BACKGROUND_EFFECT_TRANSPARENT)
    {
      gdouble saturation;

      saturation = g_settings_get_double (terminal->priv->settings,
					  "background-saturation");
      vte_terminal_set_background_saturation (vterminal, saturation);
      vte_terminal_set_opacity (vterminal, 0xffff * (1.0 - saturation));
    }
  else
    {
      vte_terminal_set_background_saturation (vterminal, 1.0);
      vte_terminal_set_opacity (vterminal, 0xffff);
    }
  /* we just need set terminal transparent when window has not rgba
   * colormap. note if window has alpha channel and we set terminal
   * background transparent, that will take root window(desktop) as
   * terminal background instead window in z-order below our window. */
  window = gtk_widget_get_toplevel (GTK_WIDGET (terminal));

  if (effect == BACKGROUND_EFFECT_TRANSPARENT &&
      GLOOK_IS_WINDOW (window) &&
      !glook_window_get_rgba_colormap (GLOOK_WINDOW (window)))
    vte_terminal_set_background_transparent (vterminal, TRUE);
  else
    vte_terminal_set_background_transparent (vterminal, FALSE);
}

/**
 * on_background_effect_settings_changed:
 */
static void
on_background_effect_settings_changed (GSettings     *settings,
				       const gchar   *key,
				       GLookTerminal *terminal)
{
  g_return_if_fail (G_IS_SETTINGS (settings));
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  glook_terminal_set_background (terminal);
}

/**
 * glook_terminal_set_profile:
 */
void
glook_terminal_set_profile (GLookTerminal *terminal,
			    const gchar   *profile)
{
  GSettings *settings;
  const PangoFontDescription *font_desc;
  gint i, j;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (profile);

  /* get settings object from profile name. */
  settings = glook_app_get_profile_settings (glook_app_get (), profile);
  if (!settings)
    return;

  /* destroy the previous settings and disconnect all signals. */
  if (terminal->priv->settings)
    {
      for (i = 0; i < G_N_ELEMENTS (terminal->priv->connect_ids); i++)
	{
	  if (terminal->priv->connect_ids[i] > 0)
	    {
	      g_signal_handler_disconnect (terminal->priv->settings,
					   terminal->priv->connect_ids[i]);
	      terminal->priv->connect_ids[i] = 0;
	    }
	}
      g_object_unref (terminal->priv->settings);
    }
  terminal->priv->settings = g_object_ref (settings);

  /* save the profile name. */
  g_free (terminal->priv->profile);
  terminal->priv->profile = g_strdup (profile);

  /* always auto hide the mouse pointer. */
  vte_terminal_set_mouse_autohide (VTE_TERMINAL (terminal), TRUE);

  /* bind settings key with terminal property. */
  for (i = 0; i < G_N_ELEMENTS (_terminal_settings); i++)
    {
      g_settings_bind (settings, _terminal_settings[i].settings_key,
		       terminal, _terminal_settings[i].prop_name,
		       G_SETTINGS_BIND_GET);
    }
  g_settings_bind_with_mapping (settings, "cursor-shape",
				terminal, "cursor-shape", G_SETTINGS_BIND_GET,
				int_to_enum_settings_mapping, NULL, NULL, NULL);
  g_settings_bind_with_mapping (settings, "cursor-blink",
				terminal, "cursor-blink-mode", G_SETTINGS_BIND_GET,
				int_to_enum_settings_mapping, NULL, NULL, NULL);
  g_settings_bind_with_mapping (settings, "backspace-binding",
				terminal, "backspace-binding", G_SETTINGS_BIND_GET,
				int_to_enum_settings_mapping, NULL, NULL, NULL);
  g_settings_bind_with_mapping (settings, "delete-binding",
				terminal, "delete-binding", G_SETTINGS_BIND_GET,
				int_to_enum_settings_mapping, NULL, NULL, NULL);
  g_settings_bind_with_mapping (settings, "font",
				terminal, "font-desc", G_SETTINGS_BIND_GET,
				font_name_to_desc_settings_mapping, NULL, settings, NULL);
  i = 0;

  /* we cannot bind use system font settings to terminal's font-desc
   * property since the gsettings doesn't enable bind more the one
   * settings to same property for same object. */
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::use-system-font",
		     G_CALLBACK (on_use_system_font_settings_changed), terminal);

  /* listen color and palette settings changed signals. */
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::use-system-color",
		     G_CALLBACK (on_color_settings_changed), terminal);
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::foreground-color",
		     G_CALLBACK (on_color_settings_changed), terminal);
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::background-color",
		     G_CALLBACK (on_color_settings_changed), terminal);
  for (j = 1; i <= 16; j++)
    {
      gchar palette_changed[32];

      g_snprintf (palette_changed, sizeof (palette_changed), "changed::palette%d", j);

      terminal->priv->connect_ids[i++] =
       g_signal_connect (settings, palette_changed,
			 G_CALLBACK (on_color_settings_changed), terminal);
    }
  /* listen background effect settings changed signals */
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::background-effect",
		     G_CALLBACK (on_background_effect_settings_changed), terminal);
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::background-image",
		     G_CALLBACK (on_background_effect_settings_changed), terminal);
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::background-scroll",
		     G_CALLBACK (on_background_effect_settings_changed), terminal);
  terminal->priv->connect_ids[i++] =
   g_signal_connect (settings, "changed::background-saturation",
		     G_CALLBACK (on_background_effect_settings_changed), terminal);

  /* get terminal color and background effect update now. */
  glook_terminal_set_colors (terminal);
  glook_terminal_set_background (terminal);

  /* save default font size for scale. */
  font_desc = vte_terminal_get_font (VTE_TERMINAL (terminal));
  terminal->priv->default_font_size = pango_font_description_get_size (font_desc);

  g_object_notify (G_OBJECT (terminal), "profile");
}

/**
 * glook_terminal_get_profile:
 */
const gchar*
glook_terminal_get_profile (GLookTerminal *terminal)
{
  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), NULL);
  return terminal->priv->profile;
}

/**
 * glook_terminal_get_popup_info:
 */
GLookTerminalPopupInfo*
glook_terminal_get_popup_info (GLookTerminal *terminal)
{
  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), NULL);
  return terminal->priv->popup_info;
}

/**
 * glook_terminal_zoom:
 */
void
glook_terminal_zoom (GLookTerminal *terminal,
		     gint           factor)
{
  const PangoFontDescription *font_desc;
  PangoFontDescription *new_font_desc;
  gint new_size;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  font_desc = vte_terminal_get_font (VTE_TERMINAL (terminal));
  new_font_desc = pango_font_description_copy (font_desc);
  new_size = terminal->priv->default_font_size * factor / 100;
  pango_font_description_set_size (new_font_desc, new_size);

  vte_terminal_set_font (VTE_TERMINAL (terminal), new_font_desc);
  pango_font_description_free (new_font_desc);
}

/**
 * get_terminal_screen_line:
 *
 * @length is characters count, not byte count.
 */
static const gchar*
get_terminal_screen_line (GLookTerminal *terminal,
			  const gchar   *text,
			  gint           length)
{
  gint columns;
  const gchar *p = text;
  gint i, j;

  g_return_val_if_fail (GLOOK_IS_TERMINAL (terminal), NULL);
  g_return_val_if_fail (text, NULL);

  if (length < 0)
    length = g_utf8_strlen (text, -1);

  /* get terminal columns */
  columns = vte_terminal_get_column_count (VTE_TERMINAL (terminal));

  /* loop until character width reach screen columns or encounter
   * a '\n' character. */
  for (i = 0, j = 0; i < columns && j < length && *p != '\0'; j++)
    {
      gunichar uch = g_utf8_get_char (p);

      i += (g_unichar_iswide_cjk (uch)) ? 2 : 1;

      if (*p == '\n')
	break;

      p = g_utf8_next_char (p);
    }
  return p;
}

/**
 * save_screen_contents_as_plain:
 *
 * save terminal screen contents to plain text file.
 */
static void
save_screen_contents_as_plain (GLookTerminal *terminal,
			       const gchar   *filename,
			       const gchar   *text,
			       gboolean       break_line)
{
  GFile *file;
  GFileOutputStream *ostream;
  const gchar *p = text;
  GError *error = NULL;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (filename);
  g_return_if_fail (text);

  /* if need not break line, we simple put all contents to file. */
  if (!break_line)
    {
      g_file_set_contents (filename, text, -1, &error);
      if (error)
	{
	  g_message (_("Unable to save file `%s': %s"), filename,
		     error->message);
	  g_error_free (error);
	}
      return;
    }
  /* create a new file or replace exists file. */
  file = g_file_new_for_path (filename);
  ostream = g_file_replace (file, NULL, FALSE,
			    G_FILE_CREATE_PRIVATE |
			    G_FILE_CREATE_REPLACE_DESTINATION,
			    NULL, &error);
  g_object_unref (file);
  if (error)
    {
      g_message (_("Unable to save file `%s': %s"), filename,
		 error->message);
      g_error_free (error);
      return;
    }
  /* write contents to file with line break. */
  while (*p != '\0')
    {
      const gchar *pn;

      /* find the line edge. line edge is determined by either
       * encounter a newline character or line length exceed the
       * screen column count. */
      pn = get_terminal_screen_line (terminal, p, -1);

      g_output_stream_write (G_OUTPUT_STREAM (ostream),
			     p, pn - p, NULL, NULL);
      g_output_stream_write (G_OUTPUT_STREAM (ostream),
			     "\n", 1, NULL, NULL);
      /* if encounter a newline character, skip it. beacuse
       * we already put a newline above. */
      if (*pn == '\n') pn++;

      /* jump to next line and rescan. */
      p = pn;
    }
  g_output_stream_close (G_OUTPUT_STREAM (ostream), NULL, NULL);
  g_object_unref (ostream);
}

/**
 * write_html_head:
 */
static void
write_html_head (GLookTerminal *terminal,
		 GOutputStream *ostream)
{
  GString *string;
#if 0
  const PangoFontDescription *fontdesc;
#endif
  const gchar *p;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (G_IS_OUTPUT_STREAM (ostream));

  string = g_string_new (NULL);

  /* doctype */
  g_string_append (string, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 "
		   "Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n");
  g_string_append (string, "<html>\n<head>\n");

  /* html head title, meta, etc. */
  p = vte_terminal_get_window_title (VTE_TERMINAL (terminal));
  g_string_append_printf (string, "<title>%s</title>\n", p);
  g_string_append_printf (string, "<meta name=\"generator\" content=\"%s/%s\">\n",
			  g_get_application_name (), VERSION);
  g_string_append (string, "<meta http-equiv=\"content-type\" "
		   "content=\"text/html; charset=UTF-8\">\n");
  g_string_append (string, "<meta name=\"keywords\" "
		   "content=\"ttyutils,look,linux,unix,terminal\">\n");
  /* style */
#if 0
  fontdesc = vte_terminal_get_font (VTE_TERMINAL (terminal));
  p = pango_font_description_get_family (fontdesc);
#else
  /* always use monospace font. */
  p = "Monospace";
#endif
  g_string_append_printf (string, "<style type=\"text/css\">\n"
			  " .body { font-family : %s; }\n"
			  "</style>\n", "Monospace");
  g_string_append (string, "</head>\n<body class=\"body\">\n");

  g_output_stream_write (ostream, string->str, strlen (string->str), NULL, NULL);
  g_string_free (string, TRUE);
}

/**
 * write_html_tail:
 */
static void
write_html_tail (GLookTerminal *terminal,
		 GOutputStream *ostream)
{
  const gchar *tail = "\n</body>\n</html>";

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (G_IS_OUTPUT_STREAM (ostream));

  g_output_stream_write (G_OUTPUT_STREAM (ostream),
			 tail, strlen (tail), NULL, NULL);
}

/**
 * terminal_attribute_equal:
 *
 * compare two vte character attributes is equal.
 */
static gboolean
terminal_attribute_equal (VteCharAttributes *a1,
			  VteCharAttributes *a2)
{
  g_return_val_if_fail (a1, FALSE);
  g_return_val_if_fail (a2, FALSE);

  return (a1->underline == a2->underline &&
	  a1->strikethrough == a2->strikethrough &&
	  gdk_color_equal (&a1->fore, &a2->fore) &&
	  gdk_color_equal (&a1->back, &a2->back));
}

/**
 * terminal_attribute_desc:
 *
 * convert a vte char attribute to HTML character
 * attribute description.
 */
static gchar*
terminal_attribute_desc (VteCharAttributes *attr)
{
  GString *string;

  g_return_val_if_fail (attr, NULL);

  string = g_string_new (NULL);

  g_string_append_printf (string, "<span style=\"color: #%02x%02x%02x; "
			  "background-color: #%02x%02x%02x;",
			  attr->fore.red / 256,
			  attr->fore.green / 256,
			  attr->fore.blue / 256,
			  attr->back.red / 256,
			  attr->back.green / 256,
			  attr->back.blue / 256);
  if (attr->underline)
    g_string_append (string, " text-decoration: underline;");
  if (attr->strikethrough)
    g_string_append (string, " text-decoration: through;");

  g_string_append (string, "\">");

  return g_string_free (string, FALSE);
}

/**
 * write_stream_with_buffer:
 */
static inline void
write_stream_with_buffer (GOutputStream *ostream,
			  GString       *string,
			  const gchar   *buffer,
			  gssize         size)
{
  g_string_append_len (string, buffer, size);

  if (string->len > 20 * 1024)
    {
      g_output_stream_write (ostream, string->str,
			     string->len, NULL, NULL);
      string->len = 0;
    }
}

/**
 * write_html_char:
 */
static void
write_html_char (const gchar       *p,
		 gunichar           uch,
		 VteCharAttributes *attr,
		 GOutputStream     *ostream,
		 GString           *string)
{
  static VteCharAttributes *prevattr = NULL;
  gchar buff[32], *span;
  const gchar *entity;

  switch (uch)
    {
    case '&':
      entity = "&amp";
      break;
    case '<':
      entity = "&lt;";
      break;
    case '>':
      entity = "&gt;";
      break;
    case '\'':
      entity = "&apos;";
      break;
    case '"':
      entity = "&quot;";
      break;
    case ' ':
      entity = "&nbsp;";
      break;
    case '\n':
      if (prevattr)
	{
	  /* we don't like the span element crossing multi-line. */
	  write_stream_with_buffer (ostream, string, "</span>", 7);
	  prevattr = NULL;
	}
      return;
    default:
      if ((0x01 <= uch && uch <= 0x08) ||
	  (0x0b <= uch && uch <= 0x0c) ||
	  (0x0e <= uch && uch <= 0x1f) ||
	  (0x7f <= uch && uch <= 0x84) ||
	  (0x86 <= uch && uch <= 0x9f))
	g_snprintf (buff, sizeof (buff), "&#x%x;", uch);
      else
	g_utf8_strncpy (buff, p, 1);

      entity = buff;
      break;
    }
  if (!prevattr)
    {
      /* start a new span element. */
      span = terminal_attribute_desc (attr);
      write_stream_with_buffer (ostream, string, span, strlen (span));
      g_free (span);

      prevattr = attr;
    }
  else
    {
      if (!terminal_attribute_equal (prevattr, attr))
	{
	  /* end previous span element and start a new span element. */
	  write_stream_with_buffer (ostream, string, "</span>", 7);
	  span = terminal_attribute_desc (attr);
	  write_stream_with_buffer (ostream, string, span, strlen (span));
	  g_free (span);

	  prevattr = attr;
	}
    }
  write_stream_with_buffer (ostream, string, entity, strlen (entity));
}

/**
 * save_screen_contents_as_html:
 */
static void
save_screen_contents_as_html (GLookTerminal *terminal,
			      const gchar   *filename,
			      const gchar   *text,
			      gboolean       break_line,
			      GArray        *attrs)
{
  GFile *file;
  GFileOutputStream *ostream;
  GString *string;
  const gchar *p = text;
  glong columns;
  gint i, j;
  GError *error = NULL;

  /* replace exists file or create new file. */
  file = g_file_new_for_path (filename);
  ostream = g_file_replace (file, NULL, FALSE,
			    G_FILE_CREATE_PRIVATE |
			    G_FILE_CREATE_REPLACE_DESTINATION,
			    NULL, &error);
  g_object_unref (file);
  if (error)
    {
      g_message (_("Unable to save file `%s': %s"), filename,
		 error->message);
      g_error_free (error);
      return;
    }
  string = g_string_sized_new (10 * 1024);

  /* put html head to begin. */
  write_html_head (terminal, G_OUTPUT_STREAM (ostream));

  columns = vte_terminal_get_column_count (VTE_TERMINAL (terminal));

  /* write contents to file. */
  for (i = 0, j = 0; i < g_utf8_strlen (text, -1); i++)
    {
      VteCharAttributes *attr;
      gunichar uch;

      uch = g_utf8_get_char (p);
      j += g_unichar_iswide_cjk (uch) ? 2 : 1;

      /* when line width exceed screen columns, we put a newline
       * sequence. */
      if (uch != '\n' && break_line && j > columns)
	{
	  write_stream_with_buffer (G_OUTPUT_STREAM (ostream), string,
				    "<br />\n", 7);
	  j = 0;
	}
      /* write html character */
      attr = &g_array_index (attrs, VteCharAttributes, p - text);
      write_html_char (p, uch, attr, G_OUTPUT_STREAM (ostream), string);

      /* when encounter a newline character, we put a newline
       * sequence. */
      if (uch == '\n')
	{
	  write_stream_with_buffer (G_OUTPUT_STREAM (ostream), string,
				    "<br />\n", 7);
	  j = 0;
	}
      p = g_utf8_next_char (p);
    }
  g_output_stream_write (G_OUTPUT_STREAM (ostream), string->str,
			 string->len, NULL, NULL);
  g_string_free (string, TRUE);

  write_html_tail (terminal, G_OUTPUT_STREAM (ostream));

  g_output_stream_close (G_OUTPUT_STREAM (ostream), NULL, NULL);
  g_object_unref (ostream);
}

/**
 * glook_terminal_save_file:
 */
void
glook_terminal_save_file (GLookTerminal *terminal,
			  const gchar   *filename,
			  const gchar   *text,
			  GArray        *attrs,
			  gint           format,
			  gboolean       break_line)
{
  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (text);

  switch (format)
    {
    case 0:	/* plain */
      save_screen_contents_as_plain (terminal, filename,
				     text, break_line);
      break;
    case 1:	/* html */
      save_screen_contents_as_html (terminal, filename, text,
				    break_line, attrs);
      break;
    default:
      g_assert_not_reached ();
    }
}

/**
 * glook_terminal_save_screen:
 * @range :  0 : current screen visible.
 *           1 : all scrollback history.
 * @format : 0 : plain text
 *           1 : html
 *
 * save terminal screen contents to file.
 */
void
glook_terminal_save_screen (GLookTerminal *terminal,
			    const gchar   *filename,
			    gint           range,
			    gint           format,
			    gboolean       break_line)
{
  gchar *text;
  GArray *attrs = NULL;
  GtkAdjustment *adj;
  glong start_row, end_row, columns;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (filename);

  /* array to store char attributes. only need for html format. */
  if (format == 1)
    attrs = g_array_new (FALSE, FALSE, sizeof (VteCharAttributes));

  /* get the text from terminal. */
  switch (range)
    {
    case 0:	/* current screen visible. */
      text = vte_terminal_get_text (VTE_TERMINAL (terminal),
				    (VteSelectionFunc) gtk_true, NULL, attrs);
      break;
    case 1:	/* all scrollback history */
      /* vte has not a interface to get scroll start and end row.
       * we make use of it's adjustment to obtain these info.
       * this is not a _standard_ way, but can work. */
      adj = vte_terminal_get_adjustment (VTE_TERMINAL (terminal));
      start_row = (glong) gtk_adjustment_get_lower (adj);
      end_row = (glong) gtk_adjustment_get_upper (adj);
      columns = vte_terminal_get_column_count (VTE_TERMINAL (terminal));

      text = vte_terminal_get_text_range (VTE_TERMINAL (terminal),
					  start_row, 0, end_row, columns,
					  (VteSelectionFunc) gtk_true,
					  NULL, attrs);
      break;
    default:
      g_critical ("Unsupported save range value %d", range);
      return;
    }
  /* save text to file with format. */
  glook_terminal_save_file (terminal, filename, text,
			    attrs, format, break_line);
  g_free (text);
  if (attrs)
    g_array_free (attrs, TRUE);
}

/**
 * on_format_combo_changed:
 */
static void
on_format_combo_changed (GtkComboBox    *combo,
			 GtkFileChooser *chooser)
{
  gint active;

  g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));

  active = gtk_combo_box_get_active (combo);
  switch (active)
    {
    case 0:
      gtk_file_chooser_set_current_name (chooser, "ttyutils-glook-screen.txt");
      break;
    case 1:
      gtk_file_chooser_set_current_name (chooser, "ttyutils-glook-screen.html");
      break;
    }
}

/**
 * glook_terminal_save_screen_dialog:
 *
 * Gtk+ 2.20.1 has bug with choose dialog in my box, when i switch
 * selection in my chooser dialog some times, the treeview will
 * dead. i must switch to other folder and back to activate it.
 * the really bad things was when I choose file A, but return
 * file B(near A) to me, or just crash with a assert(had_selection)
 * statement. my god.
 */
void
glook_terminal_save_screen_dialog (GLookTerminal *terminal,
				   GtkWidget     *window)
{
  GtkWidget *dialog;
  GtkWidget *table;
  GtkWidget *label;
  GtkComboBox *range_combo;
  GtkComboBox *format_combo;
  GtkToggleButton *break_check;
  GtkFileChooser *chooser;
  GtkFileFilter *filter;
  const gchar *dir;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));
  g_return_if_fail (GTK_IS_WINDOW (window));

  dialog = gtk_file_chooser_dialog_new (_("Save terminal screen"),
					GTK_WINDOW (window),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_SAVE,
					GTK_RESPONSE_ACCEPT,
					GTK_STOCK_CANCEL,
					GTK_RESPONSE_CANCEL, NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  gtk_file_chooser_set_do_overwrite_confirmation (chooser, TRUE);
  dir = g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS);
  gtk_file_chooser_set_current_folder (chooser, dir);
  gtk_file_chooser_set_current_name (chooser, "ttyutils-glook-screen.html");

  /* add plain text file filter */
  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("Plain text files"));
  gtk_file_filter_add_mime_type (filter, "text/plain");
  gtk_file_filter_add_pattern (filter, "*.txt");
  gtk_file_chooser_add_filter (chooser, filter);

  /* add html file filter */
  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("HTML files"));
  gtk_file_filter_add_mime_type (filter, "text/html");
  gtk_file_filter_add_pattern (filter, "*.htm");
  gtk_file_filter_add_pattern (filter, "*.html");
  gtk_file_chooser_add_filter (chooser, filter);
  gtk_file_chooser_set_filter (chooser, filter);

  /* add all files filter */
  filter = gtk_file_filter_new ();
  gtk_file_filter_set_name (filter, _("All files"));
  gtk_file_filter_add_pattern (filter, "*");
  gtk_file_chooser_add_filter (chooser, filter);

  /* add extra widgets */
  table = gtk_table_new (0, 0, FALSE);
  gtk_table_set_col_spacings (GTK_TABLE (table), 8);
  gtk_table_set_row_spacings (GTK_TABLE (table), 6);

  /* range. all of screen history vs current visible. */
  label = gtk_label_new_with_mnemonic (_("Screen ran_ge:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
		    GTK_SHRINK, GTK_FILL, 0, 0);

  range_combo = GTK_COMBO_BOX (gtk_combo_box_new_text ());
  gtk_label_set_mnemonic_widget (GTK_LABEL (label),
				 GTK_WIDGET (range_combo));
  gtk_combo_box_append_text (range_combo, _("Current visible contents"));
  gtk_combo_box_append_text (range_combo, _("All contents"));
  gtk_combo_box_set_active (range_combo, 0);
  gtk_table_attach (GTK_TABLE (table), GTK_WIDGET (range_combo), 1, 3, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);

  /* output format. plain, html. */
  label = gtk_label_new_with_mnemonic (_("Output _format:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
		    GTK_SHRINK, GTK_FILL, 0, 0);

  format_combo = GTK_COMBO_BOX (gtk_combo_box_new_text ());
  gtk_label_set_mnemonic_widget (GTK_LABEL (label),
				 GTK_WIDGET (format_combo));
  gtk_combo_box_append_text (format_combo, _("Plain text"));
  gtk_combo_box_append_text (format_combo, "HTML");
  g_signal_connect (format_combo, "changed",
		    G_CALLBACK (on_format_combo_changed), chooser);
  gtk_combo_box_set_active (format_combo, 1);
  gtk_table_attach (GTK_TABLE (table), GTK_WIDGET (format_combo), 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);

  /* break long line. */
  break_check = GTK_TOGGLE_BUTTON (gtk_check_button_new_with_mnemonic (_("Break long l_ine")));
  gtk_toggle_button_set_active (break_check, TRUE);
  gtk_widget_set_tooltip_text (GTK_WIDGET (break_check),
			       _("Break line if it's length long than screen width"));
  gtk_table_attach (GTK_TABLE (table), GTK_WIDGET (break_check), 2, 3, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);

  /* attach table to file chooser */
  gtk_widget_show_all (table);
  gtk_file_chooser_set_extra_widget (chooser, table);

  /* run as modal dialog (block). */
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      gchar *filename = gtk_file_chooser_get_filename (chooser);
      if (filename)
	{
	  GdkCursor *cursor;
	  GdkWindow *gdkwindow;

	  /* make a busy cursor, so user know we are busy. */
	  gdkwindow = gtk_widget_get_window (dialog);
	  cursor = gdk_cursor_new (GDK_WATCH);
	  gdk_window_set_cursor (gdkwindow, cursor);

	  while (gtk_events_pending ())
	    gtk_main_iteration ();

	  glook_terminal_save_screen (GLOOK_TERMINAL (terminal), filename,
				      gtk_combo_box_get_active (range_combo),
				      gtk_combo_box_get_active (format_combo),
				      gtk_toggle_button_get_active (break_check));
	  g_free (filename);
	}
    }
  gtk_widget_destroy (dialog);
}

/**
 * glook_terminal_show_not_found:
 *
 * show a server not found page like web browser.
 */
void
glook_terminal_show_not_found (GLookTerminal *terminal,
			       const gchar   *address)
{
  gchar *str, *p;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_reset (VTE_TERMINAL (terminal), TRUE, TRUE);

  vte_terminal_feed (VTE_TERMINAL (terminal), "\r\n\r\n", -1);

  p = g_strdup_printf (_("Sorry! Unable to connect server `%s'."),
		       address);
  str = g_strdup_printf ("  \033[%dm%s\033[%dm", 1, p, 0);
  g_free (p);
  vte_terminal_feed (VTE_TERMINAL (terminal), str, -1);
  g_free (str);

  str = g_strdup_printf ("\r\n\r\n  %s\r\n"
			 "    *. %s\r\n"
			 "    *. %s\r\n"
			 "    *. %s\r\n\r\n\r\n"
			 "  \033[%dmTtyutils\033[%dm - \033[%dm%s\033[%dm"
			 "\r\n",
			 _("Suggest:"),
			 _("Check the server address is right;"),
			 _("Check the server is actived;"),
			 _("Check the look service is enabled;"),
			 35, 0, 34, WEBSITE, 0);
  vte_terminal_feed (VTE_TERMINAL (terminal), str, -1);
  g_free (str);
}

/**
 * glook_terminal_show_hangup:
 */
void
glook_terminal_show_hangup (GLookTerminal *terminal)
{
  gchar *str;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_reset (VTE_TERMINAL (terminal), TRUE, TRUE);

  str = g_strdup_printf ("\r\n  \033[%dm%s\033[%dm\r\n\r\n"
			 "  \033[%dmTtyutils\033[%dm - \033[%dm%s\033[%dm\r\n",
			 1, _("Disconnect from server!"), 0,
			 35, 0, 34, WEBSITE, 0);
  vte_terminal_feed (VTE_TERMINAL (terminal), str, -1);
  g_free (str);
}

/**
 * glook_terminal_show_play_over:
 */
void
glook_terminal_show_play_over (GLookTerminal *terminal,
			       gdouble        elapsed)
{
  gchar *str, buff[128], buff2[128];

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  vte_terminal_reset (VTE_TERMINAL (terminal), TRUE, TRUE);

  g_snprintf (buff, sizeof (buff), "\033[%dm%s\033[%dm",
	      34, _("Reload"), 0);
  g_snprintf (buff2, sizeof (buff2),
	      _("Press %s to replay this file"), buff);

  str = g_strdup_printf ("\r\n  \033[%dm%s %.2f %s\033[%dm"
			 "\r\n\r\n  %s\r\n"
			 "    *. %s\r\n\r\n\r\n"
			 "  \033[%dmTtyutils\033[%dm - \033[%dm%s\033[%dm\r\n",
			 1, _("Elapsed"), elapsed, _("seconds"), 0,
			 _("Suggest:"), buff2, 35, 0, 34, WEBSITE, 0);
  vte_terminal_feed (VTE_TERMINAL (terminal), str, -1);
  g_free (str);
}

/**
 * glook_terminal_get_size:
 */
void
glook_terminal_get_size (GLookTerminal *terminal,
			 glong         *cols,
			 glong         *rows)
{
  VteTerminal *term;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  term = VTE_TERMINAL (terminal);

  if (cols) *cols = vte_terminal_get_column_count (term);
  if (rows) *rows = vte_terminal_get_row_count (term);
}

/**
 * glook_terminal_get_char_size:
 */
void
glook_terminal_get_char_size (GLookTerminal *terminal,
			      glong         *char_width,
			      glong         *char_height)
{
  VteTerminal *term;

  g_return_if_fail (GLOOK_IS_TERMINAL (terminal));

  term = VTE_TERMINAL (terminal);

  if (char_width)  *char_width  = vte_terminal_get_char_width (term);
  if (char_height) *char_height = vte_terminal_get_char_height (term);
}
