/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdlib.h>
#include <glib/gi18n.h>
#include <ttyutils/ttyutils.h>

#include "glook-uri.h"

/* below character will not be escaped when appear in the uri. */
#define _DELIMITERS	"\t :/?#[]@!$&'()*+,;=\\<>"

/* uri pattern. */
#define _SCHEME		"(?>(?<scheme>blank|agent|look|play|terminal|local):)"
#define _ATTR		"[-\\w]+=[-\\w\\pL\\pN/\\.\\\\:\t ]+"
#define _ATTRPAIR	"(?<attrs>" _ATTR "(," _ATTR ")*" ")?"
#define _URI_PATTERN	"^" _SCHEME _ATTRPAIR "$"

typedef struct _GLookUriSchemeName	GLookUriSchemeName;

struct _GLookUri
{
  gint            refcount;
  gchar          *scheme;
  GLookUriScheme  scheme_type;
  gchar          *attrs;
  GHashTable     *attrs_table;
};

struct _GLookUriSchemeName
{
  const gchar    *name;
  GLookUriScheme  scheme;
};

static GLookUriSchemeName	_scheme_supported[] =
{
    { "blank",		GLOOK_URI_SCHEME_BLANK, },
    { "look",		GLOOK_URI_SCHEME_LOOK, },
    { "agent",		GLOOK_URI_SCHEME_AGENT, },
    { "play",		GLOOK_URI_SCHEME_PLAY, },
#ifdef G_OS_UNIX
    { "terminal",	GLOOK_URI_SCHEME_TERMINAL, },
#endif
    { "local",		GLOOK_URI_SCHEME_LOCAL, },
};

/**
 * glook_uri_scheme_name_to_type:
 */
GLookUriScheme
glook_uri_scheme_name_to_type (const gchar *scheme)
{
  gint i;

  g_return_val_if_fail (scheme, GLOOK_URI_SCHEME_UNKNOWN);

  for (i = 0; i < G_N_ELEMENTS (_scheme_supported); i++)
    {
      if (g_ascii_strcasecmp (scheme, _scheme_supported[i].name) == 0)
	return _scheme_supported[i].scheme;
    }
  return GLOOK_URI_SCHEME_UNKNOWN;
}

/**
 * glook_uri_scheme_name_from_type:
 */
const gchar*
glook_uri_scheme_name_from_type (GLookUriScheme scheme_type)
{
  gint i;

  for (i = 0; i < G_N_ELEMENTS (_scheme_supported); i++)
    {
      if (scheme_type == _scheme_supported[i].scheme)
	return _scheme_supported[i].name;
    }
  return "unknown";
}

/**
 * glook_uri_scheme_name_support:
 */
gboolean
glook_uri_scheme_name_support (const gchar *scheme)
{
  return glook_uri_scheme_name_to_type (scheme) !=
         GLOOK_URI_SCHEME_UNKNOWN;
}

/**
 * glook_uri_new:
 *
 * construct a empty URI object. later we can insert
 * information to this URI.
 */
GLookUri*
glook_uri_new (void)
{
  GLookUri *uri = g_slice_new0 (GLookUri);

  uri->refcount = 1;
  uri->attrs_table =
   g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

  return uri;
}

/**
 * glook_uri_ref:
 */
GLookUri*
glook_uri_ref (GLookUri *uri)
{
  g_return_val_if_fail (uri, NULL);

  g_atomic_int_inc (&uri->refcount);
  return uri;
}

/**
 * glook_uri_unref:
 */
void
glook_uri_unref (GLookUri *uri)
{
  if (g_atomic_int_dec_and_test (&uri->refcount))
    glook_uri_free (uri);
}

/**
 * glook_uri_free:
 *
 * free URI object resources.
 */
void
glook_uri_free (GLookUri *uri)
{
  if (uri)
    {
      g_free (uri->scheme);
      g_free (uri->attrs);

      if (uri->attrs_table)
	g_hash_table_destroy (uri->attrs_table);

      g_slice_free (GLookUri, uri);
    }
}

/**
 * glook_uri_get_match_string:
 */
static gchar*
glook_uri_get_match_string (GMatchInfo  *match_info,
			    const gchar *name)
{
  gchar *match;

  match = g_match_info_fetch_named (match_info, name);
  if (!match)
    return NULL;

  /* match 0 bytes is not matched. */
  if (strlen (match) == 0)
    {
      g_free (match);
      return NULL;
    }
  return match;
}

/**
 * glook_uri_parse_attrs:
 *
 * parse attrs to a/v pairs and insert to hash table.
 */
static void
glook_uri_parse_attrs (GLookUri *uri)
{
  gchar **strv;
  gint i;

  g_return_if_fail (uri);
  g_return_if_fail (uri->attrs);

  strv = g_strsplit_set (uri->attrs, "=,", -1);
  g_assert (g_strv_length (strv) % 2 == 0);

  for (i = 0; strv[i]; i += 2)
    {
      g_hash_table_replace (uri->attrs_table, g_strdup (strv[i]),
			    g_strdup (strv[i + 1]));
    }
}

/**
 * glook_uri_new_from_string:
 *
 * try to parse uri string. if success return a new allocated
 * GLookUri struct, must free with glook_uri_free.
 */
GLookUri*
glook_uri_new_from_string (const gchar *string)
{
  GLookUri *uri;
  gchar *escape, *match;
  GRegex *regexp;
  GMatchInfo *match_info;
  GError *error = NULL;

  g_return_val_if_fail (string, NULL);

  regexp = g_regex_new (_URI_PATTERN,
			G_REGEX_CASELESS | G_REGEX_UNGREEDY,
			G_REGEX_MATCH_NOTEMPTY, &error);
  if (error)
    {
      g_error ("%s", error->message);
      g_error_free (error);
      return NULL;
    }
  /* escape string, characters in _DELIMITERS is not escaped. */
  escape = g_uri_escape_string (string, _DELIMITERS, TRUE);
  g_strstrip (escape);

  /* try to match the string with pattern. it is not always
   * set the error when match failed. */
  if (!g_regex_match_full (regexp, escape, -1, 0, 0, &match_info, &error))
    {
      _critical (_("Unable to match '%s': %s"), escape,
		 error ? error->message : _("format error."));
      if (error)
	g_error_free (error);

      g_free (escape);
      g_match_info_free (match_info);
      g_regex_unref (regexp);

      return NULL;
    }
  /* debug code, don't remove. */
#if 0
    {
      gint i, n;
      n = g_match_info_get_match_count (match_info);
      g_debug ("Matched: %d", n);

      for (i = 0; i < n; i++)
	{
	  match = g_match_info_fetch (match_info, i);
	  _debug ("Found: '%s'\n", match);
	  g_free (match);
	}
    }
#endif
  uri = glook_uri_new ();

  /* extract the matched information. this information is obtain
   * from the original string, that is, the above escape variable,
   * so don't free it before extract. */
  match = glook_uri_get_match_string (match_info, "scheme");
  /* convert the scheme to lowercase */
  uri->scheme = g_ascii_strdown (match, -1);
  uri->scheme_type = glook_uri_scheme_name_to_type (uri->scheme);
  g_free (match);

  uri->attrs = glook_uri_get_match_string (match_info, "attrs");
  if (uri->attrs)
    {
      /* parse attrs string to hash k/v pair. */
      glook_uri_parse_attrs (uri);
    }
#if 0
  _debug ("scheme: %s\n", uri->scheme);
  _debug ("attrs: %s\n", uri->attrs);
#endif

  g_free (escape);
  g_match_info_free (match_info);
  g_regex_unref (regexp);

  return uri;
}

/**
 * glook_uri_to_string:
 *
 * @userinfo: includes user info in result string.
 *
 * convert GLookUri struct to human readable string.
 * the result must free with g_free().
 */
gchar*
glook_uri_to_string (GLookUri *uri)
{
  const gchar *scheme;

  g_return_val_if_fail (uri, NULL);

  scheme = uri->scheme;
  if (!scheme)
    scheme = glook_uri_scheme_name_from_type (uri->scheme_type);

  if (uri->attrs)
    return g_strdup_printf ("%s:%s", scheme, uri->attrs);

  if (g_hash_table_size (uri->attrs_table) > 0)
    {
      GHashTableIter iter;
      GString *string;
      gpointer key, value;

      string = g_string_new ("");

      g_hash_table_iter_init (&iter, uri->attrs_table);
      while (g_hash_table_iter_next (&iter, &key, &value)) 
	{
	  g_string_append_printf (string, "%s=%s,",
				  (gchar*) key, (gchar*) value);
	}
      uri->attrs = g_string_free (string, FALSE);
      uri->attrs[strlen (uri->attrs) - 1] = '\0';

      return g_strdup_printf ("%s:%s", scheme, uri->attrs);
    }
  return g_strdup_printf ("%s:", scheme);
}

/**
 * glook_uri_get_attrs_table:
 */
GHashTable*
glook_uri_get_attrs_table (GLookUri *uri)
{
  g_return_val_if_fail (uri, NULL);
  return uri->attrs_table;
}

/**
 * glook_uri_set_scheme:
 */
void
glook_uri_set_scheme (GLookUri    *uri,
		      const gchar *scheme)
{
  gchar *newv;
  GLookUriScheme scheme_type;

  g_return_if_fail (uri);
  g_return_if_fail (scheme);

  scheme_type = glook_uri_scheme_name_to_type (scheme);
  if (scheme_type == GLOOK_URI_SCHEME_UNKNOWN)
    {
      g_warning (_("Uri scheme `%s' unrecognizable"), scheme);
      return;
    }
  newv = g_strdup (scheme);

  if (uri->scheme)
    g_free (uri->scheme);

  uri->scheme = newv;
  uri->scheme_type = scheme_type;
}

/**
 * glook_uri_set_scheme_type:
 */
void
glook_uri_set_scheme_type (GLookUri       *uri,
			   GLookUriScheme  scheme_type)
{
  g_return_if_fail (uri);
  uri->scheme_type = scheme_type;
}

/**
 * glook_uri_get_scheme:
 */
const gchar*
glook_uri_get_scheme (GLookUri *uri)
{
  g_return_val_if_fail (uri, NULL);
  return uri->scheme;
}

/**
 * glook_uri_get_scheme_type:
 */
GLookUriScheme
glook_uri_get_scheme_type (GLookUri *uri)
{
  g_return_val_if_fail (uri, GLOOK_URI_SCHEME_UNKNOWN);
  return uri->scheme_type;
}

/**
 * glook_uri_insert_attr:
 */
void
glook_uri_insert_attr (GLookUri    *uri,
		       const gchar *key,
		       const gchar *value)
{
  g_return_if_fail (uri);
  g_return_if_fail (key);
  g_return_if_fail (value);

  g_hash_table_replace (uri->attrs_table,
			g_strdup (key), g_strdup (value));
}

/**
 * glook_uri_remove_attr:
 */
void
glook_uri_remove_attr (GLookUri    *uri,
		       const gchar *key)
{
  g_return_if_fail (uri);
  g_return_if_fail (key);

  g_hash_table_remove (uri->attrs_table, key);
}

/**
 * glook_uri_lookup_attr:
 */
const gchar*
glook_uri_lookup_attr (GLookUri    *uri,
		       const gchar *key)
{
  g_return_val_if_fail (uri, NULL);
  g_return_val_if_fail (key, NULL);

  return g_hash_table_lookup (uri->attrs_table, key);
}

/**
 * glook_uri_is_equal:
 *
 * compare two GLookUri, if all field is equal, return TRUE.
 *
 * note if Uri A has a NULL value field, but Uri B has
 * same field with non-NULL value, that is unequal.
 */
gboolean
glook_uri_is_equal (GLookUri *uri1,
		    GLookUri *uri2)
{
  GHashTableIter iter;
  gpointer key, value;

  g_return_val_if_fail (uri1, FALSE);
  g_return_val_if_fail (uri2, FALSE);

  if (uri1->scheme_type != uri2->scheme_type)
    return FALSE;

  if (g_hash_table_size (uri1->attrs_table) !=
      g_hash_table_size (uri2->attrs_table))
    return FALSE;

  g_hash_table_iter_init (&iter, uri1->attrs_table);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      const gchar *value2;

      value2 = g_hash_table_lookup (uri2->attrs_table, key);
      if (!value2 || strcmp (value, value2) != 0)
	return FALSE;
    }
  return TRUE;
}

/**
 * glook_uri_get_dbus_address_from_keys:
 */
static gchar*
glook_uri_get_dbus_address_from_keys (GLookUri    *uri,
				      const gchar *family,
				      const gchar *keys[],
				      gsize        size)
{
  GString *string;
  gint i;
  GError *error = NULL;

  g_return_val_if_fail (uri, NULL);
  g_return_val_if_fail (family, NULL);

  string = g_string_new (family);

  for (i = 0; i < size; i++)
    {
      const gchar *value;

      value = g_hash_table_lookup (uri->attrs_table, keys[i]);
      if (value)
	g_string_append_printf (string, "%s=%s,", keys[i], value);
    }
  string->str[strlen (string->str) - 1] = '\0';

  if (!g_dbus_is_supported_address (string->str, &error))
    {
      g_warning ("%s", error->message);
      g_error_free (error);

      g_string_free (string, TRUE);
      return NULL;
    }
  return g_string_free (string, FALSE);
}

/**
 * glook_uri_get_dbus_address_unix:
 */
static gchar*
glook_uri_get_dbus_address_unix (GLookUri *uri)
{
  const gchar *keys[] = { "path", "abstract", "tmpdir" };

  return glook_uri_get_dbus_address_from_keys (uri, "unix:", keys,
					       G_N_ELEMENTS (keys));
}

/**
 * glook_uri_get_dbus_address_tcp:
 */
static gchar*
glook_uri_get_dbus_address_tcp (GLookUri *uri)
{
  const gchar *keys[] = { "host", "port", "family" };

  return glook_uri_get_dbus_address_from_keys (uri, "tcp:", keys,
					       G_N_ELEMENTS (keys));
}

/**
 * glook_uri_get_dbus_address_nonce_tcp:
 */
static gchar*
glook_uri_get_dbus_address_nonce_tcp (GLookUri *uri)
{
  const gchar *keys[] = { "host", "port", "family", "noncefile" };

  return glook_uri_get_dbus_address_from_keys (uri, "nonce-tcp", keys,
					       G_N_ELEMENTS (keys));
}

/**
 * glook_uri_get_dbus_address:
 *
 * get dbus address from uri.
 */
gchar*
glook_uri_get_dbus_address (GLookUri *uri,
			    GBusType *bus_type)
{
  const gchar *proto, *bus;

  if (bus_type)
    *bus_type = G_BUS_TYPE_NONE;

  /* proto must be one of unix, tcp, nonce-tcp, and bus. */
  proto = g_hash_table_lookup (uri->attrs_table, "proto");
  if (!proto)
    {
      g_critical (_("Try to get dbus address from uri, "
		    "but attribute `proto' is missing."));
      return NULL;
    }
  if (g_strcmp0 (proto, "unix") == 0)
    return glook_uri_get_dbus_address_unix (uri);

  if (g_strcmp0 (proto, "tcp") == 0)
    return glook_uri_get_dbus_address_tcp (uri);

  if (g_strcmp0 (proto, "nonce-tcp") == 0)
    return glook_uri_get_dbus_address_nonce_tcp (uri);

  if (bus_type)
    {
      if (g_strcmp0 (proto, "bus") == 0)
	{
	  bus = g_hash_table_lookup (uri->attrs_table, "bus");

	  if (g_ascii_strcasecmp (bus, "system") == 0)
	    *bus_type = G_BUS_TYPE_SYSTEM;
	  else if (g_ascii_strcasecmp (bus, "session") == 0)
	    *bus_type = G_BUS_TYPE_SESSION;
	  else
	    g_warning (_("value `%s' for attribute `bus' is invalid."), bus);

	  return NULL;
	}
    }
  return NULL;
}
