/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>

#include <gdk/gdkkeysyms.h>

#include "glook-app.h"
#include "glook-bookmark.h"
#include "glook-uri.h"
#include "glook-tab.h"
#include "glook-agent-tab.h"
#include "glook-play-tab.h"
#include "glook-look-tab.h"
#include "glook-blank-tab.h"
#include "glook-local-tab.h"
#include "glook-terminal-tab.h"
#include "glook-ui.h"
#include "glook-toolbar.h"
#include "glook-addrbar.h"
#include "glook-notebook.h"
#include "glook-statusbar.h"
#include "glook-window.h"
#include "glook-actions.h"

struct _GLookWindowPriv
{
  gboolean        rgba_colormap;
  GLookUi        *ui;
  GtkWidget      *menubar;
  GtkWidget      *toolbar;
  GtkWidget      *addrbar;
  GtkWidget      *notebook;
  GtkWidget      *statusbar;
};
#define _WINDOW_GET_PRIVATE(o)	(G_TYPE_INSTANCE_GET_PRIVATE ((o), GLOOK_TYPE_WINDOW, GLookWindowPriv))

G_DEFINE_TYPE (GLookWindow, glook_window, GTK_TYPE_WINDOW);

static gboolean	on_window_configure	(GtkWidget         *widget,
					 GdkEventConfigure *event,
					 GLookWindow       *window);

/**
 * glook_window_init:
 */
static void
glook_window_init (GLookWindow *window)
{
  GdkScreen *screen;
  GdkColormap *colormap;
  GtkWidget *vbox;
  GtkAccelGroup *accel_group;
  gchar *role;

  window->priv = _WINDOW_GET_PRIVATE (window);

  window->priv->rgba_colormap = FALSE;

  /* Set RGBA colormap if possible so VTE can use real alpha
   * channels for transparency.
   * this will modify window colormap visual depth.
   */
#ifndef G_OS_WIN32
  screen = gtk_widget_get_screen (GTK_WIDGET (window));
  colormap = gdk_screen_get_rgba_colormap (screen);
  if (colormap != NULL && gdk_screen_is_composited (screen))
    {
      gtk_widget_set_colormap (GTK_WIDGET (window), colormap);
      window->priv->rgba_colormap = TRUE;
    }
#endif
  window->priv->ui = glook_ui_new (GTK_WIDGET (window));
  window->priv->menubar = glook_ui_get_menubar (window->priv->ui);
  window->priv->toolbar = glook_toolbar_new (GTK_WIDGET (window));
  window->priv->notebook = glook_notebook_new ();
  window->priv->statusbar = glook_statusbar_new ();

  /* add ui manager's accel group. */
  accel_group = gtk_ui_manager_get_accel_group (GTK_UI_MANAGER (window->priv->ui));
  gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

  g_object_set (window, "title", "Ttyutils GLook",
		"window-position", GTK_WIN_POS_CENTER,
		"default-height", 480,
		"default-width", 600, NULL);

  /* create a box to packing the child widgets */
  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);

  gtk_box_pack_start (GTK_BOX (vbox), window->priv->menubar, FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), window->priv->toolbar, FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (vbox), window->priv->notebook, TRUE, TRUE, 0);
  gtk_box_pack_end (GTK_BOX (vbox), window->priv->statusbar, FALSE, TRUE, 0);

  /* connect toolbar signals */
  g_signal_connect (window->priv->toolbar, "zoom",
		    G_CALLBACK (on_window_toolbar_zoom), window);

  window->priv->addrbar = glook_toolbar_get_addrbar (GLOOK_TOOLBAR (window->priv->toolbar));

  /* connect address bar signals */
  g_signal_connect (window->priv->addrbar, "erase",
		    G_CALLBACK (on_window_addrbar_erase), window);
  g_signal_connect (window->priv->addrbar, "go",
		    G_CALLBACK (on_window_addrbar_go), window);
  g_signal_connect (window->priv->addrbar, "stop",
		    G_CALLBACK (on_window_addrbar_stop), window);
  g_signal_connect (window->priv->addrbar, "bookmark",
		    G_CALLBACK (on_window_addrbar_bookmark), window);

  /* connect notebook signals */
  g_signal_connect (window->priv->notebook, "page-added",
		    G_CALLBACK (on_window_notebook_page_added), window);
  g_signal_connect (window->priv->notebook, "page-removed",
		    G_CALLBACK (on_window_notebook_page_removed), window);
  g_signal_connect (window->priv->notebook, "switch-page",
		    G_CALLBACK (on_window_notebook_page_switched), window);

  g_signal_connect (window, "configure-event",
		    G_CALLBACK (on_window_configure), window);

  /* let all child widgets visible but without window itself. */
  gtk_widget_show (window->priv->toolbar);
  gtk_widget_show (window->priv->notebook);
  gtk_widget_show (window->priv->statusbar);
  gtk_widget_show (vbox);
  gtk_widget_grab_focus (window->priv->addrbar);

  role = g_strdup_printf ("ttyutils-glook-window-%d-%d-%d",
			  getpid (), g_random_int (), (gint) time (NULL));
  gtk_window_set_role (GTK_WINDOW (window), role);
  g_free (role);
}

/**
 * glook_window_dispose:
 */
static void
glook_window_dispose (GObject *object)
{
  //GLookWindow *window = GLOOK_WINDOW (object);

  G_OBJECT_CLASS (glook_window_parent_class)->dispose (object);
}

/**
 * glook_window_finalize:
 */
static void
glook_window_finalize (GObject *object)
{
  GLookWindow *window = GLOOK_WINDOW (object);

  if (window->priv->ui)
    {
      g_object_unref (window->priv->ui);
      window->priv->ui = NULL;
    }
  G_OBJECT_CLASS (glook_window_parent_class)->finalize (object);
}

/**
 * glook_window_class_init:
 */
static void
glook_window_class_init (GLookWindowClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GLookWindowPriv));

  gobject_class->dispose  = glook_window_dispose;
  gobject_class->finalize = glook_window_finalize;
}

/**
 * on_window_configure:
 *
 * called when window size or position has changed.
 */
static gboolean
on_window_configure (GtkWidget         *widget,
		     GdkEventConfigure *event,
		     GLookWindow       *window)
{
#if 0
  GdkRectangle rect;
  GdkWindow *gdkwin;
  gint x, y;
#endif
  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

#if 0
  g_debug ("x = %d, y = %d, width = %d, height = %d"
	   "bottom = %d right = %d",
	   event->x, event->y, event->width, event->height,
	   event->y + event->height,
	   event->x + event->width);
  gdkwin = gtk_widget_get_window (window);
  gdk_window_get_frame_extents (gdkwin, &rect);
  g_debug ("frame x = %d, y = %d, width = %d, height = %d",
	   rect.x, rect.y, rect.width, rect.height);
  gdk_window_get_root_origin (gdkwin, &x, &y);
  g_debug ("root x = %d, y = %d", x, y);
#endif
  return FALSE;
}

/**
 * glook_window_new:
 */
GtkWidget*
glook_window_new (void)
{
  return g_object_new (GLOOK_TYPE_WINDOW, NULL);
}

/**
 * glook_window_delete:
 */
void
glook_window_delete (GLookWindow *window)
{
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  if (GLOOK_IS_NOTEBOOK (window->priv->notebook))
    {
      g_signal_handlers_disconnect_by_func (window->priv->notebook,
					    on_window_notebook_page_added, window);
      g_signal_handlers_disconnect_by_func (window->priv->notebook,
					    on_window_notebook_page_removed, window);
      g_signal_handlers_disconnect_by_func (window->priv->notebook,
					    on_window_notebook_page_switched, window);
    }
  if (GLOOK_IS_ADDRBAR (window->priv->notebook))
    {
      g_signal_handlers_disconnect_by_func (window->priv->addrbar,
					    on_window_addrbar_erase, window);
      g_signal_handlers_disconnect_by_func (window->priv->addrbar,
					    on_window_addrbar_go, window);
      g_signal_handlers_disconnect_by_func (window->priv->addrbar,
					    on_window_addrbar_stop, window);
      g_signal_handlers_disconnect_by_func (window->priv->addrbar,
					    on_window_addrbar_bookmark, window);
    }
  if (GLOOK_IS_TOOLBAR (window->priv->notebook))
    {
      g_signal_handlers_disconnect_by_func (window->priv->toolbar,
					    on_window_toolbar_zoom, window);
    }
}

/**
 * glook_window_get_argb_colormap:
 */
gboolean
glook_window_get_rgba_colormap (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), FALSE);
  return window->priv->rgba_colormap;
}

/**
 * glook_window_get_uiman:
 */
GtkUIManager*
glook_window_get_uiman (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  return GTK_UI_MANAGER (window->priv->ui);
}

/**
 * glook_window_get_menubar:
 */
GtkWidget*
glook_window_get_menubar (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  return window->priv->menubar;
}

/**
 * glook_window_get_toolbar:
 */
GtkWidget*
glook_window_get_toolbar (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  return window->priv->toolbar;
}

/**
 * glook_window_get_addrbar:
 */
GtkWidget*
glook_window_get_addrbar (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  return window->priv->addrbar;
}

/**
 * glook_window_get_zoom_combo:
 */
GtkWidget*
glook_window_get_zoom_combo (GLookWindow *window)
{
  GLookToolbar *toolbar;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);

  toolbar = GLOOK_TOOLBAR (window->priv->toolbar);
  return glook_toolbar_get_zoom_combo (toolbar);
}

/**
 * glook_window_get_notebook:
 */
GtkWidget*
glook_window_get_notebook (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  return window->priv->notebook;
}

/**
 * glook_window_get_statusbar:
 */
GtkWidget*
glook_window_get_statusbar (GLookWindow *window)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  return window->priv->statusbar;
}

/**
 * glook_window_merge_ui:
 */
gint
glook_window_merge_ui (GLookWindow    *window,
		       GtkActionGroup *action_group,
		       GtkAccelGroup  *accel_group,
		       const gchar    *string,
		       gint            length)
{
  g_return_val_if_fail (GLOOK_IS_WINDOW (window), -1);
  g_return_val_if_fail (GTK_IS_ACTION_GROUP (action_group), -1);
  g_return_val_if_fail (string, -1);

  if (accel_group)
    gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

  return glook_ui_merge (window->priv->ui, action_group, string, length);
}

/**
 * glook_window_merge_ui:
 */
void
glook_window_unmerge_ui (GLookWindow   *window,
			 GtkAccelGroup *accel_group,
			 guint          merge_id)
{
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  if (merge_id > 0)
    glook_ui_unmerge (window->priv->ui, merge_id);

  if (accel_group)
    gtk_window_remove_accel_group (GTK_WINDOW (window), accel_group);
}

/**
 * create_or_reuse_tab:
 */
static GtkWidget*
create_or_reuse_tab (GLookWindow   *window,
		     GLookNotebook *notebook,
		     GType          newtab_type,
		     GLookUri      *uri)
{
  GtkWidget *curr_tab = NULL;
  GtkWidget *tab = NULL;
  gint n;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  g_return_val_if_fail (GLOOK_IS_NOTEBOOK (notebook), NULL);

  /* get the current tab. */
  curr_tab = glook_notebook_get_current (notebook);
  n = gtk_notebook_get_current_page (GTK_NOTEBOOK (notebook));

  /* if current tab is actived or in progress, we cannot reuse it */
  if (!curr_tab ||
      glook_tab_is_active (GLOOK_TAB (curr_tab)) ||
      glook_tab_get_progress (GLOOK_TAB (curr_tab)))
    {
      tab = g_object_new (newtab_type, NULL);
      glook_tab_set_uri (GLOOK_TAB (tab), uri);
      gtk_widget_show (tab);
      glook_notebook_insert (GLOOK_NOTEBOOK (notebook), tab, n + 1);
      glook_notebook_set_current_n (GLOOK_NOTEBOOK (notebook), n + 1);
      return tab;
    }
  /* if current tab is not actived, and has same type from
   * new tab, reuse it. */
  if (G_OBJECT_TYPE (curr_tab) == newtab_type)
    {
      glook_tab_set_uri (GLOOK_TAB (curr_tab), uri);
      return curr_tab;
    }
  /* current tab is not actived, and has differ type from
   * new tab, replace it with new tab. */
  tab = g_object_new (newtab_type, NULL);
  glook_tab_set_uri (GLOOK_TAB (tab), uri);
  gtk_widget_show (tab);
  glook_notebook_replace (GLOOK_NOTEBOOK (notebook), tab);

  return tab;
}

/**
 * on_tab_finished:
 *
 * add just open address to system recent manager.
 */
#ifdef DEBUG
static void
on_tab_finished (GLookTab    *tab,
		 GLookWindow *window)
{
  gchar *address;
  GtkRecentData info;
  gchar *groups[] = { "Ttyutils", NULL, };

  g_return_if_fail (GLOOK_IS_TAB (tab));
  g_return_if_fail (GLOOK_IS_WINDOW (window));

  address = g_object_get_data (G_OBJECT (tab), "address");
  if (!address)
    return;

  info.display_name = g_strdup_printf ("%s %s", g_get_application_name (),
				       glook_tab_get_label_text (GLOOK_TAB (tab)));
  info.description = g_strdup_printf (_("Open `%s' with %s"), address,
				      g_get_application_name ());
  info.mime_type = "application/x-ttylook";
  info.app_name = (gchar*) g_get_application_name ();
  info.app_exec = g_strdup_printf ("%s %%u", g_get_prgname ());
  info.groups = groups;
  info.is_private = TRUE;

  gtk_recent_manager_add_full (gtk_recent_manager_get_default (),
			       address, &info);
  g_free (info.display_name);
  g_free (info.description);
  g_free (info.app_exec);
}
#endif

/**
 * tab_open_uri_idle:
 */
static gboolean
tab_open_uri_idle (gpointer data)
{
  g_return_val_if_fail (GLOOK_IS_TAB (data), FALSE);

  glook_tab_open (GLOOK_TAB (data));
  return FALSE;
}

/**
 * glook_window_open_uri:
 *
 * open uri in new tab or reuse current tab.
 */
static GtkWidget*
glook_window_open_uri (GLookWindow *window,
		       GLookUri    *uri,
		       const gchar *address)
{
  GtkWidget *notebook;
  GLookUriScheme scheme;
  GtkWidget *tab = NULL;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);
  g_return_val_if_fail (uri, NULL);

  notebook = window->priv->notebook;
  scheme = glook_uri_get_scheme_type (uri);

  /* try to find a uri matched tab instead always create a new tab. */
  if (scheme == GLOOK_URI_SCHEME_AGENT ||
      scheme == GLOOK_URI_SCHEME_LOOK ||
      scheme == GLOOK_URI_SCHEME_PLAY ||
#ifdef G_OS_UNIX
      scheme == GLOOK_URI_SCHEME_TERMINAL ||
#endif
      scheme == GLOOK_URI_SCHEME_LOCAL)
    {
      gint n, i;
      GtkWidget *tab2;
      GLookUri *uri2;

      n = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
      for (i = 0; i < n; i++)
	{
	  tab2 = gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), i);
	  uri2 = glook_tab_get_uri (GLOOK_TAB (tab2));

	  if (uri2 && glook_uri_is_equal (uri, uri2))
	    {
	      gtk_notebook_set_current_page (GTK_NOTEBOOK (notebook), i);

	      if (!glook_tab_is_active (GLOOK_TAB (tab2)))
		glook_tab_open (GLOOK_TAB (tab2));
#ifdef G_OS_UNIX
	      else if (scheme == GLOOK_URI_SCHEME_TERMINAL)
		break;
#endif
	      return tab2;
	    }
	}
    }
  /* create a new tab(or reuse current tab) conform to
   * the uri scheme type. */
  switch (scheme)
    {
    case GLOOK_URI_SCHEME_AGENT:	/* agent tab */
      tab = create_or_reuse_tab (window, GLOOK_NOTEBOOK (notebook),
				 GLOOK_TYPE_AGENT_TAB, uri);
      break;
    case GLOOK_URI_SCHEME_LOOK:		/* look tab. */
      tab = create_or_reuse_tab (window, GLOOK_NOTEBOOK (notebook),
				 GLOOK_TYPE_LOOK_TAB, uri);
      break;
    case GLOOK_URI_SCHEME_PLAY:		/* play tab. */
      tab = create_or_reuse_tab (window, GLOOK_NOTEBOOK (notebook),
				 GLOOK_TYPE_PLAY_TAB, uri);
      break;
    case GLOOK_URI_SCHEME_LOCAL:	/* local tab */
      tab = create_or_reuse_tab (window, GLOOK_NOTEBOOK (notebook),
				 GLOOK_TYPE_LOCAL_TAB, uri);
      break;
#ifdef G_OS_UNIX
    case GLOOK_URI_SCHEME_TERMINAL:	/* terminal tab */
      tab = create_or_reuse_tab (window, GLOOK_NOTEBOOK (notebook),
				 GLOOK_TYPE_TERMINAL_TAB, uri);
      break;
#endif
    case GLOOK_URI_SCHEME_BLANK:	/* blank tab */
      return glook_window_add_blank_tab (window, TRUE);

    case GLOOK_URI_SCHEME_UNKNOWN:
    default:
      g_critical ("Uri unrecognizable.");
      return NULL;
    }
#ifdef DEBUG
  /* when tab open finish, we add address into system recent manager. */
  g_object_set_data_full (G_OBJECT (tab), "address",
			  g_strdup (address), g_free);
  g_signal_connect (tab, "finished", G_CALLBACK (on_tab_finished), window);
#endif

  g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, tab_open_uri_idle,
		   g_object_ref (tab), g_object_unref);
  return tab;
}

/**
 * glook_window_open_uri_string:
 */
GtkWidget*
glook_window_open_uri_string (GLookWindow *window,
			      const gchar *string)
{
  GLookBookmark *bookmark;
  GLookUri *uri;
  GtkWidget *tab;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (string, FALSE);

  uri = glook_uri_new_from_string (string);
  if (!uri)
    {
      g_critical (_("Address `%s' unrecognizable."), string);
      return FALSE;
    }
  /* update bookmark visited time. */
  bookmark = glook_app_get_bookmark (glook_app_get ());
  if (glook_bookmark_has (bookmark, string))
    glook_bookmark_update_visited (bookmark, string);

  tab = glook_window_open_uri (window, uri, string);
  glook_uri_unref (uri);

  return tab;
}

/**
 * glook_window_get_frame_extents:
 *
 * see gdk_window_get_frame_extents() for detail.
 */
gboolean
glook_window_get_frame_extents (GLookWindow  *window,
				GdkRectangle *rect)
{
  GdkWindow *gdkwindow;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (rect, FALSE);

  gdkwindow = gtk_widget_get_window (GTK_WIDGET (window));
  if (gdkwindow)
    {
      gdk_window_get_frame_extents (gdkwindow, rect);
      return TRUE;
    }
  return FALSE;
}

/**
 * glook_window_add_tab:
 */
void
glook_window_add_tab (GLookWindow *window,
		      GtkWidget   *tab,
		      gboolean     become_current)
{
  GtkNotebook *notebook;

  g_return_if_fail (GLOOK_IS_WINDOW (window));
  g_return_if_fail (GLOOK_IS_TAB (tab));

  notebook = GTK_NOTEBOOK (window->priv->notebook);

  if (gtk_notebook_get_n_pages (notebook) > 0)
    {
      /* insert new tab just after current active one. */
      gint n = gtk_notebook_get_current_page (notebook);
      glook_notebook_insert (GLOOK_NOTEBOOK (notebook), tab, n + 1);

      if (become_current)
	glook_notebook_set_current_n (GLOOK_NOTEBOOK (notebook), n + 1);
    }
  else
    {
      glook_notebook_append (GLOOK_NOTEBOOK (notebook), tab);
    }
  gtk_widget_show (tab);
}

/**
 * glook_window_add_blank_tab:
 */
GtkWidget*
glook_window_add_blank_tab (GLookWindow *window,
			    gboolean     become_current)
{
  //GtkNotebook *notebook;
  GtkWidget *tab;

  g_return_val_if_fail (GLOOK_IS_WINDOW (window), NULL);

  //notebook = GTK_NOTEBOOK (window->priv->notebook);

  tab = glook_blank_tab_new ();
  glook_blank_tab_set_uri (GLOOK_BLANK_TAB (tab));

  glook_window_add_tab (window, tab, become_current);
#if 0
  if (gtk_notebook_get_n_pages (notebook) > 0)
    {
      /* insert new tab just after current active one. */
      gint n = gtk_notebook_get_current_page (notebook);
      glook_notebook_insert (GLOOK_NOTEBOOK (notebook), tab, n + 1);

      if (become_current)
	glook_notebook_set_current_n (GLOOK_NOTEBOOK (notebook), n + 1);
    }
  else
    {
      glook_notebook_append (GLOOK_NOTEBOOK (notebook), tab);
    }
  gtk_widget_show (tab);
#endif

  return tab;
}
