/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define _XOPEN_SOURCE
#include <time.h>
#include <unistd.h>
#include <locale.h>
#include <stdlib.h>
#include <gcrypt.h>

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "glook-app.h"
#include "glook-misc.h"
#include "glook-ui.h"
#include "glook-window.h"
#include "glook-local-log.h"

static gboolean		 _opt_version = FALSE;

static GOptionEntry	_option_entries[] =
{
    { "version", 'v', 0, G_OPTION_ARG_NONE, &_opt_version,
      N_("Output version information"), NULL },

    { NULL, }
};

static gboolean	do_expire_dialog	(gpointer data);

/**
 * quit_application:
 */
static gboolean
quit_application (gpointer data)
{
  glook_app_quit (glook_app_get ());
  return FALSE;
}

/**
 * signal_terminated:
 *
 * terminate application. we first try to graceful quit.
 */
static void
signal_terminated (gint signo)
{
  static gboolean termed = FALSE;

  if (!termed)
    {
      g_idle_add (quit_application, NULL);
      termed = TRUE;
    }
  else
    {
      g_message (_("Mandatory quit."));
      exit (1);
    }
}

/**
 * cleanup_secure_memory:
 */
static void
cleanup_secure_memory (void)
{
  gcry_control (GCRYCTL_TERM_SECMEM, 0);
}

/**
 * main:
 */
gint
main (gint argc, gchar *argv[])
{
  GLookApp *app;
  GOptionContext *option;
  struct sigaction action;
  GtkWidget *window = NULL;
  const gchar *ver;
  GError *error = NULL;

  /* G_SLICE environment may cause GSettings carsh. */
  g_unsetenv ("G_SLICE");

#ifdef DEBUG
  g_setenv ("GSETTINGS_SCHEMA_DIR", "./data", FALSE);
  g_setenv ("GSETTINGS_BACKEND", "keyfile", FALSE);
#endif

  g_set_prgname (argv[0]);
  g_set_application_name ("Ttyutils GLook");

  setlocale (LC_ALL, "");
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  /* check GTK+ runtime version. */
  if (!t_check_version ())
    return EXIT_FAILURE;

  ver = gtk_check_version (GTK_MAJOR_VERSION,
			   GTK_MINOR_VERSION,
			   GTK_MICRO_VERSION);
  if (ver)
    {
      g_warning ("Gtk+ %d.%d.%d or above is requried: %s",
		 GTK_MAJOR_VERSION, GTK_MINOR_VERSION,
		 GTK_MICRO_VERSION, ver);
      return EXIT_FAILURE;
    }
  gtk_init (&argc, &argv);

  /* initialize libgcrypt subsystems. */
#ifdef DEBUG
  gcry_control (GCRYCTL_SET_VERBOSITY, 1);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1);
#else
  gcry_control (GCRYCTL_SET_VERBOSITY, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 1);
#endif

  gcry_set_log_handler (t_gcrypt_log_handler, NULL);
  if (!gcry_check_version (GCRYPT_VERSION))
    {
      g_warning (_("libgcrypt version %s or above is required."),
		 GCRYPT_VERSION);
      return EXIT_FAILURE;
    }
  gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN, NULL);
  gcry_control (GCRYCTL_INIT_SECMEM, 16 * 1024, 0);
  gcry_control (GCRYCTL_RESUME_SECMEM_WARN, NULL);
#ifdef DEBUG
  gcry_control (GCRYCTL_PRINT_CONFIG, NULL);
#endif
  gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);

  /* cleanup secure memory when exit. */
  g_atexit (cleanup_secure_memory);

  /* parse command line options */
  option = g_option_context_new ("[ADDRESS]");
  g_option_context_set_summary (option, _("Watch and control user terminal session"));
  g_option_context_set_description (option, WEBSITE);
  g_option_context_set_ignore_unknown_options (option, TRUE);
  g_option_context_add_main_entries (option, _option_entries, GETTEXT_PACKAGE);
  g_option_context_add_group (option, gtk_get_option_group (TRUE));
  g_option_context_parse (option, &argc, &argv, &error);
  if (error)
    {
      g_warning (_("Parse command line error: %s"), error->message);
      g_error_free (error);
      g_option_context_free (option);
      return EXIT_FAILURE;
    }
  g_option_context_free (option);

  if (_opt_version)
    {
      t_print_version ();
      return EXIT_SUCCESS;
    }
  memset (&action, 0, sizeof (action));
  action.sa_handler = signal_terminated;
  sigemptyset (&action.sa_mask);

  sigaction (SIGTERM, &action, 0);
  sigaction (SIGINT, &action, 0);
  sigaction (SIGHUP, &action, 0);

  /* default icon name for all windows */
  gtk_window_set_default_icon_name ("application-x-ttylook");

  /* redirect log messages */
  glook_local_log_redirect ();

  app = glook_app_new ();

  /* apply mnemonic and accelerators settings. */
  glook_app_set_accelerators (app);

  /* create window with a blank tab. */
  window = glook_app_create_window (app, FALSE);

  /* try to open address if provides. */
  if (argc > 1 && argv[1])
    {
      GLookUri *uri = glook_uri_new_from_string (argv[1]);
      if (uri)
	{
	  glook_window_open_uri_string (GLOOK_WINDOW (window), argv[1]);
	  glook_uri_free (uri);
	}
      else
	{
	  gchar *scheme, *address;

	  scheme = g_uri_parse_scheme (argv[1]);
	  if (scheme)
	    address = g_strdup_printf ("play:uri=%s", argv[1]);
	  else
	    address = g_strdup_printf ("play:path=%s", argv[1]);

	  glook_window_open_uri_string (GLOOK_WINDOW (window), address);
	  g_free (address);
	  g_free (scheme);
	}
    }
  else
    {
      GSettings *settings;
      GtkUIManager *uiman;
      GtkAction *action;
      GtkWidget *notebook;

      settings = glook_app_get_settings (app, SETTINGS_GENERAL);

      uiman = glook_window_get_uiman (GLOOK_WINDOW (window));

      /* connect agent server. */
      if (g_settings_get_boolean (settings, "startup-connect-agent"))
	{
	  action = glook_ui_get_window_action (GLOOK_UI (uiman), "Look");
	  gtk_action_activate (action);

	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	}
      /* open bookmarks. */
      if (g_settings_get_boolean (settings, "startup-open-bookmark"))
	{
	  action = glook_ui_get_window_action (GLOOK_UI (uiman), "Bookmarks");
	  gtk_action_activate (action);

	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	}
#ifdef G_OS_UNIX
      /* open terminal. */
      if (g_settings_get_boolean (settings, "startup-open-terminal"))
	{
	  action = glook_ui_get_window_action (GLOOK_UI (uiman), "Terminal");
	  gtk_action_activate (action);

	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	}
#endif
      notebook = glook_window_get_notebook (GLOOK_WINDOW (window));
      if (gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook)) == 0)
	glook_window_add_blank_tab (GLOOK_WINDOW (window), TRUE);
    }
  gtk_widget_show (window);

  if (t_expire_check ())
    g_idle_add (do_expire_dialog, window);

  /* enter main loop. */
  glook_app_run (app);
  g_object_unref (app);

  /* free log resources. */
  glook_local_log_cleanup ();
  return 0;
}

/**
 * do_expire_dialog:
 */
static gboolean
do_expire_dialog (gpointer data)
{
  GtkWidget *window = data;
  GtkWidget *dialog;
  GtkWidget *vbox;
  GtkWidget *label;
  gchar *markup;

  dialog = gtk_dialog_new_with_buttons (_("Trial edition expired"),
					GTK_WINDOW (window),
					GTK_DIALOG_MODAL |
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE,
					NULL);
  g_object_set (dialog, "default-width", 400, "default-height", 200,
		"border-width", 10, "has-separator", FALSE, NULL);
  vbox = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
  g_object_set (vbox, "spacing", 8, NULL);

  label = gtk_label_new ("");
  markup = g_markup_printf_escaped ("<span weight='bold' size='large'>%s</span>",
				    _("Trial edition expired!"));
  g_object_set (label, "xalign", 0.0, "use-markup", TRUE, "label", markup, NULL);
  g_free (markup);
  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, TRUE, 4);

  markup = g_markup_printf_escaped (_("You can purchasing Ttyutils from "
				      "<a href='%s' title='%s'>%s</a>."),
				    "http://www.ttyutils.net", "http://www.ttyutils.net",
				   _("Ttyutils website"));
  label = gtk_label_new (NULL);
  g_object_set (label, "xalign", 0.0, "label", markup, "use-markup", TRUE, NULL);
  g_free (markup);
  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, TRUE, 0);

  g_signal_connect_swapped (label, "activate-link",
			    G_CALLBACK (glook_open_url), window);

  gtk_widget_show_all (dialog);
  gtk_dialog_run (GTK_DIALOG (dialog));

  gtk_widget_destroy (dialog);
  glook_app_quit (glook_app_get ());

  return FALSE;
}
