/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>
#include <errno.h>
#include <stdlib.h>

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "look-app.h"

#undef CTRL
#define CTRL(x)	((x) & 0x1F)

struct _LookAppPrivate
{
  GMainLoop      *mainloop;
  gboolean        writable;
  gchar           quitkey;
  TLook          *look;
  TConsole       *console;
  GSList         *reginfos;
  GInetAddress   *server_addr;
  guint           step_timeout;
  GDBusProxy     *agent_proxy;
  GCancellable   *agent_cancellable;
  guint           agent_timeout_id;
};
#define LOOK_APP_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), LOOK_TYPE_APP, LookAppPrivate))

static LookApp	*_app = NULL;

G_DEFINE_TYPE (LookApp, look_app, G_TYPE_OBJECT);

/* defined in look.c */
extern time_t		look_get_uptime		(void);
extern gint		look_get_opt_timeout	(void);
extern const gchar*	look_get_opt_agent	(void);
extern void		look_redirect_log	(void);

/* console signal handlers. */
static void		on_look_connected	(TLook       *look,
						 LookApp     *app);
static void		on_look_hangup		(TLook       *look,
						 LookApp     *app);
static void		on_look_received	(TLook       *look,
						 const gchar *data,
						 guint        size,
						 LookApp     *app);


/**
 * look_app_init:
 */
void
look_app_init (LookApp *app)
{
  app->priv = LOOK_APP_PRIVATE (app);

  app->priv->mainloop = g_main_loop_new (NULL, FALSE);
  app->priv->writable = FALSE;
  app->priv->quitkey = '\\';
  app->priv->look = t_look_new ();
  app->priv->console = t_console_new ();
  app->priv->reginfos = NULL;
  app->priv->server_addr = NULL;
  app->priv->step_timeout = 0;
  app->priv->agent_proxy = NULL;
  app->priv->agent_cancellable = g_cancellable_new ();
  app->priv->agent_timeout_id = 0;
}

/**
 * look_app_dispose:
 */
void
look_app_dispose (GObject *object)
{
  LookApp *app = LOOK_APP (object);

  if (app->priv->agent_proxy)
    {
      g_object_unref (app->priv->agent_proxy);
      app->priv->agent_proxy = NULL;
    }
  if (app->priv->agent_cancellable)
    {
      g_object_unref (app->priv->agent_cancellable);
      app->priv->agent_cancellable = NULL;
    }
  if (app->priv->look)
    {
      g_object_unref (app->priv->look);
      app->priv->look = NULL;
    }
  if (app->priv->console)
    {
      g_object_unref (app->priv->console);
      app->priv->console = NULL;
    }
  if (app->priv->server_addr)
    {
      g_object_unref (app->priv->server_addr);
      app->priv->server_addr = NULL;
    }
  G_OBJECT_CLASS (look_app_parent_class)->dispose (object);
}

/**
 * look_app_finalize:
 */
void
look_app_finalize (GObject *object)
{
  LookApp *app = LOOK_APP (object);

  if (app->priv->mainloop)
    {
      g_main_loop_unref (app->priv->mainloop);
      app->priv->mainloop = NULL;
    }
  G_OBJECT_CLASS (look_app_parent_class)->finalize (object);
}

/**
 * look_app_class_init:
 */
void
look_app_class_init (LookAppClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (LookAppPrivate));

  gobject_class->dispose  = look_app_dispose;
  gobject_class->finalize = look_app_finalize;
}

/**
 * connect_step_timeout:
 */
static gboolean
connect_step_timeout (gpointer data)
{
  g_print (".");
  return TRUE;
}

/**
 * on_console_readable:
 *
 * called when console is readable.
 */
static void
on_console_readable (TConsole    *console,
		     const gchar *data,
		     guint        size,
		     LookApp     *app)
{
  static gchar quitkey = 0;

  if (quitkey == 0)
    quitkey = CTRL (app->priv->quitkey);

  if (size == 1 && data[0] == quitkey)
    {
      printf ("\r\n");
      printf (_("Connection interrupted. elapsed %ld seconds."),
	       time (NULL) - look_get_uptime ());
      printf ("\r\n");
      look_app_quit (app);
      return;
    }
  /* write to look server. */
  if (app->priv->writable)
    t_look_write (app->priv->look, data, size);
}

/**
 * on_console_hangup:
 *
 * called when console hangup.
 */
static void
on_console_hangup (TConsole *console,
		   LookApp  *app)
{
  g_return_if_fail (LOOK_IS_APP (app));
  look_app_quit (app);
}

/**
 * look_app_connect_server:
 *
 * try to connect look server.
 */
static gboolean
look_app_connect_server (LookApp *app,
			 TLook   *look)
{
  g_return_val_if_fail (LOOK_IS_APP (app), FALSE);
  g_return_val_if_fail (T_IS_LOOK (look), FALSE);

  g_print (_("Try to connect %s"), t_look_get_host (look));

  if (!t_look_connect (look))
    {
      g_warning (_("Unable to connect look server."));
      return FALSE;
    }
  /* add simple animation effect */
  app->priv->step_timeout = g_timeout_add_seconds (1, connect_step_timeout, app);

  return TRUE;
}

/**
 * look_app_apply_server_info:
 */
static void
look_app_apply_server_info (LookApp    *app,
			    TLookdInfo *server_info)
{
  guint16 rows, cols;
  const gchar *term, *charset;

  g_return_if_fail (LOOK_IS_APP (app));
  g_return_if_fail (server_info);

  /* try to change terminal size */
  t_termio_get_winsize (STDOUT_FILENO, &rows, &cols);
  if (rows != server_info->termrows || cols != server_info->termcols)
    {
      guint16 rows2, cols2;

      t_termio_set_winsize (STDOUT_FILENO, server_info->termrows,
			    server_info->termcols);
      t_termio_get_winsize (STDOUT_FILENO, &rows2, &cols2);

      if (rows2 < server_info->termrows || cols < server_info->termcols)
	g_message (_("Terminal size too small, try to change but failed."));
    }
  /* apply terminal type */
  term = g_getenv ("TERM");
  if (!term || strcmp (term, server_info->term) != 0)
    {
      g_message (_("Local terminal type %s is difference with server's %s."),
		 term ? term : "<NULL>", server_info->term);
      if (term)
	g_setenv ("TERM", server_info->term, TRUE);
    }
  /* how to change terminal charset encoding?? */
  g_get_charset (&charset);
  if (!charset || strcmp (charset, server_info->charset) != 0)
    {
      g_message (_("Local charset encoding %s is difference with server's %s."),
		 charset, server_info->charset);
    }
}

/**
 * on_look_connected:
 *
 * called when response from connect look server.
 */
static void
on_look_connected (TLook   *look,
		   LookApp *app)
{
  TLookdInfo *info;

  g_return_if_fail (T_IS_LOOK (look));
  g_return_if_fail (LOOK_IS_APP (app));

  /* remove the timeout event first */
  if (app->priv->step_timeout > 0)
    {
      g_print ("\n");
      g_source_remove (app->priv->step_timeout);
      app->priv->step_timeout = 0;
    }
  /* if connect failed. we try to connect server's really
   * address. if failed again, quit the application. */
  if (!t_look_is_connected (look))
    {
      if (app->priv->server_addr)
	{
	  /* fallback to server's really address */
	  gchar *addr = g_inet_address_to_string (app->priv->server_addr);
	  if (addr)
	    {
	      t_look_set_host (look, addr);
	      g_free (addr);
	      look_app_connect_server (app, look);
	    }
	  /* destroy the fallback address */
	  g_object_unref (app->priv->server_addr);
	  app->priv->server_addr = NULL;
	}
      else
	{
	  /* quit the application */
	  look_app_quit (app);
	}
      return;
    }
  info = t_look_get_info (look);

  /* check whether or not our control terminal is the
   * pseudo terminal of server.  we cannot continue in this case. */
  if (g_strcmp0 (info->ptsname, ttyname (STDOUT_FILENO)) == 0)
    {
      g_warning (_("Try to connect server which uses same terminal `%s' "
		   "is not allowed."), info->ptsname);
      look_app_quit (app);
      return;
    }
  /* apply server's terminal settings */
  look_app_apply_server_info (app, info);

  g_print (_("Connected. Press `Ctrl+%c' to interrupt.\n"),
	   app->priv->quitkey);

  g_signal_connect (app->priv->console, "readable",
		    G_CALLBACK (on_console_readable), app);
  g_signal_connect (app->priv->console, "hangup",
		    G_CALLBACK (on_console_hangup), app);

  /* redirect all of g_log and g_print(err) messages */
  look_redirect_log ();

  /* take terminal into raw mode */
  t_termio_makeraw (STDIN_FILENO);

  /* start read data from console. */
  if (!t_console_start (app->priv->console))
    look_app_quit (app);

  /* start read data from look server */
  t_look_read (look);
}

/**
 * on_look_received:
 *
 * receive data from look server, write to console
 */
static void
on_look_received (TLook       *look,
		  const gchar *data,
		  guint        size,
		  LookApp     *app)
{
  g_return_if_fail (T_IS_LOOK (look));
  g_return_if_fail (data);
  g_return_if_fail (LOOK_IS_APP (app));

  t_console_write (app->priv->console, data, size);
}

/**
 * on_look_hangup:
 *
 * called when disconnect with look server.
 */
static void
on_look_hangup (TLook   *look,
		LookApp *app)
{
  g_return_if_fail (LOOK_IS_APP (app));

  printf ("\r\n");
  printf (_("Disconnected from look server, quit."));
  printf ("\r\n");

  look_app_quit (app);
}


typedef struct _AgentServerEntry   AgentServerEntry;
struct _AgentServerEntry
{
  guint64  time;
  gchar   *hostname;
  gchar   *username;
  guint32  pid;
  gchar   *ttyname;
  gchar   *ptsname;
  gchar   *command;
  guint16  lookport;
};

/**
 * agent_recorder_entry_new:
 */
static AgentServerEntry *
agent_recorder_entry_new (void)
{
  return g_slice_new0 (AgentServerEntry);
}

/**
 * agent_recorder_entry_free:
 */
static void
agent_recorder_entry_free (AgentServerEntry *entry)
{
  if (entry)
    {
      g_free (entry->hostname);
      g_free (entry->username);
      g_free (entry->ttyname);
      g_free (entry->ptsname);
      g_free (entry->command);
 
      g_slice_free (AgentServerEntry, entry);
    }
}


static const gchar	*_agent_introspection_xml =
"<node>"
"  <interface name='net.ttyutils.Agent.RecorderManager'>"
"    <method name='GetList'>"
"      <arg name='entries'  type='a(tssusssq)' direction='out'/>"
"    </method>"
"  </interface>"
"</node>";


/**
 * look_agent_get_interface_info:
 */
static GDBusInterfaceInfo*
look_app_get_agent_interface_info (void)
{
  static GDBusInterfaceInfo *info = NULL;
  static gsize has_info = 0;

  if (g_once_init_enter (&has_info))
    {
      GDBusNodeInfo *data;
      data = g_dbus_node_info_new_for_xml (_agent_introspection_xml, NULL);
      info = data->interfaces[0];
      g_once_init_leave (&has_info, 1);
    }
  return info;
}

/**
 * on_agent_authorize_authenticated:
 */
static void
on_agent_authorize_authenticated (GDBusAuthObserver *observer,
				  GIOStream         *iostream,
				  GCredentials      *credentials,
				  gpointer           user_data)
{
  LookApp *app = user_data;

  g_return_if_fail (G_IS_DBUS_AUTH_OBSERVER (observer));
  g_return_if_fail (G_IS_IO_STREAM (iostream));
  g_return_if_fail (G_IS_CREDENTIALS (credentials));
  g_return_if_fail (LOOK_IS_APP (app));

  g_debug ("Auth: TODO");
}

#define exit_if_error(error)	G_STMT_START {		\
    if (error)						\
      {							\
	g_dbus_error_strip_remote_error (error);	\
	g_warning ("%s", error->message);		\
	g_error_free (error);				\
	look_app_quit (app);				\
	return;						\
      }							\
  } G_STMT_END


/**
 * get_user_select_from_stdin:
 */
static gint
get_user_select_from_stdin (gint max)
{
  gchar buff[8];
  gint choose;

  g_print (_("Please select a server to connect(`q' for quit): "));

  while (1)
    {
      memset (buff, 0, sizeof (buff));

      if (!fgets (buff, sizeof (buff), stdin))
	{
	  if (!feof (stdin))
	    g_warning ("fgets() error: %s", g_strerror (errno));

	  return -1;
	}
      if (buff[0] == 'q' || buff[0] == 'Q')
	return -1;

      choose = (gint) g_strtod (buff, NULL);
      if (choose > 0 && choose <= max)
	break;

      g_print (_("Out of range, please try again(`q' for quit): "));
    }
  return choose;
}

/**
 * on_dbus_proxy_call_get_list:
 */
static void
on_dbus_proxy_call_get_list (GObject      *source,
			     GAsyncResult *res,
			     gpointer      user_data)
{
  GDBusProxy *proxy = G_DBUS_PROXY (source);
  LookApp *app = user_data;
  GVariant *result, *child;
  GVariantIter iter;
  AgentServerEntry *entry;
  GPtrArray *entries;
  gint i = 0, choose;
  guint64 time;
  gchar *hostname, *username, *ttyname, *ptsname, *command;
  guint32 pid;
  guint16 lookport;
  GError *error = NULL;

  g_return_if_fail (LOOK_IS_APP (app));

  result = g_dbus_proxy_call_finish (proxy, res, &error);
  exit_if_error (error);

  g_variant_iter_init (&iter, result);
  child = g_variant_iter_next_value (&iter);
  g_variant_iter_init (&iter, child);

  entries = g_ptr_array_new_with_free_func ((GDestroyNotify) agent_recorder_entry_free);

  while (g_variant_iter_next (&iter, "(tssusssq)", &time, &hostname,
			      &username, &pid, &ttyname, &ptsname,
			      &command, &lookport))
    {
      g_print ("%3d. [%5d] %s@%s:%d%s \"%s\"\n", ++i,
	       pid, username,
	       hostname, lookport,
	       ptsname, command);

      entry = agent_recorder_entry_new ();

      entry->pid = pid;
      entry->time = time;
      entry->hostname = hostname;
      entry->username = username;
      entry->ttyname = ttyname;
      entry->ptsname = ptsname;
      entry->command = command;
      entry->lookport = lookport;

      g_ptr_array_add (entries, entry);
    }
  g_variant_unref (result);

  choose = get_user_select_from_stdin (i);
  if (choose <= 0)
    {
      g_ptr_array_free (entries, TRUE);
      look_app_quit (app);
      return;
    }
  entry = g_ptr_array_index (entries, choose - 1);
  if (entry)
    {
      /* set look server information */
      t_look_set_host (app->priv->look, entry->hostname);
      t_look_set_port (app->priv->look, entry->lookport);

      /* connect to look server */
      look_app_connect_server (app, app->priv->look);
    }
  g_ptr_array_free (entries, TRUE);
}

/**
 * dbus_proxy_call_get_list:
 */
static void
dbus_proxy_call_get_list (LookApp  *app)
{
  if (look_get_opt_timeout () > 0)
    g_dbus_proxy_set_default_timeout (app->priv->agent_proxy,
				      look_get_opt_timeout () * 1000);

  g_dbus_proxy_call (app->priv->agent_proxy,
		     "GetList", NULL,
		     G_DBUS_CALL_FLAGS_NO_AUTO_START,
		     -1, app->priv->agent_cancellable,
		     on_dbus_proxy_call_get_list, app);
}

/**
 * on_dbus_proxy_new_for_bus:
 */
static void
on_dbus_proxy_new_for_bus (GObject      *source,
			   GAsyncResult *res,
			   gpointer      user_data)
{
  LookApp *app = user_data;
  GError *error = NULL;

  g_return_if_fail (LOOK_IS_APP (app));

  app->priv->agent_proxy = g_dbus_proxy_new_for_bus_finish (res, &error);
  exit_if_error (error);

  dbus_proxy_call_get_list (app);
}

/**
 * on_dbus_proxy_new_for_peer:
 */
static void
on_dbus_proxy_new_for_peer (GObject      *source,
			    GAsyncResult *res,
			    gpointer      user_data)
{
  LookApp *app = user_data;
  GError *error = NULL;

  g_return_if_fail (LOOK_IS_APP (app));

  app->priv->agent_proxy = g_dbus_proxy_new_finish (res, &error);
  exit_if_error (error);

  dbus_proxy_call_get_list (app);
}

/**
 * on_dbus_connection_new_for_address:
 */
static void
on_dbus_connection_new_for_address (GObject      *source,
				    GAsyncResult *res,
				    gpointer      user_data)
{
  LookApp *app = user_data;
  GDBusConnection *connection;
  GError *error = NULL;

  g_return_if_fail (LOOK_IS_APP (app));

  /* remove all timeout event. */
  if (app->priv->step_timeout > 0)
    {
      g_print ("\n");
      g_source_remove (app->priv->step_timeout);
      app->priv->step_timeout = 0;
    }
  if (app->priv->agent_timeout_id > 0)
    {
      g_source_remove (app->priv->agent_timeout_id);
      app->priv->agent_timeout_id = 0;
    }
  connection = g_dbus_connection_new_for_address_finish (res, &error);
  exit_if_error (error);

  g_dbus_proxy_new (connection,
		    G_DBUS_PROXY_FLAGS_NONE,
		    look_app_get_agent_interface_info (),
		    NULL, "/net/ttyutils/Agent",
		    "net.ttyutils.Agent.RecorderManager",
		    app->priv->agent_cancellable,
		    on_dbus_proxy_new_for_peer, app);
  g_object_unref (connection);
}

/**
 * on_bus_watch_name_appeared:
 *
 * the agent bus name was appear on the bus daemon. this means
 * agent server is running as dbus service.
 */
static void
on_bus_watch_name_appeared (GDBusConnection *connection,
			    const gchar     *name,
			    const gchar     *name_owner,
			    gpointer         data)
{
  LookApp *app = data;
  guint watch_id;
  GBusType bus_type;

  g_return_if_fail (LOOK_IS_APP (app));

  /* we don't like to watch the name again. */
  watch_id = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (app), "watch-id"));
  g_bus_unwatch_name (watch_id);

  bus_type = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (app), "bus-type"));

  g_dbus_proxy_new_for_bus (bus_type,
			    G_DBUS_PROXY_FLAGS_DO_NOT_AUTO_START,
			    look_app_get_agent_interface_info (),
			    name, "/net/ttyutils/Agent",
			    "net.ttyutils.Agent.RecorderManager",
			    app->priv->agent_cancellable,
			    on_dbus_proxy_new_for_bus, app);
}

/**
 * on_bus_watch_name_vanished:
 *
 * this means agent is not running as dbus service(maybe not
 * startup or running as network server).
 */
static void
on_bus_watch_name_vanished (GDBusConnection *connection,
			    const gchar     *name,
			    gpointer         data)
{
  LookApp *app = data;

  g_return_if_fail (LOOK_IS_APP (app));

  g_warning (_("Agent server is not active or it is not "
	       "running as dbus service."));
  look_app_quit (app);
}

/**
 * on_dbus_connection_new_for_address_timeout:
 *
 * make connection to agent server timeout. there has many
 * reason to timeout, such as DNS resolve, network delay,
 * and so no.
 * this timeout event is only for peer connection.
 */
static gboolean
on_dbus_connection_new_for_address_timeout (gpointer data)
{
  LookApp *app = data;

  g_return_val_if_fail (LOOK_IS_APP (app), FALSE);

  g_cancellable_cancel (app->priv->agent_cancellable);
  return FALSE;
}

/**
 * on_dbus_connection_new_for_address_timeout_destroy:
 */
static void
on_dbus_connection_new_for_address_timeout_destroy (gpointer data)
{
  LookApp *app = data;

  g_return_if_fail (LOOK_IS_APP (app));

  app->priv->agent_timeout_id = 0;
  g_object_unref (app);
}

/**
 * look_app_connect_agent:
 *
 * connect to agent server, and fetch recorder list.
 */
static gboolean
look_app_connect_agent (LookApp *app)
{
  GDBusAuthObserver *observer;
  const gchar *address;
  guint watch_id;

  g_return_val_if_fail (LOOK_IS_APP (app), FALSE);

  address = look_get_opt_agent ();
  g_return_val_if_fail (address, FALSE);

  /* address may be `sessoin', `system', or any valid DBus address.
   * see DBus Specification for detail. */
  if (g_ascii_strcasecmp (address, "session") == 0)
    {
      watch_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
				   "net.ttyutils.Agent",
				   G_BUS_NAME_WATCHER_FLAGS_NONE,
				   on_bus_watch_name_appeared,
				   on_bus_watch_name_vanished,
				   app, NULL);

      g_object_set_data (G_OBJECT (app), "bus-type",
			 GUINT_TO_POINTER (G_BUS_TYPE_SESSION));
      g_object_set_data (G_OBJECT (app), "watch-id",
			 GUINT_TO_POINTER (watch_id));
    }
  else if (g_ascii_strcasecmp (address, "system") == 0)
    {
      watch_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
				   "net.ttyutils.Agent",
				   G_BUS_NAME_WATCHER_FLAGS_NONE,
				   on_bus_watch_name_appeared,
				   on_bus_watch_name_vanished,
				   app, NULL);

      g_object_set_data (G_OBJECT (app), "bus-type",
			 GUINT_TO_POINTER (G_BUS_TYPE_SYSTEM));
      g_object_set_data (G_OBJECT (app), "watch-id",
			 GUINT_TO_POINTER (watch_id));
    }
  else
    {
      GError *error = NULL;
      gint timeout;

      if (!g_dbus_is_supported_address (address, &error))
	{
	  g_message (_("Address `%s' unrecognizable: %s"),
		     address, error->message);
	  g_error_free (error);
	  return FALSE;
	}
      observer = g_dbus_auth_observer_new ();

      g_signal_connect (observer, "authorize-authenticated-peer",
			G_CALLBACK (on_agent_authorize_authenticated), app);

      g_print (_("Try to connect agent `%s'."), address);

      /* add simple animation effect */
      app->priv->step_timeout =
       g_timeout_add_seconds (1, connect_step_timeout, NULL);

      g_dbus_connection_new_for_address (address,
					 G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT,
					 observer,
					 app->priv->agent_cancellable,
					 on_dbus_connection_new_for_address, app);
      g_object_unref (observer);

      timeout = look_get_opt_timeout ();
      if (timeout > 0)
	{
	  app->priv->agent_timeout_id =
	   g_timeout_add_seconds_full (G_PRIORITY_DEFAULT, timeout,
				       on_dbus_connection_new_for_address_timeout,
				       g_object_ref (app),
				       on_dbus_connection_new_for_address_timeout_destroy);
	}
    }
  return TRUE;
}

/**
 * look_app_new:
 */
LookApp*
look_app_new (void)
{
  g_assert (!_app);
  _app = g_object_new (LOOK_TYPE_APP, NULL);
  return _app;
}

/**
 * look_app_get:
 */
LookApp*
look_app_get (void)
{
  g_assert (_app);
  return _app;
}

/**
 * look_app_set_server_address:
 */
void
look_app_set_server_address (LookApp     *app,
			     const gchar *address)
{
  g_return_if_fail (LOOK_IS_APP (app));
  g_return_if_fail (address);

  t_look_set_host (app->priv->look, address);
}

/**
 * look_app_set_writable:
 */
void
look_app_set_writable (LookApp  *app,
		       gboolean  writable)
{
  g_return_if_fail (LOOK_IS_APP (app));
  app->priv->writable = writable;
}

/**
 * look_app_set_quit_key:
 */
void
look_app_set_quit_key (LookApp  *app,
		       gchar     key)
{
  g_return_if_fail (LOOK_IS_APP (app));
  app->priv->quitkey = key;
}

/**
 * look_app_get_look:
 */
TLook*
look_app_get_look (LookApp *app)
{
  g_return_val_if_fail (LOOK_IS_APP (app), NULL);
  return app->priv->look;
}

/**
 * look_app_get_console:
 */
TConsole*
look_app_get_console (LookApp *app)
{
  g_return_val_if_fail (LOOK_IS_APP (app), NULL);
  return app->priv->console;
}

/**
 * look_app_start:
 *
 * start connect look server or agent server. and
 * enter main loop.
 */
void
look_app_start (LookApp *app)
{
  gboolean ret;

  g_return_if_fail (LOOK_IS_APP (app));

  /* apply look connection timeout */
  t_look_set_timeout (app->priv->look, look_get_opt_timeout ());

  g_signal_connect (app->priv->look, "connected",
		    G_CALLBACK (on_look_connected), app);
  g_signal_connect (app->priv->look, "received",
		    G_CALLBACK (on_look_received), app);
  g_signal_connect (app->priv->look, "hangup",
		    G_CALLBACK (on_look_hangup), app);

  /* connect to look server or agent. */
  if (t_look_get_host (app->priv->look))
    ret = look_app_connect_server (app, app->priv->look);
  else
    ret = look_app_connect_agent (app);

  if (ret)
    g_main_loop_run (app->priv->mainloop);
}

/**
 * look_app_quit_idle:
 */
static gboolean
look_app_quit_idle (gpointer data)
{
  LookApp *app = LOOK_APP (data);

  g_return_val_if_fail (LOOK_IS_APP (app), FALSE);

  return (t_console_get_buffer_size (app->priv->console) > 0);
}

/**
 * look_app_quit_idle_removed:
 */
static void
look_app_quit_idle_removed (gpointer data)
{
  LookApp *app = LOOK_APP (data);

  g_return_if_fail (LOOK_IS_APP (app));

  if (app->priv->mainloop)
    g_main_loop_quit (app->priv->mainloop);
}

/**
 * look_app_quit:
 *
 * quit application and free resources.
 */
void
look_app_quit (LookApp  *app)
{
  g_return_if_fail (LOOK_IS_APP (app));

  t_console_flush (app->priv->console);

  g_idle_add_full (G_PRIORITY_DEFAULT,
		   look_app_quit_idle, app,
		   look_app_quit_idle_removed);
}
