/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <locale.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>

#include <glib.h>
#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>
#include "look-app.h"

static struct termios    _term_attr;
static guint16		 _term_rows;
static guint16		 _term_cols;
static time_t		 _uptime = 0;

static gchar		*_opt_agent = NULL;
static gboolean		 _opt_writable = FALSE;
static gchar		*_opt_quitkey = FALSE;
static gchar		*_opt_logfile = NULL;
static gint		 _opt_timeout = 0;
static gboolean		 _opt_version = FALSE;
static gboolean		 _opt_newwin = FALSE;

static GOptionEntry	_option_entries[] =
{
    { "agent", 'a', 0, G_OPTION_ARG_STRING, &_opt_agent,
      N_("The address of agent server."), "ADDRESS" },

    { "writable", 'w', 0, G_OPTION_ARG_NONE, &_opt_writable,
      N_("Turn on writable mode(keyt-in will be send to server)"), NULL },

    { "timeout", 't', 0, G_OPTION_ARG_INT, &_opt_timeout,
      N_("Specify connection timeout"), "SECONDS" },

    { "quitkey", 'q', 0, G_OPTION_ARG_STRING, &_opt_quitkey,
      N_("Use specified quit key(default is `\\')"), "CHAR" },

    { "newwin", 'n', 0, G_OPTION_ARG_NONE, &_opt_newwin,
      N_("Startup with new terminal window"), NULL },

    { "logfile", 'l', 0, G_OPTION_ARG_FILENAME, &_opt_logfile,
      N_("Log message to file"), "FILE" },

    { "version", 'v', 0, G_OPTION_ARG_NONE, &_opt_version,
      N_("Output version information"), NULL },

    { NULL, }
};

/* access global variable from external */
time_t		look_get_uptime		(void) { return _uptime; }
gint		look_get_opt_timeout	(void) { return _opt_timeout; }
const gchar*	look_get_opt_agent	(void) { return _opt_agent; }

/**
 * quit_application:
 */
static gboolean
quit_application (gpointer data)
{
  look_app_quit (look_app_get ());
  printf ("\r\n");

  return FALSE;
}

/**
 * signal_terminate:
 */
static void
signal_terminate (gint signo)
{
  static gboolean termed = FALSE;

  if (!termed)
    {
      g_idle_add (quit_application, NULL);
      termed = TRUE;
    }
  else
    {
      printf ("\r\n");
      printf (_("Mandatory quit."));
      printf ("\r\n");
      exit (1);
    }
}

/**
 * restore_terminal_attrs:
 *
 * this function will be called when application exit.
 * see atexit () manpage.
 */
static void
restore_terminal_attrs (void)
{
  tcsetattr (STDIN_FILENO, TCSANOW, &_term_attr);
  t_termio_set_winsize (STDOUT_FILENO, _term_rows, _term_cols);
}

/**
 * main:
 */
gint
main (gint argc, gchar *argv[])
{
  GOptionContext *option;
  struct sigaction action;
  LookApp *app;
  GError *error = NULL;

  g_set_prgname (argv[0]);
  g_set_application_name ("Ttyutils Look");

  /* initial i18n support */
  setlocale (LC_ALL, "");
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  textdomain (GETTEXT_PACKAGE);

  if (!t_check_version ())
    return EXIT_FAILURE;

  /* check whether stdin/out is a really terminal device */
  if (!isatty (STDIN_FILENO) || !isatty (STDOUT_FILENO))
    {
      g_warning ("%s: stdin/out is not a tty.", g_get_prgname ());
      syslog (LOG_CRIT, "%s: stdin/stdout is not a tty.", g_get_prgname ());
      return EXIT_FAILURE;
    }
  /* initialize gthread and gobject subsystems. */
  g_thread_init (NULL);
  g_type_init ();

  /* parse command line options */
  option = g_option_context_new ("[server:port]");
  g_option_context_set_summary (option, _("Watch and control user terminal session"));
  g_option_context_set_description (option, WEBSITE);
  g_option_context_add_main_entries (option, _option_entries,
				     GETTEXT_PACKAGE);
  g_option_context_parse (option, &argc, &argv, &error);
  if (error)
    {
      g_warning (_("Parse command line error: %s"), error->message);
      g_error_free (error);
      g_option_context_free (option);
      return EXIT_FAILURE;
    }
  g_option_context_free (option);

  /* print version information and exit. */
  if (_opt_version)
    {
      t_print_version ();
      return EXIT_SUCCESS;
    }
  _uptime = time (NULL);

  /* startup with new terminal window. */
  if (_opt_newwin)
    {
      /* TODO */
    }
  /* catch system signal */
  memset (&action, 0, sizeof (action));
  action.sa_handler = signal_terminate;
  action.sa_flags = 0;
  sigemptyset (&action.sa_mask);

  /* these signals will cause program quit */
  sigaction (SIGTERM, &action, 0);
  sigaction (SIGINT,  &action, 0);
  sigaction (SIGHUP,  &action, 0);

  /* save terminal attributes and size for later restore */
  tcgetattr (STDIN_FILENO, &_term_attr);
  t_termio_get_winsize (STDOUT_FILENO, &_term_rows, &_term_cols);
  g_atexit (restore_terminal_attrs);

  /* add Lua package search path(s) */  
  t_lua_path_add (LUALIBDIR);

  /* strip '--' option if has. */
  t_strip_hyphen_options (&argc, &argv);

  app = look_app_new ();

  /* set app's options from command line */
  look_app_set_writable (app, _opt_writable);
  if (_opt_quitkey)
    {
      /* only lower or upper letter is recognizable */
      if (!g_ascii_isalpha (_opt_quitkey[0]))
	{
	  g_warning (_("Quit key `%c' unrecognizable."), _opt_quitkey[0]);
	  return EXIT_FAILURE;
	}
      look_app_set_quit_key (app, _opt_quitkey[0]);
    }
  /* we can connect to look server directly(address is given in
   * command line argument), or negotiate with agent server to get
   * look servers list. */
  if (argc > 1 && argv[1])
    {
      /* if command line has remains argument. we treat it as
       * look server address */
      look_app_set_server_address (app, argv[1]);
    }
  else
    {
      /* agent server address unspecified, fallback to default agent. */
      if (!_opt_agent)
	{
	  _opt_agent = g_strdup ("tcp:host=localhost,port=5433");

	  g_print (_("No address given, try to connect default agent "
		     "server(`%s') ...\n"), _opt_agent);
	}
    }
  if (t_expire_check ())
    g_idle_add ((GSourceFunc) raise, GUINT_TO_POINTER (SIGTERM));

  /* start watch. */
  look_app_start (app);
  g_object_unref (app);

  /* hint user we are quit. */
  g_print (_("%s quit. elapsed %ld seconds.\n"),
	   g_get_application_name (), time (NULL) - _uptime);
  return EXIT_SUCCESS;
}

/**
 * look_redirect_log:
 *
 * redirect log message to file or syslog.
 */
void
look_redirect_log (void)
{
  if (_opt_logfile)
    t_set_logfile (_opt_logfile);
  else
    openlog (g_get_prgname (), LOG_PID | LOG_CONS, LOG_USER);

  g_log_set_default_handler (t_glib_log_handler, t_get_logfile ());
  g_set_print_handler (t_glib_print_handler);
  g_set_printerr_handler (t_glib_print_handler);
}
