/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <gcrypt.h>
#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif

#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "play-app.h"

#undef CTRL
#define CTRL(x)	((x) & 0x1F)

typedef void	(*PlayCommandCB)	(PlayApp   *app);

typedef struct _PlayCommand	PlayCommand;
struct _PlayCommand
{
  gchar         ch;
  PlayCommandCB cb;
};

struct _PlayAppPrivate
{
  GMainLoop    *mainloop;
  TConsole     *console;
  TDecoder     *decoder;
  guint32       icurr;
  gboolean      paused;
  guint         play_timeout;
  guint         play_idle;
};
#define PLAY_APP_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), PLAY_TYPE_APP, PlayAppPrivate))

static PlayApp	*_app = NULL;

G_DEFINE_TYPE (PlayApp, play_app, G_TYPE_OBJECT);

/* defined in play.c */
extern void	play_redirect_log	(void);
extern guint	play_get_seek		(void);
extern guint	play_get_interval	(void);

static gboolean	play_app_play_timeout		(gpointer user_data);
static void	play_app_play_timeout_destroy	(gpointer user_data);
static gboolean	play_app_play_idle		(gpointer user_data);
static void	play_app_play_idle_destroy	(gpointer user_data);

static void	play_app_pause_cb	(PlayApp *app);
static void	play_app_rewind_cb	(PlayApp *app);
static void	play_app_forward_cb	(PlayApp *app);
static void	play_app_backward_cb	(PlayApp *app);
static void	play_app_reset_cb	(PlayApp *app);
#ifdef DEBUG
static void	play_app_refresh_cb	(PlayApp *app);
#endif

static PlayCommand	_commands[] =
{
    { 'q',		play_app_quit },
    { CTRL('c'), 	play_app_quit },
    { 32,		play_app_pause_cb },
    { 'g',		play_app_rewind_cb },
    { 'f',		play_app_forward_cb },
    { 'b',		play_app_backward_cb },
    { CTRL('s'),	play_app_reset_cb },
#ifdef DEBUG
    { CTRL('l'),	play_app_refresh_cb },
#endif
};

/**
 * play_app_init:
 */
void
play_app_init (PlayApp *app)
{
  app->priv = PLAY_APP_PRIVATE (app);

  app->priv->mainloop = g_main_loop_new (NULL, FALSE);
  app->priv->console = t_console_new ();
  app->priv->decoder = t_decoder_new ();
  app->priv->icurr = 0;
  app->priv->paused = FALSE;
  app->priv->play_timeout = 0;
  app->priv->play_idle = 0;
}

/**
 * play_app_dispose:
 */
void
play_app_dispose (GObject *object)
{
  PlayApp *app = PLAY_APP (object);

  if (app->priv->console)
    {
      g_object_unref (app->priv->console);
      app->priv->console = NULL;
    }
  if (app->priv->decoder)
    {
      g_object_unref (app->priv->decoder);
      app->priv->decoder = NULL;
    }
  G_OBJECT_CLASS (play_app_parent_class)->dispose (object);
}

/**
 * play_app_finalize:
 */
void
play_app_finalize (GObject *object)
{
  PlayApp *app = PLAY_APP (object);

  if (app->priv->mainloop)
    {
      g_main_loop_unref (app->priv->mainloop);
      app->priv->mainloop = NULL;
    }
  if (app->priv->play_timeout > 0)
    {
      g_source_remove (app->priv->play_timeout);
      app->priv->play_timeout = 0;
    }
  if (app->priv->play_idle > 0)
    {
      g_source_remove (app->priv->play_idle);
      app->priv->play_idle = 0;
    }
  G_OBJECT_CLASS (play_app_parent_class)->finalize (object);
}

/**
 * play_app_class_init:
 */
void
play_app_class_init (PlayAppClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (PlayAppPrivate));

  gobject_class->dispose  = play_app_dispose;
  gobject_class->finalize = play_app_finalize;
}

/**
 * play_app_new:
 */
PlayApp*
play_app_new (void)
{
  g_assert (!_app);
  _app = g_object_new (PLAY_TYPE_APP, NULL);
  return _app;
}

/**
 * play_app_get:
 */
PlayApp*
play_app_get (void)
{
  g_assert (_app);
  return _app;
}

/**
 * on_console_readable:
 *
 * called when console is readable.
 */
static void
on_console_readable (TConsole    *console,
		     const gchar *data,
		     guint        size,
		     PlayApp     *app)
{
  gint i;

  g_return_if_fail (T_IS_CONSOLE (console));
  g_return_if_fail (PLAY_IS_APP (app));

  if (size > 1)
    return;

  for (i = 0; i < G_N_ELEMENTS (_commands); i++)
    {
      if (_commands[i].ch == data[0])
	{
	  _commands[i].cb (app);
	  break;
	}
    }
}

/**
 * on_console_hangup:
 *
 * called when console hangup.
 */
static void
on_console_hangup (TConsole *console,
		   PlayApp  *app)
{
  g_return_if_fail (PLAY_IS_APP (app));
  play_app_quit (app);
}

/**
 * play_app_play_idle:
 */
static gboolean
play_app_play_idle (gpointer user_data)
{
  PlayApp *app = user_data;
  TRecordIndex *indexes;
  guint interval;
  guint32 index, istart, iend;

  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);

  indexes = t_decoder_get_indexes (app->priv->decoder,
				   &istart, &iend);
  index = app->priv->icurr - istart;
  interval = indexes[index].interval * 1000;

  /* don't wait for too long time. */
  if (play_get_interval () > 0)
    interval = MIN (interval, play_get_interval ());

  if (app->priv->play_timeout == 0)
    {
      app->priv->play_timeout =
       g_timeout_add_full (G_PRIORITY_DEFAULT, interval,
			   play_app_play_timeout, app,
			   play_app_play_timeout_destroy);
    }
  return FALSE;
}

/**
 * play_app_play_idle_destroy:
 */
static void
play_app_play_idle_destroy (gpointer user_data)
{
  PlayApp *app = user_data;

  g_return_if_fail (PLAY_IS_APP (app));

  app->priv->play_idle = 0;
}

/**
 * play_app_play_timeout:
 */
static gboolean
play_app_play_timeout (gpointer user_data)
{
  PlayApp *app = user_data;
  const gchar *data;
  guint16 size;
  guint32 nindexes;
  GError *error = NULL;

  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);

  nindexes = t_decoder_get_nindexes (app->priv->decoder);
  if (app->priv->icurr >= nindexes)
    {
      play_app_quit (app);
      return FALSE;
    }
  /* t_decoder_get_index_data() will auto seek if index
   * out of cache. */
  data = t_decoder_get_index_data (app->priv->decoder,
				   app->priv->icurr,
				   &size, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      play_app_quit (app);
      return FALSE;
    }
  t_console_write (app->priv->console, data, size);

  app->priv->icurr++;

  if (!app->priv->paused)
    {
      if (app->priv->play_idle == 0)
	{
	  app->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    play_app_play_idle, app,
			    play_app_play_idle_destroy);
	}
    }
  return FALSE;
}

/**
 * play_app_play_timeout_destroy:
 */
static void
play_app_play_timeout_destroy (gpointer user_data)
{
  PlayApp *app = user_data;

  g_return_if_fail (PLAY_IS_APP (app));

  app->priv->play_timeout = 0;
}

/**
 * play_app_start_play:
 */
static gboolean
play_app_start_play (PlayApp *app)
{
  guint seek, interval = 0;
  guint32 i, nindexes, istart, iend, index;
  TRecordIndex *indexes;
  const gchar *data;
  guint16 size;
  GError *error = NULL;

  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);

  /* seek to time position to playback. */
  seek = play_get_seek ();
  if (seek > 0)
    {
      nindexes = t_decoder_get_nindexes (app->priv->decoder);

      for (i = 0; i < nindexes; i++)
	{
	  if (!t_decoder_has_index (app->priv->decoder, app->priv->icurr))
	    {
	      t_decoder_seek (app->priv->decoder, app->priv->icurr, &error);
	      if (error)
		{
		  g_warning ("%s", error->message);
		  g_error_free (error);
		  return FALSE;
		}
	    }
	  indexes = t_decoder_get_indexes (app->priv->decoder, &istart, &iend);
	  index = app->priv->icurr - istart;
	  interval += indexes[index].interval * 1000;

	  /* output the data near the position. */
	  if (seek - interval < 5000)
	    {
	      data = t_decoder_get_index_data (app->priv->decoder,
					       app->priv->icurr, &size, &error);
	      if (error)
		{
		  g_warning ("%s", error->message);
		  g_error_free (error);
		  return FALSE;
		}
	      t_console_write (app->priv->console, data, size);
	    } 
	  app->priv->icurr++;

	  if (interval >= seek)
	    break;
	}
      if (i == nindexes)
	{
	  g_warning (_("Seek out of range"));
	  return FALSE;
	}
    }
  app->priv->play_idle = g_idle_add_full (G_PRIORITY_HIGH_IDLE,
					  play_app_play_idle, app,
					  play_app_play_idle_destroy);
  return TRUE;
}

/**
 * apply_terminal_settings:
 */
static void
apply_terminal_settings (PlayApp     *app,
			 TRecordHead *head)
{
  guint16 cols, rows;

  g_return_if_fail (PLAY_IS_APP (app));
  g_return_if_fail (head);

  g_setenv ("TERM", head->term, TRUE);

  t_termio_get_winsize (STDOUT_FILENO, &rows, &cols);
  if (rows < head->nrows || cols < head->ncols)
    {
      g_message (_("Terminal size small than required %dx%d"),
		 head->ncols, head->nrows);
      t_termio_set_winsize (STDOUT_FILENO, head->nrows, head->ncols);
    }
}

/**
 * play_get_password:
 *
 * read password from user console.
 */
static gboolean
play_app_get_password (TDecoder *decoder,
		       gpointer  user_data)
{
  PlayApp *app = user_data;
  gchar *passwd;
  gboolean ret = TRUE;
  GError *error = NULL;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);

  g_print (_("File is encrypted, please give the password\n"));

  passwd = gcry_malloc_secure (512);

  while (1)
    {
      gint n = t_termio_getpass (passwd, 512, &error);
      g_print ("\n");
      if (error)
	{
	  if (error->code == G_IO_ERROR_NO_SPACE)
	    {
	      g_print (_("Password too long, please try again\n"));
	      g_error_free (error);
	      error = NULL;
	      continue;
	    }
	  else if (error->code == G_IO_ERROR_INVALID_DATA)
	    {
	      g_print (_("Bad password character, please try again\n"));
	      g_error_free (error);
	      error = NULL;
	      continue;
	    }
	  g_print ("%s\n", error->message);
	  g_error_free (error);
	  ret = FALSE;
	  break;
	}
      if (n == 0)
	{
	  ret = FALSE;
	  break;
	}
      if (t_decoder_set_passwd (decoder, passwd, n))
	break;
      else
	g_print (_("Password incorrent! please try again\n"));
    }
  gcry_free (passwd);

  return ret;
}

/**
 * play_app_start:
 */
gboolean
play_app_start (PlayApp     *app,
		const gchar *pathname)
{
  TRecordHead *head;
  const gchar *charset;
  GError *error = NULL;

  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);
  g_return_val_if_fail (pathname, FALSE);

  /* open record file. */
  t_decoder_open (app->priv->decoder, pathname,
		  play_app_get_password, app, NULL, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return FALSE;
    }
  head = t_decoder_get_head (app->priv->decoder);

  g_get_charset (&charset);
  if (g_ascii_strcasecmp (charset, head->charset) != 0)
    {
      g_print (_("The record file has charset encoding `%s', "
		 "which is differ from local charset encoding `%s'.\n"),
	       head->charset, charset);
      g_print (_("Continue?[y/n]: "));

      /* need not restore after this called since application
       * quit will restore terminal attributes. */
      t_termio_set_lflag (STDIN_FILENO, ICANON, FALSE);
      if (getchar () == 'n')
	{
	  t_decoder_close (app->priv->decoder);
	  g_print ("\n");
	  return FALSE;
	}
    }
  t_term_clear_screen ();

  g_print (_("Creator: %s\n"), head->creator);
  g_print (_("Duration: %.2f seconds.\n"), head->duration);
  g_print (_("Press `q' or `CTRL+C' to quit.\n"));

  /* redirect log message. */
  play_redirect_log ();

  /* take terminal into raw mode. */
  t_termio_makeraw (STDIN_FILENO);

  apply_terminal_settings (app, head);

  g_signal_connect (app->priv->console, "readable",
		    G_CALLBACK (on_console_readable), app);
  g_signal_connect (app->priv->console, "hangup",
		    G_CALLBACK (on_console_hangup), app);
  t_console_start (app->priv->console);

  if (!play_app_start_play (app))
    return FALSE;

  g_main_loop_run (app->priv->mainloop);
  return TRUE;
}

/**
 * play_app_quit_idle:
 */
static gboolean
play_app_quit_idle (gpointer data)
{
  PlayApp *app = PLAY_APP (data);

  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);

  return (t_console_get_buffer_size (app->priv->console) > 0);
}

/**
 * play_app_quit_idle_removed:
 */
static void
play_app_quit_idle_removed (gpointer data)
{
  PlayApp *app = PLAY_APP (data);

  g_return_if_fail (PLAY_IS_APP (app));

  if (app->priv->mainloop)
    g_main_loop_quit (app->priv->mainloop);

  printf ("\r\n");
}

/**
 * play_app_quit:
 *
 * quit application and free resources.
 */
void
play_app_quit (PlayApp *app)
{
  g_return_if_fail (PLAY_IS_APP (app));

  if (app->priv->play_timeout > 0)
    g_source_remove (app->priv->play_timeout);

  if (app->priv->play_idle > 0)
    g_source_remove (app->priv->play_idle);

  t_console_flush (app->priv->console);

  g_idle_add_full (G_PRIORITY_DEFAULT,
		   play_app_quit_idle, app,
		   play_app_quit_idle_removed);
}

/**
 * play_app_pause_cb:
 */
static void
play_app_pause_cb (PlayApp *app)
{
  g_return_if_fail (PLAY_IS_APP (app));

  app->priv->paused = !app->priv->paused;

  if (!app->priv->paused)
    {
      if (app->priv->play_idle == 0)
	{
	  app->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    play_app_play_idle, app,
			    play_app_play_idle_destroy);
	}
    }
  else
    {
      if (app->priv->play_timeout > 0)
	g_source_remove (app->priv->play_timeout);

      if (app->priv->play_idle > 0)
	g_source_remove (app->priv->play_idle);
    }
}

/**
 * play_app_rewind_cb:
 */
static void
play_app_rewind_cb (PlayApp *app)
{
  g_return_if_fail (PLAY_IS_APP (app));

  if (app->priv->play_timeout > 0)
    g_source_remove (app->priv->play_timeout);

  app->priv->icurr = 0;

  t_term_clear_screen ();

  if (app->priv->play_idle == 0)
    {
      app->priv->play_idle =
       g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			play_app_play_idle, app,
			play_app_play_idle_destroy);
    }
}

/**
 * play_app_forward:
 */
static gboolean
play_app_forward (PlayApp *app,
		  guint32  step)
{
  guint32 nindexes, i;
  const gchar *data;
  guint16 size;
  GError *error = NULL;

  g_return_val_if_fail (PLAY_IS_APP (app), FALSE);

  nindexes = t_decoder_get_nindexes (app->priv->decoder);

  for (i = 0; i < step; i++, app->priv->icurr++)
    {
      if (app->priv->icurr >= nindexes)
	return FALSE;

      data = t_decoder_get_index_data (app->priv->decoder,
				       app->priv->icurr,
				       &size, &error);
      if (error)
	{
	  g_warning ("%s", error->message);
	  g_error_free (error);
	  return FALSE;
	}
      t_console_write (app->priv->console, data, size);
    }
  return TRUE;
}

/**
 * play_app_forward_cb:
 */
static void
play_app_forward_cb (PlayApp *app)
{
  guint32 step, nindexes;

  g_return_if_fail (PLAY_IS_APP (app));

  if (app->priv->play_timeout > 0)
    g_source_remove (app->priv->play_timeout);

  nindexes = t_decoder_get_nindexes (app->priv->decoder);
  step = nindexes / 100;

  if (!play_app_forward (app, MIN (step, 20)))
    {
      play_app_quit (app);
      return;
    }
  if (!app->priv->paused)
    {
      if (app->priv->play_idle == 0)
	{
	  app->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    play_app_play_idle, app,
			    play_app_play_idle_destroy);
	}
    }
}

/**
 * play_app_backward:
 */
static void
play_app_backward (PlayApp *app,
		   guint32  index)
{
  TRecordHead *head;
  TRecordIndex *indexes;
  guint32 i, j, istart, iend;
  gssize expect;
  GError *error = NULL;

  g_return_if_fail (PLAY_IS_APP (app));

  t_term_clear_screen ();

  head = t_decoder_get_head (app->priv->decoder);
  expect = head->ncols * head->nrows * 2;

  /* try to refresh screen with previous data. */
  for (i = 0, j = index; i < 300 && j > 0; i++, j--)
    {
      if (!t_decoder_has_index (app->priv->decoder, j))
	{
	  t_decoder_seek (app->priv->decoder, j, &error);
	  if (error)
	    {
	      g_warning ("%s", error->message);
	      g_error_free (error);
	      return;
	    }
	}
      indexes = t_decoder_get_indexes (app->priv->decoder,
				       &istart, &iend);
      expect -= indexes[j - istart].size;
      if (expect <= 0)
	break;
    }
  app->priv->icurr = j;

  play_app_forward (app, index - j);
}

/**
 * play_app_backward_cb:
 */
static void
play_app_backward_cb (PlayApp *app)
{
  guint32 nindexes, step;

  g_return_if_fail (PLAY_IS_APP (app));

  if (app->priv->play_timeout > 0)
    g_source_remove (app->priv->play_timeout);

  nindexes = t_decoder_get_nindexes (app->priv->decoder);
  step = MIN (nindexes / 100, 20);

  play_app_backward (app, MAX (app->priv->icurr, step) - step);

  if (!app->priv->paused)
    {
      if (app->priv->play_idle == 0)
	{
	  app->priv->play_idle =
	   g_idle_add_full (G_PRIORITY_HIGH_IDLE,
			    play_app_play_idle, app,
			    play_app_play_idle_destroy);
	}
    }
}

/**
 * play_app_refresh_cb:
 */
#ifdef DEBUG
static void
play_app_refresh_cb (PlayApp *app)
{
  g_return_if_fail (PLAY_IS_APP (app));
  t_console_refresh (app->priv->console);
}
#endif

/**
 * play_app_reset_cb:
 */
static void
play_app_reset_cb (PlayApp *app)
{
  g_return_if_fail (PLAY_IS_APP (app));
  t_term_reset ();
}
