/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <locale.h>
#include <unistd.h>
#include <termios.h>
#include <syslog.h>
#include <stdlib.h>
#include <gcrypt.h>

#include <glib.h>
#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "play-app.h"

static struct termios    _term_attr;
static guint16           _term_rows;
static guint16           _term_cols;
static time_t            _uptime;
static gboolean		 _opt_version = FALSE;
static gboolean		 _opt_summary = FALSE;
static gboolean		 _opt_commands = FALSE;
static gchar		*_opt_logfile = NULL;
static guint		 _opt_seek = 0;
static guint		 _opt_interval = 0;
static gboolean		 _opt_clear = FALSE;
static gboolean		 _opt_newwin = FALSE;

static GOptionEntry	option_entries[] =
{
    { "summary", 's', 0, G_OPTION_ARG_NONE, &_opt_summary,
      N_("Print file summary infomation, don't playback"), NULL },

    { "commands", 'm', 0, G_OPTION_ARG_NONE, &_opt_commands,
      N_("Print playback commands"), NULL },

    { "seek", 'k', 0, G_OPTION_ARG_INT, &_opt_seek,
      N_("Seek to give time position"), "MS" },

    { "interval", 'i', 0, G_OPTION_ARG_INT, &_opt_interval,
      N_("Maximum interval time, in milliseconds"), "MS" },

    { "clear", 'c', 0, G_OPTION_ARG_NONE, &_opt_clear,
      N_("Clear terminal screen when exit"), NULL },

    { "newwin", 'n', 0, G_OPTION_ARG_NONE, &_opt_newwin,
      N_("Startup with new terminal window"), NULL },

    { "logfile", 'l', 0, G_OPTION_ARG_FILENAME, &_opt_logfile,
      N_("Log message to file"), "FILE" },

    { "version", 'v', 0, G_OPTION_ARG_NONE, &_opt_version,
      N_("Output version information"), NULL },

    { NULL, }
};


/**
 * print_commands:
 */
static gint
print_commands (void)
{
  struct
    {
      const gchar *key;
      const gchar *desc;
    }
  commands[] =
    {
	{ "q, ^C ", N_("Quit") },
	{ "Space ", N_("Switch between pause and play") },
	{ " g    ", N_("Rewind") },
	{ " f    ", N_("Forward") },
	{ " b    ", N_("Backward") },
	{ "^S    ", N_("Reset terminal") },
#ifdef DEBUG
	{ "^L    ", N_("Refresh screen") },
#endif
    };
  gint i;

  g_print ("\n%s commands:\n\n", g_get_application_name ());

  for (i = 0; i < G_N_ELEMENTS (commands); i++)
    {
      g_print ("  %s : %s\n", commands[i].key,
	       gettext (commands[i].desc));
    }
  g_print (_("\n`^X' means `CTRL+X'(hold `Ctrl' key and press character `X').\n"));

  return 0;
}

/**
 * signal_terminate:
 */
static void
signal_terminate (gint signo)
{
  static gboolean termed = FALSE;

  if (!termed)
    {
      play_app_quit (play_app_get ());
      termed = TRUE;
    }
  else
    {
      printf ("\r\n");
      printf (_("Mandatory quit."));
      printf ("\r\n");
      exit (1);
    }
}

/**
 * restore_terminal_attrs:
 */
static void
restore_terminal_attrs (void)
{
  tcsetattr (STDIN_FILENO, TCSANOW, &_term_attr);
  t_termio_set_winsize (STDOUT_FILENO, _term_rows, _term_cols);
}

/**
 * decoder_passwd_func:
 *
 * always return TRUE to get record head info.
 */
static gboolean
decoder_passwd_func (TDecoder *decoder,
		     gpointer  data)
{
  return TRUE;
}

/**
 * print_summary:
 */
static gint
print_summary (const gchar *filename)
{
  TDecoder *decoder;
  TRecordHead *head;
  time_t tim;
  struct tm *tmp;
  gchar buff[128];
  GError *error = NULL;

  g_return_val_if_fail (filename, 1);

  decoder = t_decoder_new ();
  t_decoder_open (decoder, filename, decoder_passwd_func,
		  NULL, NULL, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      g_object_unref (decoder);
      return 1;
    }
  head = t_decoder_get_head (decoder);

  g_print ("\n");
  g_print (_("Filename: %s"), filename);
  g_print ("\n");
  g_print ("\n");

  g_print (_("Summary:"));
  g_print ("\n");

  g_print ("  %s : %s\n", _("Creator"), head->creator);

  tim = (time_t) head->starting;
  tmp = localtime (&tim);
  strftime (buff, sizeof (buff), "%x %X", tmp);
  g_print ("  %s : %s\n", _("Starting"), buff);

  tim = (time_t) head->duration;
  tmp = gmtime (&tim);
  strftime (buff, sizeof (buff), "%X", tmp);
  g_print ("  %s : %d %s\n", _("Duration"), tmp->tm_yday, buff);

  g_print ("\n");

  g_print (_("Detail:"));
  g_print ("\n");

  g_print ("  %s : %d\n", _("Version"), head->version);
  g_print ("  %s : %s\n", _("Charset"), head->charset);
  g_print ("  %s : %d\n", _("Compress level"), head->clevel);
  g_print ("  %s : %s %dx%d\n", _("Terminal"), head->term,
	   head->ncols, head->nrows);
  g_print ("  %s : %s\n", _("Encrypted"),
	   head->encrypted ? _("Yes") : _("No"));

  g_print ("\n");

  g_object_unref (decoder);
  return 0;
}

/**
 * start_with_new_window:
 */
static gboolean
start_with_new_window (gint   argc,
		       gchar *argv[])
{
  TDecoder *decoder;
  TRecordHead *head;
  gchar *prog;
  GString *command;
  GError *error = NULL;

  /* we can only work within window environment since
   * we need open new window.
   * a sophisticated way is needed since the way we use
   * here is too poor. */
  if (!g_getenv ("DISPLAY"))
    return FALSE;

  /* open the playback file to get require terminal size. */
  decoder = t_decoder_new ();
  t_decoder_open (decoder, argv[1], decoder_passwd_func,
		  NULL, NULL, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      g_object_unref (decoder);
      return FALSE;
    }
  head = t_decoder_get_head (decoder);

  command = g_string_new (NULL);

  /* find terminal application from system.
   * we can use gnome-terminal and vte, and other terminal which
   * support geometry command line option. */
#ifndef DEBUG
  prog = g_find_program_in_path ("gnome-terminal");
  if (prog)
    {
      g_string_append_printf (command, "%s --hide-menubar --title=\"%s\" "
			      "--geometry=%dx%d+0+0 --command=\"%s ",
			      prog, g_get_application_name (),
			      head->ncols, head->nrows, argv[0]);
      goto next;
    }
#endif
  prog = g_find_program_in_path ("vte");
  if (prog)
    {
      g_string_append_printf (command, "%s --geometry=%dx%d+0+0 "
			      "--command=\"%s ",
			      prog, head->ncols, head->nrows, argv[0]);
      goto next;
    }
  g_object_unref (decoder);
  g_string_free (command, TRUE);

  g_warning (_("No terminal application can be found."));
  return FALSE;

next:
  g_object_unref (decoder);
  g_free (prog);

  /* append ttyplay options to command line. */
  if (_opt_clear)
    g_string_append_printf (command, "-c ");
  if (_opt_seek > 0)
    g_string_append_printf (command, "-k %d ", _opt_seek);
  if (_opt_interval > 0)
    g_string_append_printf (command, "-i %d ", _opt_interval);
  if (_opt_logfile)
    g_string_append_printf (command, "-l %s ", _opt_logfile);

  g_string_append_printf (command, " %s\"", argv[1]);

  /* execute the command line async. */
  g_spawn_command_line_async (command->str, &error);
  g_string_free (command, TRUE);
  if (error)
    {
      g_warning ("%s", error->message);
      return FALSE;
    }
  return TRUE;
}

/**
 * main:
 */
gint
main (gint argc, gchar *argv[])
{
  GOptionContext *option;
  struct sigaction action;
  PlayApp *app;
  GError *error = NULL;

  g_set_prgname (argv[0]);
  g_set_application_name ("Ttyutils Player");

  setlocale (LC_ALL, "");
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  if (!t_check_version ())
    return EXIT_FAILURE;

  /* initialize gobject subsystems. */
  g_type_init ();

  /* initialize libgcrypt subsystems. */
#ifdef DEBUG
  gcry_control (GCRYCTL_SET_VERBOSITY, 1);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1);
#else
  gcry_control (GCRYCTL_SET_VERBOSITY, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 1);
#endif

  gcry_set_log_handler (t_gcrypt_log_handler, NULL);
  if (!gcry_check_version (GCRYPT_VERSION))
    {
      g_warning (_("libgcrypt version %s or above is required."),
		 GCRYPT_VERSION);
      return EXIT_FAILURE;
    }
  gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN, NULL);
  gcry_control (GCRYCTL_INIT_SECMEM, 16 * 1024, 0);
  gcry_control (GCRYCTL_RESUME_SECMEM_WARN, NULL);
  gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);

  /* parse command line options. */
  option = g_option_context_new (_("Record file"));
  g_option_context_set_summary (option, _("Playback terminal session record file"));
  g_option_context_set_description (option, WEBSITE);
  g_option_context_add_main_entries (option, option_entries, GETTEXT_PACKAGE);
  g_option_context_parse (option, &argc, &argv, &error);
  if (error)
    {
      g_warning (_("Parse command line error: %s"), error->message);
      g_error_free (error);
      g_option_context_free (option);
      return EXIT_FAILURE;
    }
  g_option_context_free (option);

  if (_opt_version)
    {
      t_print_version ();
      return EXIT_SUCCESS;
    }
  if (_opt_commands)
    {
      return print_commands ();
    }
  /* strip '--' option if has. */
  t_strip_hyphen_options (&argc, &argv);

  /* we need a file to play. */
  if (argc < 2 || !argv[1])
    {
      g_print (_("%s: No input file.\n"), g_get_prgname ());
      return EXIT_FAILURE;
    }
  if (_opt_summary)
    {
      return print_summary (argv[1]);
    }
  if (_opt_newwin)
    {
      /* continue run if cannot open the new window. */
      if (start_with_new_window (argc, argv))
	return EXIT_SUCCESS;

      g_print (_("Unable to open new terminal window, "
		 "continue with current terminal.\n"));
    }
  _uptime = time (NULL);

  /* save terminal attributes for later restore. */
  tcgetattr (STDIN_FILENO, &_term_attr);
  t_termio_get_winsize (STDOUT_FILENO, &_term_rows, &_term_cols);
  g_atexit (restore_terminal_attrs);

  memset (&action, 0, sizeof (action));
  action.sa_handler = signal_terminate;
  sigemptyset (&action.sa_mask);

  sigaction (SIGTERM, &action, 0);
  sigaction (SIGINT, &action, 0);
  sigaction (SIGHUP, &action, 0);

  if (t_expire_check ())
    g_idle_add ((GSourceFunc) raise, GUINT_TO_POINTER (SIGTERM));

  app = play_app_new ();

  /* if start failed, we don't clear the screen, even if
   * --clear option is specified. */
  if (!play_app_start (app, argv[1]))
    _opt_clear = FALSE;

  g_object_unref (app);

  if (_opt_clear)
    {
      t_term_clear_screen ();
      t_term_reset ();

      /* seem gnome-terminal need sometime to reset.
       * we wait for 1 second.
       * xterm and konsole need not wait. */
      g_usleep (500 * 1000);
    }
  printf (_("%s quit, elapsed %ld seconds."),
	  g_get_application_name (), time (NULL) - _uptime);
  printf ("\r\n");

  return EXIT_SUCCESS;
}

/**
 * play_redirect_log:
 *
 * redirect log message to file or syslog.
 */
void
play_redirect_log (void)
{
  if (_opt_logfile)
    t_set_logfile (_opt_logfile);
  else
    openlog (g_get_prgname (), LOG_PID | LOG_CONS, LOG_USER);

  g_log_set_default_handler (t_glib_log_handler, t_get_logfile ());
  g_set_print_handler (t_glib_print_handler);
  g_set_printerr_handler (t_glib_print_handler);
}

/**
 * play_get_seek:
 */
guint
play_get_seek (void)
{
  return _opt_seek;
}

/**
 * play_get_interval:
 */
guint
play_get_interval (void)
{
  return _opt_interval;
}
