/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <stdlib.h>
#include <termios.h>
#include <gcrypt.h>

#include <glib/gi18n.h>
#include <ttyutils/ttyutils.h>

#include "record-agent.h"
#include "record-app.h"
#include "record-script.h"

struct _RecordAppPrivate
{
  GMainLoop     *mainloop;
  GKeyFile      *keyfile;
  RecordLookd   *lookd;
  RecordEncoder *encoder;
  RecordProcess *process;
  TConsole      *console;
  GThread       *script_thread;
  GAsyncQueue   *script_queue;
  GList         *script_files;
  RecordAgent   *agent_proxy;
  GCancellable  *agent_cancellable;
  guint          agent_timeout;
  gboolean       agent_fail_exit;
};
#define _APP_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), RECORD_TYPE_APP, RecordAppPrivate))

G_DEFINE_TYPE (RecordApp, record_app, G_TYPE_OBJECT);

static RecordApp	*_application = NULL;
static GStaticMutex	 _script_mutex = G_STATIC_MUTEX_INIT;

extern gboolean	record_redirect_log	(void);

static void	on_console_readable	(TConsole      *console,
					 const gchar   *data,
					 guint          size,
					 RecordApp     *app);
static void	on_console_hangup	(TConsole      *console,
					 RecordApp     *app);
static void	on_process_readable	(RecordProcess *process,
					 const gchar   *data,
					 guint          size,
					 RecordApp     *app);
static void	on_process_hangup	(RecordProcess *process,
					 RecordApp     *app);
static void	on_process_child_start	(RecordProcess *process,
					 GPid           pid,
					 RecordApp     *app);
static void	on_process_child_exit	(RecordProcess *process,
					 GPid           pid,
					 gint           status,
					 RecordApp     *app);
static gboolean	record_conf_has		(RecordApp     *app,
					 const gchar   *group,
					 const gchar   *key);
static gboolean	record_conf_get_bool	(RecordApp     *app,
					 const gchar   *group,
					 const gchar   *key);
static gint	record_conf_get_int	(RecordApp     *app,
					 const gchar   *group,
					 const gchar   *key);
static gchar*	record_conf_get_string	(RecordApp     *app,
					 const gchar   *group,
					 const gchar   *key,
					 gboolean       locale);
static gchar**	record_conf_get_strlist	(RecordApp     *app,
					 const gchar   *group,
					 const gchar   *key,
					 gsize         *length);
#define record_conf_get_str(app, group, key)	record_conf_get_string(app, group, key, FALSE)
#define record_conf_get_lstr(app, group, key)	record_conf_get_string(app, group, key, TRUE)

/**
 * record_app_init:
 */
static void
record_app_init (RecordApp *app)
{
  app->priv = _APP_GET_PRIVATE (app);

  app->priv->mainloop = g_main_loop_new (NULL, FALSE);
  app->priv->process  = record_process_new ();
  app->priv->lookd    = NULL;
  app->priv->encoder  = NULL;
  app->priv->console  = t_console_new ();
  app->priv->keyfile  = g_key_file_new ();
  app->priv->agent_proxy = NULL;
  app->priv->agent_cancellable = NULL;
  app->priv->agent_timeout = 0;
  app->priv->agent_fail_exit = FALSE;
  app->priv->script_thread = NULL;
  app->priv->script_queue  = NULL;
  app->priv->script_files  = NULL;

  /* connect console signals */
  g_signal_connect (app->priv->console, "readable",
		    G_CALLBACK (on_console_readable), app);
  g_signal_connect (app->priv->console, "hangup",
		    G_CALLBACK (on_console_hangup), app);

  /* connect child process signals */
  g_signal_connect (app->priv->process, "readable",
		    G_CALLBACK (on_process_readable), app);
  g_signal_connect (app->priv->process, "hangup",
		    G_CALLBACK (on_process_hangup), app);
  g_signal_connect (app->priv->process, "child-start",
		    G_CALLBACK (on_process_child_start), app);
  g_signal_connect (app->priv->process, "child-exit",
		    G_CALLBACK (on_process_child_exit), app);
}

/**
 * record_app_dispose:
 */
static void
record_app_dispose (GObject *object)
{
  RecordApp *app = RECORD_APP (object);

  if (app->priv->agent_proxy)
    app->priv->agent_proxy = (g_object_unref (app->priv->agent_proxy), NULL);

  if (app->priv->process)
    app->priv->process = (g_object_unref (app->priv->process), NULL);

  if (app->priv->console)
    app->priv->console = (g_object_unref (app->priv->console), NULL);

  if (app->priv->lookd)
    app->priv->lookd = (g_object_unref (app->priv->lookd), NULL);

  G_OBJECT_CLASS (record_app_parent_class)->dispose (object);
}

/**
 * record_app_finalize:
 */
static void
record_app_finalize (GObject *object)
{
  RecordApp *app = RECORD_APP (object);

  if (app->priv->encoder)
    app->priv->encoder = (g_object_unref (app->priv->encoder), NULL);

  if (app->priv->keyfile)
    app->priv->keyfile = (g_key_file_free (app->priv->keyfile), NULL);

  if (app->priv->mainloop)
    app->priv->mainloop = (g_main_loop_unref (app->priv->mainloop), NULL);

  G_OBJECT_CLASS (record_app_parent_class)->finalize (object);
}

/**
 * record_app_class_init:
 */
static void
record_app_class_init (RecordAppClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (RecordAppPrivate));

  gobject_class->dispose  = record_app_dispose;
  gobject_class->finalize = record_app_finalize;
}

/**
 * record_conf_has:
 */
static inline gboolean
record_conf_has (RecordApp   *app,
		 const gchar *group,
		 const gchar *key)
{
  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  return g_key_file_has_key (app->priv->keyfile,
			     group, key, NULL);
}

/**
 * record_conf_get_bool:
 */
static inline gboolean
record_conf_get_bool (RecordApp   *app,
		      const gchar *group,
		      const gchar *key)
{
  gboolean ret;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  ret = g_key_file_get_boolean (app->priv->keyfile,
				group, key, &error);
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}

/**
 * record_conf_get_int:
 */
static inline gint
record_conf_get_int (RecordApp   *app,
		     const gchar *group,
		     const gchar *key)
{
  gint ret;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  ret = g_key_file_get_integer (app->priv->keyfile,
				group, key, &error);
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}

/**
 * record_conf_get_string:
 */
static inline gchar*
record_conf_get_string (RecordApp   *app,
			const gchar *group,
			const gchar *key,
			gboolean     locale)
{
  gchar *ret;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  if (locale)
    {
      ret = g_key_file_get_locale_string (app->priv->keyfile,
					  group, key, NULL, &error);
    }
  else
    {
      ret = g_key_file_get_string (app->priv->keyfile,
				   group, key, &error);
    }
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}

/**
 * record_conf_get_strlist:
 */
static inline gchar**
record_conf_get_strlist (RecordApp     *app,
			 const gchar   *group,
			 const gchar   *key,
			 gsize         *length)
{
  gchar **ret;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (group, FALSE);
  g_return_val_if_fail (key, FALSE);
  g_return_val_if_fail (app->priv->keyfile, FALSE);

  ret = g_key_file_get_string_list (app->priv->keyfile,
				    group, key, length, &error);
  if (error)
    {
      if (error->code != G_KEY_FILE_ERROR_GROUP_NOT_FOUND &&
	  error->code != G_KEY_FILE_ERROR_KEY_NOT_FOUND)
	{
	  g_warning (_("Bad configure entry `%s/%s': %s"),
		     group, key, error->message);
	}
      g_error_free (error);
    }
  return ret;
}
/**
 * record_app_new:
 */
RecordApp*
record_app_new (void)
{
  if (!_application)
    _application = g_object_new (RECORD_TYPE_APP, NULL);

  return _application;
}

/**
 * record_app_get:
 */
RecordApp*
record_app_get (void)
{
  return _application;
}

/**
 * record_app_load_config_file:
 */
gboolean
record_app_load_config_file (RecordApp   *app,
			     const gchar *filename)
{
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (filename, FALSE);

  g_key_file_load_from_file (app->priv->keyfile, filename,
			     G_KEY_FILE_NONE, &error);
  if (error)
    {
      g_warning (_("Failed to load configure file `%s': %s"),
		 filename, error->message);
      g_error_free (error);
      return FALSE;
    }
  return TRUE;
}

/**
 * record_app_get_keyfile:
 */
GKeyFile*
record_app_get_keyfile (RecordApp *app)
{
  g_return_val_if_fail (RECORD_IS_APP (app), NULL);
  return app->priv->keyfile;
}

/**
 * record_app_get_process:
 */
RecordProcess*
record_app_get_process (RecordApp *app)
{
  g_return_val_if_fail (RECORD_IS_APP (app), NULL);
  return app->priv->process;
}

/**
 * record_app_get_console:
 */
TConsole*
record_app_get_console (RecordApp *app)
{
  g_return_val_if_fail (RECORD_IS_APP (app), NULL);
  return app->priv->console;
}

/**
 * record_app_get_lookd:
 */
RecordLookd*
record_app_get_lookd (RecordApp *app)
{
  g_return_val_if_fail (RECORD_IS_APP (app), NULL);
  return app->priv->lookd;
}

/**
 * record_app_get_encoder:
 */
RecordEncoder*
record_app_get_encoder (RecordApp *app)
{
  g_return_val_if_fail (RECORD_IS_APP (app), NULL);
  return app->priv->encoder;
}

/**
 * on_agent_proxy_login:
 *
 * called when call proxy login method finish.
 */
static void
on_agent_proxy_login (GObject      *source,
		      GAsyncResult *res,
		      gpointer      user_data)
{
  RecordAgent *agent = RECORD_AGENT (source);
  RecordApp *app = user_data;
  gboolean ret;
  GError *error = NULL;

  g_return_if_fail (RECORD_IS_AGENT (agent));
  g_return_if_fail (RECORD_IS_APP (app));

  ret = record_agent_login_finish (agent, res, &error);
  if (error)
    {
      g_dbus_error_strip_remote_error (error);
      g_critical ("Failed to login to agent: %s", error->message);
      g_error_free (error);

      if (app->priv->agent_fail_exit)
	record_app_quit (app);

      return;
    }
}

/**
 * on_agent_proxy_created:
 *
 * called when proxy agent succeed or failure.
 */
static void
on_agent_proxy_created (GObject      *source,
			GAsyncResult *res,
			gpointer      user_data)
{
  RecordApp *app = user_data;
  gchar *hostname, *command;
  const gchar *ptsname, **argv;
  guint16 lookport;
  GError *error = NULL;

  g_return_if_fail (RECORD_IS_APP (app));

  /* get the proxy object. */
  app->priv->agent_proxy = record_agent_new_finish (res, &error);
  if (error)
    {
      g_warning ("Unable to connect agent: %s", error->message);
      g_error_free (error);

      /* quit the applicatoin if failed to make connection. */
      if (app->priv->agent_fail_exit)
	record_app_quit (app);

      return;
    }
  /* collect login information. */
  hostname = record_conf_get_str (app, "Agent", "hostname");
  if (!hostname || !hostname[0])
    {
      g_free (hostname);
      hostname = g_strdup (g_get_host_name ());
    }
  ptsname = record_process_get_ptsname (app->priv->process);

  if (app->priv->lookd)
    lookport = record_lookd_get_port (app->priv->lookd);
  else
    lookport = 0;

  argv = record_process_get_argv (app->priv->process);
  command = g_strjoinv (" ", (gchar**) argv);

  /* login to agent. */
  if (!record_agent_login (app->priv->agent_proxy, hostname,
			   ptsname, command, lookport,
			   app->priv->agent_cancellable,
			   on_agent_proxy_login, app))
    {
      if (app->priv->agent_fail_exit)
	record_app_quit (app);
    }
  g_free (hostname);
  g_free (command);
}

/**
 * record_app_create_agent_proxy:
 *
 * create agent proxy object, later we can interactive 
 * with agent server by proxy object.
 */
static void
record_app_create_agent_proxy (RecordApp       *app,
			       GDBusConnection *connection,
			       const gchar     *bus_name)
{
  gint timeout;

  g_return_if_fail (RECORD_IS_APP (app));
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));

  /* set agent proxy default timeout, in milliseconds. */
  timeout = record_conf_get_int (app, "Agent", "timeout");
  timeout = (timeout > 0) ? timeout * 1000 : -1;

  /* create dbus proxy object. */
  record_agent_new (connection, bus_name,
		    "/net/ttyutils/Agent",
		    "net.ttyutils.Agent.RecorderManager",
		    app->priv->agent_cancellable, timeout,
		    on_agent_proxy_created, app);
}

/**
 * on_watch_agent_name_appeared:
 *
 * this function only use for bus-based connection.
 */
static void
on_watch_agent_name_appeared (GDBusConnection *connection,
			      const gchar     *bus_name,
			      const gchar     *name_owned,
			      gpointer         data)
{
  RecordApp *app = data;

  g_return_if_fail (RECORD_IS_APP (data));

  if (!app->priv->agent_proxy)
    record_app_create_agent_proxy (app, connection, bus_name);
}

/**
 * on_watch_agent_name_vanished:
 *
 * this function only use for bus-based connection.
 */
static void
on_watch_agent_name_vanished (GDBusConnection *connection,
			      const gchar     *bus_name,
			      gpointer         data)
{
  RecordApp *app = data;

  g_return_if_fail (RECORD_IS_APP (data));

  /* we cannot quit after login.
   * instead, we watch the bus for appear to auto login again. */
  if (app->priv->agent_proxy)
    {
      g_object_unref (app->priv->agent_proxy);
      app->priv->agent_proxy = NULL;
    }
  else
    {
      /* the agent server doesn't startup as bus service. */
      g_warning (_("Unable to connect agent server."));
      if (app->priv->agent_fail_exit)
	record_app_quit (app);
    }
}

/**
 * on_agent_connection_closed:
 *
 * called when disconnect from agent server.
 * this function only for peer connection.
 */
static void
on_agent_connection_closed (GDBusConnection *connection,
			    gboolean         remote_peer_vanished,
			    GError          *error,
			    RecordApp       *app)
{
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));
  g_return_if_fail (RECORD_IS_APP (app));

  if (remote_peer_vanished)
    g_message ("Agent server has shutdown.");

  if (error)
    g_message ("Disconnect from agent: %s", error->message);

  if (app->priv->agent_proxy)
    {
      g_object_unref (app->priv->agent_proxy);
      app->priv->agent_proxy = NULL;
    }
}

/**
 * record_app_watch_agent_connection:
 *
 * watch connection and try to create agent proxy object.
 */
static void
record_app_watch_agent_connection (RecordApp       *app,
				   GDBusConnection *connection,
				   const gchar     *bus_name)
{
  g_return_if_fail (RECORD_IS_APP (app));
  g_return_if_fail (G_IS_DBUS_CONNECTION (connection));

  g_assert (app->priv->agent_proxy == NULL);

  /* remove resolve name timeout event. */
  if (app->priv->agent_timeout > 0)
    {
      g_source_remove (app->priv->agent_timeout);
      app->priv->agent_timeout = 0;
    }
  /* don't quit application when proxy connection been closed. */
  g_dbus_connection_set_exit_on_close (connection, FALSE);

  if (bus_name == NULL)
    {
      /* peer connection, listen closed signal.
       * we cannot watch peer appear or vanish. */
      g_signal_connect (connection, "closed",
			G_CALLBACK (on_agent_connection_closed), app);
      record_app_create_agent_proxy (app, connection, bus_name);
    }
  else
    {
      /* bus connection. watch it's appeared and vanished event. */
      g_bus_watch_name_on_connection (connection, bus_name,
				      G_BUS_NAME_WATCHER_FLAGS_AUTO_START,
				      on_watch_agent_name_appeared,
				      on_watch_agent_name_vanished, app, NULL);
    }
}

/**
 * on_agent_connection_for_address_finished:
 *
 * called when connect to agent server.
 *
 * this function only use for peer-based connection.
 */
static void
on_agent_connection_for_address_finished (GObject      *source,
					  GAsyncResult *res,
					  gpointer      user_data)
{
  RecordApp *app = RECORD_APP (user_data);
  GDBusConnection *connection;
  GError *error = NULL;

  g_return_if_fail (RECORD_IS_APP (app));

  connection = g_dbus_connection_new_for_address_finish (res, &error);
  if (error)
    {
      gchar *addr = record_conf_get_str (app, "Agent", "server-addr");
      g_warning (_("Unable to connect agent `%s': %s"),
		 addr ? addr : "<NULL>", error->message);
      g_error_free (error);
      g_free (addr);

      if (app->priv->agent_fail_exit)
	record_app_quit (app);

      return;
    }
  record_app_watch_agent_connection (app, connection, NULL);
  g_object_unref (connection);
}

/**
 * on_agent_bus_get_connection_finished:
 *
 * called when connect on dbus daemon, the next step
 * was watch net.ttyutils.Agent.
 *
 * this function only use for bus-based connection.
 */
static void
on_agent_bus_get_connection_finished (GObject      *source,
				      GAsyncResult *res,
				      gpointer      user_data)
{
  RecordApp *app = RECORD_APP (user_data);
  GDBusConnection *connection;
  GError *error = NULL;

  g_return_if_fail (RECORD_IS_APP (app));

  connection = g_bus_get_finish (res, &error);
  if (error)
    {
      g_warning (_("Unable to connect dbus daemon: %s"),
		 error->message);
      g_error_free (error);

      if (app->priv->agent_fail_exit)
	record_app_quit (app);

      return;
    }
  record_app_watch_agent_connection (app, connection,
				     "net.ttyutils.Agent");
  g_object_unref (connection);
}

/**
 * on_agent_observer_authorize_authenticated:
 *
 * authentication with agent.
 */
static gboolean
on_agent_observer_authorize_authenticated (GDBusAuthObserver *observer,
					   GIOStream         *iostream,
					   GCredentials      *credentials,
					   RecordApp       *app)
{
  g_return_val_if_fail (G_IS_DBUS_AUTH_OBSERVER (observer), FALSE);
  g_return_val_if_fail (G_IS_IO_STREAM (iostream), FALSE);
  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  g_debug ("auth: TODO");
  return TRUE;
}

/**
 * on_agent_resolve_name_timeout:
 *
 * called when reolve server host name timeout.
 */
static gboolean
on_agent_resolve_name_timeout (gpointer data)
{
  RecordApp *app = data;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  if (app->priv->agent_cancellable)
    g_cancellable_cancel (app->priv->agent_cancellable);

  return FALSE;
}

/**
 * on_agent_resolve_name_timeout_destroy:
 */
static void
on_agent_resolve_name_timeout_destroy (gpointer data)
{
  RecordApp *app = data;

  g_return_if_fail (RECORD_IS_APP (app));

  app->priv->agent_timeout = 0;
  g_object_unref (app);
}

/**
 * record_app_agent_login:
 *
 * try to login to agent.
 *
 * depend on configure agent address, we will try to
 * contact bus daemon or peer server.
 */
static void
record_app_agent_login (RecordApp *app)
{
  gint agent_type = AGENT_TYPE_PEER;
  gchar *agent_addr;
  GDBusAuthObserver *observer;

  g_return_if_fail (RECORD_IS_APP (app));

  if (app->priv->agent_cancellable)
    g_cancellable_reset (app->priv->agent_cancellable);
  else
    app->priv->agent_cancellable = g_cancellable_new ();

  /* if does not configure server type, use default type PEER. */
  if (record_conf_has (app, "Agent", "server-type"))
    {
      agent_type = record_conf_get_int (app, "Agent", "server-type");
      if (agent_type < AGENT_TYPE_PEER ||
	  agent_type > AGENT_TYPE_BUS_SYSTEM)
	{
	  g_critical ("Value of configure entry `server-type' "
		      "must between %d and %d(included).",
		      AGENT_TYPE_PEER, AGENT_TYPE_BUS_SYSTEM);
	  record_app_quit (app);
	  return;
	}
    }
  /* make connection to agent. */
  switch (agent_type)
    {
    case AGENT_TYPE_PEER:
      if (record_conf_has (app, "Agent", "server-addr"))
	{
	  agent_addr = record_conf_get_str (app, "Agent", "server-addr");
	  if (!agent_addr || !agent_addr[0])
	    {
	      g_critical (_("Agent server address unspecified, please check "
			    "configure file and try again."));
	      g_free (agent_addr);
	      record_app_quit (app);
	      return;
	    }
	}
      else
	{
	  agent_addr = g_strdup ("tcp:host=127.0.0.1,port=5433");
	  g_key_file_set_string (app->priv->keyfile, "Agent",
				 "server-addr", agent_addr);
	}
      /* authentication server. */
      observer = g_dbus_auth_observer_new ();
      g_signal_connect (observer, "authorize-authenticated-peer",
			G_CALLBACK (on_agent_observer_authorize_authenticated), app);

      /* try to create connection to agent. */
      g_dbus_connection_new_for_address (agent_addr,
					 G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT,
					 observer, app->priv->agent_cancellable,
					 on_agent_connection_for_address_finished, app);
      g_free (agent_addr);
      g_object_unref (observer);
      break;

    case AGENT_TYPE_BUS_SESSION:
      g_bus_get (G_BUS_TYPE_SESSION, app->priv->agent_cancellable,
		 on_agent_bus_get_connection_finished, app);
      break;

    case AGENT_TYPE_BUS_SYSTEM:
      g_bus_get (G_BUS_TYPE_SYSTEM, app->priv->agent_cancellable,
		 on_agent_bus_get_connection_finished, app);
      break;

    default:
      g_critical ("unsupported agent address type %d, please check "
		  "configure file and try again.", agent_type);
      record_app_quit (app);
      return;
    }
  if (record_conf_has (app, "Agent", "timeout"))
    {
      gint timeout = record_conf_get_int (app, "Agent", "timeout");
      if (timeout > 0)
	{
	  app->priv->agent_timeout =
	   g_timeout_add_seconds_full (G_PRIORITY_DEFAULT, timeout,
				       on_agent_resolve_name_timeout,
				       g_object_ref (app),
				       on_agent_resolve_name_timeout_destroy);
	}
    }
}

/**
 * on_process_hangup:
 *
 * child process disconnect.
 */
static void
on_process_hangup (RecordProcess *process,
		   RecordApp     *app)
{
  g_return_if_fail (RECORD_IS_APP (app));
  record_app_quit (app);
}

/**
 * on_process_child_start:
 *
 * called when child started.
 */
static void
on_process_child_start (RecordProcess *process,
			GPid           pid,
			RecordApp     *app)
{
  gchar *phrase;

  g_return_if_fail (RECORD_IS_APP (app));

  /* print welcome phrase. */
  if (record_conf_has (app, "Application", "phrase"))
    {
      phrase = record_conf_get_lstr (app, "Application", "phrase");
      if (phrase && phrase[0])
	printf ("%s\r\n", phrase);

      g_free (phrase);
    }
  /* start record session to file. */
  if (app->priv->encoder)
    {
      if (!record_encoder_begin (app->priv->encoder,
				 NULL, NULL, 0, 0, 0.0))
	{
	  record_app_quit (app);
	}
    }
}

/**
 * on_process_child_exit:
 *
 * called when child exited.
 */
static void
on_process_child_exit (RecordProcess *process,
		       GPid           pid,
		       gint           status,
		       RecordApp     *app)
{
  _debug ("Child process %d exit with status %d.", pid, status);

  /* close the record file. */
  if (app->priv->encoder)
    record_encoder_end (app->priv->encoder, 0.0);

  /* don't quit when child process exited beacuse child
   * exit doesn't means all child process's output already
   * been read. for example, if child process cat a large
   * file, it is maybe take ttyrcd a lots of time to read
   * after child exit. instead, we quit when receive a
   * hangup signal(see above)
   */
   // record_app_quit (app);
}

/**
 * on_process_readable:
 *
 * called when data received from child process, that is,
 * child process wanna output data to screen.
 */
static void
on_process_readable (RecordProcess *process,
		     const gchar   *data,
		     guint          size,
		     RecordApp     *app)
{
  g_return_if_fail (RECORD_IS_PROCESS (process));
  g_return_if_fail (RECORD_IS_APP (app));

  t_console_write (app->priv->console, data, size);

  /* send to look client. */
  if (app->priv->lookd)
    record_lookd_write (app->priv->lookd, data, size);

  /* record to file. */
  if (app->priv->encoder)
    record_encoder_inject (app->priv->encoder, data, size);
}

/**
 * on_console_hangup:
 *
 * called when console disconnect(e.g. terminal been closed).
 */
static void
on_console_hangup (TConsole  *console,
		   RecordApp *app)
{
  g_return_if_fail (T_IS_CONSOLE (console));
  g_return_if_fail (RECORD_IS_APP (app));

  record_app_quit (app);
}

/**
 * on_console_readable:
 *
 * called when user press the keyboard.
 */
static void
on_console_readable (TConsole    *console,
		     const gchar *data,
		     guint        size,
		     RecordApp   *app)
{
  g_return_if_fail (T_IS_CONSOLE (console));
  g_return_if_fail (RECORD_IS_APP (app));

  record_process_write (app->priv->process, data, size);
}

/**
 * on_lookd_received:
 *
 * called when data beed received from look client.
 */
static void
on_lookd_received (RecordLookd *lookd,
		   const gchar *data,
		   guint        size,
		   RecordApp   *app)
{
  g_return_if_fail (RECORD_IS_LOOKD (lookd));
  g_return_if_fail (RECORD_IS_APP (app));

  _debug ("receive %d bytes from client.", size);

  /* cannot write to child if writable is disabled */
  if (record_lookd_get_writable (lookd))
    record_process_write (app->priv->process, data, size);
}


typedef struct
{
  const gchar   *name;
  const gchar* (*func) (void);
} RecordFilenameMacro;

/**
 * record_filename_macro_pid_func:
 */
static inline const gchar*
record_filename_macro_pid_func (void)
{
  static gchar buff[32];
  g_snprintf (buff, sizeof (buff), "%d", getpid ());
  return buff;
}

/**
 * record_filename_macro_date_func:
 */
static inline const gchar*
record_filename_macro_date_func (void)
{
  static gchar buff[32];
  time_t tim = time (NULL);
  strftime (buff, sizeof (buff), "%F", localtime (&tim));
  return buff;
}

/**
 * record_filename_macro_time_func:
 */
static inline const gchar*
record_filename_macro_time_func (void)
{
  static gchar buff[32];
  time_t tim = time (NULL);
  strftime (buff, sizeof (buff), "%H-%M-%S", localtime (&tim));
  return buff;
}

/**
 * record_filename_macro_tty_func:
 */
static inline const gchar*
record_filename_macro_tty_func (void)
{
  static gchar buff[32];

  g_strlcpy (buff, ttyname (STDIN_FILENO), sizeof (buff));

  /* substitute path separator(such as `/') with `-'. */
  g_strdelimit (buff, G_DIR_SEPARATOR_S "/\\", '-');
  return buff;
}

/**
 * substitution_pathname_macros:
 */
static gchar*
substitution_pathname_macros (const gchar *string)
{
  GString *result;
  gchar name[128];
  const gchar *p, *p2, *p3;
  RecordFilenameMacro macros[] =
    {
	{ "USER",    g_get_user_name },
	{ "HOME",    g_get_home_dir },
	{ "PID",     record_filename_macro_pid_func },
	{ "DATE",    record_filename_macro_date_func },
	{ "TIME",    record_filename_macro_time_func },
	{ "TTY",     record_filename_macro_tty_func },
    };
  p = string;
  result = g_string_new (NULL);

  while (1)
    {
      if (!(p2 = g_strstr_len (p, -1, "${")))
	break;

      if (!(p3 = g_strstr_len (p2, -1, "}")))
	break;

      g_string_append_len (result, p, p2 - p);

      if (p3 - p2 - 2 == 0)
	{
	  g_string_append_len (result, p2, p3 - p2 + 1);
	}
      else
	{
	  gint i, len = p3 - p2 - 2;
	  g_strlcpy (name, p2 + 2, MIN (len + 1 , sizeof (name)));

	  for (i = 0; i < G_N_ELEMENTS (macros); i++)
	    {
	      if (g_ascii_strncasecmp (p2 + 2, macros[i].name, len) == 0)
		{
		  g_string_append (result, macros[i].func ());
		  break;
		}
	    }
	  if (i == G_N_ELEMENTS (macros))
	    {
	      g_warning (_("Unrecognizable macro substitution for `%s'"), name);
	      g_string_append_len (result, p2, p3 - p2 + 1);
	    }
	}
      p = p3 + 1;
    }
  g_string_append (result, p);

  return g_string_free (result, FALSE);
}

/**
 * record_app_start_lookd:
 *
 * start look service if enabled.
 */
static gboolean
record_app_start_lookd (RecordApp   *app,
			const gchar *ptsname)
{
  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  /* we can disable look service with `enabled' configure entry. */
  if (record_conf_has (app, "Lookd", "enabled"))
    {
      if (!record_conf_get_bool (app, "Lookd", "enabled"))
	return TRUE;
    }
  app->priv->lookd = record_lookd_new ();

  /* apply lookd configure */
  if (record_conf_has (app, "Lookd", "port"))
    {
      gint port = record_conf_get_int (app, "Lookd", "port");
      record_lookd_set_port (app->priv->lookd, port);
    }
  if (record_conf_has (app, "Lookd", "max-conn"))
    {
      gint maxconn = record_conf_get_int (app, "Lookd", "max-conn");
      record_lookd_set_maxconn (app->priv->lookd, maxconn);
    }
  if (record_conf_has (app, "Lookd", "writable"))
    {
      gboolean writable = record_conf_get_bool (app, "Lookd", "writable");
      record_lookd_set_writable (app->priv->lookd, writable);
    }
  g_signal_connect (app->priv->lookd, "received",
		    G_CALLBACK (on_lookd_received), app);

  return record_lookd_start (app->priv->lookd, ptsname);
}

/**
 * record_app_read_passwd:
 */
static gint
record_app_read_passwd (gchar *buff,
			gint   size)
{
  gint n;
  gboolean retry;
  GError *error = NULL;

  while (1)
    {
      n = t_termio_getpass (buff, size, &error);
      g_print ("\n");

      if (error)
	{
	  switch (error->code)
	    {
	    case G_IO_ERROR_NO_SPACE:
	      g_print (_("Password too long, please try again\n"));
	      retry = TRUE;
	      break;
	    case G_IO_ERROR_INVALID_DATA:
	      g_print (_("Bad password character, please try again\n"));
	      retry = TRUE;
	      break;
	    default:
	      g_print ("%s\n", error->message);
	      retry = FALSE;
	      break;
	    }
	  g_error_free (error);
	  error = NULL;

	  if (retry)
	    continue;
	  else
	    return 0;
	}
      break;
    }
  return n;
}

/**
 * record_app_record_file:
 *
 * record to file if enabled.
 */
static gboolean
record_app_record_file (RecordApp *app)
{
  gchar *string, *pathname, *creator = NULL;
  gchar *passwd = NULL;
  gint level = -1, np;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  /* filename missing. don't record. */
  if (!record_conf_has (app, "Record", "pathname"))
    return TRUE;

  /* still missing record filename. */
  string = record_conf_get_str (app, "Record", "pathname");
  if (!string || !string[0])
    {
      g_free (string);
      return FALSE;
    }
  pathname = substitution_pathname_macros (string);
  g_free (string);

  /* get password if encrypt is enabled. */
  if (record_conf_get_bool (app, "Record", "encrypt"))
    {
      passwd = gcry_malloc_secure (512);
      np = record_app_read_passwd (passwd, 512);
      if (np <= 0)
	{
	  g_print (_("No password given, skip encrypt\n"));
	  gcry_free (passwd);
	  passwd = NULL;
	}
    }
  /* compress level */
  if (record_conf_has (app, "Record", "compress-level"))
    level = record_conf_get_int (app, "Record", "compress-level");

  /* creator. */
  if (record_conf_has (app, "Record", "creator"))
    creator = record_conf_get_lstr (app, "Record", "creator");

  app->priv->encoder = record_encoder_new (pathname, level, creator);
  if (passwd)
    {
      record_encoder_set_passwd (app->priv->encoder, passwd, np);
      gcry_free (passwd);
    }
  g_free (pathname);
  g_free (creator);

  return TRUE;
}

/**
 * record_app_start_process:
 *
 * apply child process configure and start it.
 */
static gboolean
record_app_start_process (RecordApp       *app,
			  const gchar    **argv,
			  struct termios  *term_attr)
{
  gboolean login_log = FALSE;
  gchar *workdir = NULL;
  gchar **envs;
  const gchar *charset;

  /* login log(utmp/wtmp). */
  if (record_conf_has (app, "Process", "login-log"))
    login_log = record_conf_get_bool (app, "Process", "login-log");

  record_process_set_login_log (app->priv->process, login_log);

  /* child process work directory. */
  if (record_conf_has (app, "Process", "workdir"))
    workdir = record_conf_get_str (app, "Process", "workdir");

  if (workdir && workdir[0])
    {
      if (!g_file_test (workdir, G_FILE_TEST_IS_DIR))
	{
	  g_critical (_("Configure entry `%s' has invalid value `%s': %s"),
		      "workdir", workdir, _("is not a directory."));
	  return FALSE;
	}
       record_process_set_workdir (app->priv->process, workdir);
    }
  g_free (workdir);

  /* child process environments */
  if (record_conf_has (app, "Process", "environ"))
    {
      gsize length, i;

      envs = record_conf_get_strlist (app, "Process", "environ", &length);
      if (envs)
	{
	  for (i = 0; i < length; i++)
	    {
	      gchar **strv = g_strsplit (envs[i], "=", 2);
	      record_process_env_set (app->priv->process, strv[0], strv[1], TRUE);
	      g_strfreev (strv);
	    }
	  g_strfreev (envs);
	}
    }
  record_process_set_utf8 (app->priv->process, g_get_charset (&charset));
  record_process_set_argv (app->priv->process, argv);

  /* start child process */
  return record_process_start (app->priv->process, term_attr);
}

/**
 * record_app_start:
 *
 * Returns: TRUE if succeed
 */
gboolean
record_app_start (RecordApp       *app,
		  const gchar    **argv,
		  struct termios  *term_attr)
{
  const gchar *ptsname;
  gboolean agent_login = TRUE;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  /* open record file. */
  if (!record_app_record_file (app))
    return FALSE;

  /* redirect log message to file or syslog */
  if (!record_redirect_log ())
    return FALSE;

  /* start listen console I/O event */
  if (!t_console_start (app->priv->console))
    return FALSE;

  /* start child process. */
  if (!record_app_start_process (app, argv, term_attr))
    return FALSE;

  /* start lookd service. */
  ptsname = record_process_get_ptsname (app->priv->process);
  if (!record_app_start_lookd (app, ptsname))
    return FALSE;

  if (record_conf_has (app, "Agent", "fail-exit"))
    {
      app->priv->agent_fail_exit =
       record_conf_get_bool (app, "Agent", "fail-exit");
    }
  /* login to agent if enabled. */
  if (record_conf_has (app, "Agent", "login"))
    agent_login = record_conf_get_bool (app, "Agent", "login");

  if (agent_login)
    record_app_agent_login (app);

  /* take terminal into raw mode. */
  t_termio_makeraw (STDIN_FILENO);
  t_termio_makeraw (STDOUT_FILENO);

  /* running until quit. */
  g_main_loop_run (app->priv->mainloop);
  return TRUE;
}

/**
 * main_loop_quit_idle:
 */
static gboolean
record_app_quit_idle (gpointer data)
{
  RecordApp *app = data;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  /* wait for logout from agent. */
  if (app->priv->agent_proxy &&
      record_agent_get_login (app->priv->agent_proxy))
    return TRUE;

  /* wait until all data been actual transfer */
  return (t_console_get_buffer_size (app->priv->console) > 0);
}

/**
 * record_app_quit_idle_removed:
 */
static void
record_app_quit_idle_removed (gpointer data)
{
  RecordApp *app = data;

  g_return_if_fail (RECORD_IS_APP (app));

  while (g_main_context_pending (NULL))
    g_main_context_iteration (NULL, TRUE);

  if (g_main_loop_is_running (app->priv->mainloop))
    g_main_loop_quit (app->priv->mainloop);
}

/**
 * on_agent_proxy_logout:
 */
static void
on_agent_proxy_logout (GObject      *source,
		       GAsyncResult *res,
		       gpointer      user_data)
{
  RecordAgent *agent = RECORD_AGENT (source);
  RecordApp *app = user_data;
  GError *error = NULL;

  g_return_if_fail (RECORD_IS_AGENT (agent));
  g_return_if_fail (RECORD_IS_APP (app));

  record_agent_logout_finish (agent, res, &error);
  if (error)
    {
      g_dbus_error_strip_remote_error (error);
      g_critical ("Failed to logout from agent: %s", error->message);
      g_error_free (error);
    }
  g_object_unref (app);
}

/**
 * record_app_quit:
 *
 * quit the application.
 */
void
record_app_quit (RecordApp *app)
{
  g_return_if_fail (RECORD_IS_APP (app));

  /* close the process pseudo terminal.
   * ensure child process quit to emit child-exit signal. */
  record_process_stop (app->priv->process);

  t_console_flush (app->priv->console);

  /* logout from agent server. */
  if (app->priv->agent_proxy)
    {
      record_agent_logout (app->priv->agent_proxy,
			   app->priv->agent_cancellable,
			   on_agent_proxy_logout,
			   g_object_ref (app));
    }
  /* wait for script finish */
  if (app->priv->script_thread)
    {
      g_message (_("Wait for script finish."));
      g_thread_join (app->priv->script_thread);
      app->priv->script_thread = NULL;
    }
  g_idle_add_full (G_PRIORITY_LOW, record_app_quit_idle, app,
		   record_app_quit_idle_removed);
}

/**
 * script_split_argv:
 *
 * differ to g_strsplit(), which treat continuous blank
 * as element.
 */
static inline gchar**
script_split_argv (const gchar *args)
{
  gchar buff[PATH_MAX];
  gint i, j, blank = 0;

  for (i = 0, j = 0; i < strlen (args); i++)
    {
      if (g_ascii_isspace (args[i]))
	{
	  if (!blank)
	    {
	      buff[j++] = ' ';
	      blank = 1;
	      continue;
	    }
	}
      else
	{
	  buff[j++] = args[i];
	  blank = 0;
	}
    }
  return g_strsplit (buff, " ", -1);
}

/**
 * script_thread_func:
 *
 * a independent thread to execute script file.
 * execute more than one scripts concurrent is not allowed.
 */
static gpointer
script_thread_func (gpointer data)
{
  RecordApp *app = data;
  gboolean *ret;

  g_return_val_if_fail (RECORD_IS_APP (app), NULL);

  ret = g_malloc (sizeof (gboolean*));
  *ret = TRUE;

  while (app->priv->script_files)
    {
      gchar *script, **argv;

      g_static_mutex_lock (&_script_mutex);

      /* take the first file from list */
      script = app->priv->script_files->data;
      argv = script_split_argv (script);

      /* remove the file from list */
      app->priv->script_files = g_list_remove (app->priv->script_files,
					       script);
      g_free (script);

      g_static_mutex_unlock (&_script_mutex);

      /* execute script */
      *ret = record_script_run (app, (const gchar**) argv);
      g_strfreev (argv);

      if (*ret == FALSE)
	break;
    }
  g_async_queue_push (app->priv->script_queue, ret);
  return NULL;
}

/**
 * on_script_thread_timeout:
 */
static gboolean
on_script_thread_timeout (gpointer data)
{
  RecordApp *app = data;
  gboolean *ret;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  /* try to get the execute result */
  ret = g_async_queue_try_pop (app->priv->script_queue);
  if (ret == NULL)
    return TRUE;

  /* quit application if script failed */
  if (*ret == FALSE)
    record_app_quit (app);

  g_free (ret);
  return FALSE;
}

/**
 * on_script_thread_timeout_removed:
 */
static void
on_script_thread_timeout_removed (gpointer data)
{
  RecordApp *app = data;

  g_return_if_fail (RECORD_IS_APP (app));

  if (app->priv->script_thread)
    {
      g_thread_join (app->priv->script_thread);
      app->priv->script_thread = NULL;
    }
  g_async_queue_unref (app->priv->script_queue);
  app->priv->script_queue = NULL;
}

/**
 * record_app_run_script:
 *
 * execute script file.
 *
 * this function will create a thread to execute script.
 */
gboolean
record_app_run_script (RecordApp   *app,
		       const gchar *filename)
{
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (filename, FALSE);

  g_static_mutex_lock (&_script_mutex);

  /* append to script file queue */
  app->priv->script_files = g_list_append (app->priv->script_files,
					   g_strdup (filename));
  /* start a new thread to run script.
   * we cannot run script in main thread beacuse it will block
   * main event loop. */
  if (!app->priv->script_thread)
    {
      if (!app->priv->script_queue)
	app->priv->script_queue = g_async_queue_new_full (g_free);

      app->priv->script_thread = g_thread_create (script_thread_func,
						  app, TRUE, &error);
      if (error)
	{
	  g_warning (_("Failed to create thread: %s"), error->message);
	  g_error_free (error);
	  g_static_mutex_unlock (&_script_mutex);
	  return FALSE;
	}
      /* check whether script finished every 1/5 second. */
      g_timeout_add_full (G_PRIORITY_LOW, 200,
			  on_script_thread_timeout, app,
			  on_script_thread_timeout_removed);
    }
  g_static_mutex_unlock (&_script_mutex);

  return TRUE;
}
