/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <zip.h>
#include <gcrypt.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <ttyutils/ttyutils.h>

#include "record-encoder.h"

enum
{
  PROP_0,
  PROP_FILENAME,
  PROP_COMP_LEVEL,
  PROP_CREATOR,
};

struct _RecordEncoderPriv
{
  TRecordHead    head;
  struct zip    *archive;
  gint           comp_level;
  gchar         *filename;
  gchar         *creator;
  GOutputStream *idx_stream;
  GOutputStream *dat_stream;
  guint32        nindexes;
  guint64        ndata;
  guint32        nfiles;
  guint32        fnindex;
  guint64        fndata;
  gchar         *passwd;
  GTimer        *timer;
  gdouble        elapsed;
};
#define _ENCODER_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), RECORD_TYPE_ENCODER, RecordEncoderPriv))

G_DEFINE_TYPE (RecordEncoder, record_encoder, G_TYPE_OBJECT);

static void	record_encoder_set_comp_level	(RecordEncoder *encoder,
						 gint           comp_level);
static void	record_encoder_set_filename	(RecordEncoder *encoder,
						 const gchar   *filename);
static void	record_encoder_set_creator	(RecordEncoder *encoder,
						 const gchar   *creator);

/**
 * record_encoder_init:
 */
static void
record_encoder_init (RecordEncoder *encoder)
{
  encoder->priv = _ENCODER_GET_PRIVATE (encoder);

  encoder->priv->comp_level = -1;
  encoder->priv->filename = NULL;
  encoder->priv->creator = NULL;
  encoder->priv->archive = NULL;
  encoder->priv->idx_stream = NULL;
  encoder->priv->dat_stream = NULL;
  encoder->priv->nindexes = 0;
  encoder->priv->ndata = 0;
  encoder->priv->nfiles = 0;
  encoder->priv->fnindex = 0;
  encoder->priv->fndata = 0;
  encoder->priv->passwd = NULL;
  encoder->priv->timer = g_timer_new ();
  encoder->priv->elapsed = 0.0;

  memset (&encoder->priv->head, 0, sizeof (encoder->priv->head));
}

/**
 * record_encoder_dispose:
 */
static void
record_encoder_dispose (GObject *object)
{
  //RecordEncoder *encoder = RECORD_ENCODER (object);

  G_OBJECT_CLASS (record_encoder_parent_class)->dispose (object);
}

/**
 * record_encoder_finalize:
 */
static void
record_encoder_finalize (GObject *object)
{
  RecordEncoder *encoder = RECORD_ENCODER (object);

  if (encoder->priv->timer)
    {
      g_timer_destroy (encoder->priv->timer);
      encoder->priv->timer = NULL;
    }
  if (encoder->priv->passwd)
    {
      g_free (encoder->priv->passwd);
      encoder->priv->passwd = NULL;
    }
  if (encoder->priv->filename)
    {
      g_free (encoder->priv->filename);
      encoder->priv->filename = NULL;
    }
  if (encoder->priv->creator)
    {
      g_free (encoder->priv->creator);
      encoder->priv->creator = NULL;
    }
  G_OBJECT_CLASS (record_encoder_parent_class)->finalize (object);
}

/**
 * record_encoder_get_property:
 */
static void
record_encoder_get_property (GObject    *object,
			     guint       property_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  RecordEncoder *encoder = RECORD_ENCODER (object);

  switch (property_id)
    {
    case PROP_FILENAME:
      g_value_set_string (value, record_encoder_get_filename (encoder));
      break;
    case PROP_CREATOR:
      g_value_set_string (value, record_encoder_get_creator (encoder));
      break;
    case PROP_COMP_LEVEL:
      g_value_set_int (value, record_encoder_get_comp_level (encoder));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_encoder_set_property:
 */
static void
record_encoder_set_property (GObject      *object,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  RecordEncoder *encoder = RECORD_ENCODER (object);

  switch (property_id)
    {
    case PROP_FILENAME:
      record_encoder_set_filename (encoder, g_value_get_string (value));
      break;
    case PROP_CREATOR:
      record_encoder_set_creator (encoder, g_value_get_string (value));
      break;
    case PROP_COMP_LEVEL:
      record_encoder_set_comp_level (encoder, g_value_get_int (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_encoder_class_init:
 */
void
record_encoder_class_init (RecordEncoderClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (RecordEncoderPriv));

  gobject_class->get_property = record_encoder_get_property;
  gobject_class->set_property = record_encoder_set_property;
  gobject_class->dispose      = record_encoder_dispose;
  gobject_class->finalize     = record_encoder_finalize;

  g_object_class_install_property (gobject_class, PROP_FILENAME,
				   g_param_spec_string ("filename", "Filename", "Output filename", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB |
							G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
  g_object_class_install_property (gobject_class, PROP_COMP_LEVEL,
				   g_param_spec_int ("compress-level", "compress-level", "Compress level", -1, 9, -1,
						     G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB |
						     G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
  g_object_class_install_property (gobject_class, PROP_CREATOR,
				   g_param_spec_string ("creator", "Creator", "Creator", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB |
							G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
  if (gcry_md_test_algo (GCRY_MD_MD5) != 0)
    {
      g_error ("digest algorithm `%s' is not usable.",
	       gcry_md_algo_name (GCRY_MD_MD5));
      return;
    }
  if (gcry_md_test_algo (GCRY_MD_TIGER) != 0)
    {
      g_error ("digest algorithm `%s' is not usable.",
	       gcry_md_algo_name (GCRY_MD_TIGER));
      return;
    }
  g_assert (gcry_md_get_algo_dlen (GCRY_MD_MD5)   == 16);
  g_assert (gcry_md_get_algo_dlen (GCRY_MD_TIGER) == 24);
}

/**
 * record_encoder_new:
 */
RecordEncoder*
record_encoder_new (const gchar *pathname,
		    gint         comp_level,
		    const gchar *creator)
{
  g_return_val_if_fail (pathname, NULL);

  return g_object_new (RECORD_TYPE_ENCODER,
		       "compress-level", comp_level,
		       "creator", creator,
		       "filename", pathname, NULL);
}

/**
 * record_encoder_set_passwd:
 */
void
record_encoder_set_passwd (RecordEncoder *encoder,
			   const gchar   *passwd,
			   gssize         length)
{
  g_return_if_fail (RECORD_IS_ENCODER (encoder));
  g_return_if_fail (passwd);

  if (length < 0)
    length = strlen (passwd);

  /* we cannot accept 0-length password. */
  if (length == 0)
    return;

  g_free (encoder->priv->passwd);
  encoder->priv->passwd = g_malloc (16);

  /* we use digest of password in order to do actually
   * encrypt(not password itself). */
  gcry_md_hash_buffer (GCRY_MD_MD5, encoder->priv->passwd,
		       passwd, length);
  gcry_md_hash_buffer (GCRY_MD_TIGER, &encoder->priv->head.passwd,
		       passwd, length);
  encoder->priv->head.encrypted = 1;
}

/**
 * record_encoder_get_comp_level:
 */
gint
record_encoder_get_comp_level (RecordEncoder *encoder)
{
  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), 0);
  return encoder->priv->comp_level;
}

/**
 * record_encoder_set_comp_level:
 */
static void
record_encoder_set_comp_level (RecordEncoder *encoder,
			       gint           comp_level)
{
  g_return_if_fail (RECORD_IS_ENCODER (encoder));

  encoder->priv->comp_level = CLAMP (comp_level, -1, 9);
  g_object_notify (G_OBJECT (encoder), "compress-level");
}

/**
 * record_encoder_get_creator:
 */
const gchar*
record_encoder_get_creator (RecordEncoder *encoder)
{
  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), 0);
  return encoder->priv->creator;
}

/**
 * record_encoder_set_creator:
 *
 * open the output file.
 */
static void
record_encoder_set_creator (RecordEncoder *encoder,
			    const gchar   *creator)
{
  g_return_if_fail (RECORD_IS_ENCODER (encoder));

  g_free (encoder->priv->creator);
  if (creator)
    {
      encoder->priv->creator = g_strdup (creator);
    }
  else
    {
      encoder->priv->creator =
       g_strdup_printf ("%s %s", g_get_application_name (), VERSION);
    }
  g_object_notify (G_OBJECT (encoder), "creator");
}

/**
 * record_encoder_get_filename:
 */
const gchar*
record_encoder_get_filename (RecordEncoder *encoder)
{
  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), 0);
  return encoder->priv->filename;
}

/**
 * record_encoder_set_filename:
 *
 * open the output file.
 *
 * this function should be call only once when object construct.
 */
static void
record_encoder_set_filename (RecordEncoder *encoder,
			     const gchar   *filename)
{
  g_return_if_fail (RECORD_IS_ENCODER (encoder));
  g_return_if_fail (filename);

  g_free (encoder->priv->filename);
  encoder->priv->filename = g_strdup (filename);

  g_object_notify (G_OBJECT (encoder), "filename");
}

/**
 * record_encoder_write_head:
 *
 * write head to file without compress.
 */
static gboolean
record_encoder_write_head (RecordEncoder *encoder)
{
  struct zip_source *source;
  gint i;

  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);
  g_return_val_if_fail (encoder->priv->archive, FALSE);

  source = zip_source_buffer (encoder->priv->archive,
			      &encoder->priv->head, sizeof (TRecordHead), 0);
  if (!source)
    {
      g_warning ("zip_source_buffer() error: %s",
		 zip_strerror (encoder->priv->archive));
      return FALSE;
    }
  i = zip_name_locate (encoder->priv->archive, "HEAD", 0);
  if (i < 0)
    i = zip_add (encoder->priv->archive, "HEAD", source);
  else
    i = zip_replace (encoder->priv->archive, i, source);

  if (i == -1)
    {
      g_warning ("zip_add/replace() error: %s",
		 zip_strerror (encoder->priv->archive));
      zip_source_free (source);
      return FALSE;
    }
  zip_set_file_comment (encoder->priv->archive, i, "RecordHead",
			strlen ("RecordHead"));
  return TRUE;
}

/**
 * record_encoder_create_data_stream:
 */
static inline GOutputStream*
record_encoder_create_data_stream (RecordEncoder *encoder)
{
  GOutputStream *base_stream;
  GZlibCompressor *converter;
  GOutputStream *stream;

  /* create memory output stream. */
  base_stream = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);

  /* althought we will add the data to zip archive(them will
   * got compress), but we still compress the data before add
   * to zip archive, beacuse we need encrypt the compress data
   * (for better compressibility). */
  converter = g_zlib_compressor_new (G_ZLIB_COMPRESSOR_FORMAT_RAW,
				     encoder->priv->comp_level);

  /* create compress stream from base stream. */
  stream = g_converter_output_stream_new (G_OUTPUT_STREAM (base_stream),
					  G_CONVERTER (converter));

  /* we now can safely unref the objects since the stream
   * hold their reference.
   * we can still obtain these object from stream. */
  g_object_unref (base_stream);
  g_object_unref (converter);

  return stream;
}

/**
 * record_encoder_open:
 */
static gboolean
record_encoder_open (RecordEncoder *encoder)
{
  gint zerror;
  gchar comment[256];

  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);

  g_assert (encoder->priv->filename);
  g_assert (!encoder->priv->archive);
  g_assert (!encoder->priv->idx_stream);
  g_assert (!encoder->priv->dat_stream);

  /* create zip archive. */
  encoder->priv->archive = zip_open (encoder->priv->filename,
				     ZIP_CREATE | ZIP_EXCL, &zerror);
  if (!encoder->priv->archive)
    {
      gchar errbuf[256];
      zip_error_to_str (errbuf, sizeof (errbuf), zerror, errno);
      g_critical (_("Unable output to file `%s': %s"),
		  encoder->priv->filename, errbuf);
      return FALSE;
    }
  g_snprintf (comment, sizeof (comment), "Terminal session record by %s %s",
	      g_get_application_name (), VERSION);
  zip_set_archive_comment (encoder->priv->archive, comment, strlen (comment));

  /* create index and data stream. */
  encoder->priv->idx_stream = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
  encoder->priv->dat_stream = record_encoder_create_data_stream (encoder);

  g_assert (g_seekable_can_truncate (G_SEEKABLE (encoder->priv->idx_stream)));
  g_assert (g_seekable_can_seek (G_SEEKABLE (encoder->priv->idx_stream)));

  return TRUE;
}

/**
 * record_encoder_begin:
 *
 * start record file.
 */
gboolean
record_encoder_begin (RecordEncoder *encoder,
		      const gchar   *charset,
		      const gchar   *term,
		      guint16        termrows,
		      guint16        termcols,
		      gdouble        starting)
{
  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);

  /* open archive and stream for write. */
  if (!record_encoder_open (encoder))
    return FALSE;

  /* fill the record head field. */
  encoder->priv->head.version = T_RECORD_VERSION;
  encoder->priv->head.clevel  = encoder->priv->comp_level;

  encoder->priv->head.starting = (starting == 0.0) ?
                                 t_get_current_time () : starting;

  if (!charset) g_get_charset (&charset);
  g_strlcpy (encoder->priv->head.charset, charset,
	     sizeof (encoder->priv->head.charset));

  if (!term) term = g_getenv ("TERM");
  g_strlcpy (encoder->priv->head.term, term ? term : "dumb",
	     sizeof (encoder->priv->head.term));

  if (termrows == 0 || termcols == 0)
    t_termio_get_winsize (STDOUT_FILENO, &termrows, &termcols);

  encoder->priv->head.ncols = GUINT16_TO_LE (termcols);
  encoder->priv->head.nrows = GUINT16_TO_LE (termrows);

  if (encoder->priv->creator)
    {
      g_strlcpy (encoder->priv->head.creator,
		 encoder->priv->creator,
		 sizeof (encoder->priv->head.creator));
    }
  /* reset timer. */
  g_timer_start (encoder->priv->timer);

  return record_encoder_write_head (encoder);
}

/**
 * record_encoder_archive_add_index:
 *
 * flush index stream into archive file.
 * after flush, the stream will be truncate and seek to the begin.
 */
static gboolean
record_encoder_archive_add_index (RecordEncoder *encoder)
{
  GMemoryOutputStream *memory;
  gchar *data, name[128];
  gsize size;
  struct zip_source *source;
  gpointer buffer;
  gint i;

  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);
  g_return_val_if_fail (encoder->priv->archive, FALSE);

  /* flush the stream. */
  g_output_stream_flush (encoder->priv->idx_stream, NULL, NULL);

  memory = G_MEMORY_OUTPUT_STREAM (encoder->priv->idx_stream);
  data = g_memory_output_stream_get_data (memory);
  size = g_memory_output_stream_get_data_size (memory);

  /* alloc buffer as zip source and add into archive as file.
   * we should use malloc() instead g_malloc() since the zip library use
   * free() to free the data. */
  buffer = malloc (size);
  memcpy (buffer, data, size);
  source = zip_source_buffer (encoder->priv->archive, buffer, size, 1);

  g_snprintf (name, sizeof (name), "indexes/%d", encoder->priv->nfiles);
  i = zip_add (encoder->priv->archive, name, source);
  if (i == -1)
    {
      g_warning ("zip_add() `%s' error: %s", name,
		 zip_strerror (encoder->priv->archive));
      zip_source_free (source);
      return FALSE;
    }
  /* set file comment. */
  g_snprintf (name, sizeof (name), "%u:%u,%f",
	      encoder->priv->fnindex, encoder->priv->nindexes,
	      encoder->priv->elapsed);
  zip_set_file_comment (encoder->priv->archive, i, name, strlen (name));

  /* update file index offset. */
  encoder->priv->fnindex = encoder->priv->nindexes;

  /* truncate and reset the index stream. */
  g_seekable_truncate (G_SEEKABLE (memory), 0, NULL, NULL);
  return g_seekable_seek (G_SEEKABLE (memory), 0, G_SEEK_SET, NULL, NULL);
}

/**
 * record_encoder_encrypt_data:
 */
static inline gboolean
record_encoder_encrypt_data (RecordEncoder *encoder,
			     const gchar   *in,
			     gsize          insize,
			     gchar         *out,
			     gsize          outsize)
{
  gcry_cipher_hd_t cipher;
  gcry_error_t ge = 0;

  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);
  g_return_val_if_fail (in, FALSE);
  g_return_val_if_fail (out, FALSE);

  ge = gcry_cipher_open (&cipher, GCRY_CIPHER_BLOWFISH,
			 GCRY_CIPHER_MODE_CBC, GCRY_CIPHER_CBC_CTS);
  if (ge != 0)
    {
      g_warning ("Encrypt failed: %s/%s", gcry_strsource (ge),
		 gcry_strerror (ge));
      return FALSE;
    }
  gcry_cipher_setkey (cipher, encoder->priv->passwd, 16);

  ge = gcry_cipher_encrypt (cipher, out, outsize, in, insize);
  if (ge != 0)
    {
      g_warning ("Encrypt failed: %s/%s", gcry_strsource (ge),
		 gcry_strerror (ge));
      gcry_cipher_close (cipher);
      return FALSE;
    }
  gcry_cipher_close (cipher);

  return TRUE;
}

/**
 * record_encoder_archive_add_data:
 */
static gboolean
record_encoder_archive_add_data (RecordEncoder *encoder)
{
  GMemoryOutputStream *memory;
  GConverter *converter;
  gchar *data, name[128];
  gsize size;
  struct zip_source *source;
  gpointer buffer;
  gint i;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);

  /* flush the data stream(from zip stream to memory stream). */
  g_output_stream_flush (encoder->priv->dat_stream, NULL, &error);
  if (error)
    {
      _warning ("flush data stream error: %s", error->message);
      g_error_free (error);
      error = NULL;
    }
  g_object_get (encoder->priv->dat_stream, "base-stream", &memory, NULL);
  g_object_unref (memory);

  data = g_memory_output_stream_get_data (memory);
  size = g_memory_output_stream_get_data_size (memory);

  if (!encoder->priv->passwd)
    {
      buffer = malloc (size);
      memcpy (buffer, data, size);

      source = zip_source_buffer (encoder->priv->archive,
				  buffer, size, 1);
    }
  else
    {
      gsize block_size, encrypt_size = size;

      /* get clpher block size and padding. */
      gcry_cipher_algo_info (GCRY_CIPHER_BLOWFISH, GCRYCTL_GET_BLKLEN,
			     NULL, &block_size);
      if (size % block_size > 0)
	encrypt_size += block_size - size % block_size;

      /* ensure input data has enough space(padding). */
      while (encrypt_size > g_memory_output_stream_get_size (memory))
	{
	  g_output_stream_write (G_OUTPUT_STREAM (memory),
				 "\0", 1, NULL, NULL);
	}
      buffer = calloc (encrypt_size, 1);
      if (!record_encoder_encrypt_data (encoder, data, encrypt_size,
					buffer, encrypt_size))
	{
	  g_free (buffer);
	  return FALSE;
	}
      source = zip_source_buffer (encoder->priv->archive, buffer,
				  encrypt_size, 1);
    }
  /* add data as a file into archive. */
  g_snprintf (name, sizeof (name), "data/%d", encoder->priv->nfiles);
  i = zip_add (encoder->priv->archive, name, source);
  if (i == -1)
    {
      g_warning ("zip_add() `%s' error: %s", name,
		 zip_strerror (encoder->priv->archive));
      zip_source_free (source);
      return FALSE;
    }
  /* set file comment. */
  g_snprintf (name, sizeof (name), "%lld:%lld,%d",
	      encoder->priv->fndata, encoder->priv->ndata, size);
  zip_set_file_comment (encoder->priv->archive, i, name, strlen (name));

  /* update file index offset. */
  encoder->priv->fndata = encoder->priv->ndata;

  /* reset the zip compressor. */
  g_object_get (encoder->priv->dat_stream, "converter", &converter, NULL);
  g_converter_reset (converter);
  g_object_unref (converter);

  /* truncate memory stream. */
  g_seekable_truncate (G_SEEKABLE (memory), 0, NULL, NULL);
  return g_seekable_seek (G_SEEKABLE (memory), 0, G_SEEK_SET, NULL, NULL);
}

/**
 * record_encoder_archive_add:
 */
static gboolean
record_encoder_archive_add (RecordEncoder *encoder)
{
  gboolean ret = TRUE;

  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);
  g_return_val_if_fail (encoder->priv->archive, FALSE);

  if (!record_encoder_archive_add_index (encoder))
    ret = FALSE;

  if (!record_encoder_archive_add_data (encoder))
    ret = FALSE;

  encoder->priv->nfiles++;
  return ret;
}

/**
 * record_encoder_close:
 */
static void
record_encoder_close (RecordEncoder *encoder,
		      gdouble        duration)
{
  g_return_if_fail (RECORD_IS_ENCODER (encoder));

  /* flush remains data to archive. */
  if (encoder->priv->nindexes > encoder->priv->fnindex)
    record_encoder_archive_add (encoder);

  /* fill head field and write to file. */
  encoder->priv->head.duration = (duration > 0.0) ? duration :
                                 t_get_current_time () - encoder->priv->head.starting;
  encoder->priv->head.nindexes = GUINT32_TO_LE (encoder->priv->nindexes);
  encoder->priv->head.ndata    = GUINT64_TO_LE (encoder->priv->ndata);
  encoder->priv->head.nfiles   = GUINT32_TO_LE (encoder->priv->nfiles);

  record_encoder_write_head (encoder);

  /* close and free the stream. */
  g_object_unref (encoder->priv->idx_stream);
  encoder->priv->idx_stream = NULL;
  g_object_unref (encoder->priv->dat_stream);
  encoder->priv->dat_stream = NULL;

  /* close the archive. */
  if (zip_close (encoder->priv->archive) != 0)
    {
      _warning ("close archive error: %s",
		zip_strerror (encoder->priv->archive));
    }
  encoder->priv->archive = NULL;
}

/**
 * record_encoder_end:
 */
void
record_encoder_end (RecordEncoder *encoder,
		    gdouble        duration)
{
  g_return_if_fail (RECORD_IS_ENCODER (encoder));

  /* flush all pending data to archive and close it. */
  if (encoder->priv->archive)
    {
      record_encoder_close (encoder, duration);
    }
}

/**
 * record_encoder_inject_with_time:
 *
 * like record_encoder_inject(), but use specified interval time.
 * for convert tool use only.
 */
void
record_encoder_inject_with_time (RecordEncoder *encoder,
				 const gchar   *data,
				 gssize         size,
				 gdouble        interval)
{
  gchar buff[64];
  gsize data_size;
  GError *error = NULL;

  g_return_if_fail (RECORD_IS_ENCODER (encoder));
  g_return_if_fail (data);

  if (size < 0)
    size = strlen (data);

  /* size must can be store in a uint16 variable. */
  g_assert (size < G_MAXUINT16);

  /* write index to stream. */
  g_snprintf (buff, sizeof (buff), "%f,%d;", interval, size);
  g_output_stream_write (encoder->priv->idx_stream,
			 buff, strlen (buff), NULL, &error);
  if (error)
    {
      _critical ("Failed to write index stream: %s", error->message);
      g_error_free (error);
      error = NULL;
    }
  encoder->priv->nindexes++;
  encoder->priv->elapsed += interval;

  /* write data to stream. */
  g_output_stream_write (encoder->priv->dat_stream, data, size, NULL, &error);
  if (error)
    {
      _critical ("Failed to write data stream: %s", error->message);
      g_error_free (error);
      error = NULL;
    }
  encoder->priv->ndata += size;

  /* flush the stream if index stream too long.
   * this parameter is important since it decide the memory
   * require of encoder and decoder.
   * 100K index need ~1.0M memory to encode and deocde. */
  g_object_get (encoder->priv->idx_stream, "data-size", &data_size, NULL);
  if (data_size > 1024 * 100)
    {
      record_encoder_archive_add (encoder);
      encoder->priv->elapsed = 0.0;
      return;
    }
}

/**
 * record_encoder_inject:
 */
void
record_encoder_inject (RecordEncoder *encoder,
		       const gchar   *data,
		       gssize         size)
{
  gdouble interval;

  g_return_if_fail (RECORD_IS_ENCODER (encoder));
  g_return_if_fail (data);

  /* get interval since last inject and reset timer. */
  interval = g_timer_elapsed (encoder->priv->timer, NULL);
  g_timer_start (encoder->priv->timer);

  record_encoder_inject_with_time (encoder, data, size, interval);
}

/**
 * record_encoder_inject_winsize:
 */
void
record_encoder_inject_winsize (RecordEncoder *encoder,
			       guint16        rows,
			       guint16        cols)
{
  gchar seq[32];

  g_return_if_fail (RECORD_IS_ENCODER (encoder));

  g_snprintf (seq, sizeof (seq), "\033[8;%d;%dt", rows, cols);
  record_encoder_inject (encoder, seq, -1);
}
