/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define _GNU_SOURCE

#include <sys/wait.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <pwd.h>
#include <signal.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#ifdef HAVE_PTY_H
# include <pty.h>
#endif
#ifdef HAVE_RESOURCE_H
# include <sys/resource.h>
#endif

#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <ttyutils/ttyutils.h>

#include "record-utmp.h"
#include "record-process.h"

enum
{
  PROP_0,
  PROP_WORKDIR,
  PROP_LOGIN_LOG,
};

struct _RecordProcessPriv
{
  gchar        *workdir;
  gboolean      login_log;
  gchar       **argv;
  GList        *env_list;
  GPid          child_pid;
  gint          master;
  gchar        *ptsname;
  GIOChannel   *channel;
  gint          read_watch_id;
  gint          write_watch_id;
  TRing        *write_ring;
  GMutex       *write_ring_mutex;
  guint         workdir_watch_id;
};
#define _PROCESS_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), RECORD_TYPE_PROCESS, RecordProcessPriv))

G_DEFINE_TYPE (RecordProcess, record_process, G_TYPE_OBJECT);

#ifndef HAVE_FDWALK
static int	fdwalk		(int (*cb) (void *data, int fd), void *data);
#endif
static int	set_cloexec	(void *data, gint fd);

/**
 * record_process_init:
 */
static void
record_process_init (RecordProcess *process)
{
  process->priv = _PROCESS_GET_PRIVATE (process);

  process->priv->workdir = NULL;
  process->priv->login_log = FALSE;
  process->priv->argv = NULL;
  process->priv->env_list = NULL;
  process->priv->child_pid = 0;
  process->priv->master = -1;
  process->priv->ptsname = NULL;
  process->priv->channel = NULL;
  process->priv->read_watch_id = -1;
  process->priv->write_watch_id = -1;
  process->priv->write_ring = t_ring_new ();
  process->priv->write_ring_mutex = g_mutex_new ();
  process->priv->workdir_watch_id = 0;
}

/**
 * record_process_env_free:
 */
static void
record_process_env_free (RecordProcessEnv *env)
{
  if (env)
    {
      g_free (env->name);
      g_free (env->value);
      g_free (env);
    }
}

/**
 * record_process_finalize:
 */
static void
record_process_finalize (GObject *object)
{
  RecordProcess *process = RECORD_PROCESS (object);

  if (process->priv->workdir_watch_id > 0)
    g_source_remove (process->priv->workdir_watch_id);

  if (process->priv->read_watch_id > 0)
    g_source_remove (process->priv->read_watch_id);

  if (process->priv->write_watch_id > 0)
    g_source_remove (process->priv->write_watch_id);

  if (process->priv->channel)
    {
      g_io_channel_unref (process->priv->channel);
      process->priv->channel = NULL;
    }
  if (process->priv->workdir)
    process->priv->workdir = (g_free (process->priv->workdir), NULL);

  if (process->priv->argv)
    process->priv->argv = (g_strfreev (process->priv->argv), NULL);

  if (process->priv->ptsname)
    process->priv->ptsname = (g_free (process->priv->ptsname), NULL);

  if (process->priv->env_list)
    {
      g_list_foreach (process->priv->env_list,
		      (GFunc) record_process_env_free, NULL);
      g_list_free (process->priv->env_list);
    }
  if (process->priv->write_ring)
    {
      t_ring_free (process->priv->write_ring);
      process->priv->write_ring = NULL;
    }
  if (process->priv->write_ring_mutex)
    {
      g_mutex_free (process->priv->write_ring_mutex);
      process->priv->write_ring_mutex = NULL;
    }
  G_OBJECT_CLASS (record_process_parent_class)->finalize (object);
}

/**
 * record_process_get_property:
 */
static void
record_process_get_property (GObject    *object,
			     guint       property_id,
			     GValue     *value,
			     GParamSpec *pspec)
{
  RecordProcess *process = RECORD_PROCESS (object);

  switch (property_id)
    {
    case PROP_WORKDIR:
      g_value_set_string (value, record_process_get_workdir (process));
      break;
    case PROP_LOGIN_LOG:
      g_value_set_boolean (value, record_process_get_login_log (process));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_process_set_property:
 */
static void
record_process_set_property (GObject      *object,
			     guint         property_id,
			     const GValue *value,
			     GParamSpec   *pspec)
{
  RecordProcess *process = RECORD_PROCESS (object);

  switch (property_id)
    {
    case PROP_WORKDIR:
      record_process_set_workdir (process, g_value_get_string (value));
      break;
    case PROP_LOGIN_LOG:
      record_process_set_login_log (process, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * record_process_hangup_real:
 */
static void
record_process_hangup_real (RecordProcess *process)
{
  if (process->priv->read_watch_id > 0)
    g_source_remove (process->priv->read_watch_id);

  if (process->priv->write_watch_id > 0)
    g_source_remove (process->priv->write_watch_id);

  if (process->priv->channel)
    {
      g_io_channel_unref (process->priv->channel);
      process->priv->channel = NULL;
    }
  close (process->priv->master);
  process->priv->master = -1;
}

/**
 * record_process_child_exit_real:
 */
static void
record_process_child_exit_real (RecordProcess *process,
				GPid           pid,
				gint           status)
{
  if (status != 0)
    {
      gchar reason[256];

      if (WIFEXITED (status))
	{
	  g_snprintf (reason, sizeof (reason),
		      _("exit() with code %d"), WEXITSTATUS (status));
	}
      else if (WIFSIGNALED (status))
	{
	  g_snprintf (reason, sizeof (reason),
		      _("terminate by unprocessed signal %d(%s)"),
		      WTERMSIG (status),
		      g_strsignal (WTERMSIG (status)));
	}
      else if (WCOREDUMP (status))
	{
	  g_strlcpy (reason, "core dump", sizeof (reason));
	}
      else if (WIFSTOPPED (status))
	{
	  g_snprintf (reason, sizeof (reason),
		      _("terminate by signal %d(%s)"),
		      WSTOPSIG (status),
		      g_strsignal (WSTOPSIG (status)));
	}
      g_message (_("Child process %d exit with status %d beacuse process %s"),
		 pid, status, reason);
    }
  process->priv->child_pid = 0;
}

/**
 * record_process_class_init:
 */
static void
record_process_class_init (RecordProcessClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (RecordProcessPriv));

  gobject_class->set_property = record_process_set_property;
  gobject_class->get_property = record_process_get_property;
  gobject_class->finalize = record_process_finalize;

  klass->hangup = record_process_hangup_real;
  klass->child_exit = record_process_child_exit_real;

  g_object_class_install_property (gobject_class, PROP_WORKDIR,
				   g_param_spec_string ("workdir", "workdir", "Work Directory", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB |
							G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_LOGIN_LOG,
				   g_param_spec_boolean ("login-log", "login-log", "Log utmp, wtmp, lastlog", FALSE,
							 G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB |
							 G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_signal_new ("readable",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		t_marshal_VOID__STRING_UINT,
		G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_UINT);
  g_signal_new ("hangup",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_FIRST,
		G_STRUCT_OFFSET (RecordProcessClass, hangup),
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
  g_signal_new ("child-start",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1, G_TYPE_INT);
  g_signal_new ("child-exit",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_LAST,
		G_STRUCT_OFFSET (RecordProcessClass, child_exit),
		NULL, NULL,
		t_marshal_VOID__INT_INT,
		G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_INT);
}

/**
 * record_process_new:
 */
RecordProcess*
record_process_new (void)
{
  return g_object_new (RECORD_TYPE_PROCESS, NULL);
}

/**
 * record_process_get_ptsname:
 */
const gchar*
record_process_get_ptsname (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), NULL);
  return process->priv->ptsname;
}

/**
 * record_process_get_child_pid:
 */
GPid
record_process_get_child_pid (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), 0);
  return process->priv->child_pid;
}

/**
 * record_process_get_workdir:
 */
const gchar*
record_process_get_workdir (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), NULL);
  return process->priv->workdir;
}

/**
 * record_process_set_workdir:
 */
void
record_process_set_workdir (RecordProcess *process,
			    const gchar   *workdir)
{
  gchar *newdir = NULL;

  g_return_if_fail (RECORD_IS_PROCESS (process));

  if (workdir)
    newdir = g_strdup (workdir);

  g_free (process->priv->workdir);
  process->priv->workdir = newdir;

  g_object_notify (G_OBJECT (process), "workdir");
}

/**
 * record_process_get_argv:
 */
const gchar**
record_process_get_argv (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), NULL);
  return (const gchar**) process->priv->argv;
}

/**
 * record_process_set_argv:
 */
void
record_process_set_argv (RecordProcess *process,
			 const gchar  **argv)
{
  g_return_if_fail (RECORD_IS_PROCESS (process));

  if (process->priv->argv)
    g_strfreev (process->priv->argv);

  if (argv)
    process->priv->argv = g_strdupv ((gchar**) argv);
  else
    process->priv->argv = NULL;
}

/**
 * find_env_by_name:
 */
static gint
find_env_by_name (RecordProcessEnv *env,
		  const gchar      *name)
{
  g_return_val_if_fail (env, 1);
  g_return_val_if_fail (name, 1);

  return strcmp (env->name, name);
}

/**
 * record_process_set_env:
 */
gboolean
record_process_env_set (RecordProcess *process,
			const gchar   *name,
			const gchar   *value,
			gboolean       overwrite)
{
  GList *node;
  RecordProcessEnv *env;

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);
  g_return_val_if_fail (name, FALSE);
  g_return_val_if_fail (value, FALSE);

  node = g_list_find_custom (process->priv->env_list, name,
			     (GCompareFunc) find_env_by_name);
  if (node)
    {
      env = node->data;

      g_free (env->value);
      env->value = g_strdup (value);
      env->overwrite = overwrite;
    }
  else
    {
      env = g_new (RecordProcessEnv, 1);
      env->name = g_strdup (name);
      env->value = g_strdup (value);
      env->overwrite = overwrite;

      process->priv->env_list =
       g_list_prepend (process->priv->env_list, env);
    }
  return TRUE;
}

/**
 * record_process_unset_env:
 */
void
record_process_env_unset (RecordProcess *process,
			  const gchar   *name)
{
  GList *node;

  g_return_if_fail (RECORD_IS_PROCESS (process));
  g_return_if_fail (name);

  node = g_list_find_custom (process->priv->env_list, name,
			     (GCompareFunc) find_env_by_name);
  if (node)
    {
      record_process_env_free (node->data);

      process->priv->env_list =
       g_list_delete_link (process->priv->env_list, node);
    }
}

/**
 * record_process_get_env_list:
 */
const GList*
record_process_env_list (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), NULL);
  return process->priv->env_list;
}

/**
 * record_process_set_login_log:
 */
void
record_process_set_login_log (RecordProcess *process,
			      gboolean       login_log)
{
  g_return_if_fail (RECORD_IS_PROCESS (process));

  process->priv->login_log = login_log;
  g_object_notify (G_OBJECT (process), "login-log");
}

/**
 * record_process_get_login_log:
 */
gboolean
record_process_get_login_log (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);
  return process->priv->login_log;
}

/**
 * record_process_create_channel:
 */
static GIOChannel*
record_process_create_channel (RecordProcess *process)
{
  GIOChannel *channel;
  GError *error = NULL;

  g_return_val_if_fail (RECORD_IS_PROCESS (process), NULL);

  channel = g_io_channel_unix_new (process->priv->master);
  if (!channel)
    {
      g_critical (_("Failed to create pty channel."));
      return NULL;
    }
  g_io_channel_set_flags (channel, G_IO_FLAG_NONBLOCK, &error);
  if (error)
    {
      g_warning (_("Unable to set pty channel's non-block flags: %s"),
		 error->message);
      g_error_free (error);
      return NULL;
    }
  g_io_channel_set_encoding (channel, NULL, &error);
  if (error)
    {
      g_warning (_("Unable to set pty channel's encoding: %s"),
		 error->message);
      g_error_free (error);
      return NULL;
    }
  g_io_channel_set_buffered (channel, FALSE);

  return channel;
}
/**
 * record_process_channel_readable:
 */
gboolean
record_process_channel_readable (GIOChannel   *channel,
				 GIOCondition  cond,
				 gpointer      data)
{
  RecordProcess *process = RECORD_PROCESS (data);

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  if (cond & G_IO_IN)
    {
      gchar buff[4096];
      gsize bytes_read = 0;
      GIOStatus status;
      GError *error = NULL;

      status = g_io_channel_read_chars (channel, buff, sizeof (buff),
					&bytes_read, &error);
      switch (status)
	{
	case G_IO_STATUS_ERROR:
	case G_IO_STATUS_EOF:
	  if (error)
	    {
	      g_warning (_("Failed to read pty channel: %s"), error->message);
	      g_error_free (error);
	    }
	  g_signal_emit_by_name (process, "hangup");
	  return FALSE;

	case G_IO_STATUS_NORMAL:
	  break;
	case G_IO_STATUS_AGAIN:
	  g_usleep (1000 * 10);
	  break;
	default:
	  g_assert_not_reached ();
	}
      if (bytes_read > 0)
	{
	  g_signal_emit_by_name (process, "readable", buff, bytes_read);
	  return TRUE;
	}
    }
  if (cond & G_IO_HUP || cond & G_IO_ERR)
    {
      _debug ("pty io channel hungup or error.");
      g_signal_emit_by_name (process, "hangup");
      return FALSE;
    }
  return TRUE;
}

/**
 * record_process_channel_readable_removed:
 */
void
record_process_channel_readable_removed (gpointer data)
{
  RecordProcess *process = RECORD_PROCESS (data);

  g_return_if_fail (RECORD_IS_PROCESS (process));

  process->priv->read_watch_id = -1;
}

/**
 * record_process_readable_watch:
 */
static gboolean
record_process_readable_watch (RecordProcess *process)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), -1);

  g_assert (process->priv->channel);

  if (process->priv->read_watch_id > 0)
    return TRUE;

  process->priv->read_watch_id =
   g_io_add_watch_full (process->priv->channel,
			G_PRIORITY_DEFAULT,
			G_IO_IN | G_IO_ERR | G_IO_HUP,
			record_process_channel_readable,
			process,
			record_process_channel_readable_removed);

  return (process->priv->read_watch_id > 0);
}

/**
 * on_child_exit:
 */
static void
on_child_exit (GPid     pid,
	       gint     status,
	       gpointer data)
{
  RecordProcess *process = RECORD_PROCESS (data);

  g_return_if_fail (RECORD_IS_PROCESS (process));

  g_signal_emit_by_name (process, "child-exit", pid, status);
  g_spawn_close_pid (pid);
}

/**
 * child_process_setup:
 *
 * this function will execute in the context of child process.
 */
static gboolean
child_process_setup (RecordProcess *process)
{
  GList *node;

  for (node = process->priv->env_list; node; node = node->next)
    {
      RecordProcessEnv *env = node->data;

      g_setenv (env->name, env->value, env->overwrite);
    }
  if (process->priv->workdir)
    {
      if (g_chdir (process->priv->workdir) < 0)
	{
	  g_warning ("chdir error: %s", g_strerror (errno));
	  return FALSE;
	}
    }
  return TRUE;
}

/**
 * record_process_sync_workdir:
 */
static gboolean
record_process_sync_workdir (gpointer data)
{
  RecordProcess *process = data;
  gchar buf[PATH_MAX+1], *path;
  gint len = 0;

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  path = g_object_get_data (G_OBJECT (process), "cwd-path");

  len = readlink (path, buf, sizeof (buf) - 1);
  if (len > 0 && buf[0] == '/')
    {
      buf[len] = '\0';

      /* change current work directory. */
      if (g_chdir (buf) != 0)
	g_warning ("chdir() error: %s", g_strerror (errno));
    }
  else
    {
      /* On Solaris, readlink returns an empty string, but the
       * link can be used as a directory, including as a target
       * of chdir(). */
      if (g_chdir (path) != 0)
	g_warning ("chdir() error: %s", g_strerror (errno));
    }
  return TRUE;
}

/**
 * record_process_sync_workdir_destroy:
 */
static void
record_process_sync_workdir_destroy (gpointer data)
{
  RecordProcess *process = data;

  g_return_if_fail (RECORD_IS_PROCESS (process));

  g_object_set_data (G_OBJECT (process), "cwd-path", NULL);
  process->priv->workdir_watch_id = 0;
}

/**
 * record_process_workdir_watch:
 *
 * watch child process work directory change and
 * sync ttyrcd work directory.
 *
 * this feature is useful for gnome-terminal or like to
 * search child process work directory in order to set
 * child process initial work directory. see gnome-terminal
 * sources for detail.
 */
static void
record_process_workdir_watch (RecordProcess *process)
{
  static const gchar *patterns[] =
    {
      "/proc/%d/cwd",         /* Linux */
      "/proc/%d/path/cwd",    /* Solaris >= 10 */
    };
  guint i;

  g_return_if_fail (RECORD_IS_PROCESS (process));
  g_return_if_fail (process->priv->child_pid > 0);
  g_return_if_fail (process->priv->workdir_watch_id == 0);

  for (i = 0; i < G_N_ELEMENTS (patterns); ++i)
    {
      gchar *path;

      /* seem we cannot use GFileMonitor for monitor symlink.
       * test is not pass.
       */
      path = g_strdup_printf (patterns[i], process->priv->child_pid);
      if (g_file_test (path, G_FILE_TEST_IS_SYMLINK))
	{
	  /* save the path. */
	  g_object_set_data_full (G_OBJECT (process), "cwd-path", path, g_free);

	  /* scheduling to update work directory. */
	  process->priv->workdir_watch_id =
	   g_timeout_add_seconds_full (G_PRIORITY_DEFAULT, 3,
				       record_process_sync_workdir, process,
				       record_process_sync_workdir_destroy);
	  return;
	}
      g_free (path);
    }
  g_warning ("Unabel to sync child process work directory.");
}

/**
 * record_process_start:
 *
 * fork child process, and execute the argv setup before.
 */
gboolean
record_process_start (RecordProcess  *process,
		      struct termios *term_attr)
{
  struct winsize wsize;
  GPid pid;
  gint pipefds[2];
  gchar buff[32];

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  /* if argv is not set, we try to execute user login shell. */
  if (!process->priv->argv)
    {
      struct passwd *pswd = getpwuid (getuid ());
      if (!pswd)
	{
	  g_critical ("getpwuid error: %s", g_strerror (errno));
	  return FALSE;
	}
      process->priv->argv = g_new0 (char*, 2);
      process->priv->argv[0] = g_strdup (pswd->pw_shell);
    }
  /* get terminal size. */
  if (ioctl (STDOUT_FILENO, TIOCGWINSZ, &wsize) != 0)
    {
      wsize.ws_row = 24;
      wsize.ws_col = 80;
    }
  /* bring these environments into child process. */
  g_snprintf (buff, sizeof (buff), "%d", getpid ());
  record_process_env_set (process, "TTYRCD_PID", buff, TRUE);
  record_process_env_set (process, "TTYRCD_TTY",
			  ttyname (STDOUT_FILENO), TRUE);

  /* keep parent/child sync (need???) */
  if (pipe (pipefds) != 0)
    {
      g_warning ("pipe error: %s", g_strerror (errno));
      return FALSE;
    }
  /* fork a child process on the pseudo terminal slave. */
  pid = forkpty (&process->priv->master, NULL, term_attr, &wsize);
  if (pid < 0)
    {
      g_warning ("forkpty error: %s", g_strerror (errno));
      return FALSE;
    }
  if (pid == 0)		/* child */
    {
      if (!child_process_setup (process))
	exit (1);

      /* now we under pseudo terminal slave */
      g_setenv ("TTYRCD_PTS", ttyname (STDOUT_FILENO), TRUE);

      /* record login log. */
      if (process->priv->login_log)
	{
	  /* we have to call this function in child process
	   * since this function need use STDIN_FILENO to get
	   * ttyname. */
	  record_utmp_record_login_log ();
	}
      /* close all open files. */
      fdwalk (set_cloexec, GINT_TO_POINTER (3));

      /* wait for parent. */
      while (1)
	{
	  if (read (pipefds[0], &buff, 2) == 2)
	    break;
	}
      close (pipefds[0]);
      close (pipefds[1]);

      if (execvp (process->priv->argv[0], process->priv->argv) < 0)
	g_warning ("execvp() error: %s", g_strerror (errno));

      exit (0);
    }
  else if (pid > 0)	/* parent */
    {
      process->priv->child_pid = pid;

      if (write (pipefds[1], "3e", 2) < 0)
	{
	  g_warning ("write() pipe error: %s", g_strerror (errno));
	  return FALSE;
	}
      process->priv->ptsname = g_strdup (ptsname (process->priv->master));

      process->priv->channel = record_process_create_channel (process);
      if (!process->priv->channel)
	return FALSE;

      /* watch child process output. */
      if (!record_process_readable_watch (process))
	return FALSE;

      g_signal_emit_by_name (process, "child-start", pid);

      /* get notify when child exit. */
      g_child_watch_add (pid, on_child_exit, process);

      close (pipefds[0]);
      close (pipefds[1]);

      /* sync work directory with child process. */
      record_process_workdir_watch (process);
    }
  return TRUE;
}

/**
 * kill_child:
 */
static gboolean
kill_child (gpointer data)
{
  RecordProcess *process = data;
  guint signo;

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  signo = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (process), "signo"));
  if (signo == SIGKILL)
    return FALSE;

  if (process->priv->child_pid > 0)
    {
      kill (process->priv->child_pid, signo);

      g_object_set_data (G_OBJECT (process), "signo",
			 GUINT_TO_POINTER (SIGKILL));
      g_timeout_add (1000, kill_child, process);
    }
  return FALSE;
}

/**
 * record_process_stop:
 */
void
record_process_stop (RecordProcess *process)
{
  g_return_if_fail (RECORD_IS_PROCESS (process));

  /* clear login log. */
  if (process->priv->login_log)
    record_utmp_clear_login_log (process->priv->ptsname);

  if (process->priv->master)
    {
      close (process->priv->master);
      process->priv->master = -1;
    }
  g_object_set_data (G_OBJECT (process), "signo",
		     GUINT_TO_POINTER (SIGTERM));
  g_timeout_add (300, kill_child, process);
}

/**
 * record_process_channel_writable:
 */
gboolean
record_process_channel_writable (GIOChannel   *channel,
				 GIOCondition  cond,
				 gpointer      data)
{
  RecordProcess *process = data;
  gchar buff[1024];
  gint size;

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  size = t_ring_read (process->priv->write_ring, buff, sizeof (buff));
  if (size > 0)
    {
      gsize written;
      GError *error = NULL;
      GIOStatus status;
      gchar *datap = buff;

again:
      status = g_io_channel_write_chars (channel, datap, size, &written, &error);
      switch (status)
	{
	case G_IO_STATUS_ERROR:
	  if (error)
	    {
	      g_warning ("Failed to write pty channel: %s", error->message);
	      g_error_free (error);
	    }
	  return FALSE;

	case G_IO_STATUS_AGAIN:
	  if (size - written > 0)
	    {
	      if (tcdrain (g_io_channel_unix_get_fd (channel)) != 0)
		{
		  g_warning ("tcdrain error: %s", g_strerror (errno));
		  return FALSE;
		}
	      datap += written;
	      size -= written;
	      goto again;
	    }
	  break;

	case G_IO_STATUS_NORMAL:
	  break;
	default:
	  g_assert_not_reached ();
	}
    }
  return (t_ring_get_size (process->priv->write_ring) > 0);
}

/**
 * record_process_channel_writable_removed:
 */
void
record_process_channel_writable_removed (gpointer data)
{
  RecordProcess *process = RECORD_PROCESS (data);

  g_return_if_fail (RECORD_IS_PROCESS (process));

  process->priv->write_watch_id = -1;
}

/**
 * write_watch_add_idle:
 */
static gboolean
write_watch_add_idle (gpointer data)
{
  RecordProcess *process = data;

  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  if (process->priv->write_watch_id <= 0)
    {
      process->priv->write_watch_id =
       g_io_add_watch_full (process->priv->channel,
			    G_PRIORITY_DEFAULT, G_IO_OUT,
			    record_process_channel_writable,
			    process,
			    record_process_channel_writable_removed);
    }
  return FALSE;
}

/**
 * record_process_write:
 */
gboolean
record_process_write (RecordProcess *process,
		      const gchar   *buff,
		      gint           size)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);
  g_return_val_if_fail (buff, FALSE);

  if (!process->priv->channel)
    return FALSE;

  g_mutex_lock (process->priv->write_ring_mutex);

  t_ring_write (process->priv->write_ring, buff, size);

  /* switch to main thread */
  g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		   write_watch_add_idle, process, NULL);

  g_mutex_unlock (process->priv->write_ring_mutex);

  return TRUE;
}

/**
 * record_process_set_size:
 */
gboolean
record_process_set_size (RecordProcess *process,
			 guint16        lines,
			 guint16        columns)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  if (process->priv->master > 0)
    {
      return t_termio_set_winsize (process->priv->master,
				   lines ? lines : 24,
				   columns ? columns : 80);
    }
  return FALSE;
}

/**
 * record_process_get_size:
 */
gboolean
record_process_get_size (RecordProcess *process,
			 guint16       *lines,
			 guint16       *columns)
{
  g_return_val_if_fail (RECORD_IS_PROCESS (process), FALSE);

  if (process->priv->master > 0)
    {
      return t_termio_get_winsize (process->priv->master,
				   lines, columns);
    }
  return FALSE;
}

/**
 * record_process_set_utf8:
 */
void
record_process_set_utf8 (RecordProcess *process,
			 gboolean       utf8)
{
#ifdef IUTF8
  struct termios tio;

  g_return_if_fail (RECORD_IS_PROCESS (process));

  if (process->priv->master < 0)
    return;

  if (tcgetattr (process->priv->master, &tio) != -1)
    {
      tcflag_t saved_cflag = tio.c_iflag;

      if (utf8)
	tio.c_iflag |= IUTF8;
      else
	tio.c_iflag &= ~IUTF8;

      if (saved_cflag != tio.c_iflag)
	tcsetattr (process->priv->master, TCSANOW, &tio);
    }
#endif
}

/**
 * fdwalk:
 *
 * this function take from glib source
 *
 * we hope child process close all fds open by parent,
 * this function provide a effective and _portable_ way
 * to do so.
 */
#ifndef HAVE_FDWALK
static int
fdwalk (int (*cb) (void *data, int fd), void *data)
{
  gint open_max;
  gint fd;
  gint res = 0;
#ifdef HAVE_SYS_RESOURCE_H
  struct rlimit rl;
#endif

#ifdef __linux__  
  DIR *d;

  if ((d = opendir ("/proc/self/fd")))
    {
      struct dirent *de;

      while ((de = readdir (d)))
	{
	  glong l;
	  gchar *e = NULL;

	  if (de->d_name[0] == '.')
	    continue;

	  errno = 0;
	  l = strtol (de->d_name, &e, 10);
	  if (errno != 0 || !e || *e)
	    continue;

	  fd = (gint) l;

	  if ((glong) fd != l)
	    continue;

	  if (fd == dirfd (d))
	    continue;

	  if ((res = cb (data, fd)) != 0)
	    break;
	}
      closedir (d);
      return res;
    }
  /* If /proc is not mounted or not accessible we fall back to the old
   * rlimit trick */
#endif

#ifdef HAVE_SYS_RESOURCE_H
  if (getrlimit (RLIMIT_NOFILE, &rl) == 0 && rl.rlim_max != RLIM_INFINITY)
    open_max = rl.rlim_max;
  else
#endif
    open_max = sysconf (_SC_OPEN_MAX);

  for (fd = 0; fd < open_max; fd++)
    {
      if ((res = cb (data, fd)) != 0)
	break;
    }
  return res;
}
#endif

/**
 * set_cloexec:
 *
 * this function take from glib source
 */
static int
set_cloexec (void *data, gint fd)
{
  if (fd >= GPOINTER_TO_INT (data))
    {
      fcntl (fd, F_SETFD, FD_CLOEXEC);
    }
  return 0;
}
