/* Copyright (C) 2010 xiaohu<xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib/gi18n.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

#include <ttyutils/ttyutils.h>

#include "record-app.h"
#include "record-script.h"

/* lua register key for app object and process string buffer */
static const gchar  *_app_object_key = "recorader-app";
static const gchar  *_process_string = "process-string";

/* keep process string buffer thread-safe */
static GStaticMutex  _process_string_mutex = G_STATIC_MUTEX_INIT;

/* defined in record.c */
extern void	record_options_register	(lua_State    *L);

/* macros to set/get the RecordApp object from lua state */
#define script_set_app_object(L, app)	G_STMT_START {		\
    lua_pushlightuserdata (L, (gpointer) &_app_object_key);	\
    lua_pushlightuserdata (L, (gpointer) app);			\
    lua_settable (L, LUA_REGISTRYINDEX);			\
  } G_STMT_END

#define script_get_app_object(L, app)	G_STMT_START {		\
    lua_pushlightuserdata (L, (gpointer) &_app_object_key);	\
    lua_gettable (L, LUA_REGISTRYINDEX);			\
    app = lua_touserdata (L, -1);				\
    g_assert (RECORD_IS_APP (app));				\
    lua_pop (L, 1);						\
  } G_STMT_END

/* macros to set/get the GString buffer from lua state */
#define script_set_process_string(L, string)	G_STMT_START {	\
    lua_pushlightuserdata (L, (gpointer) &_process_string);	\
    lua_pushlightuserdata (L, (gpointer) string);		\
    lua_settable (L, LUA_REGISTRYINDEX);			\
  } G_STMT_END

#define script_get_process_string(L, string)	G_STMT_START {	\
    lua_pushlightuserdata (L, (gpointer) &_process_string);	\
    lua_gettable (L, LUA_REGISTRYINDEX);			\
    string = lua_touserdata (L, -1);				\
    lua_pop (L, 1);						\
  } G_STMT_END

/**
 * script_quit:
 */
static gint
script_quit (lua_State *L)
{
  RecordApp *app;
  script_get_app_object (L, app);
  record_app_quit (app);
  return 0;
}

/**
 * on_process_readable:
 *
 * called when child process output data.
 * we cache the output here for later to implement
 * expect interface.
 */
static void
on_process_readable (RecordProcess *process,
		     const gchar   *data,
		     gint           size,
		     lua_State     *L)
{
  GString *string;

  g_return_if_fail (RECORD_IS_PROCESS (process));

  g_static_mutex_lock (&_process_string_mutex);

  script_get_process_string (L, string);
  if (string)
    {
      /* prevent unlimit enlarge */
      if (string->len > 1024 * 20000)
	{
	  g_critical ("Cache size exceed 20M, mandatory clear.");
	  g_string_free (string, TRUE);
	  string = NULL;
	}
      else
	{
	  g_string_append_len (string, data, size);

	  if (string->len > 1024 * 5000)
	    g_warning ("Cache size exceed 5M.");
	}
    }
  if (!string)
    {
      string = g_string_new_len (data, size);
      script_set_process_string (L, string);
    }
  g_static_mutex_unlock (&_process_string_mutex);
}

/**
 * process_write:
 *
 * write data to child process
 */
static gint
script_feed (lua_State *L)
{
  RecordApp *app;
  RecordProcess *process;

  script_get_app_object (L, app);

  process = record_app_get_process (app);
  if (process && RECORD_IS_PROCESS (process))
    {
      record_process_write (process,
			    luaL_checkstring (L, 1),
			    luaL_optint (L, 2, -1));
    }
  return 0;
}

/**
 * script_expect:
 *
 * wait for expected string.
 * this function should be called in independent
 * thread(script thread, see record-app.c for detail).
 */
static gint
script_expect (lua_State *L)
{
  RecordApp *app;
  GRegex *regexp;
  const gchar *charset;
  guint compile_flags = 0;
  gchar *matched = NULL;
  gint start_pos = 0, end_pos = 0;
  GError *error = NULL;
  const gchar *expect = luaL_checkstring (L, 1);
  glong timeout = 1000 * 1000 * luaL_optint (L, 2, 5);

  script_get_app_object (L, app);

  if (!g_get_charset (&charset))
    compile_flags |= G_REGEX_RAW;

  regexp = g_regex_new (expect, compile_flags,
			G_REGEX_MATCH_NOTEMPTY, &error);
  if (error)
    {
      g_warning (_("Unable to compile regular expression `%s': %s"),
		 expect, error->message);
      g_free (error);
      return luaL_error (L, _("Failed to compile `%s'"), expect);
    }

#define _TIMEOUT_STEP	(1000 * 100)	/* 1/10 seconds */
  /* wait until timeout */
  while (timeout > 0)
    {
      GString *string;

      g_static_mutex_lock (&_process_string_mutex);

      script_get_process_string (L, string);
      if (string && string->str)
	{
	  GMatchInfo *match_info;

	  if (g_regex_match (regexp, string->str, 0, &match_info))
	    {
	      matched = g_match_info_fetch (match_info, 0);
	      g_match_info_fetch_pos (match_info, 0, &start_pos, &end_pos);

	      g_match_info_free (match_info);
	      g_static_mutex_unlock (&_process_string_mutex);
	      break;
	    }
	  g_match_info_free (match_info);
	}
      g_static_mutex_unlock (&_process_string_mutex);

      g_usleep (_TIMEOUT_STEP);
      timeout -= _TIMEOUT_STEP;
    }
  g_regex_unref (regexp);

  if (matched)
    {
      lua_pushstring (L, matched);
      g_free (matched);

      lua_pushinteger (L, start_pos);
      lua_pushinteger (L, end_pos);
      return 3;
    }
  return 0;
}

/**
 * script_erase:
 * remove from or clear process buffer
 */
static gint
script_erase (lua_State *L)
{
  GString *string;

  g_static_mutex_lock (&_process_string_mutex);

  script_get_process_string (L, string);
  if (string)
    {
      gint pos = 0;
      gchar *p;

      switch (lua_type (L, 1))
	{
	case LUA_TSTRING:	/* discard to first match(includes) */
	  p = g_strstr_len (string->str, string->len, lua_tostring (L, 1));
	  if (p) pos = p - string->str + strlen (lua_tostring (L, 1));
	  break;
	case LUA_TNUMBER:	/* discard to position */
	  pos = lua_tointeger (L, 1);
	  break;
	default:
	  break;
	}
      if (pos > 0)
	{
	  g_string_erase (string, 0, (pos >= string->len) ? -1 : pos);
	}
      else
	{
	  g_string_free (string, TRUE);
	  script_set_process_string (L, NULL);
	}
    }
  g_static_mutex_unlock (&_process_string_mutex);

  return 0;
}

/**
 * script_suspend:
 * lock keyboard
 */
static gint
script_suspend (lua_State *L)
{
  RecordApp *app;
  TConsole *console;

  script_get_app_object (L, app);
  console = record_app_get_console (app);

  if (T_IS_CONSOLE (console))
    t_console_suspend (console, CONSOLE_SUSPEND_STDIN, TRUE);

  return 0;
}

/**
 * script_resume:
 * unlock keyboard
 */
static gint
script_resume (lua_State *L)
{
  RecordApp *app;
  TConsole *console;

  script_get_app_object (L, app);
  console = record_app_get_console (app);

  if (T_IS_CONSOLE (console))
    t_console_resume (console, CONSOLE_SUSPEND_STDIN, TRUE);

  return 0;
}

/**
 * process_snapshot:
 *
 * take screen snapshot.
 * this interface need mvt plugin
 */
static gint
script_snapshot (lua_State *L)
{
  RecordApp *app;
  script_get_app_object (L, app);
  return 0;
}

/**
 * script_match:
 */
static gint
script_match (lua_State *L)
{
  RecordApp *app;
  script_get_app_object (L, app);
  return 0;
}

static luaL_Reg	_lua_record[] =
{
    { "quit",		script_quit },
    { "feed",		script_feed },
    { "expect",		script_expect },
    { "erase",		script_erase },
    { "snap",		script_snapshot },
    { "match",		script_match },
    { "suspend",	script_suspend },
    { "resume",		script_resume },

    { NULL, NULL },
};

/**
 * record_script_register:
 */
static void
record_script_register (lua_State    *L,
			RecordApp    *app,
			const gchar **argv)
{
  gint i;

  g_return_if_fail (RECORD_IS_APP (app));

  script_set_app_object (L, app);

  /* construct arg table */
  lua_newtable (L);

  for (i = 0; i < g_strv_length ((gchar**) argv); i++)
    {
      if (strlen (argv[i]) > 0)
	{
	  lua_pushnumber (L, i);
	  lua_pushstring (L, argv[i]);
	  lua_settable (L, -3);
	}
    }
  lua_setglobal (L, "arg");

  luaL_register (L, "record", _lua_record);
}

/**
 * record_script_run:
 *
 * run script file.
 * this function should been called in independent thread,
 * see record-app.c for detail.
 */
gboolean
record_script_run (RecordApp    *app,
		   const gchar **argv)
{
  lua_State *L;
  gboolean ret = TRUE;
  RecordProcess *process;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);
  g_return_val_if_fail (argv && argv[0], FALSE);

  /* create lua state and open standard modules */
  L = luaL_newstate ();
  if (!L)
    {
      g_warning ("luaL_newstate error.");
      return FALSE;
    }
  luaL_openlibs (L);

  /* register record module */
  record_script_register (L, app, argv);

  /* register command line options */
  record_options_register (L);

  /* save process output for later match.
   * expect() command need these data.
   * Note the callback will been running in main
   * thread(beacuse the signal is emit in main thread).
   * the callback must take care of thread-safe problem. */
  process = record_app_get_process (app);
  g_signal_connect_after (process, "readable",
			  G_CALLBACK (on_process_readable), L);

  if (luaL_dofile (L, argv[0]) != 0)
    {
      g_warning (_("Failed to run script file `%s': %s"),
		 argv[0], lua_tostring (L, -1));
      ret = FALSE;
    }
  g_signal_handlers_disconnect_by_func (process, on_process_readable, L);
  script_erase (L);
  lua_close (L);

  return ret;
}

/**
 * record_lua_dump_stack:
 *
 * take from `Lua Programming', for debug
 */
void
record_lua_dump_stack (lua_State *L)
{
  gint i, top = lua_gettop (L);

  /* repeat for each level */
  for (i = 1; i <= top; i++)
    {
      gint t = lua_type (L, i);
      switch (t)
	{
	case LUA_TSTRING:	/* strings */
	  g_print ("`%s'", lua_tostring (L, i));
	  break;
	case LUA_TBOOLEAN:	/* booleans */
	  g_print (lua_toboolean (L, i) ? "true" : "false");
	  break;
	case LUA_TNUMBER:	/* numbers */
	  g_print ("%g", lua_tonumber (L, i));
	  break;
	default:		/* other values */
	  g_print ("%s", lua_typename (L, t));
	  break;
	}
      g_print (" "); /* put a separator */
    }
  g_print ("\n"); /* end the listing */
}
