/* Copyright (C) 2010 xiaohu<xiaou417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <locale.h>
#include <unistd.h>
#include <termios.h>
#include <errno.h>
#include <stdio.h>
#include <execinfo.h>
#include <syslog.h>
#include <stdlib.h>
#include <gcrypt.h>

#include <glib.h>
#include <glib/gi18n.h>

#include <ttyutils/ttyutils.h>

#include "record-app.h"
#include "record-script.h"

static struct termios     _stdin_term_attr;
static struct termios     _stdout_term_attr;
static time_t             _uptime;

static gchar		 *_opt_config = NULL;
static gchar		 *_opt_logfile = NULL;
static gchar		**_opt_scripts = NULL;
static gchar		 *_opt_script_args = NULL;
static gboolean		  _opt_recursion = FALSE;
static gboolean		  _opt_version = FALSE;

static GOptionEntry	_option_entries[] =
{
    { "config", 'c', 0, G_OPTION_ARG_FILENAME, &_opt_config,
      N_("Use specified configure file"), "FILE" },

    { "logfile", 'l', 0, G_OPTION_ARG_FILENAME, &_opt_logfile,
      N_("Logged message to file"), "FILE" },

    { "script", 's', 0, G_OPTION_ARG_FILENAME_ARRAY, &_opt_scripts,
      N_("Execute script file when application ready"), "FILEs" },

    { "script-args", 'a', 0, G_OPTION_ARG_STRING, &_opt_script_args,
      N_("Pass arguments to script"), "ARGs" },

    { "recursion", 'r', 0, G_OPTION_ARG_NONE, &_opt_recursion,
      N_("Don't check whether terminal is pseudo terminal's slave device"), NULL },

    { "version", 'v', 0, G_OPTION_ARG_NONE, &_opt_version,
      N_("Output version information"), NULL },

    { NULL, }
};

/**
 * record_get_uptime:
 */
time_t
record_get_uptime (void)
{
  return _uptime;
}

/**
 * restore_terminal_attrs:
 *
 * this function will be called when application quit(that is,
 * register with atexit()).
 */
static void
restore_terminal_attrs (void)
{
  tcsetattr (STDIN_FILENO,  TCSANOW, &_stdin_term_attr);
  tcsetattr (STDOUT_FILENO, TCSANOW, &_stdout_term_attr);
}

/**
 * cleanup_secure_memory:
 */
static void
cleanup_secure_memory (void)
{
  gcry_control (GCRYCTL_TERM_SECMEM, 0);
}

/**
 * quit_application:
 */
static gboolean
quit_application (gpointer data)
{
  record_app_quit (record_app_get ());
  printf ("\r\n");

  return FALSE;
}

/**
 * signal_terminate:
 * handler of SIGTERM, SIGINT, SIGHUP system signal.
 */
static void
signal_terminate (gint signo)
{
  static gboolean termed = FALSE;

  if (!termed)
    {
      /* call record_app_quit() in signal handler may
       * cause coredump. */
      g_idle_add (quit_application, NULL);
      termed = TRUE;
    }
  else
    {
      printf ("\r\n");
      printf (_("Mandatory quit."));
      printf ("\r\n");
      exit (1);
    }
}

/**
 * winsize_changed_idle:
 */
#ifdef SIGWINCH
static gboolean
winsize_changed_idle (gpointer data)
{
  RecordApp *app = record_app_get ();
  guint16 rows, cols;

  if (!RECORD_IS_APP (app))
    return FALSE;

  if (t_termio_get_winsize (STDOUT_FILENO, &rows, &cols))
    {
      RecordProcess *process;
      RecordLookd *lookd;
      RecordEncoder *encoder;

      /* change child process window size. */
      process = record_app_get_process (app);
      if (RECORD_IS_PROCESS (process))
	record_process_set_size (process, rows, cols);

      /* notify client size changed. */
      lookd = record_app_get_lookd (app);
      if (RECORD_IS_LOOKD (lookd))
	record_lookd_resize (lookd, rows, cols);

      /* inject new size to record file. */
      encoder = record_app_get_encoder (app);
      if (RECORD_IS_ENCODER (encoder))
	record_encoder_inject_winsize (encoder, rows, cols);
    }
  return FALSE;
}
#endif

/**
 * signal_winsize_changed:
 * handler of SIGWINCH system signal.
 */
#ifdef SIGWINCH
static void
signal_winsize_changed (gint signo)
{
  g_idle_add (winsize_changed_idle, NULL);
}
#endif

/**
 * run_script_idle:
 *
 * a idle function for execute script file, we cannot
 * execute script before run main loop.
 */
static gboolean
run_script_idle (gpointer data)
{
  RecordApp *app = data;
  gint i;

  g_return_val_if_fail (RECORD_IS_APP (app), FALSE);

  /* scripts is a vector array(not just a string) */
  for (i = 0; _opt_scripts[i]; i++)
    {
      /* pass arguments to script.
       * you don't need use `--script-args' option to pass
       * arguments to script if you run script like below:
       *   $ ttyrcd -s "script args" ...
       * but if you like #!-style script, ttyrcd cannot
       * passes option to it(beacuse you hope options pass to
       * ttyrcd itself). */
      if (_opt_script_args)
	{
	  gchar *script = g_strjoin (" ", _opt_scripts[i],
				     _opt_script_args, NULL);
	  record_app_run_script (app, script);
	  g_free (script);
	}
      else
	{
	  record_app_run_script (app, _opt_scripts[i]);
	}
    }
  return FALSE;
}

/**
 * record_lua_path_add:
 *
 * append a lua module search path by modify LUA_PATH
 * and LUA_CPATH environment variable
 */
void
record_lua_path_add (const gchar *path)
{
  const gchar *env;
  gchar lpath[1024], *p; 

  g_return_if_fail (path);

  p = g_build_filename (path, "?", NULL);

  env = g_getenv ("LUA_PATH");
  g_snprintf (lpath, sizeof (lpath), "%s;%s.lua;%s.luc;%s",
	      p, p, p, env ? env : "");
  g_setenv ("LUA_PATH", lpath, TRUE);

  env = g_getenv ("LUA_CPATH");
  g_snprintf (lpath, sizeof (lpath), "%s.so;%s", p, env ? env : "");
  g_setenv ("LUA_CPATH", lpath, TRUE);

  g_free (p);
}

/**
 * main:
 *
 * the way to the hell.
 */
gint
main (gint argc, gchar *argv[])
{
  RecordApp *app;
  GOptionContext *option;
  struct sigaction action;
  GError *error = NULL;

  /* i18n */
  setlocale (LC_ALL, "");
  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  /* set application name for use in other place such as library. */
  g_set_prgname (argv[0]);
  g_set_application_name ("Ttyutils Recorder");

  if (!t_check_version ())
    return EXIT_FAILURE;

  /* we wish to be running under a really terminal device. */
  if (!isatty (STDIN_FILENO) || !isatty (STDOUT_FILENO))
    {
      g_warning ("%s: stdin/stdout is not a tty.", g_get_prgname ());
      syslog (LOG_CRIT, "%s: stdin/stdout is not a tty.", g_get_prgname ());
      return EXIT_FAILURE;
    }
  /* initial gthread system and gobject type system.
   * this is required before create any gobject-derived objects. */
  g_thread_init (NULL);
  g_type_init ();

  /* initialize libgcrypt subsystems. */
#ifdef DEBUG
  gcry_control (GCRYCTL_SET_VERBOSITY, 1);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1);
#else
  gcry_control (GCRYCTL_SET_VERBOSITY, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 0);
  gcry_control (GCRYCTL_CLEAR_DEBUG_FLAGS, 1);
#endif

  gcry_set_log_handler (t_gcrypt_log_handler, NULL);
  if (!gcry_check_version (GCRYPT_VERSION))
    {
      g_warning (_("libgcrypt version %s or above is required."),
		 GCRYPT_VERSION);
      return EXIT_FAILURE;
    }
  gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN, NULL);
  gcry_control (GCRYCTL_INIT_SECMEM, 16 * 1024, 0);
  gcry_control (GCRYCTL_RESUME_SECMEM_WARN, NULL);
#ifdef DEBUG
  gcry_control (GCRYCTL_PRINT_CONFIG, NULL);
#endif
  gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);

  /* cleanup secure memory when exit. */
  g_atexit (cleanup_secure_memory);

  /* install system signal handler. */
  memset (&action, 0, sizeof (action));
  action.sa_handler = signal_terminate;
  action.sa_flags = 0;
  sigemptyset (&action.sa_mask);

  /* SIGTERM, SIGINT, SIGHUP will terminate application. */
  sigaction (SIGTERM, &action, 0);
  sigaction (SIGINT, &action, 0);
  sigaction (SIGHUP, &action, 0);

  /* catch terminal winsize change signal. */
#ifdef SIGWINCH
  action.sa_handler = signal_winsize_changed;
  sigaction (SIGWINCH, &action, 0);
#endif

  /* ignore SIGUSR1(external program maybe send SIGUSR1 to
   * test whether ttyrcd is actually running) */
  action.sa_handler = SIG_IGN;
  sigaction (SIGUSR1, &action, 0);

  /* parse command line options */
  option = g_option_context_new ("[command [args ...]]");
  g_option_context_set_summary (option, _("Record terminal session"));
  g_option_context_set_description (option, WEBSITE);
  g_option_context_add_main_entries (option, _option_entries,
				     GETTEXT_PACKAGE);
  g_option_context_parse (option, &argc, &argv, &error);
  if (error)
    {
      g_warning (_("Parse command line error: %s"), error->message);
      g_error_free (error);
      g_option_context_free (option);
      return EXIT_FAILURE;
    }
  g_option_context_free (option);

  if (_opt_version)
    {
      t_print_version ();
      return EXIT_SUCCESS;
    }
  /* check whether we are running under a ttyrcd's slave device.
   * see record_process_start(). */
  if (!_opt_recursion)
    {
      if (g_getenv ("TTYRCD_PTS") &&
	  g_getenv ("TTYRCD_TTY") &&
	  g_getenv ("TTYRCD_PID") &&
	  g_file_test (g_getenv ("TTYRCD_TTY"), G_FILE_TEST_EXISTS) &&
	  strcmp (g_getenv ("TTYRCD_PTS"), ttyname (STDIN_FILENO)) == 0)
	{
	  gint pid = g_strtod (g_getenv ("TTYRCD_PID"), NULL);
	  if (pid > 0)
	    {
	      gint ret = kill (pid, SIGUSR1);
	      if (ret == 0 || errno != ESRCH)
		{
		  g_warning (_("Try to running %s under a pseudo terminal "
			       "slave device.(recursion)"), g_get_prgname ());
		  return EXIT_FAILURE;
		}
	    }
	}
    }
  _uptime = time (NULL);

  /* save terminal attributes for later restore. */
  tcgetattr (STDIN_FILENO,  &_stdin_term_attr);
  tcgetattr (STDOUT_FILENO, &_stdout_term_attr);
  g_atexit (restore_terminal_attrs);

  /* add Lua package search path(s) */
  record_lua_path_add (LUALIBDIR);

  /* create application object. */
  app = record_app_new ();

  /* load configure from file.
   *
   * if `--config' option is not appear, will try to open below
   * configure file, in that order:
   *
   *  - $SYSCONFDIR/ttyutils/ttyrcd.conf
   *  - $HOME/.config/ttyutils/ttyrcd.conf
   *
   * Note all of above files is optional(missing is not error,
   * just use built-in default value). and if more than one of
   * above files exists, all exists files will be load(by overwrite
   * previous configure).
   *
   * if `--logfile' option is appear, it may be a filename(just load
   * it) or `-'(no configure file at all).
   */
  if (!_opt_config)
    {
      gchar *files[3];
      gint i;

      files[0] = g_build_filename (SYSCONFDIR, "ttyutils", "ttyrcd.conf", NULL);
      files[1] = g_build_filename (g_get_user_config_dir (), "ttyutils",
				   "ttyrcd.conf", NULL);
      files[2] = NULL;

      for (i = 0; files[i]; i++)
	{
	  /* if file exists, but load failed, means configure file has
	   * syntax or similar error occured, we cannot continue. */
	  if (g_file_test (files[i], G_FILE_TEST_EXISTS))
	    {
	      if (!record_app_load_config_file (app, files[i]))
		{
		  g_object_unref (app);
		  return EXIT_FAILURE;
		}
	      break;
	    }
	}
      g_free (files[0]);
      g_free (files[1]);
    }
  else if (g_strcmp0 (_opt_config, "-") != 0)
    {
      /* open user specified configure file. */
      if (!record_app_load_config_file (app, _opt_config))
	{
	  g_object_unref (app);
	  return EXIT_FAILURE;
	}
    }
  /* please pay for my arduous work. */
  if (t_expire_check ())
    g_idle_add ((GSourceFunc) raise, GUINT_TO_POINTER (SIGTERM));

  /* create a idle event to ensure main loop has running before
   * script executed. */
  if (_opt_scripts)
    {
      /* don't show welcome phrase when execute script. */
      g_key_file_remove_key (record_app_get_keyfile (app),
			     "Application", "phrase", NULL);
      g_idle_add_full (G_PRIORITY_HIGH_IDLE, run_script_idle, app, NULL);
    }
  /* strip `--' options if has. */
  if (argc > 1 && argv[1])
    {
      if (strcmp (argv[1], "--") == 0)
	{
	  argv[1] = argv[0];
	  argv = &argv[1];
	  argc--;
	}
    }
  record_app_start (app, argc > 1 ? (const gchar**) &argv[1] : NULL,
		    &_stdin_term_attr);
  g_object_unref (app);

  _message ("Recorder elapse %ld seconds\n", time (NULL) - _uptime);
  return EXIT_SUCCESS;
}

/**
 * record_redirect_log:
 */
gboolean
record_redirect_log (void)
{
  if (_opt_logfile)
    {
      /* open log file. */
      if (!t_set_logfile (_opt_logfile))
	return FALSE;
    }
  else
    {
      openlog (g_get_prgname (), LOG_PID | LOG_CONS, LOG_USER);
    }
  g_log_set_default_handler (t_glib_log_handler, t_get_logfile ());
  g_set_print_handler (t_glib_print_handler);
  g_set_printerr_handler (t_glib_print_handler);

  return TRUE;
}

/**
 * lua_set_string_field:
 */
static void
lua_set_string_field (lua_State   *L,
		      const gchar *index,
		      const gchar *value)
{
  lua_pushstring (L, index);
  lua_pushstring (L, value);
  lua_settable(L, -3);
}

/**
 * lua_register_command_options:
 *
 * register command options as a table to lua state
 */
void
record_options_register (lua_State *L)
{
  lua_newtable (L);

  if (_opt_config)
    lua_set_string_field (L, "config", _opt_config);

  if (_opt_logfile)
    lua_set_string_field (L, "logfile", _opt_logfile);

  if (_opt_script_args)
    lua_set_string_field (L, "script-args", _opt_script_args);

  if (_opt_scripts)
    {
      gchar *str = g_strjoinv (" ", _opt_scripts);
      lua_set_string_field (L, "scripts", str);
      g_free (str);
    }
  if (_opt_recursion)
    {
      lua_pushstring (L, "recursion");
      lua_pushboolean (L, _opt_recursion);
      lua_settable(L, -3);
    }
  lua_setglobal (L, "options");
}
