#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <string.h>

#include "ttyutils/ttyutils.h"

#define CHUNKSIZE	(4096 - sizeof (char*) * 2)

/**
 * test_ring_quick:
 */
static void
test_ring_quick (void)
{
  TRing *ring = t_ring_new ();
  gchar buff[256];
  gint n;

  g_assert_cmpint (t_ring_get_size (ring), ==, 0);
  g_assert (t_ring_can_write (ring, 1024));
  g_assert (!t_ring_can_write (ring, CHUNKSIZE * 1280 + 1));
  g_assert (t_ring_can_read (ring, 0));
  g_assert (!t_ring_can_read (ring, 1));

  t_ring_write (ring, "helll", 5);

  g_assert_cmpint (t_ring_get_size (ring), ==, 5);
  g_assert (!t_ring_can_write (ring, CHUNKSIZE * 1280 - 4));
  g_assert (t_ring_can_read (ring, 5));
  g_assert (!t_ring_can_read (ring, 6));

  n = t_ring_read (ring, buff, sizeof (buff));

  g_assert_cmpint (n, ==, 5);
  g_assert (strncmp (buff, "helll", 5) == 0);
  g_assert_cmpint (t_ring_get_size (ring), ==, 0);
  g_assert (t_ring_read (ring, buff, sizeof (buff)) == 0);

  t_ring_free (ring);
}

/**
 * test_ring_performance:
 */
static void
test_ring_performance (void)
{
  TRing *ring = t_ring_new ();
  gint i;

  g_test_timer_start ();

  for (i = 0; i < 100000; i++)
    g_assert_cmpint (t_ring_write (ring, "helll", 5), ==, 5);

  g_test_message ("write 100,000 times(5 bytes each time) elapse %lf seconds",
		  g_test_timer_elapsed ());

  t_ring_clear (ring);
  g_test_message ("clear");

  g_test_timer_start ();

  for (i = 0; i < 100000; i++)
    g_assert_cmpint (t_ring_write (ring, "helll", 5), ==, 5);

  g_test_message ("write 100,000 times(5 bytes each time) elapse %lf seconds",
		  g_test_timer_elapsed ());

  t_ring_clear (ring);
  g_test_message ("clear");

  g_test_timer_start ();

  for (i = 0; i < 200000; i++)
    g_assert_cmpint (t_ring_write (ring, "helll", 5), ==, 5);

  g_test_message ("write 200,000 times(5 bytes each time) elapse %lf seconds",
		  g_test_timer_elapsed ());

  t_ring_clear (ring);
  g_test_message ("clear");

  g_test_timer_start ();

  for (i = 0; i < 200000; i++)
    g_assert_cmpint (t_ring_write (ring, "helll", 5), ==, 5);

  g_test_message ("write 200,000 times(5 bytes each time) elapse %lf seconds",
		  g_test_timer_elapsed ());

  t_ring_free (ring);
}

/**
 * test_ring_slow:
 */
static void
test_ring_slow (void)
{
  TRing *ring = t_ring_new ();
  gchar buff[256];
  gint i;

  t_ring_write (ring, "helll", 5);
  g_assert (t_ring_get_nchunks (ring) == 1);
  t_ring_clear (ring);

  for (i = 0; i < 100000; i++)
    t_ring_write (ring, "helll", 5);

  g_assert (t_ring_get_nchunks (ring) == 500000 / CHUNKSIZE + 1);
  g_assert (t_ring_get_size (ring) == 500000);

  for (i = 0; i < 50000; i++)
    {
      g_assert_cmpint (t_ring_read (ring, buff, 10), ==, 10);
      g_assert (strncmp (buff, "helllhelll", 10) == 0);
    }
  g_assert_cmpint (t_ring_read (ring, buff, sizeof (buff)), ==, 0);
  g_assert_cmpint (t_ring_get_nchunks (ring), ==, 1);

  for (i = 0; i < 10000; i++)
    {
      t_ring_write (ring, "helll", 5);
      g_assert (t_ring_read (ring, buff, 1) == 1);
    }
  g_assert (t_ring_get_size (ring) == 40000);
  t_ring_clear (ring);
  g_assert (t_ring_get_size (ring) == 0);

  /* read line */
  t_ring_write (ring, "helll\n", 6);
  g_assert (t_ring_can_read_line (ring, 6));
  g_assert (!t_ring_can_read_line (ring, 5));

  g_assert_cmpint (t_ring_read_line (ring, buff, sizeof (buff)), ==, 6);
  g_assert_cmpint (strncmp (buff, "helll\n", 6), ==, 0);
  g_assert_cmpint (t_ring_get_size (ring), ==, 0);

  /* read to */
  t_ring_write (ring, "helll\r\n", -1);
  g_assert_cmpint (t_ring_read_to (ring, buff, 0, sizeof (buff), NULL, 0), ==, 0);
  g_assert_cmpint (t_ring_read_to (ring, buff, 0, sizeof (buff), "\r\n", -1), ==, 5);
  g_assert_cmpint (t_ring_get_size (ring), ==, 2);

  /* discard to */
  t_ring_write (ring, "helll\r\n", -1);
  g_assert_cmpint (t_ring_discard_to (ring, 0, 100, "abcde", -1), ==, 0);
  g_assert_cmpint (t_ring_discard_to (ring, 2, 100, "abcde", -1), ==, 2);
  g_assert_cmpint (t_ring_discard_to (ring, 2, 100, "\r\n", -1), ==, 5);
  g_assert_cmpint (t_ring_get_size (ring), ==, 2);

  /* discard with */
  t_ring_clear (ring);
  t_ring_write (ring, "helll\r\n", -1);
  g_assert_cmpint (t_ring_get_size (ring), ==, 7);
  g_assert_cmpint (t_ring_discard_with (ring, 0, 100, "helll", -1), ==, 5);
  g_assert_cmpint (t_ring_discard_with (ring, 0, 100, "\r\n", -1), ==, 2);
  g_assert_cmpint (t_ring_get_size (ring), ==, 0);

  t_ring_clear (ring);
  t_ring_write (ring, "\033", -1);
  g_assert_cmpint (t_ring_get_size (ring), ==, 1);
  g_assert_cmpint (t_ring_discard_to (ring, 0, 100, "\033", -1), ==, 0);
  t_ring_write (ring, "\033", -1);
  g_assert_cmpint (t_ring_get_size (ring), ==, 2);
  g_assert_cmpint (t_ring_discard_to (ring, 1, 100, "\033", -1), ==, 1);
  g_assert_cmpint (t_ring_get_size (ring), ==, 1);
  t_ring_write (ring, "abcde\033", -1);
  t_ring_write (ring, "abcde\033", -1);
  t_ring_write (ring, "abcde\033", -1);
  t_ring_write (ring, "abcde\033", -1);
  g_assert_cmpint (t_ring_discard_to (ring, 1, 100, "\033", -1), ==, 6);
  g_assert_cmpint (t_ring_discard_to (ring, 7, 100, "\033", -1), ==, 12);
  g_assert_cmpint (t_ring_get_size (ring), ==, 7);
  t_ring_write (ring, "abcde\033", -1);
  g_assert_cmpint (t_ring_discard_with (ring, 7, 100, "\033", -1), ==, 13);
  g_assert_cmpint (t_ring_get_size (ring), ==, 0);

  t_ring_free (ring);
}

/**
 * test_ring:
 */
void
test_ring (void)
{
  if (g_test_quick ())
    test_ring_quick ();

  if (g_test_perf ())
    test_ring_performance ();

  if (g_test_slow ())
    test_ring_slow ();

  if (g_test_thorough ())
    {
      test_ring_quick ();
      test_ring_slow ();
      test_ring_performance ();
    }
}
