/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define _XOPEN_SOURCE
#include <locale.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <zip.h>
#include <gcrypt.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gtk/gtk.h>

#include <ttyutils/ttyutils.h>

/* we have use record's code do encode. */
#include "record-encoder.h"

#ifdef DEBUG
#define CONVERT_GLADE_FILE	SRCDIR G_DIR_SEPARATOR_S "assayer-convert.xml"
#else
#define CONVERT_GLADE_FILE	GLADEDIR G_DIR_SEPARATOR_S "assayer-convert.xml"
#endif

/* remember last access file. */
static gchar	*_last_typescript = NULL;
static gchar	*_last_timestamp  = NULL;

/* struct to hold convert script parameters. */
typedef struct
{
  gdouble  starting;
  gdouble  duration;
  gchar   *typescript;
  gchar   *timestamp;
  gchar   *password;
  gchar   *filename;
  gint     comp_level;
  gchar   *charset;
  gchar   *term;
  guint16  termrows;
  guint16  termcols;
  GString *messages;
} ConvertParams;

/**
 * convert_params_free:
 */
static void
convert_params_free (ConvertParams *params)
{
  if (params)
    {
      g_free (params->typescript);
      g_free (params->timestamp);
      g_free (params->password);
      g_free (params->filename);
      g_free (params->charset);
      g_free (params->term);

      if (params->messages)
	g_string_free (params->messages, TRUE);

      g_free (params);
    }
}

/**
 * convert_open_input_file:
 */
static inline GDataInputStream*
convert_open_input_file (const gchar *filename,
			 GError     **error)
{
  GFile *file;
  GFileInputStream *stream;
  GDataInputStream *data_stream;

  g_return_val_if_fail (filename, NULL);

  /* open file for read. */
  file = g_file_new_for_path (filename);
  stream = g_file_read (file, NULL, error);
  if (*error)
    {
      g_prefix_error (error, _("Unable to open `%s': "), filename);
      g_object_unref (file);
      return NULL;
    }
  g_object_unref (file);

  /* create data stream from file stream. */
  data_stream = g_data_input_stream_new (G_INPUT_STREAM (stream));
  g_object_unref (stream);

  return data_stream;
}

/**
 * convert_open_timestamp:
 */
static GDataInputStream*
convert_open_timestamp (const gchar *filename,
			GError     **error)
{
  g_return_val_if_fail (filename, NULL);

  return convert_open_input_file (filename, error);
}

/**
 * convert_open_typescript:
 */
static GDataInputStream*
convert_open_typescript (const gchar *filename,
			 gdouble     *starting,
			 GError     **error)
{
  const gchar *magic = "Script started on ";
  GDataInputStream *data_stream;
  gchar *line, *basename;

  g_return_val_if_fail (filename, NULL);
  g_return_val_if_fail (error, NULL);

  /* open typescript file to get data stream. */
  data_stream = convert_open_input_file (filename, error);
  if (!data_stream)
    return NULL;

  /* read and discard first line. */
  line = g_data_input_stream_read_line (data_stream, NULL, NULL, error);
  if (*error || !line)
    {
      basename = g_path_get_basename (filename);

      if (*error)
	g_prefix_error (error, _("Unable to read line from `%s': "), basename);
      else
	g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		     _("Unable to read line from `%s'"), basename);

      g_free (basename);
      g_object_unref (data_stream);
      return NULL;
    }
  /* typescript first line has fix string. */
  if (strncmp (line, magic, strlen (magic)) != 0)
    {
      basename = g_path_get_basename (filename);

      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Format of `%s' unrecognizable"), basename);
      g_free (basename);
      g_free (line);
      g_object_unref (data_stream);

      return NULL;
    }
  /* parse starting time(typescript first record the starting time). */
  if (starting)
    {
      struct tm tm;

      /* we should not return error when parse failed, beacuse
       * current locale maybe differ from record one. */
      if (!strptime (line + strlen (magic), "%c", &tm))
	{
	  *starting = 0.0;
	  g_warning ("Unable to parse starting time, the starting time may be lost.");
	}
      else
	{
	  *starting = mktime (&tm);
	}
    }
  g_free (line);

  return data_stream;
}


/**
 * convert_set_input_file_page_complete:
 */
static void
convert_set_input_file_page_complete (GtkAssistant *assistant)
{
  GtkBuilder *builder;
  GObject *chooser;
  GObject *page;
  GObject *typescript_label;
  GObject *timestamp_label;
  ConvertParams *params;
  GDataInputStream *stream;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  builder = g_object_get_data (G_OBJECT (assistant), "builder");
  page = gtk_builder_get_object (builder, "ConvertInputFilePage");

  params = g_object_get_data (G_OBJECT (assistant), "params");
  g_assert (params);

  typescript_label = gtk_builder_get_object (builder, "TypescriptErrorLabel");
  timestamp_label = gtk_builder_get_object (builder, "TimestampErrorLabel");
  gtk_widget_hide (GTK_WIDGET (typescript_label));
  gtk_widget_hide (GTK_WIDGET (timestamp_label));

  /* validate typescript file. */
  chooser = gtk_builder_get_object (builder, "TypescriptChooserButton");
  _last_typescript = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
  if (_last_typescript)
    {
      /* validate the typescript file. */
      stream = convert_open_typescript (_last_typescript, NULL, &error);
      if (error)
	{
	  gtk_label_set_label (GTK_LABEL (typescript_label), error->message);
	  g_error_free (error);
	  g_free (_last_typescript);
	  _last_typescript = NULL;

	  gtk_widget_show (GTK_WIDGET (typescript_label));
	  gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);
	  return;
	}
      g_object_unref (stream);
    }
  else
    {
      gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);
      return;
    }
  /* validate timestamp file. */
  chooser = gtk_builder_get_object (builder, "TimestampChooserButton");
  _last_timestamp = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
  if (_last_timestamp)
    {
      gchar *line;
      gdouble interval;
      gsize size;
      gchar *errmsg, *basename;

      basename = g_path_get_basename (_last_timestamp);
      stream = convert_open_timestamp (_last_timestamp, &error);
      if (error)
	{
	  gtk_label_set_label (GTK_LABEL (timestamp_label), error->message);
	  g_error_free (error);

	  gtk_widget_show (GTK_WIDGET (timestamp_label));
	  gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);

	  g_free (_last_timestamp);
	  _last_timestamp = NULL;
	  return;
	}
      line = g_data_input_stream_read_line (stream, NULL, NULL, &error);
      if (error)
	{
	  errmsg = g_strdup_printf (_("Unable read from `%s': %s"),
				    basename, error->message);
	  g_free (basename);
	  g_error_free (error);

	  gtk_label_set_label (GTK_LABEL (timestamp_label), errmsg);
	  g_free (errmsg);

	  gtk_widget_show (GTK_WIDGET (timestamp_label));
	  gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);

	  g_object_unref (stream);
	  g_free (_last_timestamp);
	  _last_timestamp = NULL;
	  return;
	}
      if (sscanf (line, "%lf %u", &interval, &size) != 2)
	{
	  errmsg = g_strdup_printf (_("Format of `%s' unrecognizable"),
				    basename);
	  g_free (basename);

	  gtk_label_set_label (GTK_LABEL (timestamp_label), errmsg);
	  g_free (errmsg);

	  gtk_widget_show (GTK_WIDGET (timestamp_label));
	  gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);

	  g_object_unref (stream);
	  g_free (_last_timestamp);
	  _last_timestamp = NULL;
	  return;
	}
      g_free (basename);
      g_object_unref (stream);
    }
  else
    {
      gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);
      return;
    }
  gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), TRUE);
}

/**
 * on_typescript_file_chooser_selection_changed:
 */
void
on_typescript_chooser_selection_changed (GtkFileChooser *chooser,
					 GtkAssistant   *assistant)
{
  g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));
  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  convert_set_input_file_page_complete (assistant);
}

/**
 * on_timestamp_chooser_selection_changed:
 */
void
on_timestamp_chooser_selection_changed (GtkFileChooser *chooser,
					GtkAssistant   *assistant)
{
  g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));
  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  convert_set_input_file_page_complete (assistant);
}

/**
 * convert_set_password_page_complete:
 */
static void
convert_set_password_page_complete (GtkAssistant *assistant)
{
  GtkBuilder *builder;
  GObject *page;
  GObject *entry, *entry2;

  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  builder = g_object_get_data (G_OBJECT (assistant), "builder");
  page = gtk_builder_get_object (builder, "ConvertPasswordPage");
  entry  = gtk_builder_get_object (builder, "ConvertPasswordEntry");
  entry2 = gtk_builder_get_object (builder, "ConvertPasswordEntry2");

  if (g_strcmp0 (gtk_entry_get_text (GTK_ENTRY (entry)),
		 gtk_entry_get_text (GTK_ENTRY (entry2))) == 0)
    gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), TRUE);
  else
    gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);
}

/**
 * on_script_password_entry_changed:
 */
void
on_script_password_entry_changed (GtkEntry     *entry,
				  GtkAssistant *assistant)
{
  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  convert_set_password_page_complete (assistant);
}

/**
 * on_script_password_entry2_changed:
 */
void
on_script_password_entry2_changed (GtkEntry     *entry,
				   GtkAssistant *assistant)
{
  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  convert_set_password_page_complete (assistant);
}

/**
 * collect_convert_params:
 */
static void
collect_convert_params (GtkAssistant  *assistant,
			GtkBuilder    *builder,
			ConvertParams *params)
{
  GObject *entry, *chooser, *adjust, *combo;
  GtkTreeModel *model;
  GtkTreeIter iter;
  const gchar *text;

  g_return_if_fail (GTK_IS_ASSISTANT (assistant));
  g_return_if_fail (params);

  /* input typescript file. */
  chooser = gtk_builder_get_object (builder, "TypescriptChooserButton");
  g_free (params->typescript);
  params->typescript = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));

  /* input time stamp file. */
  chooser = gtk_builder_get_object (builder, "TimestampChooserButton");
  g_free (params->timestamp);
  params->timestamp = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));

  /* password. */
  g_free (params->password);
  params->password = NULL;
  entry = gtk_builder_get_object (builder, "ConvertPasswordEntry");
  if (gtk_entry_get_text_length (GTK_ENTRY (entry)) > 0)
    params->password = g_strdup (gtk_entry_get_text (GTK_ENTRY (entry)));

  /* output filename */
  entry = gtk_builder_get_object (builder, "OutputFilenameEntry");
  text = gtk_entry_get_text (GTK_ENTRY (entry));
  g_free (params->filename);
  if (g_path_is_absolute (text))
    params->filename = g_strdup (text);
  else
    params->filename = g_build_filename (g_get_current_dir (), text, NULL);

  /* compress level */
  adjust = gtk_builder_get_object (builder, "OutputCompressAdjustment");
  params->comp_level = gtk_adjustment_get_value (GTK_ADJUSTMENT (adjust));

  /* charset. */
  combo = gtk_builder_get_object (builder, "ParamsCharsetCombo");
  gtk_combo_box_get_active_iter (GTK_COMBO_BOX (combo), &iter);
  model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
  g_free (params->charset);
  gtk_tree_model_get (model, &iter, 1, &params->charset, -1);

  /* terminal */
  combo = gtk_builder_get_object (builder, "ParamsTerminalCombo");
  gtk_combo_box_get_active_iter (GTK_COMBO_BOX (combo), &iter);
  model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
  g_free (params->term);
  gtk_tree_model_get (model, &iter, 0, &params->term, -1);

  /* terminal columns */
  adjust = gtk_builder_get_object (builder, "TermColumnsAdjustment");
  params->termcols = gtk_adjustment_get_value (GTK_ADJUSTMENT (adjust));

  /* terminal rows */
  adjust = gtk_builder_get_object (builder, "TermRowsAdjustment");
  params->termrows = gtk_adjustment_get_value (GTK_ADJUSTMENT (adjust));
}

/**
 * convert_show_confirm_info:
 */
static void
convert_show_confirm_info (GtkAssistant  *assistant,
			   GtkBuilder    *builder,
			   ConvertParams *params)
{
  GObject *label;
  gchar buff[128];

  label = gtk_builder_get_object (builder, "TypescriptConfirmLabel");
  gtk_label_set_label (GTK_LABEL (label), params->typescript);

  label = gtk_builder_get_object (builder, "TimestampConfirmLabel");
  gtk_label_set_label (GTK_LABEL (label), params->timestamp);

  label = gtk_builder_get_object (builder, "CharsetConfirmLabel");
  gtk_label_set_label (GTK_LABEL (label), params->charset);

  label = gtk_builder_get_object (builder, "TerminalConfirmLabel");
  gtk_label_set_label (GTK_LABEL (label), params->term);

  label = gtk_builder_get_object (builder, "TerminalSizeConfirmLabel");
  g_snprintf (buff, sizeof (buff), "%d x %d",
	      params->termcols, params->termrows);
  gtk_label_set_label (GTK_LABEL (label), buff);

  label = gtk_builder_get_object (builder, "EncryptionConfirmLabel");
  gtk_label_set_label (GTK_LABEL (label), params->password ? "Yes" : "No");

  label = gtk_builder_get_object (builder, "FilenameConfirmLabel");
  gtk_label_set_label (GTK_LABEL (label), params->filename);

  label = gtk_builder_get_object (builder, "CompressConfirmLabel");
  g_snprintf (buff, sizeof (buff), "%d", params->comp_level);
  gtk_label_set_label (GTK_LABEL (label), buff);
}

/**
 * on_convert_script_prepare:
 */
void
on_convert_script_prepare (GtkAssistant *assistant,
			   GtkWidget    *page,
			   gpointer      user_data)
{
  GtkBuilder *builder;
  ConvertParams *params;

  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  builder = g_object_get_data (G_OBJECT (assistant), "builder");
  params = g_object_get_data (G_OBJECT (assistant), "params");
  g_assert (params);

  switch (gtk_assistant_get_page_type (assistant, page))
    {
    case GTK_ASSISTANT_PAGE_CONFIRM:
      collect_convert_params (assistant, builder, params);
      convert_show_confirm_info (assistant, builder, params);
      break;

    case GTK_ASSISTANT_PAGE_SUMMARY:
      if (params->messages)
	{
	  GObject *buffer;

	  buffer = gtk_builder_get_object (builder, "ConvertFinishBuffer");
	  gtk_text_buffer_set_text (GTK_TEXT_BUFFER (buffer),
				    params->messages->str, params->messages->len);
	}
      break;
    default:
      break;
    }
}


/**
 * convert_convert:
 *
 * read from typescript and inject to record file.
 */
static gboolean
convert_convert_line (GDataInputStream *time_stream,
		      GDataInputStream *type_stream,
		      RecordEncoder    *encoder,
		      gdouble          *interval,
		      GError          **error)
{
  gchar *line;
  gsize nread, size;
  gchar *buffer;

  g_return_val_if_fail (G_IS_DATA_INPUT_STREAM (time_stream), FALSE);
  g_return_val_if_fail (G_IS_DATA_INPUT_STREAM (type_stream), FALSE);
  g_return_val_if_fail (RECORD_IS_ENCODER (encoder), FALSE);

  /* read timestamp and size of data.
   * each line has two field: interval and size. */
  line = g_data_input_stream_read_line (time_stream, &nread, NULL, error);
  if (*error)
    {
      g_prefix_error (error, _("Failed to read timestamp: "));
      return FALSE;
    }
  /* end of timestamp file, return FALSE without set error. */
  if (nread == 0)
    {
      g_free (line);
      return FALSE;
    }
  /* convert to binary format. */
  if (sscanf (line, "%lf %u", interval, &size) != 2)
    {
      g_free (line);
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Timestamp has invalid format."));
      return FALSE;
    }
  g_free (line);

  /* read data from stream. */
  buffer = g_malloc (size);
  nread = g_input_stream_read (G_INPUT_STREAM (type_stream),
			       buffer, size, NULL, error);
  if (*error)
    {
      g_free (buffer);
      g_prefix_error (error, _("Failed to read typescript: "));
      return FALSE;
    }
  /* end of data file, partial input error. */
  if (nread == 0)
    {
      g_free (buffer);
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_PARTIAL_INPUT,
		   _("typescript has not enough data."));
      return FALSE;
    }
  /* inject to record file. */
  record_encoder_inject_with_time (encoder, buffer, size, *interval);
  g_free (buffer);

  return TRUE;
}

/**
 * convert_script_file:
 *
 * convert script file into ttyutils record file.
 */
static gboolean
convert_script_file (ConvertParams *params)
{
  RecordEncoder *encoder;
  GDataInputStream *time_stream;
  GDataInputStream *type_stream;
  gboolean retval = TRUE;
  GError *error = NULL;

  g_return_val_if_fail (params, FALSE);

  /* open timestamp and typescript file. */
  time_stream = convert_open_timestamp (params->timestamp, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      return FALSE;
    }
  type_stream = convert_open_typescript (params->typescript,
					 &params->starting, &error);
  if (error)
    {
      g_warning ("%s", error->message);
      g_error_free (error);
      g_object_unref (time_stream);
      return FALSE;
    }
  /* create encoder object. */
  encoder = record_encoder_new (params->filename, params->comp_level,
				g_get_application_name ());
  if (params->password)
    record_encoder_set_passwd (encoder, params->password, -1);

  /* set parameters and start encoder. */
  if (!record_encoder_begin (encoder, params->charset, params->term,
			     params->termrows, params->termcols,
			     params->starting))
    {
      g_object_unref (type_stream);
      g_object_unref (time_stream);
      g_object_unref (encoder);
      return FALSE;
    }
  params->duration = 0.0;

  /* read from input file, and encode it. */
  while (1)
    {
      gdouble interval;
      GError *error = NULL;

      if (!convert_convert_line (time_stream, type_stream,
				 encoder, &interval, &error))
	{
	  if (error)
	    {
	      g_warning ("%s", error->message);
	      g_error_free (error);
	      retval = FALSE;
	    }
	  break;
	}
      params->duration += interval;
    }
  record_encoder_end (encoder, params->duration);

  g_object_unref (encoder);
  g_object_unref (type_stream);
  g_object_unref (time_stream);

  return retval;
}

/**
 * convert_log_handler:
 */
static void
convert_log_handler (const gchar    *domain,
		     GLogLevelFlags  level,
		     const gchar    *message,
		     gpointer        user_data)
{
  GString *messages = user_data;

  g_string_append_printf (messages, "%s\n", message);
}

/**
 * on_convert_script_apply:
 */
void
on_convert_script_apply (GtkAssistant *assistant,
			 gpointer      user_data)
{
  ConvertParams *params;

  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  params = g_object_get_data (G_OBJECT (assistant), "params");
  g_assert (params);

  /* buffer to save output messages. */
  if (params->messages)
    g_string_truncate (params->messages, 0);
  else
    params->messages = g_string_new (NULL);

  /* redirect log, so we can show the all of the output messages
   * from encoder to window. */
  g_log_set_default_handler (convert_log_handler, params->messages);

  /* check whether output filename exists. */
  if (g_file_test (params->filename, G_FILE_TEST_EXISTS))
    {
      if (g_unlink (params->filename) != 0)
	{
	  g_warning ("unlink `%s' error: %s", params->filename,
		     g_strerror (errno));
	  goto end;
	}
    }
  /* do convert. */
  if (convert_script_file (params))
    {
      g_string_append (params->messages, "----\n");
      g_string_append (params->messages, _("Succeed"));
    }
  else
    {
      g_string_append (params->messages, "----\n");
      g_string_append (params->messages, _("Failed"));
    }

end:
  /* reset default log handler. */
  g_log_set_default_handler (g_log_default_handler, NULL);
}

/**
 * on_output_filename_entry_changed:
 */
void
on_output_filename_entry_changed (GtkEntry     *entry,
				  GtkAssistant *assistant)
{
  GtkBuilder *builder;
  GObject *page;

  g_return_if_fail (GTK_IS_ENTRY (entry));
  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  builder = g_object_get_data (G_OBJECT (assistant), "builder");
  page = gtk_builder_get_object (builder, "ConvertOutputPage");

  if (gtk_entry_get_text_length (entry) > 0)
    gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), TRUE);
  else
    gtk_assistant_set_page_complete (assistant, GTK_WIDGET (page), FALSE);
}

/**
 * on_output_filename_button_clicked:
 */
void
on_output_filename_button_clicked (GtkButton    *button,
				   GtkAssistant *assistant)
{
  GtkWidget *dialog;
  GtkFileChooser *chooser;

  g_return_if_fail (GTK_IS_BUTTON (button));
  g_return_if_fail (GTK_IS_ASSISTANT (assistant));

  dialog = gtk_file_chooser_dialog_new ("Select output file",
					GTK_WINDOW (assistant),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					NULL);
  chooser = GTK_FILE_CHOOSER (dialog);

  gtk_file_chooser_set_do_overwrite_confirmation (chooser, TRUE);
  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
      GtkBuilder *builder;
      GObject *entry;
      gchar *filename;

      builder = g_object_get_data (G_OBJECT (assistant), "builder");
      entry = gtk_builder_get_object (builder, "OutputFilenameEntry");

      filename = gtk_file_chooser_get_filename (chooser);
      gtk_entry_set_text (GTK_ENTRY (entry), filename);
      g_free (filename);
    }
  gtk_widget_destroy (dialog);
}

/**
 * assay_convert_assistant_run:
 */
void
assay_convert_assistant_run (GtkWidget *window)
{
  static GObject *assistant = NULL;
  GtkBuilder *builder;
  GObject *chooser;
  ConvertParams *params;
  GError *error = NULL;

  g_return_if_fail (GTK_IS_WINDOW (window));

  if (assistant)
    {
      gtk_window_present (GTK_WINDOW (assistant));
      return;
    }
  builder = gtk_builder_new ();
  gtk_builder_add_from_file (builder, CONVERT_GLADE_FILE, &error);
  if (error)
    {
      g_critical ("%s", error->message);
      g_error_free (error);
      return;
    }
  assistant = gtk_builder_get_object (builder, "ConvertAssistant");
  g_signal_connect (assistant, "destroy",
		    G_CALLBACK (gtk_widget_destroyed), &assistant);

  if (_last_typescript)
    {
      chooser = gtk_builder_get_object (builder, "TypescriptChooserButton");
      gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (chooser),
				     _last_typescript);
    }
  if (_last_timestamp)
    {
      chooser = gtk_builder_get_object (builder, "TimestampChooserButton");
      gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (chooser),
				     _last_timestamp);
    }
  params = g_new0 (ConvertParams, 1);
  params->comp_level = -1;

  g_object_set_data_full (G_OBJECT (assistant), "params", params,
			  (GDestroyNotify) convert_params_free);
  g_object_set_data_full (G_OBJECT (assistant), "builder", builder,
			  (GDestroyNotify) g_object_unref);
  gtk_builder_connect_signals (builder, assistant);

  gtk_widget_show_all (GTK_WIDGET (assistant));
}
