/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define _BSD_SOURCE

#include <sys/types.h>
#include <pwd.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_FILIO_H
# include <sys/filio.h>
#endif
#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif
#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif

#include <glib/gi18n-lib.h>

#include "tmacros.h"
#include "tmarshal.h"
#include "tmisc.h"
#include "ttermio.h"
#include "ttermcap.h"
#include "tring.h"
#include "tconsole.h"

enum
{
  PROP_0,
  PROP_CACHE_SIZE,
};

struct _TConsolePrivate
{
  GIOChannel   *in_channel;
  GIOChannel   *out_channel;
  gint          in_id;
  gint          out_id;
  TRing        *out_ring;
  TRing        *cache;
  gsize         cache_size;
  gboolean      out_suspend;
  GMutex       *out_mutex;
};
#define _CONSOLE_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), T_TYPE_CONSOLE, TConsolePrivate))

G_DEFINE_TYPE (TConsole, t_console, G_TYPE_OBJECT);

static gboolean		t_console_read_channel		(TConsole     *console,
							 GIOChannel   *channel);
static gboolean		t_console_stdin_watch_incoming	(GIOChannel   *channel,
							 GIOCondition  cond,
							 gpointer      data);
static void		t_console_watch_stdin_channel	(TConsole     *console);

static gboolean		t_console_write_channel		(TConsole     *console,
							 GIOChannel   *channel);
static gboolean		t_console_stdout_watch_outgoing	(GIOChannel   *channel,
							 GIOCondition  cond,
							 gpointer      data);
static void		t_console_watch_stdout_channel	(TConsole     *console);

/** 
 * t_console_init:
 */
static void
t_console_init (TConsole *console)
{
  console->priv = _CONSOLE_GET_PRIVATE (console);

  console->priv->in_channel  = g_io_channel_unix_new (STDIN_FILENO);
  console->priv->out_channel = g_io_channel_unix_new (STDOUT_FILENO);
  console->priv->in_id  = -1;
  console->priv->out_id = -1;
  console->priv->out_ring = t_ring_new ();
  console->priv->cache = NULL;
  console->priv->cache_size = 0;
  console->priv->out_suspend = FALSE;
  console->priv->out_mutex = g_mutex_new ();
}

/** 
 * t_console_finalize:
 */
static void
t_console_finalize (GObject *object)
{
  TConsole *console = T_CONSOLE (object);

  if (console->priv->out_ring)
    {
      t_console_flush (console);
      t_ring_free (console->priv->out_ring);
      console->priv->out_ring = NULL;
    }
  if (console->priv->cache)
    {
      t_ring_free (console->priv->cache);
      console->priv->cache = NULL;
    }
  if (console->priv->out_id > 0)
    {
      g_source_remove (console->priv->out_id);
      console->priv->out_id = -1;
    }
  if (console->priv->in_id > 0)
    {
      g_source_remove (console->priv->in_id);
      console->priv->in_id = -1;
    }
  if (console->priv->in_channel)
    {
      g_io_channel_unref (console->priv->in_channel);
      console->priv->in_channel = NULL;
    }
  if (console->priv->out_channel)
    {
      g_io_channel_unref (console->priv->out_channel);
      console->priv->out_channel = NULL;
    }
  if (console->priv->out_mutex)
    {
      g_mutex_free (console->priv->out_mutex);
      console->priv->out_mutex = NULL;
    }
  G_OBJECT_CLASS (t_console_parent_class)->finalize (object);
}

/**
 * t_console_get_property:
 */
static void
t_console_get_property (GObject    *object,
			guint       property_id,
			GValue     *value,
			GParamSpec *pspec)
{
  TConsole *console = T_CONSOLE (object);

  switch (property_id)
    {
    case PROP_CACHE_SIZE:
      g_value_set_uint (value, t_console_get_cache_size (console));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * t_console_set_property:
 */
static void
t_console_set_property (GObject      *object,
			guint         property_id,
			const GValue *value,
			GParamSpec   *pspec)
{
  TConsole *console = T_CONSOLE (object);

  switch (property_id)
    {
    case PROP_CACHE_SIZE:
      t_console_set_cache_size (console, g_value_get_uint (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * t_console_readable_real:
 */
static void
t_console_readable_real (TConsole    *console,
			 const gchar *string,
			 guint        length)
{
}

/**
 * t_console_class_init:
 */
static void
t_console_class_init (TConsoleClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (TConsolePrivate));

  object_class->get_property = t_console_get_property;
  object_class->set_property = t_console_set_property;
  object_class->finalize     = t_console_finalize;

  klass->readable = t_console_readable_real;

  g_object_class_install_property (object_class, PROP_CACHE_SIZE,
				   g_param_spec_uint ("cache-size", "cache-size", "stdout cache size",
						      0, G_MAXUINT32, 0,
						      G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
						      G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_signal_new ("readable",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_FIRST,
		G_STRUCT_OFFSET (TConsoleClass, readable),
		NULL, NULL,
		t_marshal_VOID__STRING_UINT,
		G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_UINT);
  g_signal_new ("hangup",
		G_OBJECT_CLASS_TYPE (klass),
		G_SIGNAL_RUN_FIRST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
}

/** 
 * t_console_read_channel:
 *
 * read data from stdin
 */
gboolean
t_console_read_channel (TConsole   *console,
			GIOChannel *channel)
{
  gchar buff[4096];
  gsize bytes_read = 0;
  GIOStatus status;
  GError *error = NULL;

  status = g_io_channel_read_chars (channel, buff, sizeof (buff),
				    &bytes_read, &error);
  switch (status)
    {
    case G_IO_STATUS_ERROR:
      if (error)
	{
	  g_warning (_("Failed to read stdin: %s"), error->message);
	  g_error_free (error);
	}
      return FALSE;

    case G_IO_STATUS_EOF:
      g_message (_("Failed to read stdin: end of file(errno=%d(%s))"),
		 errno, g_strerror (errno));
      return FALSE;

    case G_IO_STATUS_NORMAL:
    case G_IO_STATUS_AGAIN:
      break;
    default:
      g_assert_not_reached ();
    }
  if (bytes_read > 0)
    g_signal_emit_by_name (console, "readable", buff, bytes_read, 0);

  return TRUE;
}

/** 
 * t_console_stdin_watch_incoming:
 *
 * this function will be called when stdin is readable.
 */
gboolean
t_console_stdin_watch_incoming (GIOChannel   *channel,
				GIOCondition  cond,
				gpointer      data)
{
  TConsole *console = T_CONSOLE (data);

  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  if (cond & G_IO_HUP || cond & G_IO_ERR)
    {
      g_critical ("stdin %s", cond & G_IO_ERR ? "error" : "hangup");
      g_signal_emit_by_name (console, "hangup", 0);
      return FALSE;
    }
  if (!t_console_read_channel (console, channel))
    {
      g_signal_emit_by_name (console, "hangup", 0);
      return FALSE;
    }
  return TRUE;
}

/** 
 * t_console_stdin_watch_removed:
 *
 * this function will be called after io channel's
 * readable event processor been removed.
 */
void
t_console_stdin_watch_removed (gpointer data)
{
  TConsole *console = T_CONSOLE (data);

  console->priv->in_id = -1;
}

/** 
 * t_console_watch_stdin_channel:
 *
 * watch io channel's readable event
 */
void
t_console_watch_stdin_channel (TConsole *console)
{
  g_return_if_fail (T_IS_CONSOLE (console));

  if (console->priv->in_id < 0)
    {
      console->priv->in_id =
       g_io_add_watch_full (console->priv->in_channel,
			    G_PRIORITY_DEFAULT,
			    G_IO_IN | G_IO_ERR | G_IO_HUP,
			    t_console_stdin_watch_incoming, console,
			    t_console_stdin_watch_removed);
    }
}

/** 
 * t_console_write_channel:
 *
 * we only write a few of bytes at once, to prevent take a lots of
 * system resources(mostly is CPU).
 */
static gboolean
t_console_write_channel (TConsole   *console,
			 GIOChannel *channel)
{
  gint size;
  gchar buff[512];

  /* read data from ring buffer, all data that write to terminal
   * is write to the ring buffer first. */
  size = t_ring_read (console->priv->out_ring, buff, sizeof (buff));
  while (size > 0)
    {
      GIOStatus status;
      gchar *pdata = buff;
      gsize written = 0;
      GError *error = NULL;

      status = g_io_channel_write_chars (channel, pdata, size,
					 &written, &error);
      size  -= written;
      pdata += written;

      switch (status)
	{
	case G_IO_STATUS_NORMAL:
	  break;

	case G_IO_STATUS_ERROR:
	  if (error)
	    {
	      g_warning (_("Failed to write stdout: %s"), error->message);
	      g_error_free (error);
	    }
	  return FALSE;

	case G_IO_STATUS_AGAIN:		/* device busy */
	  if (size > 0)
	    {
	      if (isatty (STDOUT_FILENO))
		{
		  /* waits until all output written to stdout has been transmitted. */
		  if (tcdrain (g_io_channel_unix_get_fd (channel)) != 0)
		    {
		      g_warning ("tcdrain error: %s", g_strerror (errno));
		      return FALSE;
		    }
		}
	      else	/* is not a tty? */
		{
		  g_message ("stdout device too busy, sleep 1/10000th of seconds");

		  if (fflush (stdout) != 0)
		    {
		      g_warning ("fflush error: %s", g_strerror (errno));
		      return FALSE;
		    }
		  g_usleep (100);
		}
	      continue;
	    }
	  break;

	default:
	  g_assert_not_reached ();
	}
      g_io_channel_flush (channel, NULL);
    }
  return TRUE;
}

/** 
 * t_console_stdout_watch_outgoing:
 *
 * this function will be called when data is ready to
 * write to terminal, and io channel is free.
 */
gboolean
t_console_stdout_watch_outgoing (GIOChannel  *channel,
				 GIOCondition cond,
				 gpointer     data)
{
  TConsole *console = T_CONSOLE (data);
  gint size;

  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  g_mutex_lock (console->priv->out_mutex);

  if (!t_console_write_channel (console, channel))
    {
      g_mutex_unlock (console->priv->out_mutex);
      return FALSE;
    }
  size = t_ring_get_size (console->priv->out_ring);

  g_mutex_unlock (console->priv->out_mutex);

  return (size > 0);
}

/** 
 * t_console_stdout_watch_removed:
 *
 * this function will be called after all data been written.
 */
void
t_console_stdout_watch_removed (gpointer data)
{
  TConsole *console = T_CONSOLE (data);

  /* set out_id to -1, so next write event can be watch,
   * see t_console_watch_stdout_channel() */
  console->priv->out_id = -1;
}

/** 
 * t_console_watch_stdout_channel:
 */
void
t_console_watch_stdout_channel (TConsole *console)
{
  /* if stdout is suspend, skip */
  if (console->priv->out_suspend)
    return;

  /* watch io channel's writable event, to write data
   * to terminal */
  if (console->priv->out_id <= 0)
    {
      console->priv->out_id =
       g_io_add_watch_full (console->priv->out_channel,
			    G_PRIORITY_DEFAULT, G_IO_OUT,
			    t_console_stdout_watch_outgoing, console,
			    t_console_stdout_watch_removed);
    }
}

/** 
 * t_console_new:
 *
 * create a new TConsole object
 */
TConsole*
t_console_new (void)
{
  return g_object_new (T_TYPE_CONSOLE, NULL);
}

/**
 * t_console_start:
 *
 * setup io channel, and start read from stdin
 */
gboolean
t_console_start (TConsole *console)
{
  GError *error = NULL;

  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  g_io_channel_set_flags (console->priv->in_channel,
			  G_IO_FLAG_NONBLOCK, &error);
  if (error)
    {
      g_critical (_("Failed to set stdin's nonblock flags: %s"),
		  error->message);
      g_error_free (error);
      return FALSE;
    }
  g_io_channel_set_encoding (console->priv->in_channel, NULL, &error);
  if (error)
    {
      g_critical (_("Failed to set stdin's encoding: %s"),
		  error->message);
      g_error_free (error);
      return FALSE;
    }
  g_io_channel_set_buffered (console->priv->in_channel, FALSE);

  g_io_channel_set_flags (console->priv->out_channel,
			  G_IO_FLAG_NONBLOCK, &error);
  if (error)
    {
      g_critical (_("Failed to set stdout's nonblock flags: %s"),
		  error->message);
      g_error_free (error);
      return FALSE;
    }
  g_io_channel_set_encoding (console->priv->out_channel, NULL, &error);
  if (error)
    {
      g_critical (_("Failed to set stdout's encoding: %s"),
		  error->message);
      g_error_free (error);
      return FALSE;
    }
  g_io_channel_set_buffered (console->priv->out_channel, FALSE);

  t_console_watch_stdin_channel (console);
  return TRUE;
}

/**
 * t_console_get_cache_size:
 *
 * cache is use for refresh screen.
 *
 * Returns: the cache size
 */
gsize
t_console_get_cache_size (TConsole *console)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), 0);
  return console->priv->cache_size;
}

/**
 * t_console_set_cache_size:
 *
 * when write data to terminal, we can keep a copied
 * of these data, later, we can use the copied to
 * refresh screen.
 *
 * this function set the cache size, set to 0 to disable
 * cache at all.
 */
void
t_console_set_cache_size (TConsole *console,
			  gsize     size)
{
  g_return_if_fail (T_IS_CONSOLE (console));

  g_mutex_lock (console->priv->out_mutex);

  if (size == 0)
    {
      t_ring_free (console->priv->cache);
      console->priv->cache = NULL;
    }
  else if (console->priv->cache == NULL)
    {
      console->priv->cache = t_ring_new ();
    }
  console->priv->cache_size = size;

  g_mutex_unlock (console->priv->out_mutex);

  g_object_notify (G_OBJECT (console), "cache-size");
}

/**
 * watch_stdout_channel_idle:
 *
 * write buffer to device
 */
static gboolean
watch_stdout_channel_idle (gpointer data)
{
  TConsole *console = data;

  if (t_ring_get_size (console->priv->out_ring) > 0)
    t_console_watch_stdout_channel (console);

  return FALSE;
}

/** 
 * t_console_write:
 *
 * write data to terminal
 */
gint
t_console_write (TConsole    *console,
		 const gchar *buff,
		 gssize       bufflen)
{
  gint nwritten;

  g_return_val_if_fail (T_IS_CONSOLE (console), -1);
  g_return_val_if_fail (buff, -1);

  if (bufflen < 0)
    bufflen = strlen (buff);

  g_mutex_lock (console->priv->out_mutex);

  /* write to a ring buffer first(later will write to device) */
  if (!t_ring_can_write (console->priv->out_ring, bufflen))
    {
      g_mutex_unlock (console->priv->out_mutex);
      t_console_flush (console);
      g_mutex_lock (console->priv->out_mutex);
    }
  nwritten = t_ring_write (console->priv->out_ring, buff, bufflen);

  /* if cache is enabled, save data to cache(for later screen refresh) */
  if (console->priv->cache_size > 0 && console->priv->cache)
    {
      gint size = t_ring_get_size (console->priv->cache);
      if (size + bufflen > console->priv->cache_size)
	{
	  gint discard = size + bufflen - console->priv->cache_size;

	  /* discard to next ESC key(maybe a Escape sequence) */
	  t_ring_discard_to (console->priv->cache,
			     discard, discard + 100, "\033", 1);
	}
      t_ring_write (console->priv->cache, buff, bufflen);
    }
  g_idle_add (watch_stdout_channel_idle, console);

  g_mutex_unlock (console->priv->out_mutex);

  return nwritten;
}

/** 
 * t_console_flush:
 *
 * write all internal buffer to terminal
 */
void
t_console_flush (TConsole *console)
{
  g_return_if_fail (T_IS_CONSOLE (console));

  if (console->priv->out_suspend)
    return;

  g_mutex_lock (console->priv->out_mutex);

  while (t_ring_get_size (console->priv->out_ring) > 0)
    {
      if (!t_console_write_channel (console, console->priv->out_channel))
	{
	  g_mutex_unlock (console->priv->out_mutex);
	  return;
	}
      if (isatty (STDOUT_FILENO))
	{
	  if (tcdrain (g_io_channel_unix_get_fd (console->priv->out_channel)) != 0)
	    {
	      _warning ("tcdrain error: %s", g_strerror (errno));
	      g_mutex_unlock (console->priv->out_mutex);
	      return;
	    }
	}
      else
	{
	  fflush (stdout);
	  fsync (STDOUT_FILENO);
	}
    }
  g_mutex_unlock (console->priv->out_mutex);
}

/** 
 * t_console_get_buffer_size:
 *
 * data write to stdout is not directly, it is write to internal
 * ring buffer first.
 *
 * this function return the internal ring buffer size.
 */
gint
t_console_get_buffer_size (TConsole *console)
{
  gint size;

  g_return_val_if_fail (T_IS_CONSOLE (console), 0);

  g_mutex_lock (console->priv->out_mutex);

  size = t_ring_get_size (console->priv->out_ring);

  g_mutex_unlock (console->priv->out_mutex);

  return size;
}

/** 
 * t_console_clear_buffer:
 *
 * clear internal ring buffer(discard all data that write to stdout),
 * see t_console_get_buffer_size().
 */
void
t_console_clear_buffer (TConsole *console)
{
  g_return_if_fail (T_IS_CONSOLE (console));

  g_mutex_lock (console->priv->out_mutex);

  t_ring_clear (console->priv->out_ring);

  g_mutex_unlock (console->priv->out_mutex);
}

/**
 * t_console_clear_cache:
 *
 * TConsole keep a internal cache for screen refresh.
 * the cache has fixed size, see t_console_set_cache_size().
 *
 * note the cache is differ with output buffer above.
 */
void
t_console_clear_cache (TConsole *console)
{
  g_return_if_fail (T_IS_CONSOLE (console));

  g_mutex_lock (console->priv->out_mutex);

  if (console->priv->cache)
    t_ring_clear (console->priv->cache);

  g_mutex_unlock (console->priv->out_mutex);
}

/** 
 * t_console_refresh:
 *
 * refresh screen
 */
void
t_console_refresh (TConsole *console)
{
  gint size;
  gint retry = 0;

  g_return_if_fail (T_IS_CONSOLE (console));

  /* cannot refresh if console is not a tty device */
  if (!isatty (STDOUT_FILENO))
    return;

  /* cannot refresh when stdout is suspend. */
  if (console->priv->out_suspend)
    return;

  /* cannot refresh if cache was disabled. */
  if (!console->priv->cache || console->priv->cache <= 0)
    return;

  g_mutex_lock (console->priv->out_mutex);

  size = t_ring_get_size (console->priv->cache);
  if (size <= 0)
    {
      g_mutex_unlock (console->priv->out_mutex);
      return;
    }
  /* clear pending output and clear screen */
  t_console_clear_buffer (console);
  t_term_clear_screen ();

  while (size > 0)
    {
      gchar buff[512], *p;
      gint nread, toread;
      gsize nwritten;
      GIOStatus status;
      GError *error = NULL;

      toread = MIN (size, sizeof (buff));

      nread = t_ring_read (console->priv->cache, buff, toread);
      size -= nread;

      if (nread < 0)
	break;

      /* append to end of cache for next refresh */
      t_ring_write (console->priv->cache, buff, nread);

      p = buff;

      /* ensure data been actually written */
      while (nread > 0)
	{
	  status = g_io_channel_write_chars (console->priv->out_channel,
					     p, nread, &nwritten, &error);
	  if (status == G_IO_STATUS_ERROR)
	    {
	      /* error occured */
	      if (error)
		{
		  g_warning (_("Failed to write stdout: %s"), error->message);
		  g_error_free (error);
		  error = NULL;
		}
	      if (retry++ < 3)
		{
		  tcdrain (g_io_channel_unix_get_fd (console->priv->out_channel));
		  g_usleep (1000);
		}
	      else
		{
		  g_mutex_unlock (console->priv->out_mutex);
		  return;
		}
	    }
	  else if (status == G_IO_STATUS_AGAIN)
	    {
	      /* wait transfer finish and try again */
	      tcdrain (g_io_channel_unix_get_fd (console->priv->out_channel));
	      g_usleep (1000);
	    }
	  else
	    {
	      g_io_channel_flush (console->priv->out_channel, NULL);
	      /* wait for a while to prevent take lots of system resources */
	      g_usleep (1);
	    }
	  nread -= nwritten;
	  p += nwritten;
	}
    }
  /* wait for transfer */
  tcdrain (g_io_channel_unix_get_fd (console->priv->out_channel));

  g_mutex_unlock (console->priv->out_mutex);
}

/** 
 * t_console_suspend_stdin:
 *
 * *lock* the keyboard
 */
static gboolean
t_console_suspend_stdin (TConsole *console)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

#if 0
  t_console_set_attr (STDIN_FILENO, IXOFF, TRUE);

  /* transmits a STOP character, which stops the terminal device from
   * transmitting data to the system */
  if (tcflow (g_io_channel_unix_get_fd (console->priv->in_channel),
	      TCIOFF) != 0)
    {
      g_warning ("tcflow error: %s", g_strerror (errno));
      return FALSE;
    }
#endif
  /* remove stdin watch */
  if (console->priv->in_id > 0)
    {
      g_source_remove (console->priv->in_id);
    }
  return TRUE;
}

/** 
 * t_console_resume_stdin:
 *
 * *unlock* the keyboard
 */
static gboolean
t_console_resume_stdin (TConsole *console,
			gboolean  clear_pending)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  if (clear_pending)
    {
      if (tcflush (g_io_channel_unix_get_fd (console->priv->in_channel),
		   TCIFLUSH) != 0)
	{
	  g_warning ("tcflush : %s", g_strerror (errno));
	  return FALSE;
	}
    }
  t_console_watch_stdin_channel (console);

  /* transmits a START character, which starts the terminal device
   * transmitting data to the system. */
#if 0
  if (tcflow (g_io_channel_unix_get_fd (console->priv->in_channel),
	      TCION) != 0)
    {
      g_warning ("tcflow error: %s", g_strerror (errno));
      return FALSE;
    }
  t_console_set_attr (STDIN_FILENO, IXOFF, FALSE);
#endif

  return TRUE;
}

/** 
 * t_console_suspend_stdout:
 */
static gboolean
t_console_suspend_stdout (TConsole *console,
			  gboolean  suspend_transmis)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  t_termio_set_iflag (STDOUT_FILENO, IXON, TRUE);

  if (suspend_transmis)
    {
      if (tcflow (g_io_channel_unix_get_fd (console->priv->out_channel),
		  TCOOFF) != 0)
	{
	  g_warning ("tcflow error: %s", g_strerror (errno));
	  return FALSE;
	}
    }
  if (console->priv->out_id > 0)
    {
      g_source_remove (console->priv->out_id);
    }
  console->priv->out_suspend = TRUE;
  return TRUE;
}

/** 
 * t_console_resume_stdout:
 */
static gboolean
t_console_resume_stdout (TConsole *console,
			 gboolean  clear_pending)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  console->priv->out_suspend = FALSE;

  /* cleanup buffer if need */
  if (clear_pending)
    {
      g_mutex_lock (console->priv->out_mutex);

      t_ring_clear (console->priv->out_ring);

      g_mutex_unlock (console->priv->out_mutex);

      if (tcflush (g_io_channel_unix_get_fd (console->priv->out_channel),
		   TCOFLUSH) != 0)
	{
	  g_warning ("tcflush error: %s", g_strerror (errno));
	  return FALSE;
	}
    }
  /* restart transmission of output.  */
  if (tcflow (g_io_channel_unix_get_fd (console->priv->out_channel),
	      TCOON) != 0)
    {
      g_warning ("tcflow error: %s", g_strerror (errno));
      return FALSE;
    }
  t_termio_set_iflag (STDOUT_FILENO, IXON, FALSE);

  g_mutex_lock (console->priv->out_mutex);

  if (t_ring_get_size (console->priv->out_ring) > 0)
    {
      t_console_watch_stdout_channel (console);
    }
  g_mutex_unlock (console->priv->out_mutex);

  return TRUE;
}

/**
 * console_suspend_stdin_idle:
 */
static gboolean
console_suspend_stdin_idle (gpointer data)
{
  t_console_suspend_stdin (T_CONSOLE (data));
  return FALSE;
}

/**
 * console_suspend_stdout_idle:
 */
static gboolean
console_suspend_stdout_idle (gpointer data)
{
  gboolean suspend_transmis =
   GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (data),
					"suspend-transmis"));

  t_console_suspend_stdout (T_CONSOLE (data),
			    suspend_transmis);
  return FALSE;
}

/**
 * console_resume_stdin_idle:
 */
static gboolean
console_resume_stdin_idle (gpointer data)
{
  gboolean clear_pending =
   GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (data),
					"clear-pending"));
  _debug ("clear pending = %d\n", clear_pending);

  t_console_resume_stdin (T_CONSOLE (data),
			  clear_pending);
  return FALSE;
}

/**
 * console_resume_stdout_idle:
 */
static gboolean
console_resume_stdout_idle (gpointer data)
{
  gboolean clear_pending =
   GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (data),
					"clear-pending"));

  t_console_resume_stdout (T_CONSOLE (data),
			   clear_pending);
  return FALSE;
}

/**
 * t_console_suspend:
 */
gboolean
t_console_suspend (TConsole       *console,
		   TConsoleSuspend suspend,
		   gboolean        suspend_transmis)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  if (suspend & CONSOLE_SUSPEND_STDIN)
    {
      g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		       console_suspend_stdin_idle,
		       console, NULL);
    }
  if (suspend & CONSOLE_SUSPEND_STDOUT)
    {
      g_object_set_data (G_OBJECT (console), "suspend-transmis",
			 GUINT_TO_POINTER (suspend_transmis));
      g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		       console_suspend_stdout_idle,
		       console, NULL);
    }
  return TRUE;
}

/**
 * t_console_resume:
 */
gboolean
t_console_resume (TConsole       *console,
		  TConsoleSuspend resume,
		  gboolean        clear_pending)
{
  g_return_val_if_fail (T_IS_CONSOLE (console), FALSE);

  g_object_set_data (G_OBJECT (console), "clear-pending",
		     GUINT_TO_POINTER (clear_pending));

  if (resume & CONSOLE_SUSPEND_STDIN)
    {
      g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		       console_resume_stdin_idle,
		       console, NULL);
    }
  if (resume & CONSOLE_SUSPEND_STDIN)
    {
      g_idle_add_full (G_PRIORITY_HIGH_IDLE,
		       console_resume_stdout_idle,
		       console, NULL);
    }
  return TRUE;
}
