/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <errno.h>
#include <zip.h>
#include <gcrypt.h>

#include <glib/gi18n-lib.h>
#include <gio/gio.h>

#include "tmacros.h"
#include "trecord.h"
#include "tdecoder.h"

enum
{
  PROP_0,
  PROP_FILENAME,
};

struct _TDecoderPrivate
{
  gchar        *filename;
  struct zip   *archive;
  TRecordHead   head;
  TRecordIndex *indexes;
  guint32       istart;
  guint32       iend;
  gchar        *data;
  guint64       dstart;
  guint64       dend;
  gchar        *passwd;
};
#define _DECODER_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), T_TYPE_DECODER, TDecoderPrivate))

G_DEFINE_TYPE (TDecoder, t_decoder, G_TYPE_OBJECT);


static void	t_decoder_set_filename		(TDecoder     *decoder,
						 const gchar  *filename);

/** 
 * t_decoder_init:
 */
static void
t_decoder_init (TDecoder *decoder)
{
  decoder->priv = _DECODER_GET_PRIVATE (decoder);

  decoder->priv->filename = NULL;
  decoder->priv->archive = NULL;
  decoder->priv->indexes = NULL;
  decoder->priv->istart = 0;
  decoder->priv->iend = 0;
  decoder->priv->data = NULL;
  decoder->priv->dstart = 0;
  decoder->priv->dend = 0;
  decoder->priv->passwd = NULL;
}

/** 
 * t_decoder_finalize:
 */
static void
t_decoder_finalize (GObject *object)
{
  TDecoder *decoder = T_DECODER (object);

  t_decoder_close (decoder);

  if (decoder->priv->filename)
    {
      g_free (decoder->priv->filename);
      decoder->priv->filename = NULL;
    }
  G_OBJECT_CLASS (t_decoder_parent_class)->finalize (object);
}

/**
 * t_decoder_get_property:
 */
static void
t_decoder_get_property (GObject    *object,
			guint       property_id,
			GValue     *value,
			GParamSpec *pspec)
{
  TDecoder *decoder = T_DECODER (object);

  switch (property_id)
    {
    case PROP_FILENAME:
      g_value_set_string (value, t_decoder_get_filename (decoder));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * t_decoder_set_property:
 */
static void
t_decoder_set_property (GObject      *object,
			guint         property_id,
			const GValue *value,
			GParamSpec   *pspec)
{
  TDecoder *decoder = T_DECODER (object);

  switch (property_id)
    {
    case PROP_FILENAME:
      t_decoder_set_filename (decoder, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * t_decoder_class_init:
 */
static void
t_decoder_class_init (TDecoderClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (TDecoderPrivate));

  object_class->get_property = t_decoder_get_property;
  object_class->set_property = t_decoder_set_property;
  object_class->finalize     = t_decoder_finalize;

  g_object_class_install_property (object_class, PROP_FILENAME,
				   g_param_spec_string ("filename", "filename", "File to decode", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));

  if (gcry_md_test_algo (GCRY_MD_MD5) != 0)
    {
      g_error ("digest algorithm `%s' is not usable.",
	       gcry_md_algo_name (GCRY_MD_MD5));
      return;
    }
  if (gcry_md_test_algo (GCRY_MD_TIGER) != 0)
    {
      g_error ("digest algorithm `%s' is not usable.",
	       gcry_md_algo_name (GCRY_MD_TIGER));
      return;
    }
  g_assert (24 == gcry_md_get_algo_dlen (GCRY_MD_TIGER));
  g_assert (16 == gcry_md_get_algo_dlen (GCRY_MD_MD5));
}

/** 
 * t_decoder_new:
 *
 * create a new TDecoder object
 */
TDecoder*
t_decoder_new (void)
{
  return g_object_new (T_TYPE_DECODER, NULL);
}

/**
 * t_decoder_set_passwd:
 *
 * this function should be called in open password
 * callback, see t_decoder_open() for detail.
 */
gboolean
t_decoder_set_passwd (TDecoder    *decoder,
		      const gchar *passwd,
		      gssize       length)
{
  gchar buff[24];

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (passwd, FALSE);
  g_return_val_if_fail (decoder->priv->archive, FALSE);

  if (length < 0)
    length = strlen (passwd);

  /* we cannot accept 0-length password. */
  if (length == 0)
    return FALSE;

  /* we need validate the password depend on record head.
   * so we cannot set password before read archive head. */
  gcry_md_hash_buffer (GCRY_MD_TIGER, buff, passwd, length);
  if (memcmp (buff, decoder->priv->head.passwd, 24) != 0)
    return FALSE;

  g_free (decoder->priv->passwd);
  decoder->priv->passwd = g_malloc (16);

  /* get decode password. */
  gcry_md_hash_buffer (GCRY_MD_MD5, decoder->priv->passwd,
		       passwd, length);
  return TRUE;
}

/**
 * t_decoder_read_record_head:
 *
 * read record head into memory.
 */
static gboolean
t_decoder_read_record_head (TDecoder    *decoder,
			    const gchar *filename,
			    GError     **error)
{
  struct zip_file *zfile;
  gint i, nread;
  struct zip_stat sb;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (decoder->priv->archive, FALSE);

  i = zip_name_locate (decoder->priv->archive, "HEAD", 0);
  if (i == -1)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
		   _("Format of `%s' invalid: %s"),
		   filename, _("Missing record head"));
      return FALSE;
    }
  zip_stat_index (decoder->priv->archive, i, 0, &sb);
  if (sizeof (TRecordHead) != (gsize) sb.size)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Format of `%s' invalid: %s"),
		   filename, _("Bad record head"));
      return FALSE;
    }
  /* open and read head file. */
  zfile = zip_fopen_index (decoder->priv->archive, i, 0);
  if (!zfile)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Failed to open record head: %s"),
		   zip_strerror (decoder->priv->archive));
      return FALSE;
    }
  nread = zip_fread (zfile, &decoder->priv->head, sb.size);
  if (nread != sizeof (TRecordHead))
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Failed to read record head: %s"),
		   g_strerror (errno));
      zip_fclose (zfile);
      return FALSE;
    }
  zip_fclose (zfile);

  /* check record version. */
  if (decoder->priv->head.version > T_RECORD_VERSION)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Version of `%s' dismatch."), filename);
      return FALSE;
    }
  /* convert head field to native byte order. */
  decoder->priv->head.nrows    = GUINT16_FROM_LE (decoder->priv->head.nrows);
  decoder->priv->head.ncols    = GUINT16_FROM_LE (decoder->priv->head.ncols);
  decoder->priv->head.nindexes = GUINT32_FROM_LE (decoder->priv->head.nindexes);
  decoder->priv->head.ndata    = GUINT64_FROM_LE (decoder->priv->head.ndata);
  decoder->priv->head.nfiles   = GUINT32_FROM_LE (decoder->priv->head.nfiles);

  return TRUE;
}

/**
 * t_decoder_get_archive_file:
 *
 * get archive index and data file by record index.
 */
static gint
t_decoder_get_archive_file (TDecoder    *decoder,
			    const gchar *filename,
			    guint32      index,
			    guint32     *istart,
			    guint32     *iend,
			    gint        *di,
			    guint64     *dstart,
			    guint64     *dend,
			    guint32     *dsize,
			    GError     **error)
{
  gint n, i;
  gdouble elapsed;

  g_return_val_if_fail (T_IS_DECODER (decoder), -1);
  g_return_val_if_fail (decoder->priv->archive, -1);
  g_return_val_if_fail (istart, -1);
  g_return_val_if_fail (iend, -1);
  g_return_val_if_fail (dstart, -1);
  g_return_val_if_fail (dend, -1);

  n = zip_get_num_files (decoder->priv->archive);
  for (i = 0; i < n; i++)
    {
      const gchar *p;
      gchar buff[64];
      gint len;

      /* skip non-indexes file. */
      p = zip_get_name (decoder->priv->archive, i, 0);
      if (!g_str_has_prefix (p, "indexes"))
	continue;

      /* get index offset and count from file comment. */
      p = zip_get_file_comment (decoder->priv->archive, i, &len, 0);
      if (!p)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Format of `%s' invalid: %s"),
		       filename, _("Missing file comment"));
	  return -1;
	}
      g_assert (len + 1 < sizeof (buff));
      g_strlcpy (buff, p, len + 1);
      if (sscanf (buff, "%u:%u,%lf", istart, iend, &elapsed) != 3)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Format of `%s' invalid: %s"),
		       filename, _("Invalid file comment"));
	  return -1;
	}
      /* index file found, get associated data file info. */
      if (index >= *istart && index < *iend)
	{
	  gchar *base;

	  /* get associated data file name. */
	  p = zip_get_name (decoder->priv->archive, i, 0);
	  base = g_path_get_basename (p);
	  g_snprintf (buff, sizeof (buff), "data/%s", base);
	  g_free (base);

	  /* get associated data file index. */
	  *di = zip_name_locate (decoder->priv->archive, buff, 0);
	  if (*di == -1)
	    {
	      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
			   _("Format of `%s' invalid: %s"),
			   filename, _("Missing required file"));
	      return -1;
	    }
	  /* get data offset and size from file comment. */
	  p = zip_get_file_comment (decoder->priv->archive, *di, &len, 0);
	  if (!p)
	    {
	      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
			   _("Format of `%s' invalid: %s"),
			   filename, _("Missing file comment"));
	      return -1;
	    }
	  g_assert (len + 1 < sizeof (buff));
	  g_strlcpy (buff, p, len + 1);
	  if (sscanf (buff, "%lld:%lld,%d", dstart, dend, dsize) != 3)
	    {
	      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
			   _("Format of `%s' invalid: %s"),
			   filename, _("Invalid file comment"));
	      return -1;
	    }
	  return i;
	}
    }
  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
	       _("Index `%d' out of range"), index);
  return -1;
}

/**
 * t_decoder_read_archive_file:
 *
 * read archive file specified by @ifile into memory.
 *
 * Returns: new alloced memory, must free with g_free().
 */
static gpointer
t_decoder_read_archive_file (TDecoder *decoder,
			     gint      ifile,
			     guint32  *size,
			     GError  **error)
{
  struct zip_file *zfile;
  struct zip_stat sb;
  gpointer buffer;

  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);

  zip_stat_index (decoder->priv->archive, ifile, 0, &sb);
  g_assert (sb.size > 0);

  zfile = zip_fopen_index (decoder->priv->archive, ifile, 0);
  if (!zfile)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Failed to open archive file: %s"),
		   zip_strerror (decoder->priv->archive));
      return NULL;
    }
  /* alloc memory to read file. */
  buffer = g_malloc (sb.size);

  if (zip_fread (zfile, buffer, sb.size) != sb.size)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_PARTIAL_INPUT,
		   _("Failed to read archive file: %s"),
		   g_strerror (errno));
      g_free (buffer);
      zip_fclose (zfile);
      return NULL;
    }
  zip_fclose (zfile);

  if (size) *size = sb.size;
  return buffer;
}

/**
 * t_decoder_decode_indexes:
 *
 * decode index contents(string format) into
 * binary TRecordIndex struct.
 */
static gboolean
t_decoder_decode_indexes (TDecoder     *decoder,
			  const gchar  *data,
			  guint32       size,
			  TRecordIndex *indexes,
			  guint32       nindexes,
			  GError      **error)
{
  guint32 i;
  const gchar *p = data, *p2;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (data, FALSE);
  g_return_val_if_fail (indexes, FALSE);

  for (i = 0; i < nindexes && size > (p - data); i++)
    {
      p2 = strchr (p, ';');
      if (!p2)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Failed to read archive index file: %s"),
		       _("Format incorrect"));
	  return FALSE;
	}
      if (sscanf (p, "%lf,%hd", &indexes[i].interval, &indexes[i].size) != 2)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Failed to read archive index file: %s"),
		       _("Format incorrect"));
	  return FALSE;
	}
      p = p2 + 1;
    }
  return TRUE;
}

/**
 * t_decoder_decrypt_data:
 */
static gboolean
t_decoder_decrypt_data (TDecoder    *decoder,
			gchar       *data,
			gsize        size,
			GError     **error)
{
  gcry_cipher_hd_t cipher;
  gcry_error_t ge = 0;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (data, FALSE);

  ge = gcry_cipher_open (&cipher, GCRY_CIPHER_BLOWFISH,
			 GCRY_CIPHER_MODE_CBC, GCRY_CIPHER_CBC_CTS);
  if (ge != 0)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   "Decrypt failed: %s/%s", gcry_strsource (ge),
		   gcry_strerror (ge));
      return FALSE;
    }
  gcry_cipher_setkey (cipher, decoder->priv->passwd, 16);

  ge = gcry_cipher_decrypt (cipher, data, size, NULL, 0);
  if (ge != 0)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   "Decrypt failed: %s/%s", gcry_strsource (ge),
		   gcry_strerror (ge));
      gcry_cipher_close (cipher);
      return FALSE;
    }
  gcry_cipher_close (cipher);

  return TRUE;
}

/**
 * t_decoder_decompress_data:
 */
static gchar*
t_decoder_decompress_data (TDecoder     *decoder,
			   GInputStream *memory,
			   guint64       size,
			   GError      **error)
{
  GZlibDecompressor *converter;
  GInputStream *stream;
  gchar *result;

  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);
  g_return_val_if_fail (G_IS_INPUT_STREAM (memory), NULL);
  g_return_val_if_fail (size > 0, NULL);

  if (size > 1024 * 1024 * 20)
    g_message ("try to alloc %lld bytes memory.", size);

  result = g_malloc (size);
  if (!result)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NO_SPACE,
		   _("Unable to alloc %lld bytes memory : %s"),
		   size, g_strerror (errno));
      return NULL;
    }
  converter = g_zlib_decompressor_new (G_ZLIB_COMPRESSOR_FORMAT_RAW);
  stream = g_converter_input_stream_new (memory, G_CONVERTER (converter));
  g_object_unref (converter);

  /* read compressed data from stream. */
  if (g_input_stream_read (stream, result, size, NULL, error) < 0)
    {
      g_prefix_error (error, _("Unable to read record data: "));
      g_object_unref (stream);
      g_free (result);
      return NULL;
    }
  g_object_unref (stream);

  return result;
}

/**
 * t_decoder_read_archive:
 */
static gboolean
t_decoder_read_archive (TDecoder    *decoder,
			guint32      index,
			const gchar *filename,
			GError     **error)
{
  gint ii, di;
  TRecordIndex *indexes;
  guint32 istart, iend;
  guint32 size, dsize;
  gpointer data;
  guint64 dstart, dend;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (decoder->priv->archive, FALSE);

  /* locate archive file by index. */
  if ((ii = t_decoder_get_archive_file (decoder, filename, index,
					&istart, &iend, &di,
					&dstart, &dend, &dsize, error)) < 0)
    return FALSE;

  /* read index file into memory. */
  data = t_decoder_read_archive_file (decoder, ii, &size, error);
  if (!data)
    return FALSE;

  /* convert to TRecordIndex struct. */
  indexes = g_new (TRecordIndex, iend - istart);
  if (!t_decoder_decode_indexes (decoder, data, size, indexes,
				 iend - istart, error))
    {
      g_free (data);
      g_free (indexes);
      return FALSE;
    }
  g_free (data);

  /* read data file into memory.
   * if data is encrypted and no passwd is given. we skip read data. */
  if (decoder->priv->head.encrypted && !decoder->priv->passwd)
    {
      g_free (decoder->priv->data);

      decoder->priv->data   = NULL;
      decoder->priv->dstart = 0;
      decoder->priv->dend   = 0;
    }
  else
    {
      GInputStream *memory;

      data = t_decoder_read_archive_file (decoder, di, &size, error);
      if (!data)
	{
	  g_free (indexes);
	  return FALSE;
	}
      /* decrypt on the fly. */
      if (decoder->priv->head.encrypted)
	{
	  gsize block_size;

	  /* get clpher block size.
	   * data file size must be a multiple of the cipher block size. */
	  gcry_cipher_algo_info (GCRY_CIPHER_BLOWFISH, GCRYCTL_GET_BLKLEN,
				 NULL, &block_size);
	  /* archive size must >= dsize. */
	  if (size % block_size != 0 || dsize > size)
	    {
	      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
			   _("Format of `%s' invalid: %s"),
			   filename, _("Bad data file size"));
	      g_free (data);
	      g_free (indexes);
	      return FALSE;
	    }
	  if (!t_decoder_decrypt_data (decoder, data, size, error))
	    {
	      g_free (data);
	      g_free (indexes);
	      return FALSE;
	    }
	}
      else
	{
	  /* archive file size must == dsize. */
	  if (dsize != size)
	    {
	      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
			   _("Format of `%s' invalid: %s"),
			   filename, _("Bad data file size"));
	      g_free (data);
	      g_free (indexes);
	      return FALSE;
	    }
	}
      /* decompress. */
      memory = g_memory_input_stream_new_from_data (data, dsize, g_free);
      data = t_decoder_decompress_data (decoder, memory, dend - dstart, error);
      if (!data)
	{
	  g_object_unref (memory);
	  g_free (indexes);
	  return FALSE;
	}
      g_object_unref (memory);

      g_free (decoder->priv->data);

      decoder->priv->data   = data;
      decoder->priv->dstart = dstart;
      decoder->priv->dend   = dend;
    }
  g_free (decoder->priv->indexes);

  decoder->priv->indexes = indexes;
  decoder->priv->istart  = istart;
  decoder->priv->iend    = iend;

  return TRUE;
}

/**
 * t_decoder_set_filename:
 */
void
t_decoder_set_filename (TDecoder     *decoder,
			const gchar  *filename)
{
  g_return_if_fail (T_IS_DECODER (decoder));
  g_return_if_fail (filename);

  g_free (decoder->priv->filename);
  decoder->priv->filename = g_strdup (filename);

  g_object_notify (G_OBJECT (decoder), "filename");
}

/**
 * t_decoder_open:
 *
 * must set filename before open.
 */
gboolean
t_decoder_open (TDecoder           *decoder,
		const gchar        *filename,
		TDecoderPasswdFunc  passwd_func,
		gpointer            user_data,
		GDestroyNotify      destroy,
		GError            **error)
{
  struct zip *archive;
  gint ze;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (filename, FALSE);

  /* save for restore. */
  archive = decoder->priv->archive;

  decoder->priv->archive = zip_open (filename, ZIP_CHECKCONS, &ze);
  if (!decoder->priv->archive)
    {
      gchar errbuf[256];

      if (ze == ZIP_ER_NOZIP)
	g_strlcpy (errbuf, _("Format unrecognizable."), sizeof (errbuf));
      else
	zip_error_to_str (errbuf, sizeof (errbuf), ze, errno);

      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Failed to open `%s': %s"), filename, errbuf);
      decoder->priv->archive = archive;
      return FALSE;
    }
  /* read and check record file head. */
  if (!t_decoder_read_record_head (decoder, filename, error))
    goto failed;

  /* check whether or not encrypted. */
  if (decoder->priv->head.encrypted)
    {
      /* chance to set decrypt password.
       * here we leave the decoder->priv->passwd check. so tools
       * like ttyassayer can use a always TRUE passwd_func
       * in order to read record head and index info. */
      if (!passwd_func || !passwd_func (decoder, user_data))
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
		       _("`%s' need password in order to decode."), filename);
	  goto failed;
	}
    }
  /* read the first index and data file into memory. */
  if (!t_decoder_read_archive (decoder, 0, filename, error))
    goto failed;

  /* close the previous archive if it is open. */
  if (archive)
    zip_close (archive);

  /* destroy user data. */
  if (user_data && destroy)
    destroy (user_data);

  t_decoder_set_filename (decoder, filename);
  return TRUE;

failed:
  if (user_data && destroy)
    destroy (user_data);

  zip_close (decoder->priv->archive);

  /* restore state. */
  decoder->priv->archive = archive;
  if (decoder->priv->archive)
    {
      t_decoder_read_record_head (decoder, decoder->priv->filename, NULL);
      t_decoder_read_archive (decoder, decoder->priv->istart,
			      decoder->priv->filename, NULL);
    }
  return FALSE;
}

/**
 * t_decoder_close:
 */
void
t_decoder_close (TDecoder *decoder)
{
  g_return_if_fail (T_IS_DECODER (decoder));

  if (decoder->priv->archive)
    {
      zip_close (decoder->priv->archive);
      decoder->priv->archive = NULL;
    }
  if (decoder->priv->indexes)
    {
      g_free (decoder->priv->indexes);
      decoder->priv->indexes = NULL;
    }
  if (decoder->priv->data)
    {
      g_free (decoder->priv->data);
      decoder->priv->data = NULL;
    }
  if (decoder->priv->passwd)
    {
      g_free (decoder->priv->passwd);
      decoder->priv->passwd = NULL;
    }
  memset (&decoder->priv->head, 0, sizeof (TRecordHead));

  g_free (decoder->priv->filename);
  decoder->priv->filename = NULL;
}

/**
 * t_decoder_get_filename:
 */
const gchar*
t_decoder_get_filename (TDecoder *decoder)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);
  return decoder->priv->filename;
}

/**
 * t_decoder_get_comment:
 */
const gchar*
t_decoder_get_comment (TDecoder *decoder,
		       gint     *length)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);
  g_return_val_if_fail (decoder->priv->archive, NULL);

  return zip_get_archive_comment (decoder->priv->archive, length, 0);
}

/**
 * t_decoder_get_archive:
 */
gpointer
t_decoder_get_archive (TDecoder *decoder)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);
  return decoder->priv->archive;
}

/**
 * t_decoder_get_head:
 */
TRecordHead*
t_decoder_get_head (TDecoder *decoder)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);
  return &decoder->priv->head;
}

/**
 * t_decoder_get_nindexes:
 */
guint32
t_decoder_get_nindexes (TDecoder *decoder)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), 0);
  return decoder->priv->head.nindexes;
}

/**
 * t_decoder_get_indexes:
 */
TRecordIndex*
t_decoder_get_indexes (TDecoder *decoder,
		       guint32  *istart,
		       guint32  *iend)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);

  if (istart) *istart = decoder->priv->istart;
  if (iend)   *iend = decoder->priv->iend;

  return decoder->priv->indexes;
}

/**
 * t_decoder_get_data:
 */
const gchar*
t_decoder_get_data (TDecoder *decoder,
		    guint64  *dstart,
		    guint64  *dend)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), NULL);

  if (dstart) *dstart = decoder->priv->dstart;
  if (dend)   *dend = decoder->priv->dend;

  return decoder->priv->data;
}

/**
 * t_decoder_has_index:
 */
gboolean
t_decoder_has_index (TDecoder *decoder,
		     guint32   index)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (decoder->priv->archive, FALSE);

  return (index >= decoder->priv->istart &&
	  index < decoder->priv->iend);
}

/**
 * t_decoder_seek:
 */
gboolean
t_decoder_seek (TDecoder *decoder,
		guint32   index,
		GError  **error)
{
  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (decoder->priv->archive, FALSE);

  /* need not op. */
  if (t_decoder_has_index (decoder, index))
    return TRUE;

  if (index > decoder->priv->head.nindexes)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		   _("Index `%d' out of range"), index);
      return FALSE;
    }
  return t_decoder_read_archive (decoder, index, decoder->priv->filename, error);
}

/**
 * t_decoder_get_index_data:
 *
 * @index is from begin of archive, not from current cache.
 */
const gchar*
t_decoder_get_index_data (TDecoder  *decoder,
			  guint32    index,
			  guint16   *size,
			  GError   **error)
{
  guint64 doffset = 0;
  guint32 i, j;

  g_return_val_if_fail (T_IS_DECODER (decoder), FALSE);
  g_return_val_if_fail (decoder->priv->archive, FALSE);

  /* index out of range. */
  if (index >= decoder->priv->head.nindexes)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVAL,
		   _("Index `%d' out of range."), index);
      return NULL;
    }
  /* cannot read encrypted data without password. */
  if (decoder->priv->head.encrypted && !decoder->priv->passwd)
    {
      g_assert (decoder->priv->data == NULL);
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
		   _("Data is encrypted and no password given."));
      return NULL;
    }
  /* index out of cache. */
  if (!t_decoder_has_index (decoder, index))
    {
      if (!t_decoder_seek (decoder, index, error))
	return NULL;
    }
  g_assert (t_decoder_has_index (decoder, index));

  /* get data offset. */
  for (i = decoder->priv->istart, j = 0; i < index; i++, j++)
    doffset += decoder->priv->indexes[j].size;

  *size = decoder->priv->indexes[j].size;

  return decoder->priv->data + doffset;
}

/**
 * t_decoder_get_index_file_elapsed:
 */
static gdouble
t_decoder_get_index_file_elapsed (TDecoder  *decoder,
				  guint32    index,
				  guint32   *istart,
				  guint32   *iend,
				  gint      *file_index,
				  GError   **error)
{
  gdouble result = 0.0, elapsed;
  gchar buff[64];
  const gchar *p;
  gint n, len;

  g_return_val_if_fail (T_IS_DECODER (decoder), 0.0);

  for (n = 0; n < decoder->priv->head.nfiles; n++)
    {
      g_snprintf (buff, sizeof (buff), "indexes/%d", n);
      *file_index = zip_name_locate (decoder->priv->archive, buff, 0);
      if (*file_index == -1)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Format of `%s' invalid: %s"),
		       decoder->priv->filename, _("Missing required file"));
	  return -1.0;
	}
      p = zip_get_file_comment (decoder->priv->archive, *file_index, &len, 0);
      if (!p)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Format of `%s' invalid: %s"),
		       decoder->priv->filename, _("Missing file comment"));
	  return -1.0;
	}
      g_assert (len + 1 < sizeof (buff));
      g_strlcpy (buff, p, len + 1);
      if (sscanf (buff, "%u:%u,%lf", istart, iend, &elapsed) != 3)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Format of `%s' invalid: %s"),
		       decoder->priv->filename, _("Invalid file comment"));
	  return -1.0;
	}
      if (index >= *istart && index < *iend)
	break;

      result += elapsed;
    }
  return result;
}

/**
 * t_decoder_get_index_elapsed:
 */
gdouble
t_decoder_get_index_elapsed (TDecoder  *decoder,
			     guint32    index,
			     GError   **error)
{
  gdouble elapsed = 0.0;
  guint32 istart, iend, i, size;
  gint file_index = 0;
  TRecordIndex *indexes;
  gchar *data;

  g_return_val_if_fail (T_IS_DECODER (decoder), 0.0);

  if (!decoder->priv->archive)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_CLOSED, _("Not open"));
      return -1.0;
    }
  /* index out of range. */
  if (index >= decoder->priv->head.nindexes)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND,
		   _("Index %d out of range"), index);
      return -1.0;
    }
  /* current cache has the index, and it is the first
   * index file. so we need not to read archive. */
  if (decoder->priv->istart == 0 &&
      t_decoder_has_index (decoder, index))
    {
      for (i = 0; i <= index; i++)
	elapsed += decoder->priv->indexes[i].interval;

      return elapsed;
    }
  /* we have to read the index file comment to get
   * the previous elapsed time. */
  elapsed = t_decoder_get_index_file_elapsed (decoder, index,
					      &istart, &iend,
					      &file_index, error);
  if (elapsed < 0)
    return -1.0;

  /* ok, our cache has the indexes data. */
  if (t_decoder_has_index (decoder, index))
    {
      for (i = 0; i <= index - istart; i++)
	elapsed += decoder->priv->indexes[i].interval;

      return elapsed;
    }
  /* we need load the index file into memory. */
  data = t_decoder_read_archive_file (decoder, file_index, &size, error);
  if (!data)
    return -1.0;

  /* convert to TRecordIndex struct. */
  indexes = g_new (TRecordIndex, iend - istart);
  if (!t_decoder_decode_indexes (decoder, data, size, indexes,
				 iend - istart, error))
    {
      g_free (data);
      g_free (indexes);
      return -1.0;
    }
  g_free (data);

  for (i = 0; i <= index - istart; i++)
    elapsed += indexes[i].interval;

  g_free (indexes);

  return elapsed;
}
