/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/socket.h>

#include <glib/gi18n.h>
#include <gio/gio.h>

#include "tmarshal.h"
#include "tmacros.h"
#include "tring.h"
#include "tlook.h"

enum
{
  PROP_0,
  PROP_HOST,
  PROP_PORT,
  PROP_TIMEOUT,
  PROP_SUSPENSION,
};

struct _TLookPriv
{
  GSocketConnection *connection;
  GCancellable      *cancellable;
  TLookdInfo         info;
  gchar             *host;
  guint16            port;
  guint              timeout;
  guint              connect_timeout;
  gchar              read_buff[512];
  TRing             *write_ring;
  gboolean           suspension;
  TRing             *cache_ring;
};
#define _LOOK_GET_PRIVATE(o)  (G_TYPE_INSTANCE_GET_PRIVATE ((o), T_TYPE_LOOK, TLookPriv))

G_DEFINE_TYPE (TLook, t_look, G_TYPE_OBJECT);

/**
 * t_look_init:
 */
static void
t_look_init (TLook *look)
{
  look->priv = _LOOK_GET_PRIVATE (look);

  look->priv->connection = NULL;
  look->priv->cancellable = g_cancellable_new ();
  look->priv->host = NULL;
  look->priv->port = 0;
  look->priv->timeout = 0;
  look->priv->connect_timeout = 0;
  look->priv->write_ring = t_ring_new ();
  look->priv->suspension = FALSE;
  look->priv->cache_ring = t_ring_new ();
}

/**
 * t_look_dispose:
 */
static void
t_look_dispose (GObject *object)
{
  TLook *look = T_LOOK (object);

  if (look->priv->cancellable)
    {
      g_object_unref (look->priv->cancellable);
      look->priv->cancellable = NULL;
    }
  if (look->priv->connection)
    {
      while (g_io_stream_has_pending (G_IO_STREAM (look->priv->connection)))
	g_main_context_iteration (NULL, FALSE);

      g_io_stream_close (G_IO_STREAM (look->priv->connection), NULL, NULL);
      g_object_unref (look->priv->connection);
      look->priv->connection = NULL;
    }
  G_OBJECT_CLASS (t_look_parent_class)->dispose (object);
}

/**
 * t_look_finalize:
 */
static void
t_look_finalize (GObject *object)
{
  TLook *look = T_LOOK (object);

  _debug ("look finalized.");

  if (look->priv->host)
    {
      g_free (look->priv->host);
      look->priv->host = NULL;
    }
  if (look->priv->write_ring)
    {
      t_ring_free (look->priv->write_ring);
      look->priv->write_ring = NULL;
    }
  if (look->priv->cache_ring)
    {
      t_ring_free (look->priv->cache_ring);
      look->priv->cache_ring = NULL;
    }
  G_OBJECT_CLASS (t_look_parent_class)->finalize (object);
}

/**
 * t_look_get_property:
 */
static void
t_look_get_property (GObject    *object,
		     guint       property_id,
		     GValue     *value,
		     GParamSpec *pspec)
{
  TLook *look = T_LOOK (object);

  switch (property_id)
    {
    case PROP_HOST:
      g_value_set_string (value, t_look_get_host (look));
      break;
    case PROP_PORT:
      g_value_set_uint (value, t_look_get_port (look));
      break;
    case PROP_TIMEOUT:
      g_value_set_uint (value, t_look_get_timeout (look));
      break;
    case PROP_SUSPENSION:
      g_value_set_boolean (value, t_look_is_suspension (look));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * t_look_set_property:
 */
static void
t_look_set_property (GObject      *object,
		     guint         property_id,
		     const GValue *value,
		     GParamSpec   *pspec)
{
  TLook *look = T_LOOK (object);

  switch (property_id)
    {
    case PROP_HOST:
      t_look_set_host (look, g_value_get_string (value));
      break;
    case PROP_PORT:
      t_look_set_port (look, g_value_get_uint (value));
      break;
    case PROP_TIMEOUT:
      t_look_set_timeout (look, g_value_get_uint (value));
      break;
    case PROP_SUSPENSION:
      t_look_set_suspension (look, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

/**
 * t_look_hangup_real:
 */
static void
t_look_hangup_real (TLook *look)
{
  if (look->priv->connection)
    {
      g_object_unref (look->priv->connection);
      look->priv->connection = NULL;
    }
}

/**
 * t_look_class_init:
 */
static void
t_look_class_init (TLookClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (TLookPriv));

  gobject_class->get_property = t_look_get_property;
  gobject_class->set_property = t_look_set_property;
  gobject_class->dispose      = t_look_dispose;
  gobject_class->finalize     = t_look_finalize;

  klass->hangup = t_look_hangup_real;

  g_object_class_install_property (gobject_class, PROP_HOST,
				   g_param_spec_string ("host", "Host", "Server host", NULL,
							G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_PORT,
				   g_param_spec_uint ("port", "Port", "Listen port", 0, G_MAXUINT16, 0,
						      G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
						      G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_TIMEOUT,
				   g_param_spec_uint ("timeout", "Timeout", "Connect timeout", 0, G_MAXUINT, 0,
						      G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
						      G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_object_class_install_property (gobject_class, PROP_SUSPENSION,
				   g_param_spec_boolean ("suspension", "Suspension", "Discard received data", FALSE,
							 G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
							 G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
  g_signal_new ("connected", T_TYPE_LOOK, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
  g_signal_new ("received", T_TYPE_LOOK, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		t_marshal_VOID__POINTER_UINT,
		G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_UINT);
  g_signal_new ("written", T_TYPE_LOOK, G_SIGNAL_RUN_LAST,
		0, NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
  g_signal_new ("hangup", T_TYPE_LOOK, G_SIGNAL_RUN_FIRST,
		G_STRUCT_OFFSET (TLookClass, hangup),
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID,
		G_TYPE_NONE, 0);
}

/**
 * t_look_new:
 */
TLook*
t_look_new (void)
{
  return g_object_new (T_TYPE_LOOK, NULL);
}

/**
 * t_look_get_host:
 */
const gchar*
t_look_get_host (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), NULL);
  return look->priv->host;
}

/**
 * t_look_set_host:
 *
 * set server address, @host may be a hostname. a IP
 * address(both IPv4 or IPv6)
 */
void
t_look_set_host (TLook       *look,
		 const gchar *host)
{
  gchar *new_host;

  g_return_if_fail (T_IS_LOOK (look));
  g_return_if_fail (host);

  new_host = g_strdup (host);
  g_free (look->priv->host);
  look->priv->host = new_host;
  g_object_notify (G_OBJECT (look), "host");
}

/**
 * t_look_get_port:
 */
guint16
t_look_get_port (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), 0);
  return look->priv->port;
}

/**
 * t_look_set_port:
 *
 * set server listen port.
 *
 * this function is just for convenience. beacuse
 * you can set port by t_look_set_host() with below format:
 *
 *  host:port
 */
void
t_look_set_port (TLook   *look,
		 guint16  port)
{
  g_return_if_fail (T_IS_LOOK (look));

  look->priv->port = port;
  g_object_notify (G_OBJECT (look), "port");
}

/**
 * t_look_get_timeout:
 */
guint
t_look_get_timeout (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), 0);
  return look->priv->timeout;
}

/**
 * t_look_set_timeout:
 *
 * set connect and read timeout, in seconds.
 */
void
t_look_set_timeout (TLook *look,
		    guint  timeout)
{
  g_return_if_fail (T_IS_LOOK (look));

  look->priv->timeout = timeout;
  g_object_notify (G_OBJECT (look), "timeout");
}

/**
 * t_look_get_cancellable:
 */
GCancellable*
t_look_get_cancellable (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), NULL);
  return look->priv->cancellable;
}

/**
 * t_look_get_info:
 *
 * get server information, this information
 * is invaild until connection been created.
 */
TLookdInfo*
t_look_get_info (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), NULL);
  return &look->priv->info;
}

/**
 * t_look_read_info_ready:
 */
static void
t_look_read_info_ready (GObject      *object,
			GAsyncResult *res,
			gpointer      data)
{
  TLook *look = data;
  GInputStream *istream = G_INPUT_STREAM (object);
  gsize nread;
  GError *error = NULL;

  g_return_if_fail (G_IS_INPUT_STREAM (istream));
  g_return_if_fail (T_IS_LOOK (look));

  nread = g_input_stream_read_finish (istream, res, &error);
  if (error)
    {
      g_warning (_("Unable to read look server information: %s"),
		 error->message);
      g_error_free (error);
      goto bad;
    }
  if (nread < sizeof (look->priv->info))
    {
      _warning ("connection been closed.");
      goto bad;
    }
  /* convert to host byte order */
  look->priv->info.termrows = g_ntohs (look->priv->info.termrows);
  look->priv->info.termcols = g_ntohs (look->priv->info.termcols);

  g_signal_emit_by_name (look, "connected", 0);
  g_object_unref (look);
  return;

bad:
  /* unref the connection before emit signal */
  g_object_unref (look->priv->connection);
  look->priv->connection = NULL;

  g_signal_emit_by_name (look, "connected", 0);
  g_object_unref (look);
}

/**
 * t_look_connect_ready:
 *
 * read server information
 */
static void
t_look_connect_ready (GObject       *object,
		      GAsyncResult  *res,
		      gpointer       data)
{
  TLook *look = data;
  GSocketClient *client = G_SOCKET_CLIENT (object);
  GSocketConnection *connection;
  GInputStream *istream;
  GSocket *socket;
  GError *error = NULL;

  g_return_if_fail (G_IS_SOCKET_CLIENT (client));
  g_return_if_fail (T_IS_LOOK (look));

  /* remove connecting timeout event first */
  if (look->priv->connect_timeout > 0)
    g_source_remove (look->priv->connect_timeout);

  /* finish the async connect operation */
  connection = g_socket_client_connect_to_host_finish (client, res, &error);
  g_object_unref (client);
  if (error)
    {
      g_warning (_("Failed to connect `%s': %s"), look->priv->host,
		 error->message);
      g_error_free (error);

      g_signal_emit_by_name (look, "connected", 0);
      g_object_unref (look);
      return;
    }
  /* save the connection */
  look->priv->connection = connection;

  /* graceful disconnect means before closing the connection, we signal
   * the receiving end the connection will be terminate, and wait for
   * it to close the connection. */
  g_tcp_connection_set_graceful_disconnect (G_TCP_CONNECTION (connection), TRUE);

  /* tell system attempt to verify that the remote socket endpoint is
   * still present if a sufficiently long period of time passes with no data
   * being exchanged. */
  socket = g_socket_connection_get_socket (connection);
  g_socket_set_keepalive (socket, TRUE);

  /* set communcation timeout */
  g_socket_set_timeout (socket, look->priv->timeout);

  istream = g_io_stream_get_input_stream (G_IO_STREAM (connection));

  /* read server infomation */
  g_input_stream_read_async (istream,
			     &look->priv->info,
			     sizeof (TLookdInfo),
			     G_PRIORITY_DEFAULT,
			     look->priv->cancellable,
			     t_look_read_info_ready, look);
}

/**
 * t_look_connect_timeout:
 */
static gboolean
t_look_connect_timeout (gpointer data)
{
  TLook *look = T_LOOK (data);

  g_return_val_if_fail (T_IS_LOOK (look), FALSE);

  if (look->priv->cancellable)
    g_cancellable_cancel (look->priv->cancellable);

  return FALSE;
}

/**
 * t_look_connect_timeout_destroy:
 */
static void
t_look_connect_timeout_destroy (gpointer data)
{
  TLook *look = T_LOOK (data);

  g_return_if_fail (T_IS_LOOK (look));

  look->priv->connect_timeout = 0;
  g_object_unref (look);
}

/**
 * t_look_connect:
 *
 * try to connect look server
 */
gboolean
t_look_connect (TLook *look)
{
  GSocketClient *client;

  g_return_val_if_fail (T_IS_LOOK (look), FALSE);
  g_return_val_if_fail (look->priv->host, FALSE);

  client = g_socket_client_new ();
  g_socket_client_set_socket_type (client, G_SOCKET_TYPE_STREAM);

  if (look->priv->connection)
    {
      g_object_unref (look->priv->connection);
      look->priv->connection = NULL;
    }
  /* reset cancellable to normal state. */
  g_cancellable_reset (look->priv->cancellable);

  /* clear suspension state. */
  look->priv->suspension = FALSE;

  /* connect server asynchronous.
   * here we increase look object's refcount. so when connect finish
   * (either connected or error occur), we must decrease the refcount,
   * otherwise, the look object will never been free. */
  g_socket_client_connect_to_host_async (client,
					 look->priv->host,
					 look->priv->port,
					 look->priv->cancellable,
					 t_look_connect_ready,
					 g_object_ref (look));
  /* set connect timeout */
  if (look->priv->timeout > 0)
    {
      look->priv->connect_timeout =
       g_timeout_add_seconds_full (G_PRIORITY_HIGH,
				   look->priv->timeout,
				   t_look_connect_timeout,
				   g_object_ref (look),
				   t_look_connect_timeout_destroy);
    }
  return TRUE;
}

/**
 * t_look_disconnect:
 */
void
t_look_disconnect (TLook *look)
{
  g_return_if_fail (T_IS_LOOK (look));

  g_cancellable_cancel (look->priv->cancellable);

  if (look->priv->connection)
    g_io_stream_close (G_IO_STREAM (look->priv->connection), NULL, NULL);
}

/**
 * t_look_is_connected:
 */
gboolean
t_look_is_connected (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), FALSE);

  if (!look->priv->connection)
    return FALSE;

  return !g_io_stream_is_closed (G_IO_STREAM (look->priv->connection));
}

/**
 * t_look_written_ready:
 */
static void
t_look_written_ready (GOutputStream *ostream,
		      GAsyncResult  *res,
		      TLook         *look)
{
  GError *error = NULL;
  gsize nwritten;

  g_return_if_fail (G_IS_OUTPUT_STREAM (ostream));
  g_return_if_fail (T_IS_LOOK (look));

  nwritten = g_output_stream_write_finish (ostream, res, &error);
  if (error)
    {
      g_warning (_("Send data to look server error: %s"),
		 error->message);
      g_error_free (error);

      g_signal_emit_by_name (look, "written", 0);
      g_object_unref (look);
      return;
    }
  /* continue send cached data */
  if (t_ring_get_size (look->priv->write_ring) > 0)
    {
      gchar buff[256];
      gint nread;

      nread = t_ring_read (look->priv->write_ring, buff, sizeof (buff));

      g_output_stream_write_async (ostream, buff, nread,
				   G_PRIORITY_DEFAULT,
				   look->priv->cancellable,
				   (GAsyncReadyCallback) t_look_written_ready,
				   look);
      return;
    }
  g_signal_emit_by_name (look, "written", 0);
  g_object_unref (look);
}

/**
 * t_look_write:
 *
 * write data to server.
 *
 * this is a async operation, data will been save in cache
 * first, and later actual write.
 *
 * if you care when data been actual transfer, you can listen
 * the written signal.
 */
gboolean
t_look_write (TLook        *look,
	      const gchar  *data,
	      gssize        size)
{
  GOutputStream *ostream;

  g_return_val_if_fail (T_IS_LOOK (look), FALSE);
  g_return_val_if_fail (data, FALSE);

  if (!look->priv->connection)
    return FALSE;

  if (size < 0)
    size = strlen (data);

  if (size == 0)
    return TRUE;

  ostream = g_io_stream_get_output_stream (G_IO_STREAM (look->priv->connection));

  /* if data too large(exceed the ring maximum size), skip cache. */
  if (!t_ring_can_write (look->priv->write_ring, size))
    {
      if (!g_output_stream_has_pending (ostream))
	{
	  /* write to server async */
	  g_output_stream_write_async (ostream, data, size,
				       G_PRIORITY_DEFAULT,
				       look->priv->cancellable,
				       (GAsyncReadyCallback) t_look_written_ready,
				       look);
	}
      else
	{
	  g_warning (_("Unable to send data to look server, lose %d bytes"), size);
	  return FALSE;
	}
    }
  /* save to ring buffer */
  t_ring_write (look->priv->write_ring, data, size);

  if (!g_output_stream_has_pending (ostream))
    {
      gchar buff[256];
      gint nread;

      nread = t_ring_read (look->priv->write_ring, buff, sizeof (buff));

      /* write to server async */
      g_output_stream_write_async (ostream, buff, nread,
				   G_PRIORITY_DEFAULT,
				   look->priv->cancellable,
				   (GAsyncReadyCallback) t_look_written_ready,
				   g_object_ref (look));
    }
  return TRUE;
}

/**
 * t_look_read_ready:
 */
static void
t_look_read_ready (GInputStream *istream,
		   GAsyncResult *res,
		   TLook        *look)
{
  gsize nread;
  GError *error = NULL;

  g_return_if_fail (T_IS_LOOK (look));
  g_return_if_fail (G_IS_INPUT_STREAM (istream));

  nread = g_input_stream_read_finish (istream, res, &error);
  if (error)
    {
      if (!g_cancellable_is_cancelled (look->priv->cancellable))
	{
	  g_warning (_("Read data from look server error: %s"),
		     error->message);
	}
      else
	{
	  _message ("read data from look erver been cancelled.");
	}
      g_error_free (error);
      g_signal_emit_by_name (look, "hangup", 0);

      /* decrease look object's refount. */
      g_object_unref (look);
      return;
    }
  /* end of file(disconnected) */
  if (nread == 0)
    {
      g_signal_emit_by_name (look, "hangup", 0);
      g_object_unref (look);
      return;
    }
  if (look->priv->suspension)
    {
      /* when suspend, save received data to cache. */
      if (!t_ring_can_write (look->priv->cache_ring, nread))
	t_ring_discard_to (look->priv->cache_ring, nread, nread, NULL, 0);

      t_ring_write (look->priv->cache_ring, look->priv->read_buff, nread);
    }
  else
    {
      /* refresh cache data first. */
      while (t_ring_get_size (look->priv->cache_ring) > 0)
	{
	  gchar buff[128];
	  gint nbytes;

	  nbytes = t_ring_read (look->priv->cache_ring, buff, sizeof (buff));
	  g_signal_emit_by_name (look, "received", buff, nbytes);
	}
      g_signal_emit_by_name (look, "received", look->priv->read_buff, nread);
    }
  /* continue read until connection been close or
   * error occur. */
  g_input_stream_read_async (istream,
			     look->priv->read_buff,
			     sizeof (look->priv->read_buff),
			     G_PRIORITY_DEFAULT,
			     look->priv->cancellable,
			     (GAsyncReadyCallback) t_look_read_ready,
			     look);
}

/**
 * t_look_read:
 *
 * read data from look server, this function just need
 * call once, it will continue read from server until
 * connection been close or error occur.
 */
gboolean
t_look_read (TLook *look)
{
  GInputStream *istream;
  GSocket *socket;

  g_return_val_if_fail (T_IS_LOOK (look), FALSE);

  if (!look->priv->connection)
    return FALSE;

  /* disable communcation timeout */
  socket = g_socket_connection_get_socket (look->priv->connection);
  g_socket_set_timeout (socket, 0);

  istream = g_io_stream_get_input_stream (G_IO_STREAM (look->priv->connection));

  /* read data from server. we increase look object's
   * refcount here. so when callback event destroy need
   * decrease it's refount. */
  g_input_stream_read_async (istream,
			     look->priv->read_buff,
			     sizeof (look->priv->read_buff),
			     G_PRIORITY_DEFAULT,
			     look->priv->cancellable,
			     (GAsyncReadyCallback) t_look_read_ready,
			     g_object_ref (look));
  return TRUE;
}

/**
 * t_look_set_suspension:
 */
void
t_look_set_suspension (TLook    *look,
		       gboolean  suspension)
{
  g_return_if_fail (T_IS_LOOK (look));

  if (suspension == FALSE)
    {
      gchar buff[512];
      gint nbytes;

      while (t_ring_get_size (look->priv->cache_ring) > 0)
	{
	  nbytes = t_ring_read (look->priv->cache_ring,
				buff, sizeof (buff));
	  g_signal_emit_by_name (look, "received", buff, nbytes);
	}
    }
  look->priv->suspension = suspension;
}

/**
 * t_look_is_suspension:
 */
gboolean
t_look_is_suspension (TLook *look)
{
  g_return_val_if_fail (T_IS_LOOK (look), FALSE);
  return look->priv->suspension;
}
