/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#define _GNU_SOURCE
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>

#include <glib/gi18n-lib.h>

#include "tmisc.h"

/* global log file handle */
static FILE	*_logfile = NULL;

/**
 * t_print_version:
 *
 * output applicatoin version information
 */
void
t_print_version (void)
{
  g_print ("%s %s\n%s %s  %s\n\n", g_get_application_name (),
	   VERSION, __DATE__, __TIME__, BUILD_OS);
  g_print (_("Copyright (C) 2010 Xiaohu Wu<xiaohu417@gmail.com>,\n"
	     "All rights reserved."));
  g_print ("\n\n%s\n\n", WEBSITE);
}

/**
 * app_glib_log_handler:
 * @data: a FILE handle or NULL.
 *
 * redirect glib's log message to file or syslog
 */
void
t_glib_log_handler (const gchar    *domain,
		    GLogLevelFlags  level,
		    const gchar    *message,
		    gpointer        data)
{
  gint priority = LOG_ALERT;
  gchar *level_str = NULL;

  switch (level)
    {
    case G_LOG_LEVEL_DEBUG:
      level_str = _("Debug");
      priority = LOG_DEBUG;
      break;
    case G_LOG_LEVEL_INFO:
      level_str = _("Info");
      priority = LOG_INFO;
      break;
    case G_LOG_LEVEL_MESSAGE:
      level_str = _("Message");
      priority = LOG_NOTICE;
      break;
    case G_LOG_LEVEL_WARNING:
      level_str = _("Warning");
      priority = LOG_WARNING;
      break;
    case G_LOG_LEVEL_CRITICAL:
      level_str = _("Critical");
      priority = LOG_CRIT;
      break;
    case G_LOG_LEVEL_ERROR:
    default:
      level_str = _("Error");
      priority = LOG_ERR;
      break;
    }
  /* if @data is a FILE handle, write message to this file,
   * otherwise write message to syslog */
  if (data)
    {
      FILE *file = (FILE*) data;
      time_t tm;
      struct tm *tmp;
      gchar times[256];

      /* record message time and pid to file beacuse log file
       * is open with append mode, and more than one process
       * maybe use same log file */
      time (&tm);
      tmp = localtime (&tm);
      strftime (times, sizeof (times), "%F-%T", tmp);
      fprintf (file, "%s[%s(%d)] <%s>%s: %s\n",
	       times, g_get_prgname (), getpid (),
	       domain ? domain : "NULL", level_str, message);
      fflush (file);
    }
  else
    {
      gchar buffer[1024];

      g_snprintf (buffer, sizeof (buffer), "<%s-%s> %s",
		  domain ? domain : "NULL", level_str, message);
      syslog (priority, buffer);
    }
}

/**
 * app_glib_print_handler:
 *
 * redirect g_print(err) message to file or syslog
 */
void
t_glib_print_handler (const gchar *message)
{
  if (_logfile)
    {
      time_t tm;
      struct tm *tmp;
      gchar times[256];

      time (&tm);
      tmp = localtime (&tm);
      strftime (times, sizeof (times), "%F-%T", tmp);

      fprintf (_logfile, "%s[%s(%d)] %s",
	       times, g_get_prgname (), getpid (), message);
      fflush (_logfile);
    }
  else
    {
      syslog (LOG_INFO, message);
    }
}

/**
 * t_gcrypt_log_handler:
 *
 * redirect gcrypt log to glib log system.
 * see gcry_set_log_handler().
 */
void
t_gcrypt_log_handler (gpointer     user_data,
		      gint         level,
		      const gchar *format,
		      va_list      args)
{
  g_logv ("GCrypt", G_LOG_LEVEL_MESSAGE, format, args);
}

/**
 * close_logfile:
 */
static void
close_logfile (void)
{
  if (_logfile)
    {
      fclose (_logfile);
      _logfile = NULL;
    }
}

/**
 * t_set_logfile:
 */
gboolean
t_set_logfile (const gchar *filename)
{
  FILE *file;
  static gsize once = 0;
  mode_t mask;

  g_return_val_if_fail (filename, FALSE);

  /* if file exists and has .conf or .lua suffix.
   * we make a backup(to prevent confuse command option
   * --logfile(-l) and --config(-c) */
  if (g_str_has_suffix (filename, ".conf") ||
      g_str_has_suffix (filename, ".lua"))
    {
      if (g_file_test (filename, G_FILE_TEST_EXISTS))
	{
	  gchar *contents;
	  gsize length;

	  if (g_file_get_contents (filename, &contents, &length, NULL))
	    {
	      gchar *bakfile = g_strdup_printf ("%s.bak", filename);

	      g_file_set_contents (bakfile, contents, length, NULL);

	      g_free (contents);
	      g_free (bakfile);
	    }
	}
    }
  /* modify new file permissions to that not accessable for
   * other users. usable just for new created file */
  mask = umask (S_IRWXG | S_IRWXO);

  /* open file in append mode */
  file = fopen (filename, "ab");
  umask (mask);

  if (file == NULL)
    {
      g_warning (_("Failed to open `%s' : %s"), filename,
		 g_strerror (errno));
      return FALSE;
    }
  if (_logfile)
    {
      fclose (_logfile);
    }
  _logfile = file;

  if (g_once_init_enter (&once))
    {
      g_atexit (close_logfile);
      g_once_init_leave (&once, 1);
    }
  return TRUE;
}

/**
 * app_get_logfile:
 */
FILE*
t_get_logfile (void)
{
  return _logfile;
}

/**
 * app_lua_path_add:
 *
 * append a lua module search path by modify LUA_PATH
 * and LUA_CPATH environment variable
 */
void
t_lua_path_add (const gchar *path)
{
  const gchar *env;
  gchar lpath[1024], *p; 

  g_return_if_fail (path);

  p = g_build_filename (path, "?", NULL);

  env = g_getenv ("LUA_PATH");
  g_snprintf (lpath, sizeof (lpath), "%s;%s.lua;%s.luc;%s",
	      p, p, p, env ? env : "");
  g_setenv ("LUA_PATH", lpath, TRUE);

  env = g_getenv ("LUA_CPATH");
  g_snprintf (lpath, sizeof (lpath), "%s.so;%s", p, env ? env : "");
  g_setenv ("LUA_CPATH", lpath, TRUE);

  g_free (p);
}

/**
 * on_quit_timeout:
 *
 * quit the application.
 */
#ifdef TRIAL_EDITION
static gboolean
on_quit_timeout (gpointer data)
{
  raise (SIGTERM);
  g_timeout_add_seconds (3, (GSourceFunc) exit, NULL);
  return FALSE;
}
#endif

/**
 * t_expire_check:
 *
 * Returns: TRUE if expired.
 */
gboolean
t_expire_check (void)
{
/* pass -DTRIAL_EDTION to CFLAGS when configure. */
#ifdef TRIAL_EDITION
#define QUIT_TIMEOUT	60	/* in minutes */
  struct tm tm, *tmp;
  time_t curr_time, expire_time;
  gchar buff[128];

  memset (&tm, 0, sizeof (tm));
  strptime (TRIAL_EDITION_EXPIRE_DATE, "%Y-%m-%d", &tm);
  expire_time = mktime (&tm);
  curr_time = time (NULL);

  if (curr_time < expire_time)
    {
      g_warning (_("This trial edition will be expired in %s, "
		   "You can purchasing Ttyutils from %s"),
		 TRIAL_EDITION_EXPIRE_DATE, EBAY_STORE_WEBSITE);

      curr_time += QUIT_TIMEOUT * 60;
      tmp = localtime (&curr_time);
      strftime (buff, sizeof (buff), "%x %X", tmp);
      g_message (_("Application will be quit at %s."), buff);
      g_timeout_add_seconds (QUIT_TIMEOUT * 60, on_quit_timeout, NULL);
      return FALSE;
    }
  else
    {
      g_critical (_("This trial edition was expired. "
		    "You can purchasing Ttyutils from %s"), EBAY_STORE_WEBSITE);
      return TRUE;
    }
#endif
  return FALSE;
}

/**
 * t_get_current_time:
 *
 * like gettimeofday(2) but return a double value.
 */
gdouble
t_get_current_time (void)
{
  GTimeVal timeval;

  g_get_current_time (&timeval);
  return (gdouble) timeval.tv_sec + (gdouble) timeval.tv_usec / 1000000;
}

/**
 * t_strip_hyphen_option:
 *
 * strip '--' command line option.
 *
 * This function will modify @argc and @argv on the fly.
 *
 * @argv must at least have @argv entries, otherwise will coredump.
 */
void
t_strip_hyphen_options (gint   *argc,
			gchar **argv[])
{
  gint i;

  g_return_if_fail (argc);
  g_return_if_fail (argv);
  g_return_if_fail (*argv);

again:
  for (i = 0; i < *argc; i++)
    {
      if (g_strcmp0 ((*argv)[i], "--") == 0)
	{
	  for (; i < *argc; i++)
	    (*argv)[i] = (*argv)[i + 1];

	  *argc -= 1;
	  goto again;
	}
    }
}

/**
 * t_check_version:
 *
 * check runtime library version.
 *
 * Returns: TRUE if succeed.
 */
gboolean
t_check_version (void)
{
  const gchar *ver;

  /* glib. */
  ver = glib_check_version (GLIB_MAJOR_VERSION,
			    GLIB_MINOR_VERSION,
			    GLIB_MICRO_VERSION);
  if (ver)
    {
      g_warning ("GLib %d.%d.%d or above is requried: %s",
		 GLIB_MAJOR_VERSION, GLIB_MINOR_VERSION,
		 GLIB_MICRO_VERSION, ver);
      return FALSE;
    }
  return TRUE;
}
