/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include <glib/gi18n-lib.h>

#include "tring.h"

#define RING_CHUNK_SIZE		(4096 - sizeof (char*) * 2)
#define RING_MAX_CHUNKS		 1280  /* ~5M */

typedef struct _TRingChunk	TRingChunk;

struct _TRingChunk
{
  gchar  buffer[RING_CHUNK_SIZE];
  gchar *read_ptr;
  gchar *write_ptr;
};

struct _TRing
{
  GList      *chunks;
  TRingChunk *read_chunk;
  TRingChunk *write_chunk;
  TRingChunk *cache_chunk;
  gint        total_size;
};

static gint	t_ring_offset		(TRing *ring,
					 gint   size,
					 gchar *str,
					 gint   tomatch);
static gint	t_ring_line_size	(TRing *ring,
					 gint   size);

/**
 * t_ring_new:
 *
 * create a TRing with a chunk
 */
TRing*
t_ring_new (void)
{
  TRing *ring;
  TRingChunk *chunk;

  ring = g_slice_new0 (TRing);

  chunk = g_slice_new0 (TRingChunk);
  chunk->read_ptr = chunk->write_ptr = chunk->buffer;

  ring->chunks = g_list_append (ring->chunks, chunk);
  ring->read_chunk = ring->write_chunk = chunk;

  return ring;
}

/** 
 * t_ring_chunk_free:
 *
 * free a TRingChunk
 */
static void
t_ring_chunk_free (gpointer data, gpointer user_data)
{
  TRingChunk *chunk = (TRingChunk*) data;

  g_slice_free (TRingChunk, chunk);
}

/**
 * t_ring_free:
 */
void
t_ring_free (TRing *ring)
{
  if (ring)
    {
      if (ring->chunks)
	{
	  g_list_foreach (ring->chunks, t_ring_chunk_free, NULL);
	  g_list_free (ring->chunks);
	}
      if (ring->cache_chunk)
	{
	  g_slice_free (TRingChunk, ring->cache_chunk);
	}
      g_slice_free (TRing, ring);
    }
}

/** 
 * t_ring_clear:
 */
void
t_ring_clear (TRing *ring)
{
  TRingChunk *newchunk;

  g_return_if_fail (ring);

  if (ring->chunks)
    {
      g_list_foreach (ring->chunks, t_ring_chunk_free, NULL);
      g_list_free (ring->chunks);
      ring->chunks = NULL;
    }
  newchunk = ring->cache_chunk;
  ring->cache_chunk = NULL;

  if (!ring->cache_chunk)
    newchunk = g_slice_new0 (TRingChunk);

  newchunk->read_ptr = newchunk->write_ptr = newchunk->buffer;

  ring->read_chunk = ring->write_chunk = newchunk;
  ring->chunks = g_list_append (ring->chunks, newchunk);
  ring->total_size = 0;
}

/** 
 * t_ring_get_nchunks:
 */
gint
t_ring_get_nchunks (TRing *ring)
{
  g_return_val_if_fail (ring, -1);

  if (!ring->chunks)
    return 0;

  return g_list_length (ring->chunks);
}

/**
 * t_ring_get_size:
 */
gint
t_ring_get_size (TRing *ring)
{
  g_return_val_if_fail (ring, -1);
  return ring->total_size;
}

/** 
 * t_ring_can_write:
 */
gboolean
t_ring_can_write (TRing *ring, gint size)
{
  gint nchunks, nremain, nwritable;

  g_return_val_if_fail (ring, FALSE);
  g_return_val_if_fail (size >= 0, FALSE);

  nchunks = size / RING_CHUNK_SIZE;
  nremain = size % RING_CHUNK_SIZE;

  nwritable = RING_CHUNK_SIZE - (ring->write_chunk->write_ptr -
				 ring->write_chunk->buffer);
  if (nremain > nwritable)
    nchunks += 1;

  return (g_list_length (ring->chunks) + nchunks < RING_MAX_CHUNKS);
}

/**
 * t_ring_write:
 */
gint
t_ring_write (TRing *ring, const gchar *data, gint size)
{
  gchar *pwrite;
  const gchar *pdata;
  gint size_written, nremain, nbytes;

  g_return_val_if_fail (ring, -1);
  g_return_val_if_fail (data, -1);

  g_assert (ring->write_chunk);

  if (size < 0)
    size = strlen (data);

  if (!t_ring_can_write (ring, size))
    return -1;

  pdata = data;
  size_written = size;

  while (size > 0)
    {
      pwrite = ring->write_chunk->write_ptr;
      nremain = RING_CHUNK_SIZE - (pwrite - ring->write_chunk->buffer);
      nbytes = MIN (size, nremain);
      memcpy (pwrite, pdata, nbytes);
      size -= nbytes;
      pdata += nbytes;

      ring->write_chunk->write_ptr += nbytes;

      if (ring->write_chunk->write_ptr ==
	  ring->write_chunk->buffer + RING_CHUNK_SIZE)
	{
	  TRingChunk *newchunk;

	  newchunk = ring->cache_chunk;
	  ring->cache_chunk = NULL;

	  if (!ring->cache_chunk)
	    newchunk = g_slice_new0 (TRingChunk);

	  newchunk->read_ptr = newchunk->write_ptr = newchunk->buffer;
	  ring->chunks = g_list_append (ring->chunks, newchunk);
	  ring->write_chunk = newchunk;
	}
    }
  ring->total_size += size_written;

  return size_written;
}

/** 
 * t_ring_can_read:
 */
gboolean
t_ring_can_read (TRing *ring, gint size)
{
  g_return_val_if_fail (ring, FALSE);

  return ring->total_size >= size;
}

/** 
 * t_ring_read:
 */
gint
t_ring_read (TRing *ring, gchar *buff, gint size)
{
  gchar *pread, *pbuff;
  gint size_readed, nremain, nbytes;

  g_return_val_if_fail (ring, -1);
  g_return_val_if_fail (size >= 0, -1);

  size_readed = size = MIN (ring->total_size, size);
  pbuff = buff;

  while (size > 0)
    {
      pread = ring->read_chunk->read_ptr;
      nremain = ring->read_chunk->write_ptr - ring->read_chunk->read_ptr;
      nbytes = MIN (nremain, size);
      if (pbuff)
	{
	  memcpy (pbuff, pread, nbytes);
	  pbuff += nbytes;
	}
      size -= nbytes;

      ring->read_chunk->read_ptr += nbytes;

      if (ring->read_chunk->read_ptr ==
	  ring->read_chunk->buffer + RING_CHUNK_SIZE)
	{
	  g_assert (ring->chunks->next);

	  ring->chunks = g_list_remove (ring->chunks, ring->read_chunk);

	  /* save for later use */
	  if (!ring->cache_chunk)
	    ring->cache_chunk = ring->read_chunk;
	  else
	    g_slice_free (TRingChunk, ring->read_chunk);

	  ring->read_chunk = (TRingChunk*) ring->chunks->data;
	}
    }
  ring->total_size -= size_readed;

  return size_readed;
}

/** 
 * t_ring_offset:
 */
gint
t_ring_offset (TRing *ring, gint size, gchar *str, gint tomatch)
{
  gchar *pread;
  GList *node;
  gint nret = 0;
  gint nmatch;

  g_return_val_if_fail (ring, -1);
  g_return_val_if_fail (str, -1);
  g_return_val_if_fail (size > 0, -1);

  node = g_list_find (ring->chunks, ring->read_chunk);
  g_assert (node);

  if (tomatch < 0)
    tomatch = strlen (str);

  nmatch = 0;

  for (; node && nret < size; node = node->next)
    {
      TRingChunk *chunk = (TRingChunk*) node->data;

      for (pread = chunk->read_ptr;
	   pread < chunk->write_ptr && nret < size;
	   pread++, nret++)
	{
	  if (str[nmatch] == *pread)
	    {
	      if (++nmatch == tomatch)
		return nret - nmatch + 1;
	    }
	  else
	    {
	      nmatch = 0;
	    }
	}
    }
  return -1;
}

/** 
 * t_ring_line_size:
 */
gint
t_ring_line_size (TRing *ring, gint size)
{
  gint nret;

  /* if terminal transfer `\r\n' as newline character sequence,
   * set IGNCR flag of terminal input mode.
   * if terminal transfer `\r' as newline character,
   * set ICRNL flag of terminal input mode.
   * if terminal transfer `\n' as newline character,
   * don't set INLCR flag of terminal input mode. */
  nret = t_ring_offset (ring, size, "\n", 1);

  /* include `\n' itself */
  return (nret > 0) ?  nret + 1 : nret;
}

/** 
 * t_ring_can_read_line:
 */
gboolean
t_ring_can_read_line (TRing *ring, int size)
{
  return (t_ring_line_size (ring, size) > 0);
}

/** 
 * t_ring_read_line:
 */
gint
t_ring_read_line (TRing *ring, gchar *buff, gint size)
{
  gint length;

  length = t_ring_line_size (ring, size);

  return (length > 0) ? t_ring_read (ring, buff, length) : -1;
}

/**
 * t_ring_read_to:
 */
gint
t_ring_read_to (TRing *ring,
		gchar *buff,
		gint   atleast,
		gint   maxsize,
		gchar *str,
		gint   nstr)
{
  gint nread, nret = 0;

  g_return_val_if_fail (ring, -1);
  g_return_val_if_fail (atleast >= 0, -1);

  if (atleast >= maxsize)
    {
      g_critical ("arguments of ring_read_to() incorrect, atleast %d must "
		  "less than or equal maxsize %d", atleast, maxsize);
      return -1;
    }
  if (atleast > 0)
    {
      nret = t_ring_read (ring, buff, atleast);
      if (nret < atleast)
	return nret;
    }
  if (!str)
    return nret;

  nread = t_ring_offset (ring, maxsize - atleast, str, nstr);
  if (nread >= 0)
    {
      nread = t_ring_read (ring, buff, nread);
      if (nread > 0)
	nret += nread;
    }
  return nret;
}

/** 
 * t_ring_read_with:
 */
gint
t_ring_read_with (TRing *ring,
		  gchar *buff,
		  gint   atleast,
		  gint   maxsize,
		  gchar *str,
		  gint   nstr)
{
  gint nread, nret = 0;

  g_return_val_if_fail (ring, -1);
  g_return_val_if_fail (atleast >= 0, -1);

  if (atleast >= maxsize)
    {
      g_critical ("arguments of ring_read_with() incorrect, atleast %d must "
		  "less than or equal maxsize %d", atleast, maxsize);
      return -1;
    }
  if (atleast > 0)
    {
      nret = t_ring_read (ring, buff, atleast);
      if (nret < atleast)
	return nret;
    }
  if (!str)
    return nret;

  nread = t_ring_offset (ring, maxsize - atleast, str, nstr);
  if (nread >= 0)
    {
      if (nstr < 0)
	{
	  nstr = strlen (str);
	}
      nread = t_ring_read (ring, buff, nread + nstr);
      if (nread > 0)
	nret += nread;
    }
  return nret;
}

/** 
 * t_ring_discard_to:
 */
gint
t_ring_discard_to (TRing *ring,
		   gint   atleast,
		   gint   maxsize,
		   gchar *str,
		   gint   nstr)
{
  g_return_val_if_fail (ring, -1);

  return t_ring_read_to (ring, NULL, atleast, maxsize, str, nstr);
}

/** 
 * t_ring_discard_with:
 */
gint
t_ring_discard_with (TRing *ring,
		     gint   atleast,
		     gint   maxsize,
		     gchar *str,
		     gint   nstr)
{
  g_return_val_if_fail (ring, -1);

  return t_ring_read_with (ring, NULL, atleast, maxsize, str, nstr);
}
