/* Copyright (C) 2010 xiaohu <xiaohu417@gmail.com>,
 * All rights reserved.
 */
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <sys/ioctl.h>
#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif
#include <errno.h>
#include <unistd.h>

#include <glib/gi18n-lib.h>
#include <gio/gio.h>

#include "ttermio.h"

#undef CTRL
#define CTRL(x)	((x) & 0x1F)


/**
 * t_termio_set_iflag:
 */
void
t_termio_set_iflag (gint      fd,
		    guint     attr,
		    gboolean  setting)
{
  struct termios ttmode;

  g_return_if_fail (isatty (fd));

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return;
    }
  if (setting)
    ttmode.c_iflag |= attr;
  else
    ttmode.c_iflag &= ~attr;

  tcsetattr (fd, TCSANOW, &ttmode);
}

/**
 * t_termio_get_iflag:
 */
gboolean
t_termio_get_iflag (gint  fd,
		    guint attr)
{
  struct termios ttmode;

  g_return_val_if_fail (isatty (fd), FALSE);

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return FALSE;
    }
  return ttmode.c_iflag & attr;
}

/**
 * t_termio_set_oflag:
 */
void
t_termio_set_oflag (gint      fd,
		    guint     attr,
		    gboolean  setting)
{
  struct termios ttmode;

  g_return_if_fail (isatty (fd));

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return;
    }
  if (setting)
    ttmode.c_oflag |= attr;
  else
    ttmode.c_oflag &= ~attr;

  tcsetattr (fd, TCSANOW, &ttmode);
}

/**
 * t_termio_get_oflag:
 */
gboolean
t_termio_get_oflag (gint  fd,
		    guint attr)
{
  struct termios ttmode;

  g_return_val_if_fail (isatty (fd), FALSE);

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return FALSE;
    }
  return ttmode.c_oflag & attr;
}

/**
 * t_termio_set_lflag:
 */
void
t_termio_set_lflag (gint      fd,
		    guint     attr,
		    gboolean  setting)
{
  struct termios ttmode;

  g_return_if_fail (isatty (fd));

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return;
    }
  if (setting)
    ttmode.c_lflag |= attr;
  else
    ttmode.c_lflag &= ~attr;

  tcsetattr (fd, TCSANOW, &ttmode);
}

/**
 * t_termio_get_lflag:
 */
gboolean
t_termio_get_lflag (gint  fd,
		    guint attr)
{
  struct termios ttmode;

  g_return_val_if_fail (isatty (fd), FALSE);

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return FALSE;
    }
  return ttmode.c_lflag & attr;
}

/**
 * t_termio_set_cflag:
 */
void
t_termio_set_cflag (gint      fd,
		    guint     attr,
		    gboolean  setting)
{
  struct termios ttmode;

  g_return_if_fail (isatty (fd));

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return;
    }
  if (setting)
    ttmode.c_cflag |= attr;
  else
    ttmode.c_cflag &= ~attr;

  tcsetattr (fd, TCSANOW, &ttmode);
}

/**
 * t_termio_get_cflag:
 */
gboolean
t_termio_get_cflag (gint  fd,
		    guint attr)
{
  struct termios ttmode;

  g_return_val_if_fail (isatty (fd), FALSE);

  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr() error: %s", g_strerror (errno));
      return FALSE;
    }
  return ttmode.c_cflag & attr;
}

/**
 * t_termio_makeraw:
 */
void
t_termio_makeraw (gint fd)
{
  struct termios ttmode;

  g_return_if_fail (isatty (fd));

#ifdef HAVE_CFMAKERAW
  cfmakeraw (&ttmode);
#else
  if (tcgetattr (fd, &ttmode) != 0)
    {
      g_warning ("tcgetattr error: %s", g_strerror (errno));
      return;
    }
  ttmode.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP |
		      INLCR | IGNCR | ICRNL | IXON);
  ttmode.c_oflag &= ~OPOST;
  ttmode.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
  ttmode.c_cflag &= ~(CSIZE | PARENB);
  ttmode.c_cflag |= CS8;
#endif

  tcsetattr (fd, TCSANOW, &ttmode);
}

/**
 * t_termio_getpass:
 *
 * get password from console.
 */
gint
t_termio_getpass (gchar   *buff,
		  gint     size,
		  GError **error)
{
  struct termios ttmode;
  const gchar *prompt;
  gint i, n;

  g_return_val_if_fail (buff, -1);
  g_return_val_if_fail (size > 0, -1);

  g_return_val_if_fail (isatty (STDIN_FILENO), -1);
  g_return_val_if_fail (isatty (STDOUT_FILENO), -1);

  if (tcgetattr (STDIN_FILENO, &ttmode) != 0)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
		   "tcgetattr() error: %s", g_strerror (errno));
      return -1;
    }
  t_termio_makeraw (STDIN_FILENO);

  prompt = _("Password: ");
  if (write (STDOUT_FILENO, prompt, strlen (prompt)) < 0)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
		   "write() error: %s", g_strerror (errno));
      return -1;
    }
  for (i = 0; i < size - 1; i++)
    {
      n = read (STDIN_FILENO, &buff[i], 1);
      if (n < 0)
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
		       "read() error: %s", g_strerror (errno));
	  return -1;
	}
      if (buff[i] == '\r' || buff[i] == '\n')
	break;

      /* give up. */
      if (buff[i] == CTRL('c') || buff[i] == CTRL('d'))
	{
	  tcsetattr (STDIN_FILENO, TCSANOW, &ttmode);
	  return 0;
	}
      /* control character is not acceptable. */
      if (g_ascii_iscntrl (buff[i]))
	{
	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA,
		       _("Invalid password character."));
	  tcsetattr (STDIN_FILENO, TCSANOW, &ttmode);
	  return -1;
	}
    }
  if (i == size - 1)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NO_SPACE,
		   _("No space to store password."));
      tcsetattr (STDIN_FILENO, TCSANOW, &ttmode);
      return -1;
    }
  buff[i] = '\0';

  tcsetattr (STDIN_FILENO, TCSANOW, &ttmode);
  return strlen (buff);
}

/**
 * t_termio_set_winsize:
 *
 * set terminal screen size.
 */
gboolean
t_termio_set_winsize (gint    fd,
		      guint16 rows,
		      guint16 cols)
{
  struct winsize wsize;

  memset (&wsize, 0, sizeof (wsize));

  wsize.ws_row = rows;
  wsize.ws_col = cols;

  if (ioctl (fd, TIOCSWINSZ, &wsize) != 0)
    {
      g_warning ("ioctl(TIOCSWINSZ) error: %s", g_strerror (errno));
      return FALSE;
    }
  return TRUE;
}

/**
 * t_termio_get_winsize:
 *
 * get terminal screen size
 */
gboolean
t_termio_get_winsize (gint     fd,
		      guint16 *rows,
		      guint16 *cols)
{
  struct winsize wsize;

  g_return_val_if_fail (rows, FALSE);
  g_return_val_if_fail (cols, FALSE);

  memset (&wsize, 0, sizeof (wsize));

  if (ioctl (fd, TIOCGWINSZ, &wsize) != 0)
    {
      g_warning ("ioctl(TIOCGWINSZ) error: %s", g_strerror (errno));
      return FALSE;
    }
  *cols = wsize.ws_col;
  *rows = wsize.ws_row;

  return TRUE;
}

